#include "api/df_setup_lists.h"
#include "datafox_library_iv.h"

#if defined(_MSC_VER) || defined(__MINGW32__)
#define snprintf	    _snprintf
#endif

DF_API_EXTERN_DECORATE int DF_API_CALL DFCLoadListenbeschreibung( int connectionNumber, int deviceAddress, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber != nullptr )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    int l_error_number = IV_ERROR_SUCCESS;
    int iRes = channelBridge->getTableDescription( l_device_address, 1, &l_error_number );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCListBCnt( int connectionNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, "" );

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    int iRes = channelBridge->GetTableCount( 1 );

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCListBDatensatz( int connectionNumber, int number, char *name, int *fieldCount, int *unused_setTo0 )
{
    (void)unused_setTo0;

    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, "\t%i", number );

    // zulaessige Bereiche pruefen
    if ( ((number >= 0 && number < MAX_RECORD_DESCRIPTION)
           || (number >= 100 && number < 104)
           || (number >= 200 && number < 208)) == false
         || name == nullptr || fieldCount == nullptr )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t\t\t\t(%i, %i)", 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    std::string szTableName;
    unsigned char ucFieldCount, ucVariable;

    int iRes = channelBridge->GetTableRecord( number, szTableName, &ucFieldCount, &ucVariable, 1 );

    if ( iRes )
    {
        strcpy( name, szTableName.c_str() );
        *fieldCount = ucFieldCount;
        // Variable nicht verwendet, kann nullptr uebergeben
    }

    channelBridge->logEnd( "\t%s\t%i\t0\t(%i, %i)", name, *fieldCount, iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCListBFeld( int connectionNumber, int listBNumber, int number, char *name, int *type, int *size )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, "\t%i\t%i", listBNumber, number );

    // zulaessige Bereiche pruefen
    if ( ((listBNumber >= 0 && listBNumber < MAX_RECORD_DESCRIPTION)
           || (listBNumber >= 100 && listBNumber < 104)
           || (listBNumber >= 200 && listBNumber < 208)) == false
         || (number >= 0 && number < MAX_FIELDS) == false || name == nullptr || type == nullptr || size == nullptr )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t\t\t\t(%i, %i)", 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    std::string szFieldName;
    unsigned char ucType, ucLength;

    int iRes = channelBridge->GetTableField( listBNumber, number, szFieldName,
                                             &ucType, &ucLength, 1 );

    if ( iRes )
    {
        strcpy( name, szFieldName.c_str() );
        *type = ucType;
        *size = ucLength;
    }

    channelBridge->logEnd( "\t%s\t%i\t%i\t(%i, %i)", name, *type, *size, iRes, channelBridge->getLastErrorNumber( 0 ) );
    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCMakeListe( int connectionNumber, int number, int recordCount, int size,
                                                     const unsigned char *buffer, int unused_setTo0 )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, "\t%i\t%i\t%i\t\t%i", number, recordCount, size, unused_setTo0 );

    // zulaessige Bereiche pruefen
    if ( (number >= 0 && number < MAX_RECORD_DESCRIPTION) == false
         || (buffer == nullptr && (recordCount != 0 || size != 0)) )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Bei mindestens einer aktiven Blockuebertragung nicht moeglich.
    if ( channelBridge->getBlockTransferActiveCount( ControllerSegment::btt_list ) )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_BLOCK_TRANSFER_ACTIVE );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    int iRes = channelBridge->MakeListe( number, recordCount, size, buffer );

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE void DF_API_CALL DFCClrListenBuffer( int connectionNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return;
    }

    channelBridge->logBegin( __func__, "" );

    // Bei mindestens einer aktiven Blockuebertragung nicht moeglich.
    if ( channelBridge->getBlockTransferActiveCount( ControllerSegment::btt_list ) )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_BLOCK_TRANSFER_ACTIVE );
        channelBridge->logEnd( "\t(-, %i)", channelBridge->getLastErrorNumber( 0 ) );
        return;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->ClearListBuffer();

    channelBridge->logEnd( "\t(-, %i)", channelBridge->getLastErrorNumber( 0 ) );
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCLoadListen( int connectionNumber, int deviceAddress, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber != nullptr )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Geräteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Anderer Blocktransfer besteht bereits?
    if ( channelBridge->getBlockTransferState( l_device_address ) == ControllerSegment::BlockTransferState::active
         && channelBridge->getBlockTransferType( l_device_address ) != ControllerSegment::btt_list )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_BLOCK_TRANSFER_ACTIVE, errorNumber );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    int iRes;
    // Listendaten wegschreiben.
    if ( channelBridge->logIsWriteList()
         && channelBridge->getBlockTransferState( l_device_address ) == ControllerSegment::BlockTransferState::inactive )
    {
        char fileNameBin[_MAX_PATH];
        char basePath[340];
        int listNum;
        time_t t;
        tm *ltm;

        t = time( nullptr );
        ltm = localtime( &t );

        for ( listNum = 0; listNum < MAX_RECORD_DESCRIPTION; listNum++ )
        {
            // Dateiname der Logdatei
            std::string logBasePath = channelBridge->logGetBasePath( basePath );
            int res = snprintf( fileNameBin, sizeof( fileNameBin ) - 1, LOG_LIST_FILENAME_BIN, logBasePath.c_str(),
                      ltm->tm_year % 100, ltm->tm_mon + 1, ltm->tm_mday,
                      ltm->tm_hour, ltm->tm_min, ltm->tm_sec, connectionNumber, 'L', listNum );

            if ( res > static_cast<int>(sizeof( fileNameBin )) )
            {
                assert( false );
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FILE_OPEN, errorNumber );
                channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
                return 0;
            }

            // Schreiben der Listendaten in die Datei.
            channelBridge->WriteListToBinFile( listNum, fileNameBin );
        }
    }

    // Listendaten wegschreiben.
    if ( channelBridge->logIsWriteList() > 1
         && channelBridge->getBlockTransferState( l_device_address ) == ControllerSegment::BlockTransferState::inactive )
    {
        char fileNameTxt[_MAX_PATH];
        char basePath[340];
        std::string listName;
        unsigned char fieldCount, varList;
        int listNum;
        time_t t;
        tm *ltm;
        int l_error_number = IV_ERROR_SUCCESS;

        // Laden der Listenbeschreibungen
        iRes = channelBridge->getTableDescription( l_device_address, 1, &l_error_number );
        // Nach Kommunikationsfehler, Kanal bereinigen
        if ( iRes == 0 )
        {
            channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber );
            channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
            channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );
            return iRes;
        }

        t = time( nullptr );
        ltm = localtime( &t );

        for ( listNum = 0; listNum < MAX_RECORD_DESCRIPTION; listNum++ )
        {
            if ( !channelBridge->GetTableRecord( listNum, listName, &fieldCount, &varList, 1 ) )
            {
                break;
            }

            // Dateiname der Logdatei
            std::string logBasePath = channelBridge->logGetBasePath( basePath );
            int res = snprintf( fileNameTxt, sizeof( fileNameTxt ) - 1, LOG_LIST_FILENAME_TXT, logBasePath.c_str(),
                                ltm->tm_year % 100, ltm->tm_mon + 1, ltm->tm_mday,
                                ltm->tm_hour, ltm->tm_min, ltm->tm_sec, connectionNumber, 'L', listName.c_str() );

            if ( res > static_cast<int>(sizeof( fileNameTxt )) )
            {
                assert( false );
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FILE_OPEN, errorNumber );
                channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
                return 0;
            }
                
            // Schreiben der Daten in die Datei.
            channelBridge->WriteListToTextFile( listNum, fileNameTxt );
        }
    }

    int l_error_number = IV_ERROR_SUCCESS;
    iRes = channelBridge->writeListData( l_device_address, &l_error_number );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );

        if ( channelBridge->getLastErrorNumber( l_device_address ) != IV_ERROR_BLOCK_TRANSFER_BREAK )
        {
            channelBridge->blockTransferDiscard( l_device_address );
        }
    }
#ifdef _DEBUG
    // Listendaten nach der Sortierung wegschreiben.
    if ( channelBridge->logIsWriteList()
         && channelBridge->getBlockTransferState( l_device_address ) == ControllerSegment::BlockTransferState::inactive )
    {
        char fileNameBin[_MAX_PATH];
        char basePath[340];
        int listNum;
        time_t t;
        tm *ltm;

        t = time( nullptr );
        ltm = localtime( &t );
        // Änderung des Dateinamens sicherstellen.
        ltm->tm_sec++;

        for ( listNum = 0; listNum < MAX_RECORD_DESCRIPTION; listNum++ )
        {
            // Dateiname der Logdatei
            snprintf( fileNameBin, sizeof( fileNameBin ) - 1, LOG_LIST_FILENAME_BIN, channelBridge->logGetBasePath( basePath ),
                      ltm->tm_year % 100, ltm->tm_mon + 1, ltm->tm_mday,
                      ltm->tm_hour, ltm->tm_min, ltm->tm_sec, connectionNumber, 'L', listNum );

            channelBridge->WriteListToBinFile( listNum, fileNameBin );
        }
    }

    // Listendaten wegschreiben.
    if ( channelBridge->logIsWriteList() > 1
         && channelBridge->getBlockTransferState( l_device_address ) == ControllerSegment::BlockTransferState::inactive )
    {
        char fileNameTxt[_MAX_PATH];
        char basePath[340];
        std::string listName;
        unsigned char fieldCount, varList;
        int listNum;
        time_t t;
        tm *ltm;

        // Datensatzbeschreibung ist noch von vorher geladen.

        t = time( nullptr );
        ltm = localtime( &t );
        // Änderung des Dateinamens sicherstellen.
        ltm->tm_sec++;

        for ( listNum = 0; listNum < MAX_RECORD_DESCRIPTION; listNum++ )
        {
            if ( !channelBridge->GetTableRecord( listNum, listName, &fieldCount, &varList, 1 ) )
            {
                break;
            }

            // Dateiname der Logdatei
            snprintf( fileNameTxt, sizeof( fileNameTxt ) - 1, LOG_LIST_FILENAME_TXT, channelBridge->logGetBasePath( basePath ),
                      ltm->tm_year % 100, ltm->tm_mon + 1, ltm->tm_mday,
                      ltm->tm_hour, ltm->tm_min, ltm->tm_sec + 1, connectionNumber, 'L', listName.c_str() );


                  // Daten in die Datei schreiben.
            channelBridge->WriteListToTextFile( listNum, fileNameTxt );
        }
    }
#endif
    channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}
