/****************************************************************************
**
** Copyright (C) Datafox GmbH.
** All rights reserved.
** Contact: Datafox GmbH (develop@datafox.de)
**
** This file provides a collection of helper functions.
**
****************************************************************************/
#pragma once

#include <string>
#include <codecvt>
#include <vector>
#include <iostream>

#if defined(_MSC_VER)

    #ifndef NDEBUG
        #define DF_ASSERT(cond, msg) { \
            if ( !cond ) { \
                OutputDebugStringA("ERROR: "); \
                OutputDebugStringA(msg); \
                OutputDebugStringA("\n"); \
            } \
            BOOST_ASSERT(cond); \
        }
    #else
        #define DF_ASSERT(cond, msg) void()
    #endif

#else

    #ifndef NDEBUG
        #define DF_ASSERT(cond, msg) { \
            BOOST_ASSERT(cond); \
        }
    #else
        #define DF_ASSERT(cond, msg) void()
    #endif

#endif

namespace Utils
{
    namespace File
    {
        bool exists( const std::string &name );

        const std::string extension( const std::string &filePath );
        const std::string name( const std::string &filePath );
        const std::string dir( const std::string &filePath );

        bool copy( const std::string &srcPath, const std::string &dstPath );

        const std::string &replaceDirSeparator( std::string &path );

        const std::string uniqueFileName( const std::string &prefix = "", const std::string &extension = "" );
        const std::string tempFileName( const std::string &prefix = "" );

        const std::string uniqueTempFileName( const std::string &prefix = "", const std::string &extension = ".tmp" );
    }

    const std::string tempDirectoryPath();
    const std::string uniqueTempDirectoryPath( const std::string &prefix = "" );

    const std::string applicationDirectoryPath();

    constexpr char hexmap[] = { '0', '1', '2', '3', '4', '5', '6', '7',
                                '8', '9', 'a', 'b', 'c', 'd', 'e', 'f' };

    // string manipulation
    std::string toHexString( const uint8_t *data, uint32_t lenght );
    std::string toHexString( const uint32_t &data, bool withPrefix = false );
    
    std::vector<uint8_t> fromHexString( const std::string &data, bool swapOrder = true );
    
    template<typename T>
    T fromHexString_uint( const std::string &data, const T &dummy )
    {
        T result{};
        std::vector<uint8_t> tmp = fromHexString( data );

        if ( tmp.size() >= sizeof( result ) )
        {
            for ( size_t i = 0; i < sizeof( result ); ++i )
            {
                result += tmp[i] << (8 * i);
            }
        }
        return result;
    }

    std::string unicodeToUTF8( const std::wstring &wstr );
    std::wstring utf8ToUnicode( const std::string &str );

    std::string unicodeToANSI( const std::wstring &wstr );
    std::wstring ansiToUnicode( const std::string &str );

    std::string to_utf8( const std::string &str );
    std::string from_utf8( const std::string &str );
    std::string from_utf8( const char *str );

    const std::string replaceAll( const std::string &value, const std::string &search, const std::string &replace );

    namespace Debug
    {
        void printToConsole( const char *format, ... );
    }
}
