#ifndef CONTROLLER_SEGMENT_H
#define CONTROLLER_SEGMENT_H

#include <cstdint>
#include <chrono>
#include <string>
#include <vector>

struct ControllerSegment
{
    ControllerSegment();
    void init();
    void reset();
    enum class BlockTransferState : int32_t
    {
        inactive,
        active
    };
    enum BlockTransferType
    {
        btt_none                 = 0x00000000,
        btt_list                 = 0x00000001,
        btt_access_control_list  = 0x00000002,
        btt_write_file           = 0x00000004,
        btt_read_file            = 0x00000008,
        btt_write_setup_file     = 0x00000010,
        btt_read_setup_file      = 0x00000020,
        btt_mask_known           = 0x0000003F
    };
    struct Flags
    {
        uint32_t settingsFromGlobal : 1;
    };

    BlockTransferState transferState;
    BlockTransferType transferType;
    Flags transferFlags;
    uint32_t allowedTransferTypes;
    std::chrono::system_clock::duration duration;
    std::chrono::system_clock::time_point beginTimePoint;
    std::chrono::system_clock::time_point currentTimeSlotTimePoint;
    std::chrono::system_clock::time_point endTimePoint;
    uint16_t updateVariation;
    uint32_t number;
    uint32_t maxNumber;
    uint32_t address;
    uint32_t size;
    uint32_t crc;
    bool resume;
    bool discard;
    std::string filePath;
    std::vector<uint8_t> fileData;
};

#endif
