#include "precompiled.h"
#include "file/interface/studio_setup_classes.h"
#ifdef DATAFOX_STUDIO_IV
#include "DatafoxMFC/Control/DFTreeCtrlDef.h"
#else
#define ICON_DEVICE_PZE               0
#define ICON_DEVICE_PZE_SEL           1
#define ICON_DEVICE_ZK                2
#define ICON_DEVICE_ZK_SEL            3
#define ICON_DEVICE_BDE               4
#define ICON_DEVICE_BDE_SEL           5
#define ICON_DEVICE_MOBIL             6
#define ICON_DEVICE_MOBIL_SEL         7
#define ICON_DEVICE_MDE               8
#define ICON_DEVICE_MDE_SEL           9
#define ICON_DEVICE_AE                10
#define ICON_DEVICE_AE_SEL            11
#define ICON_TABLE_RECORD             12
#define ICON_TABLE_RECORD_SEL         13
#define ICON_TABLE_LIST               14
#define ICON_TABLE_LIST_SEL           15
#define ICON_TABLE                    16
#define ICON_TABLE_SEL                17
#define ICON_TABLE_FIELD              18
#define ICON_TABLE_FIELD_SEL          19
#define ICON_HANDLING                 20
#define ICON_HANDLING_SEL             21
#define ICON_MENU                     22
#define ICON_MENU_SEL                 23
#define ICON_MENU_CHOICE              24
#define ICON_MENU_CHOICE_SEL          25
#define ICON_INPUT_CHAIN              26
#define ICON_INPUT_CHAIN_SEL          27
#define ICON_INPUT_CHAIN_FIELD        28
#define ICON_INPUT_CHAIN_FIELD_SEL    29
#define ICON_MDE                      30
#define ICON_MDE_SEL                  31
#define ICON_IOMODULE                 32
#define ICON_IOMODULE_SEL             33
#define ICON_DIGITAL_INPUTS           34
#define ICON_DIGITAL_INPUTS_SEL       35
#define ICON_DIGITAL_INPUT            36
#define ICON_DIGITAL_INPUT_SEL        37
#define ICON_ANALOG_INPUTS            38
#define ICON_ANALOG_INPUTS_SEL        39
#define ICON_ANALOG_INPUT             40
#define ICON_ANALOG_INPUT_SEL         41
#define ICON_TIMERS                   42
#define ICON_TIMERS_SEL               43
#define ICON_TIMER                    44
#define ICON_TIMER_SEL                45
#define ICON_INPUT_CHAINS             46
#define ICON_INPUT_CHAINS_SEL         47
#define ICON_ACCESS_CONTROL           48
#define ICON_ACCESS_CONTROL_SEL       49
#define ICON_DEVICE_TIMEBOY           50
#define ICON_DEVICE_TIMEBOY_SEL       51
#define ICON_DEVICE_EXKLUSIV          52
#define ICON_DEVICE_EXKLUSIV_SEL      53
#define ICON_DEVICE_FLEX              54
#define ICON_DEVICE_FLEX_SEL          55
#define ICON_ENERGIE_SCHEMES          56
#define ICON_ENERGIE_SCHEMES_SEL      57
#define ICON_MULTI_INPUT_CHAIN_FIELD  58
#define ICON_MULTI_INPUT_CHAIN_FIELD_SEL    59
#define ICON_INPUT_CHAIN_FIELD_DARK         60
#define ICON_INPUT_CHAIN_FIELD_DARK_SEL     61

#endif
#include "datafox/communication/module_hardware.h"
#include "common/software_version.h"

#ifdef DATAFOX_STUDIO_IV
#include "common/df4_key_box_internal.h"
#endif

#include "common/text_translate.h"

// Ersetzt das derzeitig noch nicht unterstützte std::as_const, weil dieses noch nicht vom verwendeten Compiler gcc unterstützt wird.
#define as_const(x)   x

// ########################################################
// Member: EntryBase

// Offset fuer Einzutragende Adressen die ueber Flash- oder Extended - Mem hinausgehen.
unsigned int EntryBase::m_flashOffset = 0;

// Konstruktor
EntryBase::EntryBase( EntryBase *parent, EntryType type )
    : m_uuid( boost::uuids::random_generator()() )
{
    // Grundinformationen
    m_type = type;
    m_parent = parent;
    m_refCount = 0;

    // Flags auf 0, dann setzen
    setInsertEnabled( false );
    setDeleteEnabled( false );
    setCopyEnabled( false );
    setVisible( true );
    setModified( false );

    m_handleItem = nullptr;
    m_iconIndex = 0;
    m_iconIndexSelected = 0;

    m_name = "";
    m_displayNameFormat = "";

    m_pageIndexes.clear();
}

EntryBase::EntryBase( const EntryBase &entry )
{
    // Member kopieren
    m_uuid = entry.m_uuid;
    m_type = entry.m_type;
    m_parent = entry.m_parent;      // nach Kopie zu pruefen / anzupassen !
    m_refCount = 0;                  // nach Kopie zu pruefen / anzupassen !
    // m_ulFlashOffset ist statisch
    m_typeFlags = entry.m_typeFlags;

    m_handleItem = entry.m_handleItem; // nach Kopie zu pruefen / anzupassen !
    m_iconIndex = entry.m_iconIndex;
    m_iconIndexSelected = entry.m_iconIndexSelected;

    m_name = entry.name();
    m_displayNameFormat = entry.displayNameFormat();

    m_pageIndexes = entry.m_pageIndexes;

    // Auflistung "childs" in jeweiliger Ableitung kopiert.
}

// Destruktor
EntryBase::~EntryBase()
{
    // Weil sich Eingabeketten auf die Tabellenobjekte beziehen,
    // wird von hinten nach vorn geloescht.
    while ( !empty() )
        delete takeLast();
}

/////////////////////////
// Elterneintrag

void EntryBase::setParent( EntryBase *parent )
{
    m_parent = parent; // Parent uebernehmen

    auto iterChild = constBegin();
    for ( ; iterChild != constEnd(); ++iterChild )
    {
        (*iterChild)->setChildsParent( this ); // Unterbaum angleichen
    }

    iterChild = constBegin();
    for ( ; iterChild != constEnd(); ++iterChild )
    {
        (*iterChild)->checkChildsMembers(); // Members der Childs pruefen
    }
}

/////////////////////////
// Elterneintraege

EntryDevice &EntryBase::root()
{
    EntryBase *entry = this;
    while ( !entry->isType( EntryType::Device ) )
    {
        entry = entry->parent();
    }

    return (EntryDevice &)*entry;
}

EntryTableRecord &EntryBase::rootTableRecord()
{
    return (EntryTableRecord &)*root().at( 0 ); // <- RDRoot
}

EntryTableList &EntryBase::rootTableList()
{
    return (EntryTableList &)*root().at( 1 ); // <- LDRoot
}

EntryHandling &EntryBase::rootHandling()
{
    return (EntryHandling &)*root().at( 2 ); // <- ControlRoot
}

EntrySignalProcessing &EntryBase::rootSignalProcessing()
{
    return (EntrySignalProcessing &)*root().at( 3 ); // <- MDERoot
}

EntryEventChains &EntryBase::rootEventChains()
{
    return (EntryEventChains &)*root().at( 4 ); // <- EventChainsRoot
}

EntryIOModule &EntryBase::rootIOModule()
{
    return (EntryIOModule &)*root().at( 5 ); // <- IOModuleRoot
}

EntryAccessControl2 &EntryBase::rootAccessControl()
{
    return (EntryAccessControl2 &)*root().at( 6 ); // <- Entrance2Root
}

EntryEnergieSchemes &EntryBase::rootEnergieSchemes()
{
    return (EntryEnergieSchemes &)*root().at( 7 ); // <- EnergieRoot
}

///////////////////////////////////////
// Childs

// Abfrage ob Eintrag unterhalb im Baum des aktuellen liegt
bool EntryBase::isSubChild( EntryBase *pEntry )
{
    // Jedes Child pruefen
    if ( !contains( pEntry ) )
    {
        // Childs durchsuchen
        auto iterChild = constBegin();
        for ( ; iterChild != constEnd(); ++iterChild )
        {
            if ( (*iterChild)->isSubChild( pEntry ) )
            {
                return true;
            }
        }

        return false;
    }

    return true;
}

// Durchsucht den Baum rekursiv und sammelt alle Eintraege vom uebergebenen Type in der Liste
uint32_t EntryBase::collectAllChildsByType( EntryList &childs, EntryType type )
{
    // Vom erwuenschten Type?
    if ( this->isType( type ) )
    {
        childs.push_back( this );
    }

    // Alle Untereintraege rekursiv abarbeiten!
    auto iterChild = constBegin();
    for ( ; iterChild != constEnd(); ++iterChild )
    {
        (*iterChild)->collectAllChildsByType( childs, type );
    }

    return static_cast<uint32_t>( childs.size() );
}

void EntryBase::deleteEntry( int pos )
{
    EntryBase *entry;

    if ( pos == -1 )
    {
        entry = takeLast();
    }
    else
    {
        entry = takeAt( pos );
    }

    while ( !entry->empty() )
    {
        entry->deleteEntry();
    }

    // Objekt loeschen
    delete entry;
}

void EntryBase::insertEntryFinish( EntryBase *entryNew, int pos )
{
    switch ( pos )
    {
        case 0:
            prepend( entryNew );
            break;

        case -1:
            append( entryNew );
            break;

        default:
            insert( pos, entryNew );
    }
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

void EntryBase::changeOrderInt32( unsigned char *Buf )
{
    unsigned char h;

    h = Buf[0];
    Buf[0] = Buf[3];
    Buf[3] = h;

    h = Buf[1];
    Buf[1] = Buf[2];
    Buf[2] = h;
}

void EntryBase::changeOrderInt16( unsigned char *Buf )
{
    unsigned char h;

    h = Buf[0];
    Buf[0] = Buf[1];
    Buf[1] = h;
}

unsigned short EntryBase::endianReverse( unsigned short val )
{
    return ((val & 0xFF) << 8)
        | ((val >> 8) & 0xFF);
}

unsigned int EntryBase::endianReverse( unsigned int val )
{
    return ((val & 0xFF) << 24)
        | ((val & 0xFF00) << 8)
        | ((val >> 8) & 0xFF00)
        | ((val >> 24) & 0xFF);
}

// Pruefen ob irgendein Eintrag das MODIFIED-Flag gesetzt hat
bool EntryBase::isTreeModified()
{
    // Dieser Modifiziert?
    if ( isModified() )
    {
        return true;
    }

    return isAnyChildModified();
}

bool EntryBase::isAnyChildModified()
{
    // Dieser Modifiziert?
    if ( isModified() )
    {
        return true;
    }

    // Alle Untereintraege rekursiv abarbeiten!
    auto iterChild = constBegin();
    for ( ; iterChild != constEnd(); ++iterChild )
    {
        if ( (*iterChild)->isAnyChildModified() )
        {
            return true;
        }
    }

    return false;
}

// Loeschen des MODIFIED-Flag in allen Baumknoten.
void EntryBase::treeClearModified()
{
    clearAnyChildModified();
}

void EntryBase::clearAnyChildModified()
{
    // Dieser Unmodifiziert!
    setModified( FALSE );
    // Alle Untereintraege rekursiv abarbeiten!
    auto iterChild = constBegin();
    for ( ; iterChild != constEnd(); ++iterChild )
    {
        (*iterChild)->clearAnyChildModified();
    }
}

// Uebernehmen der Parent - Pointer bei Kopie von Teilbaum
void EntryBase::setChildsParent( EntryBase *pEntry )
{
    m_parent = pEntry;

    auto iterChild = constBegin();
    for ( ; iterChild != constEnd(); ++iterChild )
    {
        (*iterChild)->setChildsParent( this );
    }
}

// Pruefen der Members nach Kopie von Teilbaum
void EntryBase::checkChildsMembers()
{
    auto iterChild = constBegin();
    for ( ; iterChild != constEnd(); ++iterChild )
    {
        (*iterChild)->checkChildsMembers();
    }
}

///////////////////////////////////
// TreeCtrl

//void EntryBase::setName( const QString &name )
//{
//    _name = name;
//}

EntryBase *EntryBase::findTreeItem( void *hSelTreeItem )
{
    if ( m_handleItem == hSelTreeItem )
    {
        return this; // Item gefunden -> suche beendet
    }
    else
    {
        // sonst: Alle Child durchsuchen
        EntryBase *Catch;
        auto i = constBegin();
        for ( ; i != constEnd(); ++i )
        {
            if ( (Catch = (*i)->findTreeItem( hSelTreeItem )) != nullptr )
            {
                return Catch; // Item gefunden -> Zeiger auf Eintrag zurückgeben
            }
        }
    }

    return nullptr;	// Item nicht gefunden
}


// ########################################################
// Member: EntryDevice

// Konstruktor
EntryDevice::EntryDevice( EntryBase *parent, EntryType type )
    : EntryBase( parent, type )
{
    // Prüfungen
    BOOST_ASSERT( sizeof( StructExtended ) == 1024 );
}

// Destruktor
EntryDevice::~EntryDevice()
{
    // Nach einer bestimmten Reihenfolge die Eintraege loeschen
    delete at( 2 ); // <- HandlingRoot
    delete at( 5 ); // <- IOModulRoot
    delete at( 3 ); // <- SignalProcessingRoot
    delete at( 6 ); // <- Entrance2Root
    delete at( 4 ); // <- EventChainsRoot
    delete at( 7 ); // <- EnergieRoot
    delete at( 0 ); // <- RDRoot
    delete at( 1 ); // <- LDRoot

    clear();
}

///////////////////////////////////
// Childs

EntryBase *EntryDevice::insertEntry( EntryType typeID, int pos )
{
    (void)typeID;
    (void)pos;

    return nullptr;
}

EntryBase *EntryDevice::cloneEntry()
{
    // Dieser Eintrag unterstützt derzeit cloneEntry nicht.
    return nullptr;
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryDevice::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();

    // Alle Indizes aus Referencen erzeugen
    convertReferenceToIndex();

    // Transponderstruktur nach gewaehltem Typ schreiben
    if ( (m_config.transponderType & 0x3F) == TRANSPONDER_MIFARE )
        memcpy( &m_config.transponderGlobal, &m_transponderMifare, sizeof( StructTransponderMifare ) );
    else
        m_config.transponderGlobal = m_transponder;


    //#################################
    // Flash-Struktur

    // Flashoffset für die folgende Serialisierung auf 0 setzen.
    file.m_flashMemOffset = 0;

    m_globalVar.ptr = sizeof( StructGlobalVar );
    m_globalVar.ptr = endianReverse( m_globalVar.ptr + FLASH_KONFIG_START );

    *reinterpret_cast<StructGlobalVar *>(&file.m_flash[file.m_flashMemOffset]) = m_globalVar;
    file.m_flashMemOffset += sizeof( StructGlobalVar );


    ////////////////////////////////////////
    // add information to xml file
    FileXML &fileXML = doc.fileXML();
    // Clear the current xml file. The following operations completely rebuild the file.
    fileXML.reset();

    // Finding the node for this structure and update parent.
    auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::ImageScope ) );
    fileXML.writeContent( FileXML::ContentType::GlobalVar, as_const( m_globalVar ) );

    /////////////////////////////////////////////
    // Die Indizes der Tabellen und deren Felder muss vor der Bedienung feststehen,
    // da Eingabeketten und deren Felder auf diese Werte angewiesen sind.

    // Sicherstellen, dass die richtigen Indizierung bei Tabellen und deren Feldern vorliegt.
    int idRoot = 0, idChild;
    // Alle Datensatzbeschreibungen durchlaufen.
    auto iterRoot = rootTableRecord().constBegin();
    for ( ; iterRoot != rootTableRecord().constEnd(); ++iterRoot )
    {
        ((EntryTable *)*iterRoot)->m_table.number = idRoot++;
        idChild = 0;
        auto iterChild = (*iterRoot)->constBegin();
        for ( ; iterChild != (*iterRoot)->constEnd(); ++iterChild )
            ((EntryTableField *)*iterChild)->m_tableField.number = idChild++;
    }

    // Alle Listenbeschreibungen durchlaufen
    iterRoot = rootTableList().constBegin();
    for ( ; iterRoot != rootTableList().constEnd(); ++iterRoot )
    {
        ((EntryTable *)*iterRoot)->m_table.number = idRoot++;
        idChild = 0;
        auto iterChild = (*iterRoot)->constBegin();
        for ( ; iterChild != (*iterRoot)->constEnd(); ++iterChild )
            ((EntryTableField *)*iterChild)->m_tableField.number = idChild++;
    }

    // Bedien - Konfiguration
    uint32_t pDat = rootHandling().writeDocument( file.m_flashMemOffset, doc );

    // Datensatzbeschreibungen
    auto pList = rootTableRecord().writeDocument( pDat, doc );

    // Listenbeschreibungen
    pFlash = rootTableList().writeDocument( pList, doc );

    // pFlash = erster freier Speicherplatz im Flash nach Setup Daten komplett

    // Reset parent node because the end of the structure 'imageScope' is reached.
    fileXML.setParentNode( oldParent );

    // Finding the node for this structure and update parent.
    oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::ExtendedScope ) );

    // Signalverarbeitung - Konfiguration
    rootSignalProcessing().writeDocument( 0, doc );

    // Ereignisketten - Konfiguration
    rootEventChains().writeDocument( 0, doc );

    // IOModule - Konfiguration
    rootIOModule().writeDocument( 0, doc );

    // Entrance2 - Arbeitet auf ExtendedFlash- und ConfigMem- Bereich.
    rootAccessControl().writeDocument( 0, doc );

    // Energieverwaltung schreiben.
    rootEnergieSchemes().writeDocument( 0, doc );

    //#################################
    // Flash-Header-Struktur

    // pDat = erster freier Speicherplatz im Flash nach Konfig-Daten = Start Datensatzbeschreibung
    file.m_flash_header.dataStruct = pDat;
    file.m_flash_header.dataStruct = endianReverse( file.m_flash_header.dataStruct + FLASH_KONFIG_START );

    // pList = erster freier Speicherplatz im Flash nach Datensatzb. = Start Listenbeschreibung
    file.m_flash_header.listStruct = pList;
    file.m_flash_header.listStruct = endianReverse( file.m_flash_header.listStruct + FLASH_KONFIG_START );

    // Headline (ab FW 03.01.22)
    std::fill_n( file.m_flash_header.headline, sizeof( file.m_flash_header.headline ), 0 );

    //#################################
    // CONFIG-Stuktur
    file.m_config = m_config;

#ifdef DATAFOX_STUDIO_IV
    // Verschlüsseln der Passwörter
    if ( m_passwordProtection )
    {
        df4_crypt_internal df4;
        df4.crypt( reinterpret_cast<unsigned char *>(&file.m_config.transponderUserPassword), sizeof( file.m_config.transponderUserPassword ) );
        df4.init();
        df4.crypt( reinterpret_cast<unsigned char *>(&file.m_config.passwordSystemMenu), sizeof( file.m_config.passwordSystemMenu ) );
    }
#endif

    // Reset parent node because the end of the structure 'imageScope' is reached.
    fileXML.setParentNode( oldParent );

    fileXML.writeContent( FileXML::ContentType::Config, as_const( m_config ) );

    //#################################
    // Extended-Flash-Struktur

    // Lesen der Extended - Struktur
    StructExtendedFlash extBegin = *reinterpret_cast<StructExtendedFlash *>(file.m_flashExtended);

    // Standardwerte fuer GV in ExtendedFlash eintragen.
    extBegin.offsetGlobalVariableValues = extBegin.offsetFreeScope;

    // Schreiben der Wertestruktur fuer Globale Variable
    *reinterpret_cast<StructValueGlobalVar *>(&file.m_flashExtended[extBegin.offsetGlobalVariableValues]) = m_valueGlobalVar;

    extBegin.offsetFreeScope += sizeof( StructValueGlobalVar );

    // Finding the node for this structure and update parent.
    oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::ExtendedScope ) );

    fileXML.writeContent( FileXML::ContentType::GlobalVarValue, as_const( m_valueGlobalVar ) );

    // Finding the node for this structure (transponderConfiguration) and update parent.
    auto oldParent2 = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::TransponderConfiguration ) );

    switch ( m_config.transponderType & 0x3F )
    {
        case TRANSPONDER_LEGIC_ADVANT:
        {
            // Benoetigte Mindestversion
            file.NeedStudioVersion( "04.02.03.02" );
            file.NeedFirmware2Version( "04.02.03.01" );
            file.UnsupportFirmware1Version();

            // Struktur STransponderLegicAdvantOld wurde ab 04.02.03 verworfen und wird nicht mehr abgelegt.

            // Gesamte Transponderkonfigurationen in den Erweitertbereich ablegen.
            // Eintragen der gesamten Konfigurationen in den Extended-Bereich.
            extBegin.offsetTransponderConfiguration = extBegin.offsetFreeScope;

            // Feste Werte sicherstellen und eintragen.
            m_transponderLegicAdvantHeader.structSize = sizeof( StructTransponderLegicAdvantHeader );
            m_transponderLegicAdvantHeader.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
            m_transponderLegicAdvantHeader.transponderType = (m_config.transponderType & 0x3F);
            // Headerstruktur eintragen
            *(StructTransponderLegicAdvantHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration] = m_transponderLegicAdvantHeader;
            fileXML.writeContent( FileXML::ContentType::TransponderLegicAdvantHeader, as_const( m_transponderLegicAdvantHeader ) );

            // Basiszeiger und Zeiger auf OffsetArray
            unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + sizeof( StructTransponderLegicAdvantHeader )];
            unsigned short *offsArray = (unsigned short *)base;
            // Zeiger auf die erste Transponderkonfiguration.
            StructTransponderLegicAdvant *transponderStruct = (StructTransponderLegicAdvant *)(base + m_transponderLegicAdvantHeader.count * sizeof( unsigned short ));

            int i;
            for ( i = 0; i < m_transponderLegicAdvantHeader.count; i++ )
            {

// Offset und Transponderstruktur eintragen.
                *offsArray = (unsigned short)((unsigned char *)transponderStruct - base);
                *transponderStruct = m_transponderLegicAdvant[i];
                fileXML.writeContent( FileXML::ContentType::TransponderLegicAdvant, as_const( m_transponderLegicAdvant[i] ) );

                // Beide Zeiger um eins weiter.
                offsArray++;
                transponderStruct++;
            }

            // Positionsoffset des naechsten freien Speichers berechnen.
            extBegin.offsetFreeScope += sizeof( StructTransponderLegicAdvantHeader ) + m_transponderLegicAdvantHeader.count * sizeof( unsigned short ) + m_transponderLegicAdvantHeader.count * sizeof( StructTransponderLegicAdvant );
            break;
        }

        case TRANSPONDER_MIFARE_DESFIRE:
        {
            // Benoetigte Mindestversion
            file.NeedStudioVersion( "04.02.03.02" );
            file.NeedFirmware2Version( "04.02.03.01" );
            file.UnsupportFirmware1Version();

            // Gesamte Transponderkonfigurationen in den Erweitertbereich ablegen.

            // Eintragen der gesamten Konfigurationen in den Extended-Bereich.
            extBegin.offsetTransponderConfiguration = extBegin.offsetFreeScope;

            // Feste Werte sicherstellen und eintragen.
            m_transponderMifareDesfireHeader.structSize = sizeof( m_transponderMifareDesfireHeader );
            m_transponderMifareDesfireHeader.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
            m_transponderMifareDesfireHeader.transponderType = (m_config.transponderType & 0x3F);
            // Headerstruktur eintragen
            *(StructTransponderMifareDesfireHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration] = m_transponderMifareDesfireHeader;
            fileXML.writeContent( FileXML::ContentType::TransponderMifareDesfireHeader, as_const( m_transponderMifareDesfireHeader ) );

            // Basiszeiger und Zeiger auf OffsetArray
            unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + sizeof( StructTransponderMifareDesfireHeader )];
            unsigned short *offsArray = (unsigned short *)base;
            // Zeiger auf die erste Transponderkonfiguration.
            StructTransponderMifareDesfire *transponderStruct = (StructTransponderMifareDesfire *)(base + m_transponderMifareDesfireHeader.count * sizeof( unsigned short ));

            int i;
            for ( i = 0; i < m_transponderMifareDesfireHeader.count; i++ )
            {

// Offset und Transponderstruktur eintragen.
                *offsArray = (unsigned short)((unsigned char *)transponderStruct - base);
                *transponderStruct = m_transponderMifareDesfire[i];
                fileXML.writeContent( FileXML::ContentType::TransponderMifareDesfire, as_const( m_transponderMifareDesfire[i] ) );

                // Beide Zeiger um eins weiter.
                offsArray++;
                transponderStruct++;
            }

            // Positionsoffset des naechsten freien Speichers berechnen.
            extBegin.offsetFreeScope += sizeof( StructTransponderMifareDesfireHeader ) + m_transponderMifareDesfireHeader.count * sizeof( unsigned short ) + m_transponderMifareDesfireHeader.count * sizeof( StructTransponderMifareDesfire );

            break;
        }

        case TRANSPONDER_MIFARE_PLUS:
        {
            // Benoetigte Mindestversion
            file.NeedStudioVersion( "04.02.03.02" );
            file.NeedFirmware2Version( "04.02.03.01" );
            file.UnsupportFirmware1Version();

            // Gesamte Transponderkonfigurationen in den Erweitertbereich ablegen.

            // Eintragen der gesamten Konfigurationen in den Extended-Bereich.
            extBegin.offsetTransponderConfiguration = extBegin.offsetFreeScope;

            // Feste Werte sicherstellen und eintragen.
            m_transponderMifarePlusHeader.structSize = sizeof( m_transponderMifarePlusHeader );
            m_transponderMifarePlusHeader.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
            m_transponderMifarePlusHeader.transponderType = (m_config.transponderType & 0x3F);
            // Headerstruktur eintragen
            *(StructTransponderMifarePlusHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration] = m_transponderMifarePlusHeader;
            fileXML.writeContent( FileXML::ContentType::TransponderMifarePlusHeader, as_const( m_transponderMifarePlusHeader ) );

            // Basiszeiger und Zeiger auf OffsetArray
            unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + sizeof( StructTransponderMifarePlusHeader )];
            unsigned short *offsArray = (unsigned short *)base;
            // Zeiger auf die erste Transponderkonfiguration.
            StructTransponderMifarePlus *transponderStruct = (StructTransponderMifarePlus *)(base + m_transponderMifarePlusHeader.count * sizeof( unsigned short ));

            int i;
            for ( i = 0; i < m_transponderMifarePlusHeader.count; i++ )
            {

// Offset und Transponderstruktur eintragen.
                *offsArray = (unsigned short)((unsigned char *)transponderStruct - base);
                *transponderStruct = m_transponderMifarePlus[i];
                fileXML.writeContent( FileXML::ContentType::TransponderMifarePlus, as_const( m_transponderMifarePlus[i] ) );

                // Beide Zeiger um eins weiter.
                offsArray++;
                transponderStruct++;
            }

            // Positionsoffset des naechsten freien Speichers berechnen.
            extBegin.offsetFreeScope += sizeof( StructTransponderMifarePlusHeader ) + m_transponderMifarePlusHeader.count * sizeof( unsigned short ) + m_transponderMifarePlusHeader.count * sizeof( StructTransponderMifarePlus );

            break;
        }

        case TRANSPONDER_MIFARE_ULTRALIGHT:
        {
            // Benoetigte Mindestversion
            file.NeedStudioVersion( "04.02.03.02" );
            file.NeedFirmware2Version( "04.02.03.01" );
            file.UnsupportFirmware1Version();

            // Gesamte Transponderkonfigurationen in den Erweitertbereich ablegen.

            // Eintragen der gesamten Konfigurationen in den Extended-Bereich.
            extBegin.offsetTransponderConfiguration = extBegin.offsetFreeScope;

            // Feste Werte sicherstellen und eintragen.
            m_transponderMifareUltralightHeader.structSize = sizeof( m_transponderMifareUltralightHeader );
            m_transponderMifareUltralightHeader.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
            m_transponderMifareUltralightHeader.transponderType = (m_config.transponderType & 0x3F);
            // Headerstruktur eintragen
            *(StructTransponderMifareUltralightHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration] = m_transponderMifareUltralightHeader;
            fileXML.writeContent( FileXML::ContentType::TransponderMifareUltralightHeader, as_const( m_transponderMifareUltralightHeader ) );

            // Basiszeiger und Zeiger auf OffsetArray
            unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + sizeof( StructTransponderMifareUltralightHeader )];
            unsigned short *offsArray = (unsigned short *)base;
            // Zeiger auf die erste Transponderkonfiguration.
            StructTransponderMifareUltralight *transponderStruct = (StructTransponderMifareUltralight *)(base + m_transponderMifareUltralightHeader.count * sizeof( unsigned short ));

            int i;
            for ( i = 0; i < m_transponderMifareUltralightHeader.count; i++ )
            {

// Offset und Transponderstruktur eintragen.
                *offsArray = (unsigned short)((unsigned char *)transponderStruct - base);
                *transponderStruct = m_transponderMifareUltralight[i];
                fileXML.writeContent( FileXML::ContentType::TransponderMifareUltralight, as_const( m_transponderMifareUltralight[i] ) );

                // Beide Zeiger um eins weiter.
                offsArray++;
                transponderStruct++;
            }

            // Positionsoffset des naechsten freien Speichers berechnen.
            extBegin.offsetFreeScope += sizeof( StructTransponderMifareUltralightHeader ) + m_transponderMifareUltralightHeader.count * sizeof( unsigned short ) + m_transponderMifareUltralightHeader.count * sizeof( StructTransponderMifareUltralight );

            break;
        }

        case TRANSPONDER_ICODE:
        {
            // Benoetigte Mindestversion
            file.NeedStudioVersion( "04.02.03.02" );
            file.NeedFirmware2Version( "04.02.03.01" );
            file.UnsupportFirmware1Version();

            // Gesamte Transponderkonfigurationen in den Erweitertbereich ablegen.

            // Eintragen der gesamten Konfigurationen in den Extended-Bereich.
            extBegin.offsetTransponderConfiguration = extBegin.offsetFreeScope;

            // Feste Werte sicherstellen und eintragen.
            m_transponderiCodeHeader.structSize = sizeof( m_transponderiCodeHeader );
            m_transponderiCodeHeader.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
            m_transponderiCodeHeader.transponderType = (m_config.transponderType & 0x3F);
            // Headerstruktur eintragen
            *(StructTransponderiCodeHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration] = m_transponderiCodeHeader;
            fileXML.writeContent( FileXML::ContentType::TransponderiCodeHeader, as_const( m_transponderiCodeHeader ) );

            // Basiszeiger und Zeiger auf OffsetArray
            unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + sizeof( StructTransponderiCodeHeader )];
            unsigned short *offsArray = (unsigned short *)base;
            // Zeiger auf die erste Transponderkonfiguration.
            StructTransponderiCode *transponderStruct = (StructTransponderiCode *)(base + m_transponderiCodeHeader.count * sizeof( unsigned short ));

            int i;
            for ( i = 0; i < m_transponderiCodeHeader.count; i++ )
            {

// Offset und Transponderstruktur eintragen.
                *offsArray = (unsigned short)((unsigned char *)transponderStruct - base);
                *transponderStruct = m_transponderiCode[i];
                fileXML.writeContent( FileXML::ContentType::TransponderiCode, as_const( m_transponderiCode[i] ) );

                // Beide Zeiger um eins weiter.
                offsArray++;
                transponderStruct++;
            }

            // Positionsoffset des naechsten freien Speichers berechnen.
            extBegin.offsetFreeScope += sizeof( StructTransponderiCodeHeader ) + m_transponderiCodeHeader.count * sizeof( unsigned short ) + m_transponderiCodeHeader.count * sizeof( StructTransponderiCode );

            break;
        }

        case TRANSPONDER_JAVACARD:
        {
            // Benoetigte Mindestversion
            file.NeedStudioVersion( "04.03.06.02" );
            file.NeedFirmware2Version( "04.03.06.01" );
            file.UnsupportFirmware1Version();

            // Gesamte Transponderkonfigurationen in den Erweitertbereich ablegen.

            // Eintragen der gesamten Konfigurationen in den Extended-Bereich.
            extBegin.offsetTransponderConfiguration = extBegin.offsetFreeScope;

            // Feste Werte sicherstellen und eintragen.
            m_transponderJavaCardHeader.structSize = sizeof( m_transponderJavaCardHeader );
            m_transponderJavaCardHeader.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
            m_transponderJavaCardHeader.transponderType = (m_config.transponderType & 0x3F);
            // Headerstruktur eintragen
            *(StructTransponderJavaCardHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration] = m_transponderJavaCardHeader;
            fileXML.writeContent( FileXML::ContentType::TransponderJavaCardHeader, as_const( m_transponderJavaCardHeader ) );

            // Basiszeiger und Zeiger auf OffsetArray
            unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + sizeof( StructTransponderJavaCardHeader )];
            unsigned short *offsArray = (unsigned short *)base;
            // Zeiger auf die erste Transponderkonfiguration.
            StructTransponderJavaCard *transponderStruct = (StructTransponderJavaCard *)(base + m_transponderJavaCardHeader.count * sizeof( unsigned short ));

            int i;
            for ( i = 0; i < m_transponderJavaCardHeader.count; i++ )
            {

// Offset und Transponderstruktur eintragen.
                *offsArray = (unsigned short)((unsigned char *)transponderStruct - base);
                *transponderStruct = m_transponderJavaCard[i];
                fileXML.writeContent( FileXML::ContentType::TransponderJavaCard, as_const( m_transponderJavaCard[i] ) );

                // Beide Zeiger um eins weiter.
                offsArray++;
                transponderStruct++;
            }

            // Positionsoffset des naechsten freien Speichers berechnen.
            extBegin.offsetFreeScope += sizeof( StructTransponderJavaCardHeader ) + m_transponderJavaCardHeader.count * sizeof( unsigned short ) + m_transponderJavaCardHeader.count * sizeof( StructTransponderJavaCard );

            break;
        }

        case TRANSPONDER_MYD:
        {
            // Benoetigte Mindestversion
            file.NeedStudioVersion( "04.02.03.02" );
            file.NeedFirmware2Version( "04.02.03.01" );
            file.UnsupportFirmware1Version();

            // Gesamte Transponderkonfigurationen in den Erweitertbereich ablegen.

            // Eintragen der gesamten Konfigurationen in den Extended-Bereich.
            extBegin.offsetTransponderConfiguration = extBegin.offsetFreeScope;

            // Feste Werte sicherstellen und eintragen.
            m_transponderMyDHeader.structSize = sizeof( m_transponderMyDHeader );
            m_transponderMyDHeader.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
            m_transponderMyDHeader.transponderType = (m_config.transponderType & 0x3F);
            // Headerstruktur eintragen
            *(StructTransponderMyDHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration] = m_transponderMyDHeader;
            fileXML.writeContent( FileXML::ContentType::TransponderMyDHeader, as_const( m_transponderMyDHeader ) );

            // Basiszeiger und Zeiger auf OffsetArray
            unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + sizeof( StructTransponderMyDHeader )];
            unsigned short *offsArray = (unsigned short *)base;
            // Zeiger auf die erste Transponderkonfiguration.
            StructTransponderMyD *transponderStruct = (StructTransponderMyD *)(base + m_transponderMyDHeader.count * sizeof( unsigned short ));

            int i;
            for ( i = 0; i < m_transponderMyDHeader.count; i++ )
            {

// Offset und Transponderstruktur eintragen.
                *offsArray = (unsigned short)((unsigned char *)transponderStruct - base);
                *transponderStruct = m_transponderMyD[i];
                fileXML.writeContent( FileXML::ContentType::TransponderMyD, as_const( m_transponderMyD[i] ) );

                // Beide Zeiger um eins weiter.
                offsArray++;
                transponderStruct++;
            }

            // Positionsoffset des naechsten freien Speichers berechnen.
            extBegin.offsetFreeScope += sizeof( StructTransponderMyDHeader ) + m_transponderMyDHeader.count * sizeof( unsigned short ) + m_transponderMyDHeader.count * sizeof( StructTransponderMyD );

            break;
        }

        case TRANSPONDER_ISO_TAGS:
        {
            // Benoetigte Mindestversion
            file.NeedStudioVersion( "04.02.03.02" );
            file.NeedFirmware2Version( "04.02.03.01" );
            file.UnsupportFirmware1Version();

            // Gesamte Transponderkonfigurationen in den Erweitertbereich ablegen.

            // Eintragen der gesamten Konfigurationen in den Extended-Bereich.
            extBegin.offsetTransponderConfiguration = extBegin.offsetFreeScope;

            // Feste Werte sicherstellen und eintragen.
            m_transponderISOTagsHeader.structSize = sizeof( m_transponderISOTagsHeader );
            m_transponderISOTagsHeader.structFlags = 0x00; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
            m_transponderISOTagsHeader.transponderType = (m_config.transponderType & 0x3F);
            // Headerstruktur eintragen
            *(StructTransponderISOTagsHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration] = m_transponderISOTagsHeader;
            fileXML.writeContent( FileXML::ContentType::TransponderISOTagsHeader, as_const( m_transponderISOTagsHeader ) );

            // Basiszeiger und Zeiger auf OffsetArray
            unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + sizeof( StructTransponderISOTagsHeader )];
            unsigned short *offsArray = (unsigned short *)base;
            // Zeiger auf die erste Transponderkonfiguration.
            StructTransponderISOTags *transponderStruct = (StructTransponderISOTags *)(base + m_transponderISOTagsHeader.count * sizeof( unsigned short ));

            int i;
            for ( i = 0; i < m_transponderISOTagsHeader.count; i++ )
            {

// Offset und Transponderstruktur eintragen.
                *offsArray = (unsigned short)((unsigned char *)transponderStruct - base);
                *transponderStruct = m_transponderISOTags[i];
                fileXML.writeContent( FileXML::ContentType::TransponderISOTags, as_const( m_transponderISOTags[i] ) );

                // Beide Zeiger um eins weiter.
                offsArray++;
                transponderStruct++;
            }

            // Positionsoffset des naechsten freien Speichers berechnen.
            extBegin.offsetFreeScope += sizeof( StructTransponderISOTagsHeader ) + m_transponderISOTagsHeader.count * sizeof( unsigned short ) + m_transponderISOTagsHeader.count * sizeof( StructTransponderISOTags );
            break;
        }

        case TRANSPONDER_CLOCK_DATA:
        case TRANSPONDER_WIEGAND:
        {
            // Benoetigte Mindestversion
//             file.NeedStudioVersion("04.02.05.01");
//             file.NeedFirmware2Version("04.02.05.01");
//             file.UnsupportFirmware1Version();

            // Gesamte Transponderkonfigurationen in den Erweitertbereich ablegen.

            // Eintragen der gesamten Konfigurationen in den Extended-Bereich.
            extBegin.offsetTransponderConfiguration = extBegin.offsetFreeScope;

            // Feste Werte sicherstellen und eintragen.
            m_transponderWiegandClockDataHeader.structSize = sizeof( m_transponderWiegandClockDataHeader );
            m_transponderWiegandClockDataHeader.structFlags = 0x00; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
            m_transponderWiegandClockDataHeader.transponderType = (m_config.transponderType & 0x3F);
            // Headerstruktur eintragen
            *(StructTransponderWiegandClockDataHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration] = m_transponderWiegandClockDataHeader;
            fileXML.writeContent( FileXML::ContentType::TransponderWiegandClockDataHeader, as_const( m_transponderWiegandClockDataHeader ) );

            // Basiszeiger und Zeiger auf OffsetArray
            unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + sizeof( StructTransponderWiegandClockDataHeader )];
            unsigned short *offsArray = (unsigned short *)base;
            // Zeiger auf die erste Transponderkonfiguration.
            StructTransponderWiegandClockData *transponderStruct = (StructTransponderWiegandClockData *)(base + m_transponderWiegandClockDataHeader.count * sizeof( unsigned short ));

            int i;
            for ( i = 0; i < m_transponderWiegandClockDataHeader.count; i++ )
            {

// Offset und Transponderstruktur eintragen.
                *offsArray = (unsigned short)((unsigned char *)transponderStruct - base);
                *transponderStruct = m_transponderWiegandClockData[i];
                fileXML.writeContent( FileXML::ContentType::TransponderWiegandClockData, as_const( m_transponderWiegandClockData[i] ) );

                // Beide Zeiger um eins weiter.
                offsArray++;
                transponderStruct++;
            }

            // Positionsoffset des naechsten freien Speichers berechnen.
            extBegin.offsetFreeScope += sizeof( StructTransponderWiegandClockDataHeader ) + m_transponderWiegandClockDataHeader.count * sizeof( unsigned short ) + m_transponderWiegandClockDataHeader.count * sizeof( StructTransponderWiegandClockData );
            break;
        }


        case TRANSPONDER_BLUEBOX_ID_TRONIC:
        {
            // Benoetigte Mindestversion
            file.NeedStudioVersion( "04.02.05.01" );
            file.NeedFirmware2Version( "04.02.05.01" );
            file.UnsupportFirmware1Version();

            // Gesamte Transponderkonfigurationen in den Erweitertbereich ablegen.

            // Eintragen der gesamten Konfigurationen in den Extended-Bereich.
            extBegin.offsetTransponderConfiguration = extBegin.offsetFreeScope;

            // Feste Werte sicherstellen und eintragen.
            m_transponderBlueboxIDTronicHeader.structSize = sizeof( m_transponderBlueboxIDTronicHeader );
            m_transponderBlueboxIDTronicHeader.structFlags = m_passwordProtection ? 0x01 : 0x00; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
            m_transponderBlueboxIDTronicHeader.transponderType = (m_config.transponderType & 0x3F);
            // Headerstruktur eintragen
            *(StructTransponderETSITagsHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration] = m_transponderBlueboxIDTronicHeader;
            fileXML.writeContent( FileXML::ContentType::TransponderETSITagsHeader, as_const( m_transponderBlueboxIDTronicHeader ) );

            // Basiszeiger und Zeiger auf OffsetArray
            unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + sizeof( StructTransponderETSITagsHeader )];
            unsigned short *offsArray = (unsigned short *)base;
            // Zeiger auf die erste Transponderkonfiguration.
            StructTransponderETSITags *transponderStruct = (StructTransponderETSITags *)(base + m_transponderBlueboxIDTronicHeader.count * sizeof( unsigned short ));

            int i;
            for ( i = 0; i < m_transponderBlueboxIDTronicHeader.count; i++ )
            {

// Offset und Transponderstruktur eintragen.
                *offsArray = (unsigned short)((unsigned char *)transponderStruct - base);
                *transponderStruct = m_transponderBlueboxIDTronic[i];

#ifdef DATAFOX_STUDIO_IV
                // Verschlüsseln des Passworts
                if ( m_passwordProtection )
                {
                    df4_crypt_internal df4;
                    df4.crypt( reinterpret_cast<unsigned char *>(&transponderStruct->password), sizeof( transponderStruct->password ) );
                }
#endif

                fileXML.writeContent( FileXML::ContentType::TransponderETSITags, as_const( *transponderStruct ) );

                // Beide Zeiger um eins weiter.
                offsArray++;
                transponderStruct++;
            }

            // Positionsoffset des naechsten freien Speichers berechnen.
            extBegin.offsetFreeScope += sizeof( StructTransponderETSITagsHeader ) + m_transponderBlueboxIDTronicHeader.count * sizeof( unsigned short ) + m_transponderBlueboxIDTronicHeader.count * sizeof( StructTransponderETSITags );
            break;
        }

        case TRANSPONDER_Q5:
        {
            // Benoetigte Mindestversion
            file.NeedStudioVersion( "04.02.05.01" );
            file.NeedFirmware2Version( "04.02.05.01" );
            file.UnsupportFirmware1Version();

            // Gesamte Transponderkonfigurationen in den Erweitertbereich ablegen.

            // Eintragen der gesamten Konfigurationen in den Extended-Bereich.
            extBegin.offsetTransponderConfiguration = extBegin.offsetFreeScope;

            // Feste Werte sicherstellen und eintragen.
            m_transponderQ5Header.structSize = sizeof( m_transponderQ5Header );
            m_transponderQ5Header.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
            m_transponderQ5Header.transponderType = (m_config.transponderType & 0x3F);
            // Headerstruktur eintragen
            *(StructTransponderQ5Header *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration] = m_transponderQ5Header;
            fileXML.writeContent( FileXML::ContentType::TransponderQ5Header, as_const( m_transponderQ5Header ) );

            // Basiszeiger und Zeiger auf OffsetArray
            unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + sizeof( StructTransponderQ5Header )];
            unsigned short *offsArray = reinterpret_cast<unsigned short *>(base);
            // Zeiger auf die erste Transponderkonfiguration.
            StructTransponderQ5 *transponderStruct = (StructTransponderQ5 *)(base + m_transponderQ5Header.count * sizeof( unsigned short ));

            int i;
            for ( i = 0; i < m_transponderQ5Header.count; i++ )
            {

// Offset und Transponderstruktur eintragen.
                *offsArray = (unsigned short)((unsigned char *)transponderStruct - base);
                *transponderStruct = m_transponderQ5[i];
                fileXML.writeContent( FileXML::ContentType::TransponderQ5, as_const( m_transponderQ5[i] ) );

                // Beide Zeiger um eins weiter.
                offsArray++;
                transponderStruct++;
            }

            // Positionsoffset des naechsten freien Speichers berechnen.
            extBegin.offsetFreeScope += sizeof( StructTransponderQ5Header ) + m_transponderQ5Header.count * sizeof( unsigned short ) + m_transponderQ5Header.count * sizeof( StructTransponderQ5 );

            break;
        }
        case TRANSPONDER_UNIQUE_FIT:
        {
            // Benoetigte Mindestversion
            file.NeedStudioVersion( "04.03.03.02" );
            file.NeedFirmware3Version( "04.03.03.04" );
            file.UnsupportFirmware2Version();
            file.UnsupportFirmware1Version();
            break;
        }

    case TRANSPONDER_HITAG1_FEIG:
    {
        // Benoetigte Mindestversion
        file.NeedStudioVersion( "04.03.15.00" );
        file.NeedFirmware3Version( "04.03.14.02" );
        file.NeedFirmware2Version( "04.02.05.61" );
        file.UnsupportFirmware1Version();

        // Gesamte Transponderkonfigurationen in den Erweitertbereich ablegen.

        // Eintragen der gesamten Konfigurationen in den Extended-Bereich.
        extBegin.offsetTransponderConfiguration = extBegin.offsetFreeScope;

        // Feste Werte sicherstellen und eintragen.
        m_transponderHitag1FeigHeader.structSize = sizeof( m_transponderHitag1FeigHeader );
        m_transponderHitag1FeigHeader.structFlags = m_passwordProtection ? 0x01 : 0x00; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
        m_transponderHitag1FeigHeader.transponderType = (m_config.transponderType & 0x3F);
        // Headerstruktur eintragen
        *(StructTransponderHitag1FeigHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration] = m_transponderHitag1FeigHeader;
        fileXML.writeContent( FileXML::ContentType::TransponderHitag1FeigHeader, as_const( m_transponderHitag1FeigHeader ) );

        // Basiszeiger und Zeiger auf OffsetArray
        unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + sizeof( StructTransponderHitag1FeigHeader )];
        unsigned short *offsArray = (unsigned short *)base;
        // Zeiger auf die erste Transponderkonfiguration.
        StructTransponderHitag1Feig *transponderStruct = (StructTransponderHitag1Feig *)(base + m_transponderHitag1FeigHeader.count * sizeof( unsigned short ));

        for ( uint8_t i = 0; i < m_transponderHitag1FeigHeader.count; i++ )
        {
            // Offset und Transponderstruktur eintragen.
            *offsArray = (unsigned short)((unsigned char *)transponderStruct - base);
            *transponderStruct = m_transponderHitag1Feig[i];
            fileXML.writeContent( FileXML::ContentType::TransponderHitag1Feig, as_const( m_transponderHitag1Feig[i] ) );

            // Beide Zeiger um eins weiter.
            offsArray++;
            transponderStruct++;
        }

        // Positionsoffset des naechsten freien Speichers berechnen.
        extBegin.offsetFreeScope += sizeof( StructTransponderHitag1FeigHeader ) + m_transponderHitag1FeigHeader.count * sizeof( unsigned short ) + m_transponderHitag1FeigHeader.count * sizeof( StructTransponderHitag1Feig );
        break;
    }

    case TRANSPONDER_PRECONFIGURED_READER:
    {
        // Benoetigte Mindestversion
        file.NeedStudioVersion( "04.03.18.00" );
        file.NeedFirmware3Version( "04.03.18.00" );
        file.UnsupportFirmware2Version();
        file.UnsupportFirmware1Version();

        // Gesamte Transponderkonfigurationen in den Erweitertbereich ablegen.

        // Eintragen der gesamten Konfigurationen in den Extended-Bereich.
        extBegin.offsetTransponderConfiguration = extBegin.offsetFreeScope;

        // Feste Werte sicherstellen und eintragen.
        m_transponderPreconfiguredReaderHeader.structSize = sizeof( m_transponderPreconfiguredReaderHeader );
        m_transponderPreconfiguredReaderHeader.structFlags = m_passwordProtection ? 0x01 : 0x00; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
        m_transponderPreconfiguredReaderHeader.transponderType = (m_config.transponderType & 0x3F);
        // Headerstruktur eintragen
        *(StructTransponderPreconfiguredReaderHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration] = m_transponderPreconfiguredReaderHeader;
        fileXML.writeContent( FileXML::ContentType::TransponderPreconfiguredReaderHeader, as_const( m_transponderPreconfiguredReaderHeader ) );

        // Basiszeiger und Zeiger auf OffsetArray
        unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + sizeof( m_transponderPreconfiguredReaderHeader )];
        unsigned short *offsArray = (unsigned short *)base;
        // Zeiger auf die erste Transponderkonfiguration.
        StructTransponderPreconfiguredReader *transponderStruct = (StructTransponderPreconfiguredReader *)(base + m_transponderPreconfiguredReaderHeader.count * sizeof( unsigned short ));

        for ( uint8_t i = 0; i < m_transponderPreconfiguredReaderHeader.count; i++ )
        {
            // Offset und Transponderstruktur eintragen.
            *offsArray = (unsigned short)((unsigned char *)transponderStruct - base);
            *transponderStruct = m_transponderPreconfiguredReader[i];
            fileXML.writeContent( FileXML::ContentType::TransponderPreconfiguredReader, as_const( m_transponderPreconfiguredReader[i] ) );

            // Beide Zeiger um eins weiter.
            offsArray++;
            transponderStruct++;
        }

        // Positionsoffset des naechsten freien Speichers berechnen.
        extBegin.offsetFreeScope += sizeof( m_transponderPreconfiguredReaderHeader ) + m_transponderPreconfiguredReaderHeader.count * sizeof( unsigned short ) + m_transponderPreconfiguredReaderHeader.count * sizeof( m_transponderPreconfiguredReader );
        break;
    }

    } // switch (Transpondertyp)

    // Reset parent node because the end of the structure 'imageScope' is reached.
    fileXML.setParentNode( oldParent );

    // Pruefung auf benoetigte Mindestversion
    if ( m_extended.dataOnCard )
    {
        // Data on Card wird erst ab Versionen x unterstuetzt.
        file.NeedStudioVersion( "04.02.03.02" );
        file.NeedFirmware2Version( "04.02.03.01" );
        file.UnsupportFirmware1Version();
    }

    // Pruefung auf benoetigte Mindestversion
    if ( m_extended.showSmallWatch )
    {
        // Zusaetzliche Anzeige einer Kleinen Uhr mit Sekunden wird erst ab Versionen x unterstuetzt.
        file.NeedStudioVersion( "04.02.04.03" );
        file.ShouldFirmware2Version( "04.02.04.10" );
    }

    // Struktur fuer Erweiterte Funktionalitaeten
    extBegin.offsetExtendedStruct = extBegin.offsetFreeScope;

#ifdef DATAFOX_STUDIO_IV
    df4_key_box_internal key_box;
    // Werte für den Passwortschutz übernehmen.
    if ( m_passwordProtection )
    {
        // Benoetigte Mindestversion
        file.NeedStudioVersion( "04.03.11.00" );
        file.NeedFirmware3Version( "04.03.11.00" );
        file.UnsupportFirmware2Version();
        file.UnsupportFirmware1Version();

        key_box.set_flag( df4_key_box_internal::flag1, true );
        key_box.set_key( m_password );
    }
    else
    {
        key_box.set_flag( df4_key_box_internal::flag1, false );
    }
    // Struktur mit den Informationen über den Passwortschutz.
    m_extended.protectionPassword = key_box.to_crypted_key_box( false );
#endif

    if ( m_extended.entranceOnline == 3 )
    {
        // Online-Offline-ZK mit Vorprüfung.
        // Benoetigte Mindestversion
        file.NeedStudioVersion( "04.03.12.00" );
        file.NeedLibraryVersion( "04.03.12.00" );
        file.NeedFirmware3Version( "04.03.12.00" );
        file.UnsupportFirmware2Version();
        file.UnsupportFirmware1Version();
    }

    if ( m_extended.timerTimeBaseDivisor > 1 )
    {
        file.NeedStudioVersion( "04.03.15.00" );
        file.NeedLibraryVersion( "04.03.15.00" );
        file.NeedFirmware3Version( "04.03.14.04" );
        file.UnsupportFirmware2Version();
        file.UnsupportFirmware1Version();
    }

    // Diese Struktur enthaelt informationen die beim Erstellen des Baumimage von anderen Eintraegen
    // eingetragen werden, somit muss sie als letzte Struktur abgelegt werden.
    StructExtended &extended = *reinterpret_cast<StructExtended *>(&file.m_flashExtended[extBegin.offsetExtendedStruct]);
    extended = m_extended;

#ifdef DATAFOX_STUDIO_IV
    // Verschlüsseln der "älteren" Transponderpasswörter
    if ( m_passwordProtection )
    {
        df4_crypt_internal df4;
        auto transponder_type = m_config.transponderType & 0x3F;
        if ( transponder_type == TRANSPONDER_HITAG1_SECURITY
             || transponder_type == TRANSPONDER_HITAG2
             || transponder_type == TRANSPONDER_TITAN )
        {
            df4.crypt( reinterpret_cast<unsigned char *>(extended.transponderPassword32), sizeof( extended.transponderPassword32 ) );
        }
        else if ( transponder_type == TRANSPONDER_MIFARE )
        {
            auto size = sizeof( unsigned char ) + sizeof( extended.transponderPasswordA0 ) * 2
                + sizeof( extended.transponderPasswordB0 ) * 3;
            // In der Erweiterten Struktur liegt transponderPasswordA0 separat und die anderen, mit einem Flagbyte vorangestellt, in einem Block vor.
            df4.crypt( reinterpret_cast<unsigned char *>(extended.transponderPasswordA0), sizeof( extended.transponderPasswordA0 ) );
            df4.crypt( reinterpret_cast<unsigned char *>(&extended.transponderPasswordFlag), size );
        }
    }
#endif
    fileXML.writeContent( FileXML::ContentType::Extended, extended );

    extBegin.offsetFreeScope += sizeof( StructExtended );

    // Schreiben der Erweiterten Struktur
    *reinterpret_cast<StructExtendedFlash *>(file.m_flashExtended) = extBegin;
    fileXML.writeContent( FileXML::ContentType::ExtendedFlash, as_const( extBegin ) );

    // write all relevant file names to xml
    fileXML.writeFileNames( file.m_listFileName, file.m_listEntrance2FileName );

    return pFlash;
}

uint32_t EntryDevice::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    unsigned int ListStart;
    unsigned int pDat;
    unsigned int pList;

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    // read file list from xml
    if ( !useAESFile )
    {
        fileXML.combineValid( fileXML.readFileNames( file.m_listFileName, file.m_listEntrance2FileName ) );
    }

    //#################################
    // Config-Stuktur
    if ( useAESFile )
    {
        m_config = file.m_config;
    }
    else
    {
        fileXML.combineValid( fileXML.readContent( FileXML::ContentType::Config, m_config ) );
    }

    // Transponderstruktur nach gewaehltem Typ lesen
    if ( (m_config.transponderType & 0x3F) == TRANSPONDER_MIFARE )
        memcpy( &m_transponderMifare, &m_config.transponderGlobal, sizeof( StructTransponderMifare ) );
    else
        m_transponder = m_config.transponderGlobal; // Transponderstruktur lesen

    // Sicherstellen das wenn keine MDE aktiv ist, dass dann auch die Flags deaktiviert sind.
    if ( m_config.operatingMode != 3 )
        m_config.lampFlagsMDE = 0x01;

    // Timeout fuer Tastenumschaltung
    if ( m_config.timeoutPZE < 5 || m_config.timeoutPZE > 60 )
        m_config.timeoutPZE = 10;

    //#################################
    // Extended-Flash-Struktur

    // Erweiterten Struktur lesen
    StructExtendedFlash extBegin{};
    if ( useAESFile )
    {
        extBegin = *reinterpret_cast<StructExtendedFlash *>(file.m_flashExtended);

        if ( extBegin.offsetGlobalVariableValues )
            m_valueGlobalVar = *(StructValueGlobalVar *)&file.m_flashExtended[extBegin.offsetGlobalVariableValues];
        else
            memset( (unsigned char *)&m_valueGlobalVar, 0, sizeof( StructValueGlobalVar ) );
    }
    else
    {
        fileXML.combineValid( fileXML.readContent( FileXML::ContentType::ExtendedFlash, extBegin ) );

        // Finding the node for this structure and update parent.
        auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::ExtendedScope ) );

        fileXML.combineValid( fileXML.readContent( FileXML::ContentType::GlobalVarValue, m_valueGlobalVar ) );

        // Reset parent node because the end of the structure 'imageScope' is reached.
        fileXML.setParentNode( oldParent );
    }

    // Lesen der erweiterten Parameter - Struktur
    // Diese Struktur enthaelt informationen auf die beim Baumaufbau von anderen Eintraegen zurueckgegriffen wird,
    // somit muss sie vor dem Aufbau geladen sein.
    if ( extBegin.offsetExtendedStruct || !useAESFile )
    {
        if ( useAESFile )
        {
            m_extended = *(StructExtended *)&file.m_flashExtended[extBegin.offsetExtendedStruct];
        }
        else
        {
            fileXML.combineValid( fileXML.readContent( FileXML::ContentType::Extended, m_extended ) );
        }

        // Bei Konfigurationsdateien, die mit einer früheren Version als 04.03.15.00 erstellt wurden, ist der Zeitteiler noch nicht gesetzt.
        if ( m_extended.timerTimeBaseDivisor == 0 )
        {
            m_extended.timerTimeBaseDivisor = 1;
        }

#ifdef DATAFOX_STUDIO_IV
        // Werte für den Passwortschutz übernehmen.
        m_passwordProtection = (m_extended.protectionPassword.flags & 0x01) == 0x01;
        m_password = m_extended.protectionPassword.key;

        if ( !std::all_of( reinterpret_cast<const unsigned char *>(&m_extended.protectionPassword)
                           , reinterpret_cast<const unsigned char *>(&m_extended.protectionPassword) + sizeof( StructKeyBox ), []( int i )
        {
            return i == 0;
        } ) )
        {
            df4_key_box_internal key_box;

            if ( key_box.from_crypted_key_box( m_extended.protectionPassword ) )
            {
                m_passwordProtection = key_box.is_flag( df4_key_box_internal::flag1 );
                m_hidePassword = m_passwordProtection;
                m_password = key_box.get_key();
            }
        }

        // Entschlüsseln der Passwörter
        if ( m_passwordProtection )
        {
            df4_crypt_internal df4;
            df4.crypt( reinterpret_cast<unsigned char *>(&m_config.transponderUserPassword), sizeof( m_config.transponderUserPassword ) );
            df4.init();
            df4.crypt( reinterpret_cast<unsigned char *>(&m_config.passwordSystemMenu), sizeof( m_config.passwordSystemMenu ) );
        }

        // Entschlüsseln der "älteren" Transponderpasswörter
        if ( m_passwordProtection )
        {
            df4_crypt_internal df4;
            auto transponder_type = m_config.transponderType & 0x3F;
            if ( transponder_type == TRANSPONDER_HITAG1_SECURITY
                 || transponder_type == TRANSPONDER_HITAG2
                 || transponder_type == TRANSPONDER_TITAN )
            {
                df4.crypt( reinterpret_cast<unsigned char *>(m_extended.transponderPassword32), sizeof( m_extended.transponderPassword32 ) );
            }
            else if ( transponder_type == TRANSPONDER_MIFARE )
            {
                auto size = sizeof( unsigned char ) + sizeof( m_extended.transponderPasswordA0 ) * 2
                    + sizeof( m_extended.transponderPasswordB0 ) * 3;
                // In der Erweiterten Struktur liegt transponderPasswordA0 separat und die anderen, mit einem Flagbyte vorangestellt, in einem Block vor.
                df4.crypt( reinterpret_cast<unsigned char *>(m_extended.transponderPasswordA0), sizeof( m_extended.transponderPasswordA0 ) );
                df4.crypt( reinterpret_cast<unsigned char *>(&m_extended.transponderPasswordFlag), size );
            }
        }
#endif
    }

    // In der Version 04.02.02 kam eine Struktur fuer Transponder Legic Advant 
    // in der Extended-Struktur zum Einsatz die ab 04.02.03 entfaellt.
    // Der dafuer und auch fuer ein paar passwörter reservierte Speicher wird
    // hier uebergangsweise neu initalisiert.
    SoftwarebuildVersion studioVersion;
    if ( useAESFile )
    {
        studioVersion = file.m_file_header.studioVersion;
    }
    else
    {
        studioVersion = fileXML.studioVersion();
    }
    if ( studioVersion >= "04.02.01.00" && studioVersion < "04.02.03.00" )
    {
        // Daten der alten Struktur und 6 passwoerter a 21Byte loeschen.
        memset( &m_extended.dataOnCard, 0, sizeof( StructTransponderLegicAdvantOld ) + 6 * 21 );
    }

    // Bereichsprüfung für GPS-Struktur
    if ( m_extended.gps.startSpeed < 5 || m_extended.gps.startSpeed > 200 )
        m_extended.gps.startSpeed = 5;

    if ( m_extended.gps.startMove < 50 || m_extended.gps.startMove > 60000 )
        m_extended.gps.startMove = 50;

    if ( m_extended.gps.stopTimeout < 3 || m_extended.gps.stopTimeout > 60000 )
        m_extended.gps.stopTimeout = 3;

    if ( m_extended.gps.stopSpeed < 5 || m_extended.gps.stopSpeed > 200 )
        m_extended.gps.stopSpeed = 5;

    if ( m_extended.gps.cycle > 60000 )
        m_extended.gps.cycle = 0;

    if ( m_extended.gps.course > 3 )
        m_extended.gps.course = 0;

    if ( m_extended.gps.move != 0 && (m_extended.gps.move < 50 || m_extended.gps.move > 60000) )
        m_extended.gps.move = 0;

    //############################
    // 2010-02-01: Passwort für Hitag2 wird auf "Standard" initialisiert, wenn noch nicht gesetzt.
    if ( m_extended.transponderPassword32[0] == 0x00 )
        strcpy( m_extended.transponderPassword32, "4D494B52" );

    // Finding the node for this structure and update parent.
    auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::ExtendedScope ) );
    fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::TransponderConfiguration ) );

    switch ( m_config.transponderType & 0x3F )
    {
        case TRANSPONDER_LEGIC_ADVANT:
        {
            // Struktur fuer Transponder Legic Advant in Extended-Struktur ablegen.
            // Ggf. eine alte Transponderkonfigurationen aus dem Erweitertbereich uebernehmen.
            // In der Version 04.02.02 kam eine Sonderstruktur zum Einsatz, diese wird ab
            // der Version 04.02.03 verworfen und durch ein flexibleres System ersetzt.
            SoftwarebuildVersion studioVersion;
            if ( useAESFile )
            {
                studioVersion = file.m_file_header.studioVersion;
            }
            else
            {
                studioVersion = fileXML.studioVersion();
            }

            if ( studioVersion >= "04.02.02.00" && studioVersion < "04.02.03.00" )
            {
                if ( useAESFile )
                {
                    size_t i{}; // Pruefen ob Daten vorliegen, ggf. bereits uebernommen wurden.
                    for ( i = 0; i < sizeof( StructTransponderLegicAdvantOld ); i++ )
                    {
                        if ( ((unsigned char *)&m_extended.dataOnCard)[i] != 0 )
                        {
                            break;
                        }
                    }

                    if ( i < sizeof( StructTransponderLegicAdvantOld ) )
                    {
                        StructTransponderLegicAdvantOld *legicAdvant = (StructTransponderLegicAdvantOld *)&m_extended.dataOnCard;
                        // Uebernahme der aelteren Strukturdaten, ab dieser Version wird nur noch mit der aktuellen Struktur gearbeitet.
                        m_transponderLegicAdvantHeader.mode = (legicAdvant->typeFlags & 0xC0) >> 6;
                        m_transponderLegicAdvantHeader.flags = (legicAdvant->typeFlags & 0x04) >> 2;
                        m_transponderLegicAdvantHeader.flags |= (legicAdvant->typeFlags & 0x02);
                        m_transponderLegicAdvantHeader.flags |= (legicAdvant->typeFlags & 0x08) >> 1;
                        m_transponderLegicAdvantHeader.flags |= (legicAdvant->flags & 0x01) << 3;

                        m_transponderLegicAdvant[0].rwType = legicAdvant->valType;
                        m_transponderLegicAdvant[0].rwFlags = (legicAdvant->typeFlags & 0x01);
                        m_transponderLegicAdvant[0].rwSegment = legicAdvant->segment;
                        m_transponderLegicAdvant[0].rwOffset = legicAdvant->pos;
                        m_transponderLegicAdvant[0].rwLength = legicAdvant->length;
                        memcpy( m_transponderLegicAdvant[0].kghSearch, legicAdvant->kghSearch, sizeof( legicAdvant->kghSearch ) );
                        m_transponderLegicAdvant[0].kghSearchLength = legicAdvant->kghSearchLength;
                        m_transponderLegicAdvant[0].crcCheck = (legicAdvant->typeFlags & 0x30) >> 4;
                        m_transponderLegicAdvant[0].crcAddress = legicAdvant->crcAddress;
                        m_transponderLegicAdvant[0].outputFormat = legicAdvant->format;
                        m_transponderLegicAdvant[0].outputLength = legicAdvant->formatLength;

                        // Daten der alten Struktur loeschen.
                        memset( &m_extended.dataOnCard, 0, sizeof( StructTransponderLegicAdvantOld ) );
                    }
                }
                else
                {
                    // The data structure 'StructTransponderLegicAdvantOld' is not supported.
                    fileXML.combineValid( false );
                }
            }

            // Liegt die Konfiguration bereits in dem Extended-Bereich vor?
            if ( extBegin.offsetTransponderConfiguration != 0 || !useAESFile )
            {

                // Ablageformat der Daten
                // m_transponderLegicAdvantHeader->structSize + m_transponderLegicAdvantHeader->count * sizeof(unsigned short) +  m_transponderLegicAdvantHeader->count * sizeof(StructTransponderLegicAdvant);

                if ( useAESFile )
                {
                    m_transponderLegicAdvantHeader = *(StructTransponderLegicAdvantHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration];

                    // Basiszeiger auf OffsetArray
                    unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + m_transponderLegicAdvantHeader.structSize];
                    // OffsetArray fuer Abruf der einzelnen Transponderstrukturen
                    unsigned short *offsArray = (unsigned short *)base;
                    StructTransponderLegicAdvant *transponderStruct;

                    int i;
                    for ( i = 0; i < m_transponderLegicAdvantHeader.count; i++ )
                    {

                        transponderStruct = (StructTransponderLegicAdvant *)(base + *offsArray);
                        m_transponderLegicAdvant[i] = *transponderStruct;
                        offsArray++;
                    }
                }
                else
                {
                    auto &header = m_transponderLegicAdvantHeader;
                    auto &content = m_transponderLegicAdvant;
                    FileXML::ContentType headerType = FileXML::ContentType::TransponderLegicAdvantHeader;
                    FileXML::ContentType type = FileXML::ContentType::TransponderLegicAdvant;

                    fileXML.combineValid( fileXML.readContent( headerType, header ) );

                    if ( fileXML.isValid() )
                    {
                        uint32_t size = fileXML.numberOfChildItems( type );

                        // Check the number of transponder configurations defined in the XML. If the number does not match the definition of the
                        // data structure 'm_transponderLegicAdvant', there is an error in the XML description.
                        fileXML.combineValid( size == header.count );
                        if ( size == header.count )
                        {
                            tinyxml2::XMLElement *lastXmlElement{};
                            for ( uint32_t i = 0; i < size; ++i )
                            {
                                fileXML.combineValid( fileXML.readContent( type, content[i], i, lastXmlElement ) );
                                if ( !fileXML.isValid() )
                                {
                                    break;
                                }
                            }
                        }
                    }
                }
            }
            break;
        }

        case TRANSPONDER_MIFARE_DESFIRE:
        {

            // Gesamte Transponderkonfigurationen aus dem Erweitertbereich lesen.

            // Liegt die Konfiguration bereits in dem Extended-Bereich vor?
            if ( extBegin.offsetTransponderConfiguration != 0 || !useAESFile )
            {

                // Ablageformat der Daten
                // m_transponderMifareDesfireHeader->structSize + m_transponderMifareDesfireHeader->count * sizeof(unsigned short) +  m_transponderMifareDesfireHeader->count * sizeof(StructTransponderMifareDesfire);
                if ( useAESFile )
                {
                    m_transponderMifareDesfireHeader = *(StructTransponderMifareDesfireHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration];

                    // Basiszeiger auf OffsetArray
                    unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + m_transponderMifareDesfireHeader.structSize];
                    // OffsetArray fuer Abruf der einzelnen Transponderstrukturen
                    unsigned short *offsArray = (unsigned short *)base;
                    StructTransponderMifareDesfire *transponderStruct;

                    int i;
                    for ( i = 0; i < m_transponderMifareDesfireHeader.count; i++ )
                    {

                        transponderStruct = (StructTransponderMifareDesfire *)(base + *offsArray);
                        m_transponderMifareDesfire[i] = *transponderStruct;
                        offsArray++;
                    }
                }
                else
                {
                    auto &header = m_transponderMifareDesfireHeader;
                    auto &content = m_transponderMifareDesfire;
                    FileXML::ContentType headerType = FileXML::ContentType::TransponderMifareDesfireHeader;
                    FileXML::ContentType type = FileXML::ContentType::TransponderMifareDesfire;

                    fileXML.combineValid( fileXML.readContent( headerType, header ) );

                    if ( fileXML.isValid() )
                    {
                        uint32_t size = fileXML.numberOfChildItems( type );

                        // Check the number of transponder configurations defined in the XML. If the number does not match the definition of the
                        // data structure 'm_transponderLegicAdvant', there is an error in the XML description.
                        fileXML.combineValid( size == header.count );
                        if ( size == header.count )
                        {
                            tinyxml2::XMLElement *lastXmlElement{};
                            for ( uint32_t i = 0; i < size; ++i )
                            {
                                fileXML.combineValid( fileXML.readContent( type, content[i], i, lastXmlElement ) );
                                if ( !fileXML.isValid() )
                                {
                                    break;
                                }
                            }
                        }
                    }
                }
            }
            break;
        }

        case TRANSPONDER_MIFARE_PLUS:
        {

            // Gesamte Transponderkonfigurationen aus dem Erweitertbereich lesen.

            // Liegt die Konfiguration bereits in dem Extended-Bereich vor?
            if ( extBegin.offsetTransponderConfiguration != 0 || !useAESFile )
            {

                // Ablageformat der Daten
                // m_transponderMifarePlusHeader->structSize + m_transponderMifarePlusHeader->count * sizeof(unsigned short) +  m_transponderMifarePlusHeader->count * sizeof(StructTransponderMifarePlus);
                if ( useAESFile )
                {
                    m_transponderMifarePlusHeader = *(StructTransponderMifarePlusHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration];

                    // Basiszeiger auf OffsetArray
                    unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + m_transponderMifarePlusHeader.structSize];
                    // OffsetArray fuer Abruf der einzelnen Transponderstrukturen
                    unsigned short *offsArray = (unsigned short *)base;
                    StructTransponderMifarePlus *transponderStruct;

                    int i;
                    for ( i = 0; i < m_transponderMifarePlusHeader.count; i++ )
                    {

                        transponderStruct = (StructTransponderMifarePlus *)(base + *offsArray);
                        m_transponderMifarePlus[i] = *transponderStruct;
                        offsArray++;
                    }
                }
                else
                {
                    auto &header = m_transponderMifarePlusHeader;
                    auto &content = m_transponderMifarePlus;
                    FileXML::ContentType headerType = FileXML::ContentType::TransponderMifarePlusHeader;
                    FileXML::ContentType type = FileXML::ContentType::TransponderMifarePlus;

                    fileXML.combineValid( fileXML.readContent( headerType, header ) );

                    if ( fileXML.isValid() )
                    {
                        uint32_t size = fileXML.numberOfChildItems( type );

                        // Check the number of transponder configurations defined in the XML. If the number does not match the definition of the
                        // data structure 'm_transponderLegicAdvant', there is an error in the XML description.
                        fileXML.combineValid( size == header.count );
                        if ( size == header.count )
                        {
                            tinyxml2::XMLElement *lastXmlElement{};
                            for ( uint32_t i = 0; i < size; ++i )
                            {
                                fileXML.combineValid( fileXML.readContent( type, content[i], i, lastXmlElement ) );
                                if ( !fileXML.isValid() )
                                {
                                    break;
                                }
                            }
                        }
                    }
                }
            }
            break;
        }

        case TRANSPONDER_MIFARE_ULTRALIGHT:
        {

            // Gesamte Transponderkonfigurationen aus dem Erweitertbereich lesen.

            // Liegt die Konfiguration bereits in dem Extended-Bereich vor?
            if ( extBegin.offsetTransponderConfiguration != 0 || !useAESFile )
            {

                // Ablageformat der Daten
                // m_transponderMifareUltralightHeader->structSize + m_transponderMifareUltralightHeader->count * sizeof(unsigned short) +  m_transponderMifareUltralightHeader->count * sizeof(StructTransponderMifareUltralight);

                if ( useAESFile )
                {
                    m_transponderMifareUltralightHeader = *(StructTransponderMifareUltralightHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration];

                    // Basiszeiger auf OffsetArray
                    unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + m_transponderMifareUltralightHeader.structSize];
                    // OffsetArray fuer Abruf der einzelnen Transponderstrukturen
                    unsigned short *offsArray = (unsigned short *)base;
                    StructTransponderMifareUltralight *transponderStruct;

                    int i;
                    for ( i = 0; i < m_transponderMifareUltralightHeader.count; i++ )
                    {

                        transponderStruct = (StructTransponderMifareUltralight *)(base + *offsArray);
                        m_transponderMifareUltralight[i] = *transponderStruct;
                        offsArray++;
                    }
                }
                else
                {
                    auto &header = m_transponderMifareUltralightHeader;
                    auto &content = m_transponderMifareUltralight;
                    FileXML::ContentType headerType = FileXML::ContentType::TransponderMifareUltralightHeader;
                    FileXML::ContentType type = FileXML::ContentType::TransponderMifareUltralight;

                    fileXML.combineValid( fileXML.readContent( headerType, header ) );

                    if ( fileXML.isValid() )
                    {
                        uint32_t size = fileXML.numberOfChildItems( type );

                        // Check the number of transponder configurations defined in the XML. If the number does not match the definition of the
                        // data structure 'm_transponderLegicAdvant', there is an error in the XML description.
                        fileXML.combineValid( size == header.count );
                        if ( size == header.count )
                        {
                            tinyxml2::XMLElement *lastXmlElement{};
                            for ( uint32_t i = 0; i < size; ++i )
                            {
                                fileXML.combineValid( fileXML.readContent( type, content[i], i, lastXmlElement ) );
                                if ( !fileXML.isValid() )
                                {
                                    break;
                                }
                            }
                        }
                    }
                }
            }
            break;
        }

        case TRANSPONDER_ICODE:
        {

            // Gesamte Transponderkonfigurationen aus dem Erweitertbereich lesen.

            // Liegt die Konfiguration bereits in dem Extended-Bereich vor?
            if ( extBegin.offsetTransponderConfiguration != 0 || !useAESFile )
            {

                // Ablageformat der Daten
                // m_transponderiCodeHeader->structSize + m_transponderiCodeHeader->count * sizeof(unsigned short) +  m_transponderiCodeHeader->count * sizeof(StructTransponderiCode);
                if ( useAESFile )
                {
                    m_transponderiCodeHeader = *(StructTransponderiCodeHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration];

                    // Basiszeiger auf OffsetArray
                    unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + m_transponderiCodeHeader.structSize];
                    // OffsetArray fuer Abruf der einzelnen Transponderstrukturen
                    unsigned short *offsArray = (unsigned short *)base;
                    StructTransponderiCode *transponderStruct;

                    int i;
                    for ( i = 0; i < m_transponderiCodeHeader.count; i++ )
                    {

                        transponderStruct = (StructTransponderiCode *)(base + *offsArray);
                        m_transponderiCode[i] = *transponderStruct;
                        offsArray++;
                    }
                }
                else
                {
                    auto &header = m_transponderiCodeHeader;
                    auto &content = m_transponderiCode;
                    FileXML::ContentType headerType = FileXML::ContentType::TransponderiCodeHeader;
                    FileXML::ContentType type = FileXML::ContentType::TransponderiCode;

                    fileXML.combineValid( fileXML.readContent( headerType, header ) );

                    if ( fileXML.isValid() )
                    {
                        uint32_t size = fileXML.numberOfChildItems( type );

                        // Check the number of transponder configurations defined in the XML. If the number does not match the definition of the
                        // data structure 'm_transponderLegicAdvant', there is an error in the XML description.
                        fileXML.combineValid( size == header.count );
                        if ( size == header.count )
                        {
                            tinyxml2::XMLElement *lastXmlElement{};
                            for ( uint32_t i = 0; i < size; ++i )
                            {
                                fileXML.combineValid( fileXML.readContent( type, content[i], i, lastXmlElement ) );
                                if ( !fileXML.isValid() )
                                {
                                    break;
                                }
                            }
                        }
                    }
                }
            }
            break;
        }

        case TRANSPONDER_JAVACARD: // ToDo HK Neuer Transponderleser
        {

            // Gesamte Transponderkonfigurationen aus dem Erweitertbereich lesen.

            // Liegt die Konfiguration bereits in dem Extended-Bereich vor?
            if ( extBegin.offsetTransponderConfiguration != 0 || !useAESFile )
            {

                // Ablageformat der Daten
                // m_transponderJavaCardHeader->structSize + m_transponderJavaCardHeader->count * sizeof(unsigned short) +  m_transponderJavaCardHeader->count * sizeof(StructTransponderJavaCard);
                if ( useAESFile )
                {
                    m_transponderJavaCardHeader = *(StructTransponderJavaCardHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration];

                    // Basiszeiger auf OffsetArray
                    unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + m_transponderJavaCardHeader.structSize];
                    // OffsetArray fuer Abruf der einzelnen Transponderstrukturen
                    unsigned short *offsArray = (unsigned short *)base;
                    StructTransponderJavaCard *transponderStruct;

                    int i;
                    for ( i = 0; i < m_transponderJavaCardHeader.count; i++ )
                    {

                        transponderStruct = (StructTransponderJavaCard *)(base + *offsArray);
                        m_transponderJavaCard[i] = *transponderStruct;
                        offsArray++;
                    }
                }
                else
                {
                    auto &header = m_transponderJavaCardHeader;
                    auto &content = m_transponderJavaCard;
                    FileXML::ContentType headerType = FileXML::ContentType::TransponderJavaCardHeader;
                    FileXML::ContentType type = FileXML::ContentType::TransponderJavaCard;

                    fileXML.combineValid( fileXML.readContent( headerType, header ) );

                    if ( fileXML.isValid() )
                    {
                        uint32_t size = fileXML.numberOfChildItems( type );

                        // Check the number of transponder configurations defined in the XML. If the number does not match the definition of the
                        // data structure 'm_transponderLegicAdvant', there is an error in the XML description.
                        fileXML.combineValid( size == header.count );
                        if ( size == header.count )
                        {
                            tinyxml2::XMLElement *lastXmlElement{};
                            for ( uint32_t i = 0; i < size; ++i )
                            {
                                fileXML.combineValid( fileXML.readContent( type, content[i], i, lastXmlElement ) );
                                if ( !fileXML.isValid() )
                                {
                                    break;
                                }
                            }
                        }
                    }
                }
            }
            break;
        }


        case TRANSPONDER_MYD:
        {

            // Gesamte Transponderkonfigurationen aus dem Erweitertbereich lesen.

            // Liegt die Konfiguration bereits in dem Extended-Bereich vor?
            if ( extBegin.offsetTransponderConfiguration != 0 || !useAESFile )
            {

                // Ablageformat der Daten
                // m_transponderMyDHeader->structSize + m_transponderMyDHeader->count * sizeof(unsigned short) +  m_transponderMyDHeader->count * sizeof(StructTransponderMyD);
                if ( useAESFile )
                {
                    m_transponderMyDHeader = *(StructTransponderMyDHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration];

                    // Basiszeiger auf OffsetArray
                    unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + m_transponderMyDHeader.structSize];
                    // OffsetArray fuer Abruf der einzelnen Transponderstrukturen
                    unsigned short *offsArray = (unsigned short *)base;
                    StructTransponderMyD *transponderStruct;

                    int i;
                    for ( i = 0; i < m_transponderMyDHeader.count; i++ )
                    {

                        transponderStruct = (StructTransponderMyD *)(base + *offsArray);
                        m_transponderMyD[i] = *transponderStruct;
                        offsArray++;
                    }
                }
                else
                {
                    auto &header = m_transponderMyDHeader;
                    auto &content = m_transponderMyD;
                    FileXML::ContentType headerType = FileXML::ContentType::TransponderMyDHeader;
                    FileXML::ContentType type = FileXML::ContentType::TransponderMyD;

                    fileXML.combineValid( fileXML.readContent( headerType, header ) );

                    if ( fileXML.isValid() )
                    {
                        uint32_t size = fileXML.numberOfChildItems( type );

                        // Check the number of transponder configurations defined in the XML. If the number does not match the definition of the
                        // data structure 'm_transponderLegicAdvant', there is an error in the XML description.
                        fileXML.combineValid( size == header.count );
                        if ( size == header.count )
                        {
                            tinyxml2::XMLElement *lastXmlElement{};
                            for ( uint32_t i = 0; i < size; ++i )
                            {
                                fileXML.combineValid( fileXML.readContent( type, content[i], i, lastXmlElement ) );
                                if ( !fileXML.isValid() )
                                {
                                    break;
                                }
                            }
                        }
                    }
                }
            }
            break;
        }

        case TRANSPONDER_ISO_TAGS:
        {

            // Gesamte Transponderkonfigurationen aus dem Erweitertbereich lesen.

            // Liegt die Konfiguration bereits in dem Extended-Bereich vor?
            if ( extBegin.offsetTransponderConfiguration != 0 || !useAESFile )
            {

                // Ablageformat der Daten
                // m_transponderISOTagsHeader->structSize + m_transponderISOTagsHeader->count * sizeof(unsigned short) +  m_transponderISOTagsHeader->count * sizeof(StructTransponderISOTags);
                if ( useAESFile )
                {
                    m_transponderISOTagsHeader = *(StructTransponderISOTagsHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration];

                    // Basiszeiger auf OffsetArray
                    unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + m_transponderISOTagsHeader.structSize];
                    // OffsetArray fuer Abruf der einzelnen Transponderstrukturen
                    unsigned short *offsArray = (unsigned short *)base;
                    StructTransponderISOTags *transponderStruct;

                    int i;
                    for ( i = 0; i < m_transponderISOTagsHeader.count; i++ )
                    {

                        transponderStruct = (StructTransponderISOTags *)(base + *offsArray);
                        m_transponderISOTags[i] = *transponderStruct;
                        offsArray++;
                    }
                }
                else
                {
                    auto &header = m_transponderISOTagsHeader;
                    auto &content = m_transponderISOTags;
                    FileXML::ContentType headerType = FileXML::ContentType::TransponderISOTagsHeader;
                    FileXML::ContentType type = FileXML::ContentType::TransponderISOTags;

                    fileXML.combineValid( fileXML.readContent( headerType, header ) );

                    if ( fileXML.isValid() )
                    {
                        uint32_t size = fileXML.numberOfChildItems( type );

                        // Check the number of transponder configurations defined in the XML. If the number does not match the definition of the
                        // data structure 'm_transponderLegicAdvant', there is an error in the XML description.
                        fileXML.combineValid( size == header.count );
                        if ( size == header.count )
                        {
                            tinyxml2::XMLElement *lastXmlElement{};
                            for ( uint32_t i = 0; i < size; ++i )
                            {
                                fileXML.combineValid( fileXML.readContent( type, content[i], i, lastXmlElement ) );
                                if ( !fileXML.isValid() )
                                {
                                    break;
                                }
                            }
                        }
                    }
                }
            }
            break;
        }

        case TRANSPONDER_WIEGAND:
        case TRANSPONDER_CLOCK_DATA:
        {

            // Gesamte Transponderkonfigurationen aus dem Erweitertbereich lesen.

            // Liegt die Konfiguration bereits in dem Extended-Bereich vor?
            if ( extBegin.offsetTransponderConfiguration != 0 || !useAESFile )
            {
                if ( useAESFile )
                {
                    // Ablageformat der Daten
                    m_transponderWiegandClockDataHeader = *(StructTransponderWiegandClockDataHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration];

                    // Basiszeiger auf OffsetArray
                    unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + m_transponderWiegandClockDataHeader.structSize];
                    // OffsetArray fuer Abruf der einzelnen Transponderstrukturen
                    unsigned short *offsArray = (unsigned short *)base;
                    StructTransponderWiegandClockData *transponderStruct;

                    int i;
                    for ( i = 0; i < m_transponderWiegandClockDataHeader.count; i++ )
                    {

                        transponderStruct = (StructTransponderWiegandClockData *)(base + *offsArray);
                        m_transponderWiegandClockData[i] = *transponderStruct;
                        offsArray++;
                    }
                }
                else
                {
                    auto &header = m_transponderWiegandClockDataHeader;
                    auto &content = m_transponderWiegandClockData;
                    FileXML::ContentType headerType = FileXML::ContentType::TransponderWiegandClockDataHeader;
                    FileXML::ContentType type = FileXML::ContentType::TransponderWiegandClockData;

                    fileXML.combineValid( fileXML.readContent( headerType, header ) );

                    if ( fileXML.isValid() )
                    {
                        uint32_t size = fileXML.numberOfChildItems( type );

                        // Check the number of transponder configurations defined in the XML. If the number does not match the definition of the
                        // data structure 'm_transponderLegicAdvant', there is an error in the XML description.
                        fileXML.combineValid( size == header.count );
                        if ( size == header.count )
                        {
                            tinyxml2::XMLElement *lastXmlElement{};
                            for ( uint32_t i = 0; i < size; ++i )
                            {
                                fileXML.combineValid( fileXML.readContent( type, content[i], i, lastXmlElement ) );
                                if ( !fileXML.isValid() )
                                {
                                    break;
                                }
                            }
                        }
                    }
                }
            }
            break;
        }



        case TRANSPONDER_BLUEBOX_ID_TRONIC:
        {

            // Gesamte Transponderkonfigurationen aus dem Erweitertbereich lesen.

            // Liegt die Konfiguration bereits in dem Extended-Bereich vor?
            if ( extBegin.offsetTransponderConfiguration != 0 || !useAESFile )
            {

                // Ablageformat der Daten
                // m_transponderBlueboxIDTronicHeader->structSize + m_transponderBlueboxIDTronicHeader->count * sizeof(unsigned short) +  m_transponderBlueboxIDTronicHeader->count * sizeof(StructTransponderETSITags);
                if ( useAESFile )
                {
                    m_transponderBlueboxIDTronicHeader = *(StructTransponderETSITagsHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration];

                    // Basiszeiger auf OffsetArray
                    unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + m_transponderBlueboxIDTronicHeader.structSize];
                    // OffsetArray fuer Abruf der einzelnen Transponderstrukturen
                    unsigned short *offsArray = (unsigned short *)base;
                    StructTransponderETSITags *transponderStruct;

                    int i;
                    for ( i = 0; i < m_transponderBlueboxIDTronicHeader.count; i++ )
                    {

                        transponderStruct = (StructTransponderETSITags *)(base + *offsArray);
                        m_transponderBlueboxIDTronic[i] = *transponderStruct;

#ifdef DATAFOX_STUDIO_IV
                        // Entschlüsseln des Passworts
                        if ( m_passwordProtection )
                        {
                            df4_crypt_internal df4;
                            df4.crypt( reinterpret_cast<unsigned char *>(&m_transponderBlueboxIDTronic[i].password), sizeof( m_transponderBlueboxIDTronic[i].password ) );
                        }
#endif
                        offsArray++;
                    }
                }
                else
                {
                    auto &header = m_transponderBlueboxIDTronicHeader;
                    auto &content = m_transponderBlueboxIDTronic;
                    FileXML::ContentType headerType = FileXML::ContentType::TransponderETSITagsHeader;
                    FileXML::ContentType type = FileXML::ContentType::TransponderETSITags;

                    fileXML.combineValid( fileXML.readContent( headerType, header ) );

                    if ( fileXML.isValid() )
                    {
                        uint32_t size = fileXML.numberOfChildItems( type );

                        // Check the number of transponder configurations defined in the XML. If the number does not match the definition of the
                        // data structure 'm_transponderLegicAdvant', there is an error in the XML description.
                        fileXML.combineValid( size == header.count );
                        if ( size == header.count )
                        {
                            tinyxml2::XMLElement *lastXmlElement{};
                            for ( uint32_t i = 0; i < size; ++i )
                            {
                                fileXML.combineValid( fileXML.readContent( type, content[i], i, lastXmlElement ) );
                                if ( !fileXML.isValid() )
                                {
                                    break;
                                }

#ifdef DATAFOX_STUDIO_IV
                                // Entschlüsseln des Passworts
                                if ( m_passwordProtection )
                                {
                                    df4_crypt_internal df4;
                                    df4.crypt( reinterpret_cast<unsigned char *>(&content[i].password), sizeof( content[i].password ) );
                                }
#endif
                            }
                        }
                    }
                }
            }
            break;
        }
        case TRANSPONDER_Q5:
        {

            // Gesamte Transponderkonfigurationen aus dem Erweitertbereich lesen.

            // Liegt die Konfiguration bereits in dem Extended-Bereich vor?
            if ( extBegin.offsetTransponderConfiguration != 0 || !useAESFile )
            {

                // Ablageformat der Daten
                // m_transponderQ5Header->structSize + m_transponderQ5Header->count * sizeof(unsigned short) +  m_transponderQ5Header->count * sizeof(StructTransponderQ5);
                if ( useAESFile )
                {
                    m_transponderQ5Header = *(StructTransponderQ5Header *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration];

                    // Basiszeiger auf OffsetArray
                    unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + m_transponderQ5Header.structSize];
                    // OffsetArray fuer Abruf der einzelnen Transponderstrukturen
                    unsigned short *offsArray = (unsigned short *)base;
                    StructTransponderQ5 *transponderStruct;

                    int i;
                    for ( i = 0; i < m_transponderQ5Header.count; i++ )
                    {

                        transponderStruct = (StructTransponderQ5 *)(base + *offsArray);
                        m_transponderQ5[i] = *transponderStruct;
                        offsArray++;
                    }
                }
                else
                {
                    auto &header = m_transponderQ5Header;
                    auto &content = m_transponderQ5;
                    FileXML::ContentType headerType = FileXML::ContentType::TransponderQ5Header;
                    FileXML::ContentType type = FileXML::ContentType::TransponderQ5;

                    fileXML.combineValid( fileXML.readContent( headerType, header ) );

                    if ( fileXML.isValid() )
                    {
                        uint32_t size = fileXML.numberOfChildItems( type );

                        // Check the number of transponder configurations defined in the XML. If the number does not match the definition of the
                        // data structure 'm_transponderLegicAdvant', there is an error in the XML description.
                        fileXML.combineValid( size == header.count );
                        if ( size == header.count )
                        {
                            tinyxml2::XMLElement *lastXmlElement{};
                            for ( uint32_t i = 0; i < size; ++i )
                            {
                                fileXML.combineValid( fileXML.readContent( type, content[i], i, lastXmlElement ) );
                                if ( !fileXML.isValid() )
                                {
                                    break;
                                }
                            }
                        }
                    }
                }
            }
            break;
        }

        case TRANSPONDER_HITAG1_FEIG:
        {

            // Gesamte Transponderkonfigurationen aus dem Erweitertbereich lesen.

            // Liegt die Konfiguration bereits in dem Extended-Bereich vor?
            if ( extBegin.offsetTransponderConfiguration != 0 || !useAESFile )
            {

                // Ablageformat der Daten
                // m_transponderHitag1FeigHeader->structSize + m_transponderHitag1FeigHeader->count * sizeof(unsigned short) +  m_transponderHitag1FeigHeader->count * sizeof(StructTransponderHitag1Feig);
                if ( useAESFile )
                {
                    m_transponderHitag1FeigHeader = *(StructTransponderHitag1FeigHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration];

                    // Basiszeiger auf OffsetArray
                    unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + m_transponderHitag1FeigHeader.structSize];
                    // OffsetArray fuer Abruf der einzelnen Transponderstrukturen
                    unsigned short *offsArray = (unsigned short *)base;
                    StructTransponderHitag1Feig *transponderStruct;

                    for ( uint8_t i = 0; i < m_transponderHitag1FeigHeader.count; i++ )
                    {

                        transponderStruct = (StructTransponderHitag1Feig *)(base + *offsArray);
                        m_transponderHitag1Feig[i] = *transponderStruct;
                        offsArray++;
                    }
                }
                else
                {
                    auto &header = m_transponderHitag1FeigHeader;
                    auto &content = m_transponderHitag1Feig;
                    FileXML::ContentType headerType = FileXML::ContentType::TransponderHitag1FeigHeader;
                    FileXML::ContentType type = FileXML::ContentType::TransponderHitag1Feig;

                    fileXML.combineValid( fileXML.readContent( headerType, header ) );

                    if ( fileXML.isValid() )
                    {
                        uint32_t size = fileXML.numberOfChildItems( type );

                        // Check the number of transponder configurations defined in the XML. If the number does not match the definition of the
                        // data structure 'm_transponderLegicAdvant', there is an error in the XML description.
                        fileXML.combineValid( size == header.count );
                        if ( size == header.count )
                        {
                            tinyxml2::XMLElement *lastXmlElement{};
                            for ( uint32_t i = 0; i < size; ++i )
                            {
                                fileXML.combineValid( fileXML.readContent( type, content[i], i, lastXmlElement ) );
                                if ( !fileXML.isValid() )
                                {
                                    break;
                                }
                            }
                        }
                    }
                }
            }
            break;
        }

        case TRANSPONDER_PRECONFIGURED_READER:
        {
            // Gesamte Transponderkonfigurationen aus dem Erweitertbereich lesen.

            // Liegt die Konfiguration bereits in dem Extended-Bereich vor?
            if ( extBegin.offsetTransponderConfiguration != 0 || !useAESFile )
            {
                // Ablageformat der Daten
                // m_transponderPreconfiguredReaderHeader->structSize + m_transponderPreconfiguredReaderHeader->count * sizeof( unsigned short ) +  m_transponderPreconfiguredReaderHeader->count * sizeof( StructTransponderPreconfiguredReader );
                if ( useAESFile )
                {
                    m_transponderPreconfiguredReaderHeader = *(StructTransponderPreconfiguredReaderHeader *)&file.m_flashExtended[extBegin.offsetTransponderConfiguration];

                    // Basiszeiger auf OffsetArray
                    unsigned char *base = &file.m_flashExtended[extBegin.offsetTransponderConfiguration + m_transponderPreconfiguredReaderHeader.structSize];
                    // OffsetArray fuer Abruf der einzelnen Transponderstrukturen
                    unsigned short *offsArray = (unsigned short *)base;
                    StructTransponderPreconfiguredReader *transponderStruct;

                    for ( uint8_t i = 0; i < m_transponderPreconfiguredReaderHeader.count; i++ )
                    {
                        transponderStruct = (StructTransponderPreconfiguredReader *)(base + *offsArray);
                        m_transponderPreconfiguredReader[i] = *transponderStruct;
                        offsArray++;
                    }
                }
                else
                {
                    auto &header = m_transponderPreconfiguredReaderHeader;
                    auto &content = m_transponderPreconfiguredReader;
                    FileXML::ContentType headerType = FileXML::ContentType::TransponderPreconfiguredReaderHeader;
                    FileXML::ContentType type = FileXML::ContentType::TransponderPreconfiguredReader;

                    fileXML.combineValid( fileXML.readContent( headerType, header ) );

                    if ( fileXML.isValid() )
                    {
                        uint32_t size = fileXML.numberOfChildItems( type );

                        // Check the number of transponder configurations defined in the XML. If the number does not match the definition of the
                        // data structure 'm_transponderLegicAdvant', there is an error in the XML description.
                        fileXML.combineValid( size == header.count );
                        if ( size == header.count )
                        {
                            tinyxml2::XMLElement *lastXmlElement{};
                            for ( uint32_t i = 0; i < size; ++i )
                            {
                                fileXML.combineValid( fileXML.readContent( type, content[i], i, lastXmlElement ) );
                                if ( !fileXML.isValid() )
                                {
                                    break;
                                }
                            }
                        }
                    }
                }
            }
            break;
        }
    } // switch (Transpondertyp)

    // Reset parent node because the end of the structure 'imageScope' is reached.
    fileXML.setParentNode( oldParent );

    //#################################
    // Flash-Header-Stuktur

    // Speicheraufteilung ermitteln
    ListStart = endianReverse( file.m_flash_header.list[0] );
    if ( ListStart != FLASH_LIST_START )
    {
        // Fehlerhafter Wert
        // Wird durch writeDocument neu gesetzt.
    }

    // Datensatzbeschreibungen lesen
    pDat = file.m_flash_header.dataStruct;
    pDat = endianReverse( pDat ) - FLASH_KONFIG_START;

    // Listenbeschreibungen lesen
    pList = file.m_flash_header.listStruct;
    pList = endianReverse( pList ) - FLASH_KONFIG_START;

    //#################################
    // Flash-Struktur

    // Flashoffset für die folgende Serialisierung auf Begin setzen.
    file.m_flashMemOffset = 0;

    // Finding the node for this structure and update parent.
    oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::ImageScope ) );

    if ( useAESFile )
    {
        m_globalVar = *reinterpret_cast<StructGlobalVar *>(&file.m_flash[file.m_flashMemOffset]);
        m_globalVar.ptr = endianReverse( m_globalVar.ptr ) - FLASH_KONFIG_START;

        // Flashoffset für die weitere Serialisierung anpassen.
        file.m_flashMemOffset = m_globalVar.ptr;

        // Bedienung lesen
        file.m_flashMemOffset = rootHandling().readDocument( file.m_flashMemOffset, doc );
    }
    else
    {
        fileXML.combineValid( fileXML.readContent( FileXML::ContentType::GlobalVar, m_globalVar ) );
        rootHandling().readDocument( 0, doc );
    }

    // Datensatzbeschreibungen
    rootTableRecord().readDocument( pDat, doc );

    // Listenbeschreibungen
    pFlash = rootTableList().readDocument( pList, doc );

    // Reset parent node because the end of the structure 'imageScope' is reached.
    fileXML.setParentNode( oldParent );


    // Finding the node for this structure and update parent.
    oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::ExtendedScope ) );

    // Signalverarbeitung - Konfiguration
    rootSignalProcessing().readDocument( 0, doc );

    // Ereignisketten - Konfiguration
    rootEventChains().readDocument( 0, doc );

    // IOModule - Konfiguration
    rootIOModule().readDocument( 0, doc );

    // Daten fuer Entrance2 - aus file.m_flashExtended und ConfigMem einlesen.
    rootAccessControl().readDocument( 0, doc );

    // Daten für die Energieverwaltung einlesen.
    rootEnergieSchemes().readDocument( 0, doc );
    // Geladene Energieschemen auf Gueltigkeit pruefen
    rootEnergieSchemes().verifyEnergieSchemes();

    // Reset parent node because the end of the structure 'imageScope' is reached.
    fileXML.setParentNode( oldParent );


    // Alle Referenzen aus Indizes erzeugen
    convertIndexToReference();

    return pFlash;
}

///////////////////////////////////////
// TabCtrl
int EntryDevice::m_pageIndex = PAGE_NO;

void EntryDevice::setPageIndex( int index )
{
    m_pageIndex = index;
}

int EntryDevice::pageIndex()
{
    return m_pageIndex;
}

//////////////////////////////////////
// Referenzierung zwischen Einträgen

void EntryDevice::convertReferenceToIndex()
{
    EntryEventChains &eventChains = rootEventChains();

    // Fuer Energie-Betriebsarten
    m_extended.energieOption.invokeRelease = (m_refInputChainEnergieOption[0] != nullptr) ? eventChains.indexOf( m_refInputChainEnergieOption[0] ) + 1 : 0;
    m_extended.energieOption.invokeGPS = (m_refInputChainEnergieOption[1] != nullptr) ? eventChains.indexOf( m_refInputChainEnergieOption[1] ) + 1 : 0;

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertReferenceToIndex();
}

void EntryDevice::convertIndexToReference()
{
    EntryEventChains &eventChains = rootEventChains();

    // Fuer Energie-Betriebsarten
    m_refInputChainEnergieOption[0] = (m_extended.energieOption.invokeRelease > 0) ? (EntryInputChain *)eventChains.at( m_extended.energieOption.invokeRelease - 1 ) : nullptr;
    if ( m_refInputChainEnergieOption[0] )
    {
        m_refInputChainEnergieOption[0]->incRefCount();
    }

    m_refInputChainEnergieOption[1] = (m_extended.energieOption.invokeGPS > 0) ? (EntryInputChain *)eventChains.at( m_extended.energieOption.invokeGPS - 1 ) : nullptr;
    if ( m_refInputChainEnergieOption[1] )
    {
        m_refInputChainEnergieOption[1]->incRefCount();
    }

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertIndexToReference();
}

///////////////////////////////////////
// Daten des Eintrags
void EntryDevice::setDefault()
{
    setDefault( SetupType::PZE_MasterIV );
}

void EntryDevice::setDefault( SetupType setupType )
{
    // Setuptyp für welche Geräte dieses Setup gilt.
    setSetupType( setupType );

    m_passwordProtection = false;
    m_hidePassword = false;

    // Initialisieren der Membervariablen
    m_config.autoIncrementFlags = 0;
    m_config.autoIncrementValue = 1;
    m_config.initialValue1 = 0;
    m_config.initialValue2 = 0;
    m_config.initialValue3 = 0;

    // Config-Struktur
    m_config.backlightLCD = FALSE;

    m_config.onlineOperation = FALSE;
    m_config.waitingTimeForReply = 3000;
    m_config.waitingTimeForReply = endianReverse( m_config.waitingTimeForReply );

    m_config.operatingMode = 0;

    m_config.passwordSystemMenu = 0;
    m_config.passwordSystemMenu = endianReverse( m_config.passwordSystemMenu );
    m_config.transponderUserPassword = 0;
    m_config.transponderUserPassword = endianReverse( m_config.transponderUserPassword );

    m_config.deactivateOnOffButton = FALSE;
    m_config.showInfoPage = TRUE;

    m_config.transponderModeUnique = TRUE;
    m_config.presentationModeUnique = 0;

    m_config.ioModuleActive = FALSE;

    m_config.transponderType = 0;
    m_config.transponderSegment = 0;

    m_config.enableEntrance = FALSE;
    m_config.enableEntrance2 = FALSE;

    m_config.showBigWatchAlways = FALSE;
    m_config.keySound = FALSE;
    m_config.signalLastFunctionKey = FALSE;
    m_config.peepFlags = 0;

    // Zuschneidefunktion fuer Legic/Mifare Leser
    m_config.transponderCutFlags = 0;
    m_config.transponderCutFrom = 1;
    m_config.transponderCutLength = 1;

    memset( (unsigned char *)&m_valueGlobalVar, 0, sizeof( m_valueGlobalVar ) );

    // Transponder
    memset( &m_transponder, 0, sizeof( StructTransponder ) );
    m_transponder.tpSegment1 = 255; // Kein
    m_transponder.tpSegment2 = 255;
    m_transponder.tpSegment3 = 255;


    // Timeout fuer Tastenumschaltung
    m_config.timeoutPZE = 10;
    // Flags fuer rote Lampensteuerung
    m_config.lampFlagsMDE = 0x01;

    // Erweiterte Parameterstruktur initialisieren
    memset( &m_extended, 0, sizeof( m_extended ) );
    // Defaultvalues
    if ( setupType == SetupType::IPC_Extension )
        m_extended.doubleInputTime = 0;
    else
        m_extended.doubleInputTime = 60;

    m_extended.confirmationTime = 3;

    // Sommer-/Winterzeit Umschaltung
    m_extended.daylightActive = 1;

    // Im Maerz von Winter- auf Sommerzeit
    m_extended.daylight1Month = 3;
    m_extended.daylight1Week = 5;
    m_extended.daylight1Day = 7;
    m_extended.daylight1Hour = 2;
    // Im Oktober von Sommer- auf Winterzeit
    m_extended.daylight2Month = 10;
    m_extended.daylight2Week = 5;
    m_extended.daylight2Day = 7;
    m_extended.daylight2Hour = 3;

    // Zeitbasis der Timer (1000ms / 1 -> 1s)
    m_extended.timerTimeBaseDivisor = 1;

    // MIFARE Transponderleser
    memset( &m_transponderMifare, 0, sizeof( StructTransponderMifare ) );
    m_transponderMifare.sectorBlock = 0; // Kein

    // Flags fuer die Transponderkonfiguration
    m_transponderActiveFlags = 0;

    // Legic Advant Headerstruct
    memset( &m_transponderLegicAdvantHeader, 0, sizeof( StructTransponderLegicAdvantHeader ) );
    m_transponderLegicAdvantHeader.structSize = sizeof( StructTransponderLegicAdvantHeader );
    m_transponderLegicAdvantHeader.count = 1; // Basiskonfiguration
    m_transponderLegicAdvantHeader.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
    m_transponderLegicAdvantHeader.mode = 0; // Nur Advant Ausweise
    m_transponderLegicAdvantHeader.flags = 0x01; // ISO14443 Ausweise zugelassen.
    m_transponderLegicAdvantHeader.flags |= 0x02; // ISO15693 Ausweise zugelassen.
    // Legic Advant Transponderleser; Basiskonfiguration + 4 moegliche Ersatzkonfigurationen
    memset( m_transponderLegicAdvant, 0, sizeof( StructTransponderLegicAdvant ) * 5 );
    // Im Standard UID lesen im Hex Format.
    m_transponderLegicAdvant[0].rwType = 0;
    m_transponderLegicAdvant[0].outputFormat = 1;

    // Mifare Desfire Headerstruct
    memset( &m_transponderMifareDesfireHeader, 0, sizeof( StructTransponderMifareDesfireHeader ) );
    m_transponderMifareDesfireHeader.structSize = sizeof( StructTransponderMifareDesfireHeader );
    m_transponderMifareDesfireHeader.count = 1; // Basiskonfiguration
    m_transponderMifareDesfireHeader.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
    // Mifare Desfire Transponderleser; Basiskonfiguration + 4 moegliche Ersatzkonfigurationen
    memset( m_transponderMifareDesfire, 0, sizeof( StructTransponderMifareDesfire ) * 5 );
    // Im Standard UID lesen im Hex gedreht Format.
    m_transponderMifareDesfire[0].rwFlags = 0x01;
    m_transponderMifareDesfire[0].outputFormat = 3;

    // Mifare Plus Headerstruct
    memset( &m_transponderMifarePlusHeader, 0, sizeof( StructTransponderMifarePlusHeader ) );
    m_transponderMifarePlusHeader.structSize = sizeof( StructTransponderMifarePlusHeader );
    m_transponderMifarePlusHeader.count = 1; // Basiskonfiguration
    m_transponderMifarePlusHeader.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
    m_transponderMifarePlusHeader.securityLevel = 1;
    m_transponderMifarePlusHeader.iso14443A = 0xFF; // Alle ISO Ausweise erlaubt.
    // Mifare Plus Transponderleser; Basiskonfiguration + 4 moegliche Ersatzkonfigurationen
    memset( m_transponderMifarePlus, 0, sizeof( StructTransponderMifarePlus ) * 5 );
    // Standard Seriennummer lesen und Hexadezimal liefern.
    m_transponderMifarePlus[0].rwFlags = 0x01;
    m_transponderMifarePlus[0].outputFormat = 1;

    // Mifare Ultralight Headerstruct
    memset( &m_transponderMifareUltralightHeader, 0, sizeof( StructTransponderMifareUltralightHeader ) );
    m_transponderMifareUltralightHeader.structSize = sizeof( StructTransponderMifareUltralightHeader );
    m_transponderMifareUltralightHeader.count = 1; // Basiskonfiguration
    m_transponderMifareUltralightHeader.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
    // Mifare Ultralight Transponderleser; Basiskonfiguration + 4 moegliche Ersatzkonfigurationen
    memset( m_transponderMifareUltralight, 0, sizeof( StructTransponderMifareUltralight ) * 5 );
    // Standard ist Seriennummer lesen und Hexadezimal liefern.
    m_transponderMifareUltralight[0].rwFlags = 0x01;
    m_transponderMifareUltralight[0].outputFormat = 1;

    // iCode Headerstruct
    memset( &m_transponderiCodeHeader, 0, sizeof( StructTransponderiCodeHeader ) );
    m_transponderiCodeHeader.structSize = sizeof( StructTransponderiCodeHeader );
    m_transponderiCodeHeader.count = 1; // Basiskonfiguration
    m_transponderiCodeHeader.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
    // iCode Transponderleser; Basiskonfiguration + 4 moegliche Ersatzkonfigurationen
    memset( m_transponderiCode, 0, sizeof( StructTransponderiCode ) * 5 );
    // Standard ist Seriennummer lesen und Hexadezimal liefern.
    m_transponderiCode[0].rwFlags = 0x01;
    m_transponderiCode[0].outputFormat = 1;

    // JavaCard Headerstruct
    memset( &m_transponderJavaCardHeader, 0, sizeof( StructTransponderJavaCardHeader ) );
    m_transponderJavaCardHeader.structSize = sizeof( StructTransponderJavaCardHeader );
    m_transponderJavaCardHeader.count = 1; // Basiskonfiguration
    m_transponderJavaCardHeader.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
    // JavaCard Transponderleser; Basiskonfiguration + 6 moegliche Ersatzkonfigurationen
    memset( m_transponderJavaCard, 0, sizeof( StructTransponderJavaCard ) * 7 );
    // Standard ist Seriennummer lesen und Hexadezimal liefern.
    m_transponderJavaCard[0].rwFlags = 0x01;
    m_transponderJavaCard[0].outputFormat = 1;

    // MyD Headerstruct
    memset( &m_transponderMyDHeader, 0, sizeof( StructTransponderMyDHeader ) );
    m_transponderMyDHeader.structSize = sizeof( StructTransponderMyDHeader );
    m_transponderMyDHeader.count = 1; // Basiskonfiguration
    m_transponderMyDHeader.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
    m_transponderMyDHeader.isoFlags = 0x01; // ISO14443 Ausweise erlaubt.
    m_transponderMyDHeader.isoFlags |= 0x02; // ISO15693 Ausweise erlaubt.
    // MyD Transponderleser; Basiskonfiguration + 4 moegliche Ersatzkonfigurationen
    memset( m_transponderMyD, 0, sizeof( StructTransponderMyD ) * 5 );
    // Standard ist Seriennummer lesen und Hexadezimal liefern.
    m_transponderMyD[0].rwFlags = 0x01;
    m_transponderMyD[0].outputFormat = 1;

    // ISOTags Headerstruct
    memset( &m_transponderISOTagsHeader, 0, sizeof( StructTransponderISOTagsHeader ) );
    m_transponderISOTagsHeader.structSize = sizeof( StructTransponderISOTagsHeader );
    m_transponderISOTagsHeader.count = 1; // Basiskonfiguration
    m_transponderISOTagsHeader.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
    // ISOTags Transponderleser; Basiskonfiguration + 4 moegliche Ersatzkonfigurationen
    memset( m_transponderISOTags, 0, sizeof( StructTransponderISOTags ) * 1 );
    // Standard Hexadezimal liefern.
    m_transponderISOTags[0].outputFormat = 1;


    // Wiegand Clock Data 
    memset( &m_transponderWiegandClockDataHeader, 0, sizeof( StructTransponderWiegandClockDataHeader ) );
    m_transponderWiegandClockDataHeader.structSize = sizeof( StructTransponderWiegandClockDataHeader );
    m_transponderWiegandClockDataHeader.count = 1; // Basiskonfiguration
    m_transponderWiegandClockDataHeader.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey 
    memset( m_transponderWiegandClockData, 0, sizeof( StructTransponderWiegandClockData ) * 1 );
    // Defaultsettings nachtragen
    m_transponderWiegandClockData[0].flags = 0; // Steuerleitung, Parität und Firmencode auf aus
    m_transponderWiegandClockData[0].clsWireActiveState = 0; // Low aktiv 
    m_transponderWiegandClockData[0].bitFormat = 1; // nicht individuell Standard 26 Bit
    m_transponderWiegandClockData[0].numberOfBits = 26; // nicht individuell Standard 26 Bit
    m_transponderWiegandClockData[0].parityFlags01 = 0;
    m_transponderWiegandClockData[0].parityFlags02 = 2;
    m_transponderWiegandClockData[0].parityBitPos01 = 1;
    m_transponderWiegandClockData[0].parityBitPos02 = 26;
    m_transponderWiegandClockData[0].parityStart01 = 2;
    m_transponderWiegandClockData[0].parityStart02 = 9;
    m_transponderWiegandClockData[0].parityEnd01 = 10;
    m_transponderWiegandClockData[0].parityEnd02 = 25;
    m_transponderWiegandClockData[0].facilityCodeFormat = 0; // Dezimal
    m_transponderWiegandClockData[0].facilityCodeBitPos = 2; // für 26 Bit Format
    m_transponderWiegandClockData[0].facilityCodeNumberOfBits = 8; // für 26 Bit Format
    m_transponderWiegandClockData[0].cardDataBitPos = 10; // für 26 Bit Format
    m_transponderWiegandClockData[0].cardDataNumberOfBits = 16; // für 26 Bit Format
    m_transponderWiegandClockData[0].outputData = 0; // nur Kartennummer 
    m_transponderWiegandClockData[0].outputFormat = 0; // Stardard Dezimal 
    m_transponderWiegandClockData[0].outputLength = 0; // keine Längenbegrenzung 

    // BlueboxIDTronic Headerstruct
    memset( &m_transponderBlueboxIDTronicHeader, 0, sizeof( StructTransponderETSITagsHeader ) );
    m_transponderBlueboxIDTronicHeader.structSize = sizeof( StructTransponderETSITagsHeader );
    m_transponderBlueboxIDTronicHeader.count = 1; // Basiskonfiguration
    m_transponderBlueboxIDTronicHeader.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
    // BlueboxIDTronic Transponderleser; Basiskonfiguration + 4 moegliche Ersatzkonfigurationen
    memset( m_transponderBlueboxIDTronic, 0, sizeof( StructTransponderETSITags ) * 1 );
    m_transponderBlueboxIDTronic[0].strength = 100; // Feldstaerke 100%
    // Standard TID liefern.
    m_transponderBlueboxIDTronic[0].rwFlags = 0x01;
    // Standard Hexadezimal liefern.
    m_transponderBlueboxIDTronic[0].outputFormat = 1;

    // Startwert bei Mifare
    memset( m_extended.transponderInitialValue, 0, sizeof( m_extended.transponderInitialValue ) );
    // Sector Tailer bei Mifare
    m_extended.transponderSectorTrailer = 0xFF078069;
    // Standardpasswoerter bei Mifare-Leser
    m_extended.transponderPasswordFlag = 0;
    strcpy( m_extended.transponderPasswordA0, "FFFFFFFFFFFF" );
    strcpy( m_extended.transponderPasswordB0, "FFFFFFFFFFFF" );
    strcpy( m_extended.transponderPasswordA1, "FFFFFFFFFFFF" );
    strcpy( m_extended.transponderPasswordB1, "FFFFFFFFFFFF" );
    strcpy( m_extended.transponderPasswordA2, "FFFFFFFFFFFF" );
    strcpy( m_extended.transponderPasswordB2, "FFFFFFFFFFFF" );

    // Q5 Headerstruct
    memset( &m_transponderQ5Header, 0, sizeof( StructTransponderQ5Header ) );
    m_transponderQ5Header.structSize = sizeof( StructTransponderQ5Header );
    m_transponderQ5Header.count = 1; // Basiskonfiguration
    m_transponderQ5Header.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
    m_transponderQ5Header.isoFlags = 0x01; // ISO14443 Ausweise erlaubt.
    m_transponderQ5Header.isoFlags |= 0x02; // ISO15693 Ausweise erlaubt.
    // Q5 Transponderleser; Basiskonfiguration + 4 moegliche Ersatzkonfigurationen
    memset( m_transponderQ5, 0, sizeof( StructTransponderQ5 ) * 5 );
    // Standard ist nicht die Seriennummer lesen, aber Hexadezimal liefern.
    m_transponderQ5[0].rwFlags = 0x00;
    m_transponderQ5[0].outputFormat = 1;

    // Hitag1Feig Headerstruct
    memset( &m_transponderHitag1FeigHeader, 0, sizeof( StructTransponderHitag1FeigHeader ) );
    m_transponderHitag1FeigHeader.structSize = sizeof( StructTransponderHitag1FeigHeader );
    m_transponderHitag1FeigHeader.count = 1; // Basiskonfiguration
    m_transponderHitag1FeigHeader.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
    // Hitag1Feig Transponderleser; Basiskonfiguration + 4 moegliche Ersatzkonfigurationen
    memset( m_transponderHitag1Feig, 0, sizeof( StructTransponderHitag1Feig ) * 1 );
    // Standard Hexadezimal liefern.
    m_transponderHitag1Feig[0].facilityCodeCount = 4;
    m_transponderHitag1Feig[0].facilityCodeStart = 1;
    m_transponderHitag1Feig[0].cardNumberCount = 8;
    m_transponderHitag1Feig[0].cardNumberStart = 5;
    m_transponderHitag1Feig[0].outputFormat = 1;

    // PreconfiguredReader Headerstruct
    memset( &m_transponderPreconfiguredReaderHeader, 0, sizeof( StructTransponderPreconfiguredReaderHeader ) );
    m_transponderHitag1FeigHeader.structSize = sizeof( StructTransponderPreconfiguredReaderHeader );
    m_transponderHitag1FeigHeader.count = 0; // Basiskonfiguration
    m_transponderHitag1FeigHeader.structFlags = 0x01; // Ggf. verwendete Passwoerter mit RC4 und DatafoxKey verschluesseln.
    // PreconfiguredReader Transponderleser; Basiskonfiguration + 4 moegliche Ersatzkonfigurationen
    memset( m_transponderPreconfiguredReader, 0, sizeof( StructTransponderPreconfiguredReader ) * 1 );

    // Betriebsart
    m_extended.energieOption.mode = 0;     // Standardbetrieb
    m_extended.energieOption.flags = 0x02; // GPS-Start/Stopp für An/Aus der Zündung nutzen

    // GPS
    m_extended.gps.startSpeed = 5;
    m_extended.gps.startMove = 50;
    m_extended.gps.stopTimeout = 3;
    m_extended.gps.stopSpeed = 5;

    // Werte für Meldung "Datenspeichern erfolgreich." initialisieren.
    m_extended.dataWriteOKTime = 500;
    m_extended.dataWriteOKFontType = 1;

    // Passwort für Hitag2 wird auf "Standard" initialisiert.
    strcpy( m_extended.transponderPassword32, "4D494B52" );

    // !!! Energieverwaltung !!!
    // Wird im Eintrag EntryEnergieSchemes::SetDefaultEnergieSchemes verwaltet.

    m_iconIndex = ICON_DEVICE_PZE;
    m_iconIndexSelected = ICON_DEVICE_PZE_SEL;
    setName( "Master IV" );

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_DEVICE_0 );
    m_pageIndexes.push_back( PAGE_DEVICE_1 );
    m_pageIndexes.push_back( PAGE_DEVICE_2 );
    m_pageIndexes.push_back( PAGE_DEVICE_3 );
    m_pageIndexes.push_back( PAGE_DEVICE_4 );
    m_pageIndexes.push_back( PAGE_DEVICE_5 );
    m_pageIndexes.push_back( PAGE_DEVICE_6 );
    m_pageIndexes.push_back( PAGE_DEVICE_7 );

    // GV's initialisieren
    memset( &m_globalVar, 0, sizeof( StructGlobalVar ) );
    strcpy( m_globalVar.names[0], "GlobVarStr1" );
    strcpy( m_globalVar.names[1], "GlobVarStr2" );

    // Referenzbezug initialisieren
    m_refInputChainEnergieOption.fill( nullptr );

    EntryTableRecord *pRDRoot = new EntryTableRecord( this );
    pRDRoot->setDefault();
    append( pRDRoot );

    EntryTableList *pLDRoot = new EntryTableList( this );
    pLDRoot->setDefault();
    append( pLDRoot );

    EntryHandling *pControlRoot = new EntryHandling( this );
    pControlRoot->setDefault();
    append( pControlRoot );

    EntrySignalProcessing *pMDERoot = new EntrySignalProcessing( this );
    pMDERoot->setDefault();
    append( pMDERoot );

    EntryEventChains *pInputChains = new EntryEventChains( this );
    pInputChains->setDefault();
    append( pInputChains );

    EntryIOModule *pIOModule = new EntryIOModule( this );
    pIOModule->setDefault();
    append( pIOModule );

    EntryAccessControl2 *pEntrance2Root = new EntryAccessControl2( this );
    pEntrance2Root->setDefault();
    append( pEntrance2Root );

    EntryEnergieSchemes *pEnergieRoot = new EntryEnergieSchemes( this );
    pEnergieRoot->setDefault();
    append( pEnergieRoot );

    setInsertEnabled( FALSE );
    setDeleteEnabled( FALSE );
}

SetupType EntryDevice::setupType() const
{
    return m_setupType;
}

void EntryDevice::setSetupType( SetupType setupType )
{
    m_setupType = setupType;
}

bool EntryDevice::isTypeDevice( enum_device_flag typeDeviceFlags )
{
    bool res = false;

    switch ( m_setupType )
    {
        case SetupType::PZE_MasterIV:
            if ( typeDeviceFlags & PZE_MasterIV )
                res = true;
            break;

        case SetupType::ZK_MasterIV:                          // ZK-MasterIV
            if ( typeDeviceFlags & ZK_MasterIV )
                res = true;
            break;

        case SetupType::BDE_MasterIV:                         // BDE-MasterIV
            if ( typeDeviceFlags & BDE_MasterIV )
                res = true;
            break;

        case SetupType::Mobil_MasterIV:                       // Mobil-MasterIV
            if ( typeDeviceFlags & Mobil_MasterIV )
                res = true;
            break;

        case SetupType::MDE_BoxIV:                         // MDE-BoxIV
            if ( typeDeviceFlags & MDE_BoxIV )
                res = true;
            break;

        case SetupType::AE_MasterIV:                          // AE-MasterIV
            if ( typeDeviceFlags & AE_MasterIV )
                res = true;
            break;

        case SetupType::Timeboy_MasterIV:                     // TimeboyIV
            if ( typeDeviceFlags & Timeboy_MasterIV )
                res = true;
            break;

        case SetupType::Exklusiv_LineIV:                    // Exklusiv-Line IV
            if ( typeDeviceFlags & Exklusiv_LineIV )
                res = true;
            break;

        case SetupType::Flex_MasterIV:                        // Flex-MasterIV
            if ( typeDeviceFlags & Flex_MasterIV )
                res = true;
            break;

        case SetupType::Timeboy_Mobil_PZE:           // Timeboy-Mobil-PZE
            if ( typeDeviceFlags & Timeboy_Mobil_PZE )
                res = true;
            break;

        case SetupType::EVO_2_8_Pure:                // EVO 2.8 Pure
            if ( typeDeviceFlags & EVO_2_8_Pure || typeDeviceFlags & EVO_3_5_Pure )
                res = true;
            break;

        case SetupType::EVO_4_3:                // EVO 4.3
            if ( typeDeviceFlags & EVO_4_3 || typeDeviceFlags & EVO_4_6_FlexKey )
                res = true;
            break;

        case SetupType::IPC_Extension:     // IPC-Extension
            if ( typeDeviceFlags & IPC_Extension_VARIO_5_7
                 || typeDeviceFlags & IPC_Extension_VARIO_10
                 || typeDeviceFlags & IPC_Extension_Q7
                 || typeDeviceFlags & Vario_10
                 || typeDeviceFlags & Vario_12
                 || typeDeviceFlags & Vario_15
                 || typeDeviceFlags & Vario_17
                 || typeDeviceFlags & IPC_EVO_12_1
                 || typeDeviceFlags & IPC_EVO_15
                 || typeDeviceFlags & IPC_EVO_18_5
                 || typeDeviceFlags & IPC_EVO_24 )
                res = true;
            break;


        case SetupType::IO_Box:                      // IO-Box
            if ( typeDeviceFlags & IO_Box )
                res = true;
            break;

        case SetupType::KYO_Inloc:                      // KYO Inloc
            if ( typeDeviceFlags & KYO_Inloc )
                res = true;
            break;

        case SetupType::Mobil_Box:                   // Mobil-Box
            if ( typeDeviceFlags & Mobil_Box )
                res = true;
            break;

        case SetupType::Timeboy_TBDock:              // Dockingstation-V2
            if ( typeDeviceFlags & Timeboy_TBDock )
                res = true;
            break;

        case SetupType::Mobil_TBDock:                // Fahrzeugdatenlogger-V2
            if ( typeDeviceFlags & Mobil_TBDock )
                res = true;
            break;

        case SetupType::KYO_Oneloc:                   // KYO Oneloc
            if ( typeDeviceFlags & KYO_Oneloc )
                res = true;
            break;

        case SetupType::EVO_3_5_Universal:           // EVO 3.5 Universal
            if ( typeDeviceFlags & EVO_3_5_Universal )
                res = true;
            break;

        case SetupType::KYO_Cenloc:                 // KYO Cenloc
            if ( typeDeviceFlags & KYO_Cenloc )
                res = true;
            break;
    }

    return res;
}

TransponderType EntryDevice::transponderType() const
{
    return static_cast<TransponderType>(m_config.transponderType & 0x3F);
}

void EntryDevice::setTransponderType( TransponderType transponderType )
{
    m_config.transponderType &= 0xC0;
    m_config.transponderType |= static_cast<uint8_t>( transponderType );
}

// Gibt zurück ob der gewählte Transponder mehrere Konfigurationen unterstützt.
bool EntryDevice::hasTranponderMultiConfigs( TransponderType transponderType )
{
    // Transponder unterstützt mehrere Konfigurationen.
    bool multiConfigs = false;

    // Verwendeter Transpondertyp auf aktuellen setzen, wenn undefiniert
    if ( transponderType == TransponderType::Undefined )
    {
        transponderType = static_cast<TransponderType>(m_config.transponderType & 0x3F);
    }

    // Je nach gewaehltem Transpondertyp
    switch ( transponderType )
    {
        case TransponderType::None:
            break;

        case TransponderType::Unique:
        case TransponderType::Hitag_2:
        case TransponderType::Legic_prime:
        case TransponderType::Hitag_1:
        case TransponderType::Hitag_S_H32:
        case TransponderType::Mifare_classic:
            break;

        case TransponderType::Q5:
            multiConfigs = true;
            break;

        case TransponderType::Hitag_S_H56:
        case TransponderType::Hitag_S_H48:
        case TransponderType::IButton:
        case TransponderType::Simons_voss:
        case TransponderType::Nedap:
        case TransponderType::Titan:
        case TransponderType::Hitag_1_crypt:
        case TransponderType::Hitag_1_security:
        case TransponderType::HID_Proxkey:
        case TransponderType::HID_IClass:
            break;

        case TransponderType::Legic_advant_prime:
        case TransponderType::Mifare_desfire:
        case TransponderType::Mifare_plus:
        case TransponderType::Mifare_ultra_light:
        case TransponderType::ICode:
        case TransponderType::MYD:
        case TransponderType::ISO_Tags:
        case TransponderType::Bluebox_id_tronic:
            multiConfigs = true;
            break;

         case TransponderType::Unique_FIT:
             break;

        case TransponderType::Java_card:
            multiConfigs = true;
            break;

        case TransponderType::OCL:
        case TransponderType::Wiegand:
        case TransponderType::Clock_data:
        case TransponderType::Hitag_1_Feig:
        case TransponderType::Preconfigured_reader:
            break;

        default:
            // Der Transpondertyp muss eingetragen sein, damit die Anzahl der aktiven Konfigurationen geliefert werden kann.
            BOOST_ASSERT( false );
    }

    return multiConfigs;
}

// Gibt die Anzahl der aktiven Transponderkonfigurationen zurück.
uint8_t EntryDevice::transponderConfigCount()
{
    // Anzahl Konfigurationen von Transponders
    uint8_t configCount = 1;

    // Verwendeter Transpondertyp
    TransponderType transponderType = static_cast<TransponderType>(m_config.transponderType & 0x3F);

    // Je nach gewaehltem Transpondertyp
    switch ( transponderType )
    {
        case TransponderType::None:
            configCount = 0;
            break;

        case TransponderType::Unique:
        case TransponderType::Hitag_2:
        case TransponderType::Legic_prime:
        case TransponderType::Hitag_1:
        case TransponderType::Hitag_S_H32:
        case TransponderType::Mifare_classic:
            break;

        case TransponderType::Q5:
            configCount = m_transponderQ5Header.count;
            break;

        case TransponderType::Hitag_S_H56:
        case TransponderType::Hitag_S_H48:
        case TransponderType::IButton:
        case TransponderType::Simons_voss:
        case TransponderType::Nedap:
        case TransponderType::Titan:
        case TransponderType::Hitag_1_crypt:
        case TransponderType::Hitag_1_security:
        case TransponderType::HID_Proxkey:
        case TransponderType::HID_IClass:
            break;

        case TransponderType::Legic_advant_prime:
            configCount = m_transponderLegicAdvantHeader.count;
            break;

        case TransponderType::Mifare_desfire:
            configCount = m_transponderMifareDesfireHeader.count;
            break;

        case TransponderType::Mifare_plus:
            configCount = m_transponderMifarePlusHeader.count;
            break;

        case TransponderType::Mifare_ultra_light:
            configCount = m_transponderMifareUltralightHeader.count;
            break;

        case TransponderType::ICode:
            configCount = m_transponderiCodeHeader.count;
            break;

        case TransponderType::MYD:
            configCount = m_transponderMyDHeader.count;
            break;

        case TransponderType::ISO_Tags:
            configCount = m_transponderISOTagsHeader.count;
            break;

        case TransponderType::Bluebox_id_tronic:
            configCount = m_transponderBlueboxIDTronicHeader.count;
            break;

        case TransponderType::Unique_FIT:
            break;

        case TransponderType::Java_card:
            configCount = m_transponderJavaCardHeader.count;
            break;

        case TransponderType::OCL:
            break;

        case TransponderType::Wiegand:
        case TransponderType::Clock_data:
            configCount = m_transponderWiegandClockDataHeader.count;
            break;

        case TransponderType::Hitag_1_Feig:
            configCount = m_transponderHitag1FeigHeader.count;
            break;

        case TransponderType::Preconfigured_reader:
            configCount = m_transponderPreconfiguredReaderHeader.count;
            break;

        default:
            // Der Transpondertyp muss eingetragen sein, damit die Anzahl der aktiven Konfigurationen geliefert werden kann.
            BOOST_ASSERT( false );
    }

    return configCount;
}

// ########################################################
// Member: EntryMenu

EntryMenu::EntryMenu( const EntryMenu &entry )
    : EntryBase( entry )
{
    // Tiefes kopieren: Member der Basisklasse kopieren
    //*((EntryBase*)this) = *((EntryBase*)&entry);

    // Member kopieren
    m_operating = entry.m_operating;
    m_menu = entry.m_menu;
    m_refJumpESC = entry.m_refJumpESC;

    EntryBase *pNew;
    // Alle Childs kopieren
    auto iterChild = ((EntryBase *)&entry)->constBegin();
    for ( ; iterChild != ((EntryBase *)&entry)->constEnd(); ++iterChild )
    {
        // neues anfuegen
        pNew = (EntryBase *)((EntryMenuChoice *)*iterChild)->cloneEntry();
        pNew->setParent( this );
        //*((EntryMenuChoice*)pNew) = *((EntryMenuChoice*)*iterChild);
        append( pNew );
    }

    //##################################################
    //Beim Kopieren der Untermenues muss der Referenzzeiger 
    //auf die neuen Adressen angepasst werden!!!

    // Anpassung bei Untermenue
    // - Referenz bei JumpESC anpassen.

    EntryDevice *device = nullptr;
    EntryBase *handling = nullptr;
    EntryBase *main_menu = nullptr;
    EntryBase *entryBase = nullptr;

    // Auf Rooteintrag wechseln.
    entryBase = ((EntryMenu *)&entry)->parent();
    while ( !entryBase->isType( EntryType::Device ) )
    {
        // Eintrag unter Handling?
        if ( entryBase->isType( EntryType::Handling ) )
        {
            handling = entryBase;
            main_menu = handling->first();
        }

        entryBase = entryBase->parent();
    }
    device = (EntryDevice *)entryBase;

    // Zeiger auf Hauptmenu sicherstellen.
    if ( !main_menu && device->rootHandling().size() )
        main_menu = device->rootHandling().first();

    // Sprungziel nach Bestaetigung mit ESC. Uebergeordnetes U-Menue, Naechste E-Kette bzw. Menue und Hauptmenue
    if ( ((EntryMenu *)&entry)->m_refJumpESC == (EntryBase *)0 ||
        ((EntryMenu *)&entry)->m_refJumpESC == (EntryBase *)-1 ||
         ((EntryMenu *)&entry)->m_refJumpESC == (EntryBase *)-2 )
    {
        // Nichts zu tun.
    }
    else
    {

        int index;
        // Versuch die Referenz zu E-Kette oder U-Menue des Uebergeordneten Menueeintrages aufzuloesen.
        index = ((EntryMenu *)&entry)->parent()->indexOf( ((EntryMenu *)&entry)->m_refJumpESC );
        if ( index == -1 )
        {

// Versuch die Referenz zu F-Taste aufzuloesen.
            if ( main_menu )
                index = main_menu->indexOf( ((EntryMenu *)&entry)->m_refJumpESC );

            if ( index == -1 )
            {
// Umstellen auf Naechste E-Kette bzw. Menue
                this->m_refJumpESC = (EntryBase *)-1;

            }
            else
            {
                     // Nichts zu tun. Referenz bleibt die gleiche.
            }
        }
        else
        {
             // Verweist auf das eingene Menue?
            if ( ((EntryInputChain *)&entry) == ((EntryInputChain *)&entry)->m_refJumpESC )
            {
                this->m_refJumpESC = this;
            }
            else
            {
                     // Umstellen auf Naechste E-Kette bzw. Menue
                this->m_refJumpESC = (EntryBase *)-1;
            }
        }
    }
}

///////////////////////////////////////
// Childs

EntryBase *EntryMenu::insertEntry( EntryType typeID, int pos )
{
    // Wenn Parent BEDIEN_ROOT sind maximal 15 Menueeintraege anlegbar.
    if ( parentType() == EntryType::Handling )
    {
        if ( size() >= 15 )
        {
            // Es sind maximal 15 Menueeinträge im Hauptmenue anlegbar.
            return nullptr;
        }
    }

    EntryMenuChoice *entryNew = new EntryMenuChoice( this );

    entryNew->setDefault( boost::str( boost::format( "%1%%2%" ) % QT_TRANSLATE_NOOP_UTF8( "EntryMenu", "Menueeintrag" ) % (size() + 1) ) );

    insertEntryFinish( entryNew, pos );

    return entryNew;
}

EntryBase *EntryMenu::cloneEntry()
{
    return reinterpret_cast<EntryBase *>(new EntryMenu( *this ));
}

void EntryMenu::deleteEntry( int pos )
{
    EntryBase *entry;

    switch ( pos )
    {
        case -1:
            entry = last();
            break;

        default:
            entry = value( pos );
    }

    // Pruefen ob Menüeintrag ohne Bezugverluste geloescht werden kann
    if ( ((EntryMenuChoice *)entry)->refCount() == 0 )
    {
        // Loeschanforderung an Basisklasse weiterleiten
        EntryBase::deleteEntry( pos );
    }
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryMenu::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    unsigned int aP[MAX_SUBPOINTS];
    int i;

    // Daten fuer Bedien - Struktur uebernehmen
    m_operating.ptr = pFlash + sizeof( StructOperating );
    m_operating.ptr = endianReverse( m_operating.ptr + FLASH_KONFIG_START );
    *reinterpret_cast<StructOperating *>(&file.m_flash[pFlash]) = m_operating;
    pFlash += sizeof( StructOperating );

    // The parent is configured outside the method and is not changed.
    fileXML.writeContent( FileXML::ContentType::Operating, as_const( m_operating ) );

    // Daten von BedienMenue uebernehmen

    // Anzahl Unterpunkte berechnen und eintragen
    m_menu.numberOfEntries = static_cast<unsigned char>(size());
    if ( m_menu.numberOfEntries > MAX_SUBPOINTS )
        m_menu.numberOfEntries = MAX_SUBPOINTS;

    // BedienMenue.ptr berechnen und eintragen
    m_menu.ptr = pFlash + sizeof( StructMenu );
    m_menu.ptr = endianReverse( m_menu.ptr + FLASH_KONFIG_START );
    *reinterpret_cast<StructMenu *>(&file.m_flash[pFlash]) = m_menu;
    pFlash += sizeof( StructMenu );

    // The parent is configured outside the method and is not changed. The tags 'operating' and 'menu'
    // are sibling elements.
    fileXML.writeContent( FileXML::ContentType::Menu, as_const( m_menu ) );

    // Pruefung auf benoetigte Mindestversion
    if ( (m_menu.behaviourAtEsc >= 112 && m_menu.behaviourAtEsc <= 127)
         || (m_menu.behaviourAtEsc >= 200 && m_menu.behaviourAtEsc <= 229)
         || m_menu.behaviourAtEsc == 254 || m_menu.behaviourAtEsc == 255 )
    {
        // ESC-Verhalten bei U-Menue wird erst ab Versionen x unterstuetzt.
        file.NeedStudioVersion( "04.02.02.08" );
        file.NeedFirmware2Version( "04.02.02.17" );
        file.UnsupportFirmware1Version();
    }

    // Erster freier Speicherplatz nach BedienMenue -> beginn Offsetarray
    aP[0] = pFlash + sizeof( unsigned int ) * m_menu.numberOfEntries; // Platz fuer Offsetarray reservieren

    // Finding the node for the child structures and update parent.
    auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::Menu ) );

    // Alle Unterpunkte in Flash eintragen und den jeweiligen Flashoffset sichern
    auto iterChild = constBegin();
    for ( i = 1; iterChild != constEnd() && (i <= m_menu.numberOfEntries); ++iterChild, ++i )
    {
        aP[i] = (*iterChild)->writeDocument( aP[i - 1], doc );
    }

    // Reset parent node because the end of the structure is reached.
    fileXML.setParentNode( oldParent );

    // Pointer auf Unterpunkte eintragen
    for ( i = 0; i < m_menu.numberOfEntries; ++i )
    {
        aP[i] = endianReverse( aP[i] + FLASH_KONFIG_START );
        *reinterpret_cast<unsigned int *>(&file.m_flash[pFlash + i * sizeof( unsigned int )]) = aP[i];
    }

    return aP[m_menu.numberOfEntries];
}

uint32_t EntryMenu::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    tinyxml2::XMLElement *lastXmlElement{};

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    if ( useAESFile )
    {
        // Daten von Bedien einlesen
        m_operating = *reinterpret_cast<StructOperating *>(&file.m_flash[pFlash]);
        m_operating.ptr = endianReverse( m_operating.ptr ) - FLASH_KONFIG_START;
        pFlash += sizeof( StructOperating );

        // Daten von BedienMenue einlesen
        m_menu = *reinterpret_cast<StructMenu *>(&file.m_flash[pFlash]);
        m_menu.ptr = endianReverse( m_menu.ptr ) - FLASH_KONFIG_START;
        pFlash += sizeof( StructMenu );
    }
    else
    {
        // pFlash contains the information about the corresponding index of the previous sibling element to be read.
        // The readDocument method searches depending on the parent or the previous element.

        // Daten von Bedien einlesen
        fileXML.combineValid( fileXML.readContent( FileXML::ContentType::Operating, m_operating, pFlash, lastXmlElement ) );

        // Daten von BedienMenue einlesen
        // Different entry types can occur. Therefore, indexing based on the type is not correct. Since each type is
        // preceded by an 'operating' element, this entry is used as the basis for the search.
        fileXML.combineValid( fileXML.readContent( FileXML::ContentType::Menu, m_menu, lastXmlElement ) );
    }

    // Name fuer Baum eintragen
    setName( m_menu.heading1 );

    // Bei einem Hauptmenue wird dieses nicht unterstuetzt!
    if ( m_menu.behaviourAtEsc != EV_UPPER_MENUE && parentType() == EntryType::Handling )
    {
        m_menu.behaviourAtEsc = EV_UPPER_MENUE;
    }

    // Finding the node for the child structures and update parent.
    auto oldParent = fileXML.setParentNode( lastXmlElement );

    unsigned int offset{};
    // Daten aller Eintraege einlesen
    for ( unsigned char i = 0; i < (int)m_menu.numberOfEntries; ++i )
    {
        EntryMenuChoice *pNeu = new EntryMenuChoice( this );
        pNeu->setDefault();

        // Flashoffset des Eintrags ermitteln
        if ( useAESFile )
        {
            offset = m_menu.ptr + i * sizeof( unsigned int );
            offset = *reinterpret_cast<unsigned int *>(&file.m_flash[offset]);
            offset = endianReverse( offset ) - FLASH_KONFIG_START;

            pNeu->readDocument( offset, doc );
        }
        else
        {
            pNeu->readDocument( i, doc );
        }
        append( pNeu );
    }

    // Reset parent node because the end of the structure is reached.
    fileXML.setParentNode( oldParent );

    return pFlash;
}

///////////////////////////////////
// Referenzierung zwischen Einträgen

void EntryMenu::convertReferenceToIndex()
{
    int index;
    EntryDevice *device = nullptr;
    EntryBase *handling = nullptr;
    EntryBase *main_menu = nullptr;
    EntryBase *entry = nullptr;

    // Auf Rooteintrag wechseln.
    entry = this->parent();
    while ( !entry->isType( EntryType::Device ) )
    {
        // Eintrag unter Handling?
        if ( entry->isType( EntryType::Handling ) )
        {
            handling = entry;
            main_menu = handling->first();
        }

        entry = entry->parent();
    }
    device = (EntryDevice *)entry;

    // Zeiger auf Hauptmenu sicherstellen.
    if ( !main_menu && device->rootHandling().size() )
        main_menu = device->rootHandling().first();

    // Sprungziel nach Enter Ermitteln und Eintragen
    if ( m_refJumpESC == (EntryBase *)0 )
    {
// Keine Auswahl oder Uebergeordnetes Untermenue gewaehlt.
        m_menu.behaviourAtEsc = EV_UPPER_MENUE;

    }
    else if ( m_refJumpESC == (EntryBase *)-1 )
    {
// Naechste Eingabekette bzw. Menue gewaehlt.
        m_menu.behaviourAtEsc = EV_NEXT_SIBLING;

    }
    else if ( m_refJumpESC == (EntryBase *)-2 )
    {
// Hauptmenue gewaehlt.
        m_menu.behaviourAtEsc = EV_MAIN_MENUE;

    }
    else
    {

     // Entweder ist das Sprungziel E-Feld, E-Kette bzw. U-Menue, F-Taste   

     // Versuch die Referenz zu E-Kette bzw. U-Menue aufzuloesen.
        index = parent()->indexOf( m_refJumpESC );
        if ( index == -1 )
        {

// Versuch die Referenz zu F-Taste aufzuloesen.
            if ( main_menu )
            {

                index = main_menu->indexOf( m_refJumpESC );
            }

            if ( index == -1 )
            {

// Von Naechster E-Kette oder Menue ausgehen.
                m_menu.behaviourAtEsc = EV_NEXT_SIBLING;
            }
            else
            {

                     // F-Taste
                m_menu.behaviourAtEsc = index + 112; // 112 => F1
            }
        }
        else
        {
             // Bis zu 30 erlaubt.
            if ( index >= 30 ) index = 0;
            // E-Kette bzw. U-Menue
            m_menu.behaviourAtEsc = index + 200;
        }
    }

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertReferenceToIndex();
}

void EntryMenu::convertIndexToReference()
{
    int index;
    EntryDevice *device = nullptr;
    EntryBase *handling = nullptr;
    EntryBase *main_menu = nullptr;
    EntryBase *entry = nullptr;

    // Auf Rooteintrag wechseln.
    entry = this->parent();
    while ( !entry->isType( EntryType::Device ) )
    {
        // Eintrag unter Handling?
        if ( entry->isType( EntryType::Handling ) )
        {
            handling = entry;
            main_menu = handling->first();
        }

        entry = entry->parent();
    }
    device = (EntryDevice *)entry;

    // Zeiger auf Hauptmenu sicherstellen.
    if ( !main_menu && device->rootHandling().size() )
        main_menu = device->rootHandling().first();

    index = m_menu.behaviourAtEsc;
    // Sprungziel nach ESC
    switch ( index )
    {
        case EV_UPPER_MENUE:
        case 1: // Ausschalten faellt weg und wird zu Uebergeordnetem Untermenue.
            m_refJumpESC = (EntryBase *)0;
            break;
        case EV_NEXT_SIBLING:
            m_refJumpESC = (EntryBase *)-1;
            break;
        case EV_MAIN_MENUE:
            m_refJumpESC = (EntryBase *)-2;
            break;
        default:
            m_refJumpESC = (EntryBase *)0;
            if ( index >= 200 && index <= 229 )
            {
// E-Kette bzw. Menue.
                m_refJumpESC = parent()->value( index - 200 );

            }
            else if ( index >= 112 && index <= 127 )
            {
// F-Taste
                if ( main_menu )
                    m_refJumpESC = main_menu->value( index - 112 ); // 112 => F1
            }
            if ( !m_refJumpESC )
            {
// Von Naechster E-Kette bzw. Menue.
                m_refJumpESC = (EntryBase *)-1;
            }
            break;
    }
    if ( m_refJumpESC != (EntryBase *)0 && m_refJumpESC != (EntryBase *)-1 && m_refJumpESC != (EntryBase *)-2 )
    {

        m_refJumpESC->incRefCount();
    }

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertIndexToReference();
}

///////////////////////////////////
// Daten des Eintrags

void EntryMenu::setDefault()
{
    setName( QT_TRANSLATE_NOOP_UTF8( "EntryMenu", "Menue" ) );

    m_iconIndex = ICON_MENU;
    m_iconIndexSelected = ICON_MENU_SEL;

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_MENU_0 );

    std::memset( &m_operating, 0, sizeof( m_operating ) );
    m_operating.type = TB_MENUE;
    m_operating.d1 = m_operating.d2 = m_operating.type;

    std::memset( &m_menu, 0, sizeof( m_menu ) );
    m_menu.flags = 0;
    m_menu.numberOfEntries = 0;
    m_menu.behaviourAtEsc = EV_NEXT_SIBLING;

    std::strncpy( m_menu.heading1, name().c_str(), sizeof( m_menu.heading1 ) - 1 );
    strcpy( m_menu.heading2, "" );

    setInsertEnabled( TRUE );
    setDeleteEnabled( TRUE );
}







// ########################################################
// Member: EntryHandling

///////////////////////////////////
// Childs

EntryBase *EntryHandling::insertEntry( EntryType typeID, int pos )
{
    // Wenn Parent ROOT ist maximal 1 Hauptmenue anlegbar.
    if ( parentType() == EntryType::Device )
    {
        if ( size() >= 1 )
        {
            // Es ist maximal 1 Hauptmenue anlegbar.
            return nullptr;
        }
    }

    EntryMenu *entryNew = new EntryMenu( this );

    entryNew->setDefault();

    insertEntryFinish( entryNew, pos );

    return entryNew;
}

EntryBase *EntryHandling::cloneEntry()
{
    // Dieser Eintrag unterstützt derzeit cloneEntry nicht.
    return nullptr;
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryHandling::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    unsigned int aP[MAX_SUBPOINTS];
    int i;

    // Daten von KonfigRoot eintragen

    // Anzahl Eintraege in Wurzel Konfigdaten
    m_configRoot.numberOfEntries = (unsigned char)size();
    if ( m_configRoot.numberOfEntries > MAX_SUBPOINTS )
        m_configRoot.numberOfEntries = MAX_SUBPOINTS;

    m_configRoot.ptr = pFlash + sizeof( StructConfigRoot );
    m_configRoot.ptr = endianReverse( m_configRoot.ptr + FLASH_KONFIG_START );

    *reinterpret_cast<StructConfigRoot *>(&file.m_flash[pFlash]) = m_configRoot;
    pFlash += sizeof( StructConfigRoot );


    ////////////////////////////////////////
    // add information to xml file
    FileXML &fileXML = doc.fileXML();

    // The parent is configured outside the method and is not changed.
    fileXML.writeContent( FileXML::ContentType::ConfigRoot, as_const( m_configRoot ) );

    // erster freier Speicherplatz nach KonfigRoot -> begin Pointerarray
    aP[0] = pFlash + sizeof( unsigned int ) * m_configRoot.numberOfEntries; // Platz fuer Pointerarray reservieren

    // Finding the node for this structure and update parent.
    auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::ConfigRoot ) );

    // Alle Unterpunkte in Flash eintragen und die jeweiligen Pointer sichern
    auto iterChild = constBegin();
    for ( i = 1; iterChild != constEnd() && (i <= m_configRoot.numberOfEntries); ++iterChild, ++i )
    {
        aP[i] = (*iterChild)->writeDocument( aP[i - 1], doc );
    }

    // Reset parent node because the end of the structure is reached.
    fileXML.setParentNode( oldParent );

    // Pointer auf Unterpunkte eintragen
    for ( i = 0; i < m_configRoot.numberOfEntries; ++i )
    {
        aP[i] = endianReverse( aP[i] + FLASH_KONFIG_START );
        *reinterpret_cast<unsigned int *>(&file.m_flash[pFlash + i * sizeof( unsigned int )]) = aP[i];
    }

    // Tastenumschaltung in EEPROM eintragen
    memcpy( &file.m_flash_header.autoKeys, m_autoKey, sizeof( m_autoKey ) );

    return aP[m_configRoot.numberOfEntries];
}

uint32_t EntryHandling::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    int i{};
    unsigned int pDat{};
    StructOperating operating{};

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    // Daten von Konfig Root einlesen
    if ( useAESFile )
    {
        m_configRoot = *reinterpret_cast<StructConfigRoot *>(&file.m_flash[pFlash]);
        m_configRoot.ptr = endianReverse( m_configRoot.ptr ) - FLASH_KONFIG_START;
        pFlash += sizeof( StructConfigRoot );

    }
    else
    {
        fileXML.combineValid( fileXML.readContent( FileXML::ContentType::ConfigRoot, m_configRoot ) );
    }

    // Finding the node for this structure and update parent.
    auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::ConfigRoot ) );

    // Daten aller Eintraege einlesen
    for ( i = 0; i < (int)m_configRoot.numberOfEntries; i++ )
    {
        if ( useAESFile )
        {
            // Pointer des Eintrags berechnen
            pDat = m_configRoot.ptr + i * sizeof( unsigned int );
            pDat = *reinterpret_cast<unsigned int *>(&file.m_flash[pDat]);
            pDat = endianReverse( pDat ) - FLASH_KONFIG_START;

            // Daten von Bedien - Struktur fuer Klassenerzeugung einlesen
            operating = *reinterpret_cast<StructOperating *>(&file.m_flash[pDat]);
        }
        else
        {
            tinyxml2::XMLElement *lastXmlItem{};
            fileXML.combineValid( fileXML.readContent( FileXML::ContentType::Operating, operating, i, lastXmlItem ) );
        }

        // Daten von Bedien - Struktur fuer Klassenerzeugung einlesen
        switch ( operating.type )
        {
            case TB_MENUE:
            {
                EntryMenu *pNeu = new EntryMenu( this );
                pNeu->setDefault();
                pNeu->readDocument( useAESFile ? pDat : i, doc );
                append( pNeu );
                break;
            }

            case TB_DATENSATZ:
            {
                EntryInputChain *pNeu = new EntryInputChain( this );
                pNeu->setDefault();
                pNeu->readDocument( useAESFile ? pDat : i, doc );
                append( pNeu );
                break;
            }

            default:
                BOOST_ASSERT( false );
                break;
        }
    }

    // Reset parent node because the end of the structure is reached.
    fileXML.setParentNode( oldParent );


    // Tastenumschaltung aus EEPROM lesen
    memcpy( m_autoKey, &file.m_flash_header.autoKeys, sizeof( m_autoKey ) );
    if ( m_autoKey[0] != 0 )
    {
        // Versionsbyte abfragen
        // Tastenumschaltung
        memset( m_autoKey, 0, sizeof( m_autoKey ) );
        // Defaultfunktion 252: Letzte Tasten F1, F2
        m_autoKey[4] = 252;
        m_autoKey[4 + 3] = 0x7F;
    }

    // Wenn keine Auswahl vorliegt, wird die default Einstellung vorgegeben.
    int count = 0;
    for ( i = 0; i < 15; ++i )
    {
        if ( m_autoKey[4 * (i + 1)] == 0 ) // Keine Funktion gewaehlt?
            count++;
    }

    // Wenn bei allen keine Funktion gewahlet ist, dann default setzen.
    if ( count == 15 )
    {
        // Tastenumschaltung
        memset( m_autoKey, 0, sizeof( m_autoKey ) );
        // Defaultfunktion 250: Hauptmenue
        m_autoKey[4] = 250;
        m_autoKey[4 + 3] = 0x7F;
    }

    return pFlash;
}

///////////////////////////////////
// Daten des Eintrags

void EntryHandling::setDefault()
{
    setName( QT_TRANSLATE_NOOP_UTF8( "EntryHandling", "Bedienung" ) );

    m_iconIndex = ICON_HANDLING;
    m_iconIndexSelected = ICON_HANDLING_SEL;

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_HANDLING_0 );

    m_configRoot.autoSwitchOff = 0;
    m_configRoot.behaviourAfterSwitchOn = EVERH_MENUE;

    // Tastenumschaltung
    memset( m_autoKey, 0, sizeof( m_autoKey ) );
    // Defaultfunktion 252: Letzte Tasten F1, F2
    m_autoKey[4] = 252;
    m_autoKey[4 + 3] = 0x7F;

    setInsertEnabled( TRUE );
    setDeleteEnabled( FALSE );
}

// ########################################################
// Member: EntryTableRecord

///////////////////////////////////
// Childs

EntryBase *EntryTableRecord::insertEntry( EntryType typeID, int pos )
{
    // Wenn Parent ROOT sind maximal 20 Datensatzbeschreibungen anlegbar.
    if ( parentType() == EntryType::Device )
    {
        if ( size() >= MAX_RECORD_DESCRIPTION )
        {
            // Es sind maximal MAX_RECORD_DESCRIPTION Datensatzbeschreibungen anlegbar.
            return nullptr;
        }
    }

    EntryTable *entryNew = new EntryTable( this );
    entryNew->setDefault( boost::str( boost::format( "%1%%2%" ) % QT_TRANSLATE_NOOP_UTF8( "EntryTableRecord", "Datensatz" ) % (size() + 1) ) );

    insertEntryFinish( entryNew, pos );

    // Allen Unterpunkten ggf. neue Indizes zuweisen.
    int i;
    auto iterChild = constBegin();
    for ( i = 0; iterChild != constEnd(); ++iterChild, ++i )
    {
        ((EntryTable *)*iterChild)->m_table.number = i; // Index der Tabelle eintragen
    }

    return entryNew;
}

EntryBase *EntryTableRecord::cloneEntry()
{
    // Dieser Eintrag unterstützt derzeit cloneEntry nicht.
    return nullptr;
}

void EntryTableRecord::deleteEntry( int pos )
{
    EntryBase *entry;

    switch ( pos )
    {
        case -1:
            entry = last();
            break;

        default:
            entry = value( pos );
    }

    // Pruefen ob Tabelle ohne Bezugverluste geloescht werden kann
    if ( ((EntryTable *)entry)->refCount() == 0 )
    {
        // Loeschanforderung an Basisklasse weiterleiten
        EntryBase::deleteEntry( pos );

        // Allen Unterpunkten ggf. neue Indizes zuweisen.
        int i;
        auto iterChild = constBegin();
        for ( i = 0; iterChild != constEnd(); ++iterChild, ++i )
        {
            ((EntryTable *)*iterChild)->m_table.number = i; // Index der Tabelle eintragen
        }
    }
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryTableRecord::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    unsigned int aP[MAX_RECORD_DESCRIPTION + 1];
    int childCount;
    int i;
    EntryDevice &device = root();

    // Anzahl Eintraege berechnen
    childCount = size();
    if ( childCount > MAX_RECORD_DESCRIPTION )
        childCount = MAX_RECORD_DESCRIPTION;

    // erster freier Speicherplatz nach Konfigdaten -> begin Pointerarray
    aP[0] = pFlash + sizeof( unsigned int ) * MAX_RECORD_DESCRIPTION; // Platz fuer Pointerarray reservieren

    // Index der Spalten fuer Prioritaet der Datensaetze loeschen.
    for ( size_t i = childCount; i < sizeof( device.m_extended.recordPriorityColumn ); ++i )
    {
        device.m_extended.recordPriorityColumn[i] = 0;
    }

    ////////////////////////////////////////
    // add information to xml file
    FileXML &fileXML = doc.fileXML();

    // Finding the node for this structure and update parent.
    auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::Records ) );

    // Alle Unterpunkte in Flash eintragen und die jeweiligen Pointer sichern
    auto iterChild = constBegin();
    for ( i = 1; iterChild != constEnd(); ++iterChild, ++i )
    {
        // Index der Tabelle eintragen
        ((EntryTable *)*iterChild)->m_table.number = i - 1;

        aP[i] = (*iterChild)->writeDocument( aP[i - 1], doc );
    }

    // Reset parent node because the end of the structure is reached.
    fileXML.setParentNode( oldParent );

    // Pointer auf Unterpunkte eintragen
    for ( i = 0; i < childCount; i++ )
    {
        aP[i] = endianReverse( aP[i] + FLASH_KONFIG_START );
        *reinterpret_cast<unsigned int *>(&file.m_flash[pFlash + i * sizeof( unsigned int )]) = aP[i];
    }
    // nicht verwendete Pointer -> nullptr
    for ( ; i < MAX_RECORD_DESCRIPTION; i++ )
    {
        *reinterpret_cast<unsigned int *>(&file.m_flash[pFlash + i * sizeof( unsigned int )]) = 0;
    }

    return aP[childCount]; // Flashposition hinter letzter Struktur
}

uint32_t EntryTableRecord::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    uint32_t i{};
    unsigned int pDat{};

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);
    bool oldValidState = fileXML.isValid();

    // Finding the node for this structure and update parent.
    auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::Records ) );
    auto numberOfDescriptions = fileXML.numberOfChildItems( FileXML::ContentType::RecordDescription );

    // Daten aller Eintraege einlesen
    for ( i = 0; i < MAX_RECORD_DESCRIPTION; ++i )
    {
        if ( useAESFile )
        {
            // Pointer des Eintrags berechnen
            pDat = pFlash + i * sizeof( unsigned int );
            pDat = *reinterpret_cast<unsigned int *>(&file.m_flash[pDat]);
            // Abbruch nach erstem Pointer mit Wert nullptr
            if ( !pDat )
                break;

            pDat = endianReverse( pDat ) - FLASH_KONFIG_START;
        }
        else
        {
            // Process only existing record descriptions.
            if ( i >= numberOfDescriptions )
            {
                break;
            }
        }

        EntryTable *pNeu = new EntryTable( this );
        pNeu->setDefault();

        // Index der Tabelle eintragen (Wird durch gekuerzte Struktur nicht mit abgelegt)
        pNeu->m_table.number = i;

        pNeu->readDocument( useAESFile ? pDat : i, doc );

        // Termination as soon as an entry does not exist or returns an error.
        // The validation state is reset.
        if ( !fileXML.isValid() )
        {
            fileXML.setValidState( oldValidState );
            break;
        }
        append( pNeu );
    }

    // Reset parent node because the end of the structure is reached.
    fileXML.setParentNode( oldParent );

    return pFlash;
}

///////////////////////////////////
// Daten des Eintrags

void EntryTableRecord::setDefault()
{
    // Icon's
    m_iconIndex = ICON_TABLE_RECORD;
    m_iconIndexSelected = ICON_TABLE_RECORD_SEL;

    // Name
    setName( QT_TRANSLATE_NOOP_UTF8( "EntryTableRecord", "Datensatztabellen für Erfassung ( Lesen )" ) );

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_TABLE_RECORD_0 );

    setInsertEnabled( TRUE );
    setDeleteEnabled( FALSE );
}

// ########################################################
// Member: EntryTableList

///////////////////////////////////
// Childs

EntryBase *EntryTableList::insertEntry( EntryType typeID, int pos )
{
    // Wenn Parent ROOT sind maximal 20 Listenbeschreibungen anlegbar.
    if ( parentType() == EntryType::Device )
    {
        if ( size() >= MAX_RECORD_DESCRIPTION )
        {
            // Es sind maximal MAX_RECORD_DESCRIPTION Listenbeschreibungen anlegbar.
            return nullptr;
        }
    }

    EntryTable *entryNew = new EntryTable( this );

    entryNew->setDefault( boost::str( boost::format( "%1%%2%" ) % QT_TRANSLATE_NOOP_UTF8( "EntryTableList", "Liste" ) % (size() + 1) ) );

    insertEntryFinish( entryNew, pos );

    // Allen Unterpunkten ggf. neue Indizes zuweisen.
    int i;
    auto iterChild = constBegin();
    for ( i = 0; iterChild != constEnd(); ++iterChild, ++i )
        ((EntryTable *)*iterChild)->m_table.number = i; // Index der Tabelle eintragen

    return entryNew;
}

EntryBase *EntryTableList::cloneEntry()
{
    // Dieser Eintrag unterstützt derzeit cloneEntry nicht.
    return nullptr;
}

void EntryTableList::deleteEntry( int pos )
{
    EntryBase *entry;

    switch ( pos )
    {
        case -1:
            entry = last();
            break;

        default:
            entry = value( pos );
    }

    // Pruefen ob Tabelle ohne Bezugverluste geloescht werden kann
    if ( ((EntryTable *)entry)->refCount() == 0 )
    {
        // Loeschanforderung an Basisklasse weiterleiten
        EntryBase::deleteEntry( pos );

        // Allen Unterpunkten ggf. neue Indizes zuweisen.
        int i;
        auto iterChild = constBegin();
        for ( i = 0; iterChild != constEnd(); ++iterChild, ++i )
        {
            ((EntryTable *)*iterChild)->m_table.number = i; // Index der Tabelle eintragen
        }
    }
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryTableList::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    unsigned int aP[MAX_RECORD_DESCRIPTION + 1];
    int childCount, i;

    // Anzahl Eintraege berechnen
    childCount = size();
    if ( childCount > MAX_RECORD_DESCRIPTION )
        childCount = MAX_RECORD_DESCRIPTION;

    // Index der Sortierspalte fuer alle Listen loeschen. Childs tragen diesen dann ein.
    for ( int i = childCount; i < MAX_RECORD_DESCRIPTION; ++i )
        file.m_flash_header.listKeyField[i] = 0;

    // erster freier Speicherplatz nach Konfigdaten -> begin Pointerarray
    aP[0] = pFlash + sizeof( unsigned int ) * MAX_RECORD_DESCRIPTION; // Platz fuer Pointerarray reservieren

    ////////////////////////////////////////
    // add information to xml file
    FileXML &fileXML = doc.fileXML();

    // Finding the node for this structure and update parent.
    auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::Lists ) );

    // Alle Unterpunkte in Flash eintragen und die jeweiligen Pointer sichern
    auto iterChild = constBegin();
    for ( i = 1; iterChild != constEnd(); ++iterChild, ++i )
    {
        // Index der Tabelle eintragen
        ((EntryTable *)*iterChild)->m_table.number = i - 1;

        aP[i] = (*iterChild)->writeDocument( aP[i - 1], doc );
    }

    // Reset parent node because the end of the structure is reached.
    fileXML.setParentNode( oldParent );

    // Pointer auf Unterpunkte eintragen
    for ( i = 0; i < childCount; ++i )
    {
        aP[i] = endianReverse( aP[i] + FLASH_KONFIG_START );
        *reinterpret_cast<unsigned int *>(&file.m_flash[pFlash + i * sizeof( unsigned int )]) = aP[i];
    }
    // nicht verwendete Pointer -> nullptr
    for ( ; i < MAX_RECORD_DESCRIPTION; ++i )
    {
        *reinterpret_cast<unsigned int *>(&file.m_flash[pFlash + i * sizeof( unsigned int )]) = 0;
    }

    return aP[childCount]; // Flashposition hinter letzter Struktur
}

uint32_t EntryTableList::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    unsigned int pDat{};

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);
    bool oldValidState = fileXML.isValid();

    // Finding the node for this structure and update parent.
    auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::Lists ) );
    auto numberOfDescriptions = fileXML.numberOfChildItems( FileXML::ContentType::RecordDescription );

    // Daten aller Eintraege einlesen
    for ( unsigned int i = 0; i < MAX_RECORD_DESCRIPTION; ++i )
    {
        if ( useAESFile )
        {
            // Pointer des Eintrags berechnen
            pDat = pFlash + i * sizeof( unsigned int );
            pDat = *reinterpret_cast<unsigned int *>(&file.m_flash[pDat]);
            // Abbruch nach erstem Pointer mit Wert nullptr
            if ( !pDat )
                break;

            pDat = endianReverse( pDat ) - FLASH_KONFIG_START;
        }
        else
        {
            // Process only existing record descriptions.
            if ( i >= numberOfDescriptions )
            {
                break;
            }
        }

        EntryTable *pNeu = new EntryTable( this );
        pNeu->setDefault( QT_TRANSLATE_NOOP_UTF8( "EntryTableList", "Listenbeschreibung" ) );

        // Index der Tabelle eintragen (Wird durch gekuerzte Struktur nicht mit abgelegt)
        pNeu->m_table.number = i;

        pNeu->readDocument( useAESFile ? pDat : i, doc );

        // Termination as soon as an entry does not exist or returns an error.
        // The validation state is reset.
        if ( !fileXML.isValid() )
        {
            fileXML.setValidState( oldValidState );
            break;
        }
        append( pNeu );
    }

    // Reset parent node because the end of the structure is reached.
    fileXML.setParentNode( oldParent );

    return pFlash;
}


///////////////////////////////////
// Daten des Eintrags

void EntryTableList::setDefault()
{
    // Icon's
    m_iconIndex = ICON_TABLE_LIST;
    m_iconIndexSelected = ICON_TABLE_LIST_SEL;

    // Name
    setName( QT_TRANSLATE_NOOP_UTF8( "EntryTableList", "Listentabellen für Datenauswahl ( Schreiben )" ) );

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_TABLE_LIST_0 );

    // Setzen der Bearbeitungsmoeglichkeit
    setInsertEnabled( TRUE );
    setDeleteEnabled( FALSE );
}

// ########################################################
// Member: EntryTable

///////////////////////////////////
// Childs

EntryBase *EntryTable::insertEntry( EntryType typeID, int pos )
{
    // Es sind maximal 25 Felder in Datensatz- oder Listenbeschreibungen anlegbar.
    if ( parentType() == EntryType::TableRecord || parentType() == EntryType::TableList )
    {
        if ( size() >= MAX_FIELDS )
        {
            // Es sind maximal MAX_FIELDS Felder anlegbar.
            return nullptr;
        }

        // Maximal kann die Summe der Feldlaengen MAX_RECORD_SIZE betragen
        if ( recordSize() > MAX_RECORD_SIZE )
        {
            // Die Maximale Größe des Datensatzes würde überschritten.
            return nullptr;
        }
    }

    EntryTableField *entryNew = new EntryTableField( this );
    entryNew->setDefault( boost::str( boost::format( "%1%%2%" ) % QT_TRANSLATE_NOOP_UTF8( "EntryTable", "Feld" ) % (size() + 1) ), TF_A_NUM, 21 );

    insertEntryFinish( entryNew, pos );

    // Allen Unterpunkten ggf. neue Indizes zuweisen.
    int i = 0;
    auto iterChild = constBegin();
    for ( ; iterChild != constEnd(); ++iterChild, ++i )
        ((EntryTableField *)*iterChild)->m_tableField.number = i; // Index der Tabelle eintragen

    return entryNew;
}

EntryBase *EntryTable::cloneEntry()
{
    // Dieser Eintrag unterstützt derzeit cloneEntry nicht.
    return nullptr;
}

void EntryTable::deleteEntry( int pos )
{
    EntryBase *entry;

    switch ( pos )
    {
        case -1:
            entry = last();
            break;

        default:
            entry = value( pos );
    }

    // Pruefen ob Tabelle ohne Bezugverluste geloescht werden kann
    if ( ((EntryTableField *)entry)->refCount() == 0 )
    {
        // Loeschanforderung an Basisklasse weiterleiten
        EntryBase::deleteEntry( pos );

        // Allen Unterpunkten ggf. neue Indizes zuweisen.
        int i;
        auto iterChild = constBegin();
        for ( i = 0; iterChild != constEnd(); ++iterChild, ++i )
        {
            ((EntryTableField *)*iterChild)->m_tableField.number = i; // Index der Tabelle eintragen
        }
    }
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryTable::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    unsigned int aP[MAX_SUBPOINTS + 1];
    int i;

    // Anzahl Eintraege berechnen
    m_table.count = (unsigned char)size();
    if ( m_table.count > MAX_SUBPOINTS )
        m_table.count = MAX_SUBPOINTS;

    m_table.ptr = pFlash + sizeof( StructRecordDescription );
    m_table.ptr += m_flashOffset;
    m_table.ptr = endianReverse( m_table.ptr + FLASH_KONFIG_START );
    *(StructRecordDescription *)&file.m_flash[pFlash] = *(StructRecordDescription *)&m_table;
    pFlash += sizeof( StructRecordDescription );

    ////////////////////////////////////////
    // add information to xml file
    FileXML &fileXML = doc.fileXML();

    // The parent is configured outside the method and is not changed.
    fileXML.writeContent( FileXML::ContentType::RecordDescription, as_const( *(StructRecordDescription *)&m_table ) );

    // erster freier Speicherplatz nach DatensatzEinlesen -> begin Pointerarray
    aP[0] = pFlash + sizeof( unsigned int ) * m_table.count; // Platz für Pointerarray reservieren

    // Finding the node for this structure and update parent.
    auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::RecordDescription ) );

    // Alle Unterpunkte in Flash eintragen und die jeweiligen Pointer sichern
    auto iterChild = constBegin();
    for ( i = 1; iterChild != constEnd() && (i <= m_table.count); ++iterChild, ++i )
    {
        // Feldnummer eintragen
        ((EntryTableField *)*iterChild)->m_tableField.number = i - 1;
        // Feldstruktur uebernehmen
        aP[i] = (*iterChild)->writeDocument( aP[i - 1], doc );
    }

    // Reset parent node because the end of the structure is reached.
    fileXML.setParentNode( oldParent );

    // Pointer auf Unterpunkte eintragen
    for ( i = 0; i < m_table.count; ++i )
    {
        // In ExtendedFlash eintragen
        aP[i] += m_flashOffset;
        aP[i] = endianReverse( aP[i] + FLASH_KONFIG_START );
        *reinterpret_cast<unsigned int *>(&file.m_flash[pFlash + i * sizeof( unsigned int )]) = aP[i];
    }

    // Bei Listen die Nummer des Schluesselfeldes im file.m_eep setzen
    if ( parentType() == EntryType::TableList )
    {
        if ( m_refKeyField )
            file.m_flash_header.listKeyField[m_table.number] = m_refKeyField->m_tableField.number + 1;
        else
            file.m_flash_header.listKeyField[m_table.number] = 0;
    }

    return aP[m_table.count]; // Flashposition hinter letzter Struktur
}

uint32_t EntryTable::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    tinyxml2::XMLElement *lastXmlElement{};

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    // Daten von DatensatzBeschreibung einlesen
    if ( useAESFile )
    {
        *(StructRecordDescription *)&m_table = *(StructRecordDescription *)&file.m_flash[pFlash];
        m_table.ptr = endianReverse( m_table.ptr ) - FLASH_KONFIG_START;
        // Offset abziehen, da Erweitertes - Flash hinter Setupdaten.
        m_table.ptr -= m_flashOffset;

        pFlash += sizeof( StructRecordDescription );
    }
    else
    {
        if ( fileXML.numberOfChildItems( FileXML::ContentType::RecordDescription ) > 0 )
        {
            fileXML.combineValid( fileXML.readContent( FileXML::ContentType::RecordDescription, *(StructRecordDescription *)&m_table, pFlash, lastXmlElement ) );
        }
    }

    // Name fuer Baum eintragen
    setName( m_table.name );

    // Finding the node for the child structures and update parent.
    auto oldParent = fileXML.setParentNode( lastXmlElement );

    unsigned int offset{};
    // Daten aller Eintraege einlesen
    for ( unsigned char i = 0; i < m_table.count; ++i )
    {
        if ( useAESFile )
        {
            offset = m_table.ptr + i * sizeof( unsigned int );	// Pointer des Eintrags berechnen

            offset = *reinterpret_cast<unsigned int *>(&file.m_flash[offset]);
            offset = endianReverse( offset ) - FLASH_KONFIG_START;
            // Bei ExtendedFlash mit Offset
            offset -= m_flashOffset;
        }

        EntryTableField *pNeu = new EntryTableField( this );
        pNeu->setDefault();

        // Index des Tabellenfeldes eintragen (Wird durch gekuerzte Struktur nicht mit abgelegt)
        pNeu->m_tableField.number = i;

        pNeu->readDocument( useAESFile ? offset : i, doc );
        append( pNeu );
    }

    // Reset parent node because the end of the structure is reached.
    fileXML.setParentNode( oldParent );

    // Bei Listen die Nummer des Schluesselfeldes mitlesen
    if ( parentType() == EntryType::TableList )
    {
        if ( (file.m_flash_header.listKeyField[m_table.number] > 0) && (file.m_flash_header.listKeyField[m_table.number] < MAX_FIELDS) )
        {
            m_refKeyField = (EntryTableField *)at( file.m_flash_header.listKeyField[m_table.number] - 1 );
            m_refKeyField->incRefCount();
        }
    }

    return pFlash;
}

void EntryTable::checkMembers()
{
    // Ist Liste unter Zutrittskontrolle
    if ( parentType() == EntryType::AccessControl )
    {
        setInsertEnabled( FALSE );
        setDeleteEnabled( FALSE );
    }
}

///////////////////////////////////////
// Referenzierung zwischen Einträgen

void EntryTable::convertReferenceToIndex()
{
    EntryDevice &device = root();

    // Fuer Sortierspalte bei Listen
    //device.m_extended.gprs.invokeInputChain = (m_refInputChainGPRS != nullptr) ? eventChains.indexOf(m_refInputChainGPRS) + 1 : 0;

    // Bei Datensaetzen die Spaltennummer des Prioritaetsfeldes uebernehmen.
    if ( parentType() == EntryType::TableRecord )
    {
        // Für Prioritaetsspalte bei Datensaetzen.
        device.m_extended.recordPriorityColumn[m_table.number] = (m_refPriorityField != nullptr) ? indexOf( m_refPriorityField ) + 1 : 0;
    }

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertReferenceToIndex();
}

void EntryTable::convertIndexToReference()
{
    EntryDevice &device = root();

    // Fuer Sortierspalte bei Listen.
    //m_refInputChainGPRS = (device.m_extended.gprs.invokeInputChain > 0) ? (EntryInputChain*)eventChains.at(device.m_extended.gprs.invokeInputChain - 1) : nullptr;
    //if ( m_refInputChainGPRS )
    //    m_refInputChainGPRS->incRefCount();

    // Bei Datensaetzen die Spaltennummer des Prioritaetsfeldes uebernehmen.
    if ( parentType() == EntryType::TableRecord )
    {
        // Fuer Prioritaetsspalte bei Datensaetzen.
        m_refPriorityField = (device.m_extended.recordPriorityColumn[m_table.number] > 0) ? reinterpret_cast<EntryTableField *>(value( device.m_extended.recordPriorityColumn[m_table.number] - 1 )) : nullptr;
        if ( m_refPriorityField )
            m_refPriorityField->incRefCount();
    }

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertIndexToReference();
}

///////////////////////////////////////
// Helpers

// Prueft die Gesamtdatensatzlaenge der Datensatz- / Listenbeschreibung
int EntryTable::recordSize()
{
    int iLength = 0;
    auto iterChild = constBegin();
    for ( ; iterChild != constEnd(); ++iterChild )
        iLength += ((EntryTableField *)*iterChild)->m_tableField.size;

    return iLength;
}

///////////////////////////////////
// Daten des Eintrags

void EntryTable::setDefault()
{
    setDefault( QT_TRANSLATE_NOOP_UTF8( "EntryTable", "Datensatz" ) );
}

void EntryTable::setDefault( const std::string &name )
{
    // Icon's
    m_iconIndex = ICON_TABLE;
    m_iconIndexSelected = ICON_TABLE_SEL;

    // Name
    setName( name );

    // Tabellenstruktur
    memset( &m_table, 0, sizeof( m_table ) );
    strncpy( (char *)m_table.name, this->name().c_str(), sizeof( m_table.name ) - 1 );

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_TABLE_0 );

    // Bezuege von Eingabeketten auf diese Tabelle
    setRefCount( 0 );

    // Setzen der Bearbeitungsmoeglichkeit
    setInsertEnabled( true );
    setDeleteEnabled( true );

    // Pruefen der Abhaengigkeit zum Parent
    checkMembers();
}


// ########################################################
// Member: EntryTableField

///////////////////////////////////
// Childs

EntryBase *EntryTableField::insertEntry( EntryType typeID, int pos )
{
    (void)typeID;
    (void)pos;

    return nullptr;
}

EntryBase *EntryTableField::cloneEntry()
{
    // Dieser Eintrag unterstützt derzeit cloneEntry nicht.
    return nullptr;
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryTableField::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    /*
    int tableNum = ((EntryTable*)GetParent())->m_table.number;
    if ( GetParent()->GetParent()->get_type() == TableList )
    sprintf( m_field.name, "Listenfeld_%02i_%02i", tableNum + 1, m_field.number + 1 );
    else
    sprintf( m_field.name, "Datsatfeld_%02i_%02i", tableNum + 1, m_field.number + 1 );

    m_field.type = 4;
    if ( m_field.number < 16 )
    m_field.size = 10;
    */

    // Daten von DatensatzFeld - Struktur uebernehmen
    *(StructRecordField *)&file.m_flash[pFlash] = *(StructRecordField *)&m_tableField;
    pFlash += sizeof( StructRecordField );

    ////////////////////////////////////////
    // add information to xml file
    FileXML &fileXML = doc.fileXML();

    // The parent is configured outside the method and is not changed.
    fileXML.writeContent( FileXML::ContentType::RecordField, as_const( *(StructRecordField *)&m_tableField ) );

    // Bei Verwendung von Feldtyp DATA, muss die Mindestversion 04.03.13 sein.
    if ( m_tableField.type == TF_DATA )
    {
        // Benoetigte Mindestversion
        file.NeedStudioVersion( "04.03.13.00" );
        file.NeedLibraryVersion( "04.03.13.00" );
        file.NeedFirmware3Version( "04.03.13.00" );
        file.UnsupportFirmware2Version();
        file.UnsupportFirmware1Version();
    }

    return pFlash;
}

uint32_t EntryTableField::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    // Automatische Umstellung von unsigned int - Werten ohne MSG durchfuehren
    static bool bNoMessage = FALSE;

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    // Daten von DatensatzFeld - uebernehmen
    if ( useAESFile )
    {
        *(StructRecordField *)&m_tableField = *(StructRecordField *)&file.m_flash[pFlash];
        pFlash += sizeof( StructRecordField );
    }
    else
    {
        tinyxml2::XMLElement *lastXmlElement{};
        fileXML.combineValid( fileXML.readContent( FileXML::ContentType::RecordField, *(StructRecordField *)&m_tableField, pFlash, lastXmlElement ) );
    }

    // Name für Baum eintragen
    setName( m_tableField.name );

    // Typ Long bei alten Setupdateien auf ASCII( NUM ) umstellen
    if ( m_tableField.type == TF_ULONG )
    {
        if ( !bNoMessage )
        {
            //QString str;
            //str = QT_TRANSLATE_NOOP_UTF8( "EntryTableField", "In der zu ladenden Setupdatei ist mindestens ein Feldtyp nicht weiter unterstützt (unsigned int).\n");
            //str += QT_TRANSLATE_NOOP_UTF8( "EntryTableField", "Um die Setupdaten mit den neueren Firmwareversionen verwenden zu können, muss eine Umstellung vorgenommen werden.\n");
            //str += QT_TRANSLATE_NOOP_UTF8( "EntryTableField", "Nach der automatischen Umstellung sollte die Setupdatei gespeichert werden.");

            //QMessageBox::warning( ownerWidget(), "DatafoxStudioIV", str );
            bNoMessage = TRUE;
        }

        m_tableField.type = TF_NUM;

        // Sonderbehandlung für alte MDE Datensaetze (auch andere Kennungsfelder)

        if ( strcmp( (const char *)m_tableField.name, "Kennung" ) == 0 )
            m_tableField.size = 6;
        else
            m_tableField.size = 9;
    }

    //BO: 04.03.08.05 vom 19.05.2016
    // Feldtyp von Action2.Elapse umstellen.
    if ( grandParentType() == EntryType::AccessControl )
    {
        if ( std::string( reinterpret_cast<EntryTable *>(parent())->m_table.name ) == "Action2"
             && std::string( m_tableField.name ) == "Elapse" && m_tableField.type != TF_A_NUM )
        {
            m_tableField.type = TF_A_NUM;
        }

        //BO: B_1038 vom 22.09.2021
        // Feldtypen bei Event.PortIn, Action.PortOut und Action2.PortOut von FT_A_NUM sicherstellen.
        // Der Feldtyp kann bei ausgelesenen Setups, wenn diese mit einer älteren Studioversion/DFCom übertragen wurden
        // noch auf FT_NUM stehen. Nachkorrektur zur Änderung vom 14.09.2018
        if ( (std::string( reinterpret_cast<EntryTable *>(parent())->m_table.name ) == "Event" && std::string( m_tableField.name ) == "PortIn" && m_tableField.type != TF_A_NUM)
            || ((std::string( reinterpret_cast<EntryTable *>(parent())->m_table.name ) == "Action" || std::string( reinterpret_cast<EntryTable *>(parent())->m_table.name ) == "Action2") 
                 && std::string( m_tableField.name ) == "PortOut" && m_tableField.type != TF_A_NUM) )
        {
            std::string tname( reinterpret_cast<EntryTable *>(parent())->m_table.name );
            m_tableField.type = TF_A_NUM;
        }
   }



    return pFlash;
}

void EntryTableField::checkMembers()
{
    // Ist Feld unter Liste - Zutrittskontrolle
    if ( grandParentType() == EntryType::AccessControl )
    {
        setDeleteEnabled( false );
    }
}

///////////////////////////////////
// Daten des Eintrags

void EntryTableField::setDefault()
{
    setDefault( QT_TRANSLATE_NOOP_UTF8( "EntryTableField", "Feld" ), TF_A_NUM, 21 );
}

void EntryTableField::setDefault( const std::string &name, unsigned char ucType, unsigned char ucLength )
{
    // Icon's
    m_iconIndex = ICON_TABLE_FIELD;
    m_iconIndexSelected = ICON_TABLE_FIELD_SEL;

    // Name
    setName( name );
    // Feldstruktur
    memset( m_tableField.name, 0, sizeof( m_tableField.name ) );
    strncpy( (char *)m_tableField.name, this->name().c_str(), sizeof( m_tableField.name ) - 1 );
    m_tableField.type = ucType;
    m_tableField.size = ucLength;

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_TABLE_FIELD_0 );

    // Bezuege von Eingabekettenfeldern auf dieses Tabellenfeld
    setRefCount( 0 );

    // Setzen der Bearbeitungsmoeglichkeit
    setInsertEnabled( false );
    setDeleteEnabled( true );

    // Pruefen der Abhaengigkeit zum Parent
    checkMembers();
}

// ########################################################
// Member: EntryMenuChoice

EntryMenuChoice::EntryMenuChoice( const EntryMenuChoice &entry )
    : EntryBase( entry )
{
    // Member kopieren
    m_menuItem = entry.m_menuItem;

    EntryBase *pNew{};
    // Alle Childs kopieren
    auto iterChild = entry.constBegin();
    for ( ; iterChild != entry.constEnd(); ++iterChild )
    {
        // neues anfuegen
        switch ( (*iterChild)->type() )
        {
            case EntryType::Menu: // Menue
                pNew = (EntryBase *)((EntryMenu *)*iterChild)->cloneEntry();
                break;

            case EntryType::InputChain: // Eingabekette
                pNew = (EntryBase *)((EntryInputChain *)*iterChild)->cloneEntry();
                break;
        }

        BOOST_ASSERT( pNew != nullptr );
        if ( pNew != nullptr )
        {
            pNew->setParent( this );
            append( pNew );
        }
    }
}

///////////////////////////////////
// Childs

EntryBase *EntryMenuChoice::insertEntry( EntryType typeID /*= Type_id::Base*/, int pos /*= -1*/ )
{
    EntryBase *entryNew;

    switch ( typeID )
    {
        case EntryType::Menu:
            entryNew = new EntryMenu( this );
            entryNew->setDefault();
            break;

        case EntryType::InputChain:
            entryNew = new EntryInputChain( this );
            ((EntryInputChain *)entryNew)->setDefault( boost::str( boost::format( "%1%%2%" ) % QT_TRANSLATE_NOOP_UTF8( "EntryMenuChoice", "Eingabekette" ) % (size() + 1) ) );
            break;

        default:
            entryNew = new EntryInputChain( this );
            entryNew->setDefault();
            break;
    }

    insertEntryFinish( entryNew, pos );

    return entryNew;
}

EntryBase *EntryMenuChoice::cloneEntry()
{
    return reinterpret_cast<EntryBase *>(new EntryMenuChoice( *this ));
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryMenuChoice::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    unsigned int aP[MAX_SUBPOINTS];
    int Unterpunkte;
    int i;

    // Bei MDE - Setup soll der Block fuer die Steuerparameter,
    // ausgegraut werden, wenn keine der untergeordneten Eingabeketten eine
    // Datensatzbeschreibung zugewiesen hat.
    // Steuerparameter automatisch auf Buchung fixieren.
    int activateSteuerparam = FALSE;
    auto iterChild = constBegin();
    for ( ; iterChild != constEnd(); ++iterChild )
    {
        // Vorsichtshalber drauf pruefen, da Untermenue moeglich
        if ( (*iterChild)->isType( EntryType::InputChain ) )
        {
            if ( ((EntryInputChain *)*iterChild)->m_readRecord.number != 0xFF )
            {
                activateSteuerparam = TRUE;
                break;
            }
        }
    }
    // Sind Childs vorhanden aber keiner mit zugeordneter Datensatzbeschreibung
    // dann Steuerparameter auf Buchung fixieren.
    if ( size() && !activateSteuerparam )
        m_menuItem.flags |= 5;

    // Anzahl Unterpunkte berechnen und eintragen
    Unterpunkte = (int)size();
    if ( Unterpunkte > MAX_SUBPOINTS )
        Unterpunkte = MAX_SUBPOINTS;

    m_menuItem.numberOfEntries = Unterpunkte;

    // MenueEintrag.ptr berechnen und eintragen
    m_menuItem.ptr = pFlash + sizeof( StructMenuItem );
    m_menuItem.ptr = endianReverse( m_menuItem.ptr + FLASH_KONFIG_START );

    // Daten von MenueEintrag eintragen
    StructMenuItem &menu_eintrag = *reinterpret_cast<StructMenuItem *>(&file.m_flash[pFlash]);
    menu_eintrag = m_menuItem;

#ifdef DATAFOX_STUDIO_IV
    // Verschlüsseln des Passworts
    if ( root().isPasswordProtection() )
    {
        df4_crypt_internal df4;
        df4.crypt( reinterpret_cast<unsigned char *>(&menu_eintrag.password), sizeof( menu_eintrag.password ) );
    }
#endif

    pFlash += sizeof( StructMenuItem );


    ////////////////////////////////////////
    // add information to xml file
    FileXML &fileXML = doc.fileXML();

    // The parent is configured outside the method and is not changed.
    fileXML.writeContent( FileXML::ContentType::MenuItem, as_const( menu_eintrag ) );

    // erster freier Speicherplatz nach MenueEintrag -> begin Pointerarray
    aP[0] = pFlash + sizeof( unsigned int ) * Unterpunkte; // Platz für Pointerarray reservieren

    // Finding the node for the child structures and update parent.
    auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::MenuItem ) );

    // Alle Unterpunkte in Flash eintragen und die jeweiligen Pointer sichern
    iterChild = constBegin();
    for ( i = 1; iterChild != constEnd() && (i <= MAX_SUBPOINTS); ++iterChild, ++i )
        aP[i] = (*iterChild)->writeDocument( aP[i - 1], doc );

    // Reset parent node because the end of the structure is reached.
    fileXML.setParentNode( oldParent );

    // Pointer auf Unterpunkte eintragen
    for ( i = 0; i < Unterpunkte; ++i )
    {
        aP[i] = endianReverse( aP[i] + FLASH_KONFIG_START );
        *reinterpret_cast<unsigned int *>(&file.m_flash[pFlash + i * sizeof( unsigned int )]) = aP[i];
    }

    return aP[Unterpunkte];
}

uint32_t EntryMenuChoice::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    int i{};
    unsigned int pDat{};
    StructOperating operating{};
    tinyxml2::XMLElement *lastXmlElement{};

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    // Daten von MenueEintrag einlesen
    if ( useAESFile )
    {
        m_menuItem = *(StructMenuItem *)&file.m_flash[pFlash];
        m_menuItem.ptr = endianReverse( m_menuItem.ptr ) - FLASH_KONFIG_START;
    }
    else
    {
        fileXML.combineValid( fileXML.readContent( FileXML::ContentType::MenuItem, m_menuItem, pFlash, lastXmlElement ) );
    }

#ifdef DATAFOX_STUDIO_IV
    // Entschlüsseln des Passworts
    if ( root().isPasswordProtection() )
    {
        df4_crypt_internal df4;
        df4.crypt( reinterpret_cast<unsigned char *>(&m_menuItem.password), sizeof( m_menuItem.password ) );
    }
#endif

    pFlash += sizeof( StructMenuItem );

    // Finding the node for the child structures and update parent.
    auto oldParent = fileXML.setParentNode( lastXmlElement );

    // Daten der Aktion einlesen
    // Bedien einlesen
    tinyxml2::XMLElement *subItem{};
    for ( i = 0; i < (int)m_menuItem.numberOfEntries; ++i )
    {
        // Bedien einlesen
        if ( useAESFile )
        {
            pDat = pFlash + i * sizeof( unsigned int );	// Pointer des Eintrags berechnen
            pDat = *(unsigned int *)(&file.m_flash[pDat]);
            pDat = endianReverse( pDat ) - FLASH_KONFIG_START;

            operating = *(StructOperating *)&file.m_flash[pDat];
        }
        else
        {
            fileXML.combineValid( fileXML.readContent( FileXML::ContentType::Operating, operating, subItem ) );
        }

        switch ( operating.type )
        {
            case TB_MENUE:
            {
                EntryMenu *pNeu = new EntryMenu( this );
                pNeu->setDefault();
                pNeu->readDocument( useAESFile ? pDat : i, doc );
                append( pNeu );
                break;
            }

            case TB_DATENSATZ:
            {
                EntryInputChain *pNeu = new EntryInputChain( this );
                pNeu->setDefault();
                pNeu->readDocument( useAESFile ? pDat : i, doc );
                append( pNeu );
                break;
            }

            default:
                BOOST_ASSERT( false );
                break;
        }
    }

    // Reset parent node because the end of the structure is reached.
    fileXML.setParentNode( oldParent );

    // Name für Baum eintragen
    setName( m_menuItem.name );

    return pFlash;
}

///////////////////////////////////
// Daten des Eintrags

void EntryMenuChoice::setDefault()
{
    setDefault( QT_TRANSLATE_NOOP_UTF8( "EntryMenuChoice", "Menueeintrag" ) );
}

void EntryMenuChoice::setDefault( const std::string &name )
{
    setName( name );

    m_iconIndex = ICON_MENU_CHOICE;
    m_iconIndexSelected = ICON_MENU_CHOICE_SEL;

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_MENU_CHOICE_0 );

    strncpy( (char *)m_menuItem.name, this->name().c_str(), sizeof( m_menuItem.name ) - 1 );
    m_menuItem.password = 0;
    m_menuItem.flags = 0;

    for ( size_t i = 0; i < sizeof( m_menuItem.unused ); ++i )
    {
        m_menuItem.unused[i] = 0;
    }

    setInsertEnabled( TRUE );
    setDeleteEnabled( TRUE );
}


// ########################################################
// Member: EntryInputChainField

EntryInputChainField::EntryInputChainField( const EntryInputChainField &entry )
    : EntryBase( entry )
{
    // Member uebernehmen
    m_textEx = entry.m_textEx;
    m_confirmationEx = entry.m_confirmationEx;
    m_numFieldEx = entry.m_numFieldEx;
    m_listSelectionEx = entry.m_listSelectionEx;
    m_dataEx = entry.m_dataEx;
    m_globVarToHIDEx = entry.m_globVarToHIDEx;
    m_switchModuleEnergieEx = entry.m_switchModuleEnergieEx;
    m_systemMessageEx = entry.m_systemMessageEx;
    m_takePhotoEx = entry.m_takePhotoEx;
    m_feverControlEx = entry.m_feverControlEx;
    m_multiInputSourceEx = entry.m_multiInputSourceEx;
    m_readField = entry.m_readField;
    
    m_multiInputChildType = entry.m_multiInputChildType;
    m_pageIndex = entry.m_pageIndex;

    m_refTableField = entry.m_refTableField;
    // Bezugszaehler erhoehen
    if ( this->m_refTableField )
    {
        this->m_refTableField->incRefCount();
    }

    m_refTableFieldIntegral = entry.m_refTableFieldIntegral;
    // Bezugszaehler erhoehen
    if ( this->m_refTableFieldIntegral )
    {
        this->m_refTableFieldIntegral->incRefCount();
    }

    m_refJumpESC = entry.m_refJumpESC;
    // m_refJumpESC -> Kann erst nach vollstaendigem Kopieren der Eingabekette ermittelt/geprueft werden.
    m_refJump = entry.m_refJump;
    // m_refJump    -> Kann erst nach vollstaendigem Kopieren der Eingabekette ermittelt/geprueft werden.
    m_refJumpSelNoEntry = entry.m_refJumpSelNoEntry;
    // m_refJumpSelNoEntry    -> Kann erst nach vollstaendigem Kopieren der Eingabekette ermittelt/geprueft werden.

    // Feldmethoden "Aus Liste waehlen" und "In Liste schreiben"
    // Referenzen uebernehmen
    if ( fieldFunction() == TBFE_LIST_SELECT_SET || fieldFunction() == TBFE_LIST_SELECT_GET || fieldFunction() == TBFE_DATA_ON_CARD )
    {
        m_refList = entry.m_refList;
        if ( m_refList )
        {
            m_refList->incRefCount();

            // Feld aus dem ausgewaehlt wird
            m_refAuswahlFeld = entry.m_refAuswahlFeld;
            if ( m_refAuswahlFeld )
                m_refAuswahlFeld->incRefCount();

            // Feld das abgespeichert wird
            m_refDatFeld = entry.m_refDatFeld;
            if ( m_refDatFeld )
                m_refDatFeld->incRefCount();

            // Erstes Selektionsfeld der Datensatzbeschreibung
            m_refSelDFeld = entry.m_refSelDFeld;
            if ( isValidSelRef( m_refSelDFeld ) )
                m_refSelDFeld->incRefCount();

            // Erstes Selektionsfeld der Liste
            m_refSelLFeld = entry.m_refSelLFeld;
            if ( m_refSelLFeld )
                m_refSelLFeld->incRefCount();

            // Zweites Selektionsfeld der Datensatzbeschreibung
            m_refSelDFeld2 = entry.m_refSelDFeld2;
            if ( isValidSelRef( m_refSelDFeld2 ) )
                m_refSelDFeld2->incRefCount();

            // Zweites Selektionsfeld der Liste
            m_refSelLFeld2 = entry.m_refSelLFeld2;
            if ( m_refSelLFeld2 )
                m_refSelLFeld2->incRefCount();
        }
    }
    else
    {
        // Referenz fuer Listenzugriff
        m_refList = nullptr;        // Liste aus der ausgewaehlt wird
        m_refAuswahlFeld = nullptr; // Feld aus dem ausgewaehlt wird
        m_refDatFeld = nullptr;     // Feld das abgespeichert wird
        m_refSelDFeld = nullptr;    // Erstes Selektionsfeld der Datensatzbeschreibung
        m_refSelLFeld = nullptr;    // Erstes Selektionsfeld der Liste
        m_refSelDFeld2 = nullptr;   // Zweites Selektionsfeld der Datensatzbeschreibung
        m_refSelLFeld2 = nullptr;   // Zweites Selektionsfeld der Liste
    }


    // Bei einer Kopie das Referenzfeld mitkopieren.
    if ( fieldFunction() == TBFE_FEVER_CONTROL )
    {
        m_refFeverControlStatusField = entry.m_refFeverControlStatusField;
        // Bezugszaehler erhoehen
        if ( this->m_refFeverControlStatusField )
        {
            this->m_refFeverControlStatusField->incRefCount();
        }
    }

    // Bei einer Kopie die Unterfunktionen mitkopieren.
    if ( fieldFunction() == TBFE_MULTI_INPUT_SOURCE )
    {
        m_refMultiInputSourceField = entry.m_refMultiInputSourceField;
        // Bezugszaehler erhoehen
        if ( this->m_refMultiInputSourceField )
        {
            this->m_refMultiInputSourceField->incRefCount();
        }

        EntryInputChainField *childCopy;
        // Alle Childs kopieren
        auto iterChild = entry.constBegin();
        for ( ; iterChild != entry.constEnd(); ++iterChild )
        {
            // neues anfuegen
            childCopy = (EntryInputChainField *)((EntryInputChainField *)*iterChild)->cloneEntry();
            childCopy->setParent( this );
            append( childCopy );
        }
    }
}

///////////////////////////////////
// Childs

EntryBase *EntryInputChainField::insertEntry( EntryType typeID, int pos )
{
    (void)typeID;
    (void)pos;

    return nullptr;
}

EntryBase *EntryInputChainField::cloneEntry()
{
    return (EntryBase *)new EntryInputChainField( *this );
}

EntryInputChainField::MultiInputChildType EntryInputChainField::multiInputChildType()
{
    return m_multiInputChildType;
}

void EntryInputChainField::setMultiInputChildType( MultiInputChildType type )
{
    m_multiInputChildType = type;
}

EntryInputChainField *EntryInputChainField::insertMultiInputChild( MultiInputChildType type )
{
    // Bei einer MultiInputSource werden Untereinträge für die Einzelnen Feldfunktionen benötigt.
    if ( fieldFunction() == TBFE_MULTI_INPUT_SOURCE )
    {
        EntryInputChainField *field = multiInputChild( type );
        if ( field )
        {
            // Darf nur einmal als Child vorhanden sein.
            return field;
        }

        switch ( type )
        {
            case MultiInputChildType::Keypad:
                // Eintrag für Tastatur.
                field = new EntryInputChainField( this );
                field->setDefault( QT_TRANSLATE_NOOP_UTF8( "EntryInputChainField", "Tastatur" ) );
                field->setMultiInputChildType( MultiInputChildType::Keypad );
                field->setIconIndexes( ICON_INPUT_CHAIN_FIELD_DARK, ICON_INPUT_CHAIN_FIELD_DARK_SEL );
                field->setFieldFunction( TBFE_NORM );
                field->m_numFieldEx.flags = NFF_TASTATUR;
                insertEntryFinish( field, 0 );
                m_multiInputSourceEx.keypadInput = true;
                break;

            case MultiInputChildType::Barcode:
            {
                // Eintrag für Barcode.
                field = new EntryInputChainField( this );
                field->setDefault( QT_TRANSLATE_NOOP_UTF8( "EntryInputChainField", "Barcode" ) );
                field->setMultiInputChildType( MultiInputChildType::Barcode );
                field->setIconIndexes( ICON_INPUT_CHAIN_FIELD_DARK, ICON_INPUT_CHAIN_FIELD_DARK_SEL );
                field->setFieldFunction( TBFE_NORM );
                field->m_numFieldEx.flags = NFF_BARCODE;
                int pos = 0;
                if ( multiInputChild( MultiInputChildType::Keypad ) )
                {
                    ++pos;
                }
                insertEntryFinish( field, pos );
                m_multiInputSourceEx.barcodeInput = true;
                break;
            }

            case MultiInputChildType::Tranponder:
            {
                // Eintrag für Transponder.
                field = new EntryInputChainField( this );
                field->setDefault( QT_TRANSLATE_NOOP_UTF8( "EntryInputChainField", "Transponder" ) );
                field->setMultiInputChildType( MultiInputChildType::Tranponder );
                field->setIconIndexes( ICON_INPUT_CHAIN_FIELD_DARK, ICON_INPUT_CHAIN_FIELD_DARK_SEL );
                field->setFieldFunction( TBFE_NORM );
                field->m_numFieldEx.flags = NFF_TRANSPONDER;
                int pos = 0;
                if ( multiInputChild( MultiInputChildType::Keypad ) )
                {
                    ++pos;
                }

                if ( multiInputChild( MultiInputChildType::Barcode ) )
                {
                    ++pos;
                }

                insertEntryFinish( field, pos );
                m_multiInputSourceEx.transponderInput = true;
                break;
            }

            case MultiInputChildType::FingerprintIdentification:
                // Eintrag für Fingerprint Identifikation.
                field = new EntryInputChainField( this );
                field->setDefault( QT_TRANSLATE_NOOP_UTF8( "EntryInputChainField", "Fingerprint Id." ) );
                field->setMultiInputChildType( MultiInputChildType::FingerprintIdentification );
                field->setIconIndexes( ICON_INPUT_CHAIN_FIELD_DARK, ICON_INPUT_CHAIN_FIELD_DARK_SEL );
                field->setFieldFunction( TBFE_FINGERPRINT_IDENTIFICATION );
                insertEntryFinish( field, -1 );
                m_multiInputSourceEx.fingerprintIdentificationInput = true;
                break;

            case MultiInputChildType::FingerprintVerification:
                // Eintrag für Fingerprint Verifikation.
                field = new EntryInputChainField( this );
                field->setDefault( QT_TRANSLATE_NOOP_UTF8( "EntryInputChainField", "Fingerprint Ve." ) );
                field->setMultiInputChildType( MultiInputChildType::FingerprintVerification );
                field->setIconIndexes( ICON_INPUT_CHAIN_FIELD_DARK, ICON_INPUT_CHAIN_FIELD_DARK_SEL );
                field->setFieldFunction( TBFE_FINGERPRINT_VERIFICATION );
                insertEntryFinish( field, -1 );
                m_multiInputSourceEx.fingerprintVerificationInput = true;
                break;
        }

        return field;
    }

    return nullptr;
}

void EntryInputChainField::removeMultiInputChild( MultiInputChildType type )
{
        // Bei einer MultiInputSource werden Untereinträge für die Einzelnen Feldfunktionen benötigt.
    if ( fieldFunction() == TBFE_MULTI_INPUT_SOURCE )
    {
        EntryInputChainField *field = multiInputChild( type );
        if ( field )
        {
            removeOne( field );

            switch ( type )
            {
                case MultiInputChildType::Keypad:
                    m_multiInputSourceEx.keypadInput = false;
                    break;

                case MultiInputChildType::Barcode:
                    m_multiInputSourceEx.barcodeInput = false;
                    break;

                case MultiInputChildType::Tranponder:
                    m_multiInputSourceEx.transponderInput = false;
                    break;

                case MultiInputChildType::FingerprintIdentification:
                    m_multiInputSourceEx.fingerprintIdentificationInput = false;
                    break;

                case MultiInputChildType::FingerprintVerification:
                    m_multiInputSourceEx.fingerprintVerificationInput = false;
                    break;
            }
        }
    }
}

EntryInputChainField *EntryInputChainField::multiInputChild( MultiInputChildType type )
{
    for ( std::size_t i = 0; i < size(); ++i )
    {
        EntryBase *base = value( i );
        if ( base->isType( EntryType::InputChainField ) )
        {
            EntryInputChainField *field = dynamic_cast<EntryInputChainField *>(base);
            if ( field->multiInputChildType() == type )
            {
                return field;
            }
        }
    }

    return nullptr;
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryInputChainField::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    StructCompare compare{};

    // Haupteintrag von Baum ermitteln.
    EntryDevice &device = root();

    uint8_t transponderConfigCount = root().transponderConfigCount();
    if ( transponderConfigCount )
    {
        switch ( root().m_config.transponderType & 0x3F )
        {
            case TRANSPONDER_HITAG1:
            case TRANSPONDER_HITAG1_SECURITY:
            case TRANSPONDER_HITAG2:
            case TRANSPONDER_HITAGS_H32:
            case TRANSPONDER_HITAGS_H48:
            case TRANSPONDER_HITAGS_H56:
            case TRANSPONDER_TITAN:
                if ( (m_numFieldEx.tag.tpMode1 & BDTP_ACTIVE) != BDTP_ACTIVE )
                {
                    // Transponderkonfiguration für EK-Feld
                    memset( &m_numFieldEx.tag, 0, sizeof( StructTransponder ) );
                    m_numFieldEx.tag.tpSegment1 = 255; // Kein
                    m_numFieldEx.tag.tpSegment2 = 255;
                    m_numFieldEx.tag.tpSegment3 = 255;
                }
                break;

            case TRANSPONDER_MIFARE:
            {
                StructTransponderMifare tag = *(StructTransponderMifare *)&m_numFieldEx.tag;
                if ( (tag.mode & BDTP_ACTIVE) != BDTP_ACTIVE )
                {
                    // Transponderkonfiguration für EK-Feld, Strukturen StructTransponderMifare entspricht im inneren Aufbau der von StructTransponder.
                    memset( &m_numFieldEx.tag, 0, sizeof( StructTransponder ) );
                    m_numFieldEx.tag.tpSegment1 = 255; // Kein
                    m_numFieldEx.tag.tpSegment2 = 255;
                    m_numFieldEx.tag.tpSegment3 = 255;
                }
                break;
            }

            default:
                if ( root().hasTranponderMultiConfigs() )
                {
                    // Gültigen Bereich Sicherstellen.
                    if ( m_numFieldEx.tag.tpSegment1 == 255
                         || m_numFieldEx.tag.tpSegment1 >= transponderConfigCount )
                    {
                        // Transponderkonfiguration für EK-Feld
                        memset( &m_numFieldEx.tag, 0, sizeof( StructTransponder ) );
                        m_numFieldEx.tag.tpSegment1 = 0; // Standard
                        m_numFieldEx.tag.tpSegment2 = 255;
                        m_numFieldEx.tag.tpSegment3 = 255;
                    }
                }
                else
                {
                    // Transponderkonfiguration für EK-Feld
                    memset( &m_numFieldEx.tag, 0, sizeof( StructTransponder ) );
                    m_numFieldEx.tag.tpSegment1 = 255; // Kein
                    m_numFieldEx.tag.tpSegment2 = 255;
                    m_numFieldEx.tag.tpSegment3 = 255;
                }
                break;
        }

        // Bei Listenfunktionen und DataOnCard
        if ( ((m_listSelectionEx.transponderConfig & 0xF0) >> 4) >= transponderConfigCount 
             || (m_listSelectionEx.transponderConfig & 0x0F) >= transponderConfigCount )
        {
            m_listSelectionEx.transponderConfig = 0;
        }
    }
    else // Der gewählte Transponder unterstützt keine überlagerte Konfiguration.
    {
        // Transponderkonfiguration für EK-Feld
        memset( &m_numFieldEx.tag, 0, sizeof( StructTransponder ) );
        m_numFieldEx.tag.tpSegment1 = 255; // Kein
        m_numFieldEx.tag.tpSegment2 = 255;
        m_numFieldEx.tag.tpSegment3 = 255;

        // Bei Listenfunktionen und DataOnCard
        m_listSelectionEx.transponderConfig = 0;
    }

    // Ausschneidefunktion nur bei diesen Feldfunktionen zulassen.
    if ( fieldFunction() != TBFE_NORM
         && fieldFunction() != TBFE_COPY_GV_TO_FIELD
         && fieldFunction() != TBFE_TIMEBOY_COPY_GV_TO_FIELD
         && fieldFunction() != TBFE_COPY_FIELD_TO_GV
         && fieldFunction() != TBFE_TIMEBOY_COPY_FIELD_TO_GV
         && fieldFunction() != TBFE_LIST_SELECT_GET
         && fieldFunction() != TBFE_FINGERPRINT_IDENTIFICATION
         && fieldFunction() != TBFE_FINGERPRINT_VERIFICATION
         && fieldFunction() != TBFE_DEBUG )
    {
        // Zuschneidefunktion deaktivieren.
        m_readField.barcodeRangeFrom = 0;
        m_readField.barcodeRangeUpTo = 0;
    }

    // Daten fuer union je nach "Einlesemethode"
    switch ( fieldFunction() )
    {
        case TBFE_NORM:
        case TBFE_GET_TIME:
        case TBFE_KONSTANTE:
        case TBFE_COPY_GV_TO_FIELD:
        case TBFE_TIMEBOY_COPY_GV_TO_FIELD:
        case TBFE_SERIENNUMMER:
        case TBFE_ANALOG_WERT:
        case TBFE_ANALOG_STATUS:
        case TBFE_DIGITAL_WERT:
        case TBFE_ZAEHLER_WERT:
        case TBFE_FUNC:
        case TBFE_SPS_WERT:
        case TBFE_ENTRANCE_ZM_NUM:
        case TBFE_ENTRANCE_READER_NUM:
        case TBFE_ENTRANCE_PNR:
        case TBFE_ENTRANCE_STATUS:
        case TBFE_DEBUG:
        case TBFE_GPRS_ALIVE_COUNTER:
        case TBFE_FIRMWARE_VERSION:
        case TBFE_ONLINE_STATUS:
        case TBFE_SUMMER_WINTER_STATUS:
        case TBFE_GPS_RECORD_FIX:
        case TBFE_GPS_RECORD_VARIABLE:
        case TBFE_ENTRANCE2_IDENTIFICATION:
        case TBFE_COPY_FIELD_TO_GV:
        case TBFE_TIMEBOY_COPY_FIELD_TO_GV:
        case TBFE_FINGERPRINT_SCANN:
        case TBFE_FINGERPRINT_CARD:
        case TBFE_FINGERPRINT_TEACH:
        case TBFE_FINGERPRINT_IDENTIFICATION:
        case TBFE_FINGERPRINT_VERIFICATION:
        case TBFE_FINGERPRINT_DELETE:
        case TBFE_OPERATION:
        case TBFE_SWITCH_RELAY:
        case TBFE_SWITCH_LED:
        case TBFE_TIMER:
        case TBFE_TRANSPONDER_WRITE:
        case TBFE_GPS_DISTANCE:
        case TBFE_MOBILE_PARAMETER:
        case TBFE_SWITCH_BUZZER:
        case TBFE_KONSTANTE_TYPE_DEVICE_NAME:
        case TBFE_MOBILMASTERIV_MENUENTRY:
        case TBFE_SWITCH_COMMUNICATION:
        case TBFE_FINGERPRINT_ADMNISTARTION:

            m_readField.ex.numField = m_numFieldEx;
            compare = m_numFieldEx.compare;

            if ( (fieldFunction() == TBFE_GPS_RECORD_FIX ||
                   fieldFunction() == TBFE_GPS_RECORD_VARIABLE) && m_numFieldEx.relayTime != 0 )
            {
                // Wartezeit auf gueltige GPS Koordinaten.
                file.NeedStudioVersion( "04.02.03.00" );
                file.NeedFirmware2Version( "04.02.03.01" );
                file.UnsupportFirmware1Version();
            }
            if ( fieldFunction() == TBFE_MOBILMASTERIV_MENUENTRY )
            {
                // Anzeige eines Menue
                file.NeedStudioVersion( "04.02.04.03" );
                file.NeedFirmware2Version( "04.02.04.10" );
                file.UnsupportFirmware1Version();
            }
            if ( fieldFunction() == TBFE_SWITCH_COMMUNICATION )
            {
                // Umschalten der Hauptkommunikation per Feldfunktion-
                file.NeedStudioVersion( "04.02.04.03" );
                file.NeedFirmware2Version( "04.02.04.10" );
                file.UnsupportFirmware1Version();
            }

            if ( fieldFunction() == TBFE_ANALOG_WERT )
            {
                // Integralwert uebernehmen?
                if ( (m_readField.variable2 & 0x80) != 0 )
                {
                    // Wird erst ab dieser Version unterstuetzt.
                    file.NeedStudioVersion( "04.02.04.03" );
                    file.NeedFirmware2Version( "04.02.04.10" );
                    file.UnsupportFirmware1Version();
                }
            }
            if ( fieldFunction() == TBFE_FINGERPRINT_ADMNISTARTION )
            {
                file.NeedStudioVersion( "04.03.10.10" );
                file.NeedFirmware3Version( "04.03.10.07" );
                file.UnsupportFirmware2Version();
                file.UnsupportFirmware1Version();
            }
            break;

        case TBFE_LIST_SELECT_GET:
        case TBFE_LIST_SELECT_SET:
        case TBFE_DATA_ON_CARD:

            m_readField.ex.listSelection = m_listSelectionEx;
            compare = m_listSelectionEx.compare;

            if ( m_listSelectionEx.selectionNoEntry == 4 )
            {
                // Wird erst ab dieser Version unterstuetzt.
                file.NeedStudioVersion( "04.03.05.04" );
                file.NeedFirmware3Version( "04.03.05.14" );
                file.UnsupportFirmware2Version();
                file.UnsupportFirmware1Version();
            }

            // Mindestversion bei DataOnCard haengt an der gewaehtlen Option unter den Grundeinstellungen.
            break;

        case TBFE_BESTAETIGUNG:

            m_readField.ex.confirmation = m_confirmationEx;
            break;

        case TBFE_SEND_SMS:
        case TBFE_SCRIPT:

            m_readField.ex.text = m_textEx;
            // Pruefung auf benoetigte Mindestversion
            file.NeedStudioVersion( "04.02.03.02" );
            file.NeedFirmware2Version( "04.02.03.01" );
            file.UnsupportFirmware1Version();

            // BOt: 27.01.2016
            // Zusätzlich wird bei folgendem Inhalt die Version 04.03.06.xx vorrausgesetzt.
            // Änderung ist für den Fahrzeugdatenlogger V2 Kompatibilität mit V1 notwendig,
            // hierdurch wird eine sonst benötigte Liste eingespart.
            if ( fieldFunction() == TBFE_SCRIPT && strcmp( m_textEx.text, "fdlflags" ) == 0 )
            {
                file.NeedStudioVersion( "04.03.06.10" );
                file.NeedFirmware3Version( "04.03.06.28" );
                file.UnsupportFirmware2Version();
            }
            break;

        case TBFE_SHOW_MESSAGE:

            m_readField.ex.text = m_textEx;
            // Pruefung auf benoetigte Mindestversion
            file.NeedStudioVersion( "04.03.05.07" );
            file.NeedFirmware3Version( "04.03.05.31" );
            file.UnsupportFirmware2Version();
            file.UnsupportFirmware1Version();
            break;

        case TBFE_SWITCH_BACKLIGHT:

            m_readField.ex.data = m_dataEx;
            if ( device.hasSetupType( { SetupType::PZE_MasterIV, SetupType::AE_MasterIV } ) )
            {
                // Pruefung auf benoetigte Mindestversion
                file.NeedStudioVersion( "04.03.15.04" );
                file.ShouldFirmware3Version( "04.03.15.01" ); // Bei PZE- und AE-MasterIV erst ab dieser Version unterstützt.
                file.NeedFirmware2Version( "04.02.05.00" ); // Hier gibt es aufgrund der unbekannten Feldfunktion keinen Absturz mehr.
                file.UnsupportFirmware1Version();
            }
            else
            {
                // Pruefung auf benoetigte Mindestversion
                file.NeedStudioVersion( "04.03.00.08" );
                file.ShouldFirmware3Version( "04.03.00.25" );
                file.NeedFirmware2Version( "04.02.05.00" ); // Hier gibt es aufgrund der unbekannten Feldfunktion keinen Absturz mehr.
                file.UnsupportFirmware1Version();
            }
            break;

        case TBFE_GV_TO_INTERFACE:

            m_readField.ex.globVarToHID = m_globVarToHIDEx;
            // Pruefung auf benoetigte Mindestversion
            file.NeedStudioVersion( "04.03.00.08" );
            file.ShouldFirmware3Version( "04.03.00.25" );
            file.NeedFirmware2Version( "04.02.05.00" );
            file.UnsupportFirmware1Version();
            break;

        case TBFE_SWITCH_MODULE_ENERGIE:

            m_readField.ex.switchModuleEnergie = m_switchModuleEnergieEx;
            // Pruefung auf benoetigte Mindestversion
            file.NeedStudioVersion( "04.03.07.10" );
            file.NeedFirmware2Version( "04.02.05.00" );
            file.UnsupportFirmware1Version();
            break;

        case TBFE_SYSTEM_MESSAGE:

            m_readField.ex.systemMessage = m_systemMessageEx;
            // Pruefung auf benoetigte Mindestversion
            file.NeedStudioVersion( "04.03.09.00" );
            file.NeedFirmware2Version( "04.02.05.00" );
            file.UnsupportFirmware1Version();
            break;

        case TBFE_TIMER_STATE:
            m_readField.ex.numField = m_numFieldEx;
           // Pruefung auf benoetigte Mindestversion
            file.NeedStudioVersion( "04.03.09.09" );
            file.NeedFirmware2Version( "04.02.05.50" );
            file.NeedFirmware3Version( "04.03.09.05" );
            file.UnsupportFirmware1Version();
            break;

        case TBFE_TAKE_PHOTO:

            m_readField.ex.takePhoto = m_takePhotoEx;
            // Pruefung auf benoetigte Mindestversion
            file.NeedStudioVersion( "04.03.09.00" );
            file.NeedFirmware2Version( "04.02.05.00" );
            file.UnsupportFirmware1Version();
            break;

        case TBFE_VIBRATION_VALUE:
            m_readField.ex.numField = m_numFieldEx;
            // Pruefung auf benoetigte Mindestversion
            file.NeedStudioVersion( "04.03.11.00" );
            file.NeedFirmware2Version( "04.02.05.50" );
            file.NeedFirmware3Version( "04.03.11.00" );
            file.UnsupportFirmware1Version();
            break;

        case TBFE_READ_VARIABLE:
        case TBFE_WRITE_VARIABLE:

            m_readField.ex.text = m_textEx;
            // Pruefung auf benoetigte Mindestversion
            file.NeedStudioVersion( "04.03.12.00" );
            file.NeedFirmware3Version( "04.03.12.00" );
            file.UnsupportFirmware2Version();
            file.UnsupportFirmware1Version();
            break;

        case TBFE_SIGNATURE:
        {
            // Zusätzliche Funktionalität deaktivieren.
            m_readField.inputTypeCount = 0;
            m_readField.variable1 = 0;
            m_readField.variable2 = 0;

            m_readField.ex.text = m_textEx;
            // Pruefung auf benoetigte Mindestversion
            file.NeedStudioVersion( "04.03.12.11" );
            file.NeedFirmware3Version( "04.03.12.06" );
            file.UnsupportFirmware2Version();
            file.UnsupportFirmware1Version();
            break;
        }

        case TBFE_FEVER_CONTROL:

            m_readField.ex.feverControl = m_feverControlEx;
            // Pruefung auf benoetigte Mindestversion
            file.NeedStudioVersion( "04.03.15.08" );
            file.NeedFirmware3Version( "04.03.15.08" );
            file.UnsupportFirmware2Version();
            file.UnsupportFirmware1Version();
            break;

        case TBFE_MULTI_INPUT_SOURCE:
        {
            uint8_t childCount = 0;
            if ( m_multiInputSourceEx.keypadInput || multiInputChild( MultiInputChildType::Keypad ) )
            {
                BOOST_ASSERT( m_multiInputSourceEx.keypadInput && multiInputChild( MultiInputChildType::Keypad ) );
                ++childCount;
            }
            if ( m_multiInputSourceEx.barcodeInput || multiInputChild( MultiInputChildType::Barcode ) )
            {
                BOOST_ASSERT( m_multiInputSourceEx.barcodeInput && multiInputChild( MultiInputChildType::Barcode ) );
                ++childCount;
            }
            if ( m_multiInputSourceEx.transponderInput || multiInputChild( MultiInputChildType::Tranponder ) )
            {
                BOOST_ASSERT( m_multiInputSourceEx.transponderInput && multiInputChild( MultiInputChildType::Tranponder ) );
                ++childCount;
            }
            if ( (m_multiInputSourceEx.fingerprintIdentificationInput || multiInputChild( MultiInputChildType::FingerprintIdentification ))
                 || (m_multiInputSourceEx.fingerprintVerificationInput || multiInputChild( MultiInputChildType::FingerprintVerification )) )
            {
                BOOST_ASSERT( (m_multiInputSourceEx.fingerprintIdentificationInput && multiInputChild( MultiInputChildType::FingerprintIdentification ))
                              || (m_multiInputSourceEx.fingerprintVerificationInput && multiInputChild( MultiInputChildType::FingerprintVerification )) );
                ++childCount;
            }
            BOOST_ASSERT( childCount == size() );

            m_multiInputSourceEx.childCount = (uint8_t)size();
            m_multiInputSourceEx.ptr = pFlash + sizeof( StructReadField );
            m_multiInputSourceEx.ptr += m_flashOffset;
            m_multiInputSourceEx.ptr = endianReverse( m_multiInputSourceEx.ptr + FLASH_KONFIG_START );

            m_readField.ex.multiInputSource = m_multiInputSourceEx;
            // Pruefung auf benoetigte Mindestversion
            file.NeedStudioVersion( "04.03.18.00" );
            file.NeedFirmware3Version( "04.03.18.00" );
            file.UnsupportFirmware2Version();
            file.UnsupportFirmware1Version();
            break;
        }

        case TBFE_LONG_BARCODE:
        {
            // Zusätzliche Funktionalität deaktivieren.
            m_readField.inputTypeCount = 0;
            m_readField.variable1 = 0;
            m_readField.variable2 = 0;

            m_readField.ex.text = m_textEx;
            // Pruefung auf benoetigte Mindestversion
            file.NeedStudioVersion( "04.03.18.00" );
            file.NeedFirmware3Version( "04.03.18.00" );
            file.UnsupportFirmware2Version();
            file.UnsupportFirmware1Version();
            break;
        }

        default:
            BOOST_ASSERT( false );
    }

    // Daten von FeldEinlesen - Struktur uebernehmen
    *(StructReadField *)&file.m_flash[pFlash] = m_readField;
    pFlash += sizeof( StructReadField );

    ////////////////////////////////////////
    // add information to xml file
    FileXML &fileXML = doc.fileXML();
    fileXML.writeContent( FileXML::ContentType::ReadField, as_const( m_readField ) );

    // Pruefung auf benoetigte Mindestversion
    if ( ((compare.jumpDestination >= 112 && compare.jumpDestination <= 127) || (compare.jumpDestination >= 200 && compare.jumpDestination <= 229) || compare.jumpDestination == 254) ||
        ((m_readField.behaviourAtEsc >= 112 && m_readField.behaviourAtEsc <= 127) || (m_readField.behaviourAtEsc >= 200 && m_readField.behaviourAtEsc <= 229) || m_readField.behaviourAtEsc == 254) )
    {
        // ESC-Verhalten bei E-Menue wird erst ab Versionen x unterstuetzt.
        file.NeedStudioVersion( "04.02.02.08" );
        file.NeedFirmware2Version( "04.02.02.17" );
        file.UnsupportFirmware1Version();
    }
    // Pruefung auf benoetigte Mindestversion
    if ( (m_readField.systemVariable & 0x0F) != 0 )
    {
        // Bit 1: BC oder TP sofort Einschalten.
        // Bit 2: BC oder TP Eingeschaltet Akustisch signalisieren.
        // Bit 3,4: Vorbelegung des Wertes aus Feld oder GV.
        file.NeedStudioVersion( "04.02.03.02" );
        file.NeedFirmware2Version( "04.02.03.01" );
        file.UnsupportFirmware1Version();
    }
    // Erweiterte Ausschneidefunktion
    if ( (m_readField.barcodeRangeFrom == 0 && m_readField.barcodeRangeUpTo != 0) ||
        (m_readField.barcodeRangeFrom == 0xFF && m_readField.barcodeRangeUpTo != 0) )
    {
        // Ausschneiden von Links und Rechts wird erst ab Versionen x unterstuetzt.
        file.NeedStudioVersion( "04.02.04.03" );
        file.NeedFirmware2Version( "04.02.04.10" );
        file.UnsupportFirmware1Version();
    }

    // Untereinträge eintragen
    if ( size() )
    {
        uint8_t childNumber = 0;
        uint32_t aP[MAX_SUBPOINTS];
        // Alle Childeintraege ablegen
        // erster freier Speicherplatz nach DatensatzEinlesen -> begin Pointerarray
        aP[childNumber] = pFlash + sizeof( uint32_t ) * size(); // Platz fuer Pointerarray reservieren

        auto iterChild = constBegin();
        for ( childNumber = 1; iterChild != constEnd(); ++iterChild, ++childNumber )
        {
            aP[childNumber] = (*iterChild)->writeDocument( aP[childNumber - 1], doc );
        }

        // Offsets auf Unterpunkte eintragen
        for ( childNumber = 0; childNumber < size(); ++childNumber )
        {
            aP[childNumber] += m_flashOffset;
            aP[childNumber] = endianReverse( aP[childNumber] + FLASH_KONFIG_START );
            *(uint32_t *)&file.m_flash[pFlash + childNumber * sizeof( uint32_t )] = aP[childNumber];
        }

        pFlash = aP[size()];
    }

    return pFlash;
}

uint32_t EntryInputChainField::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    tinyxml2::XMLElement *lastXmlElement{};

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    // Daten von FeldEinlesen einlesen
    if ( useAESFile )
    {
        m_readField = *(StructReadField *)&file.m_flash[pFlash];
        pFlash += sizeof( StructReadField );
    }
    else
    {
        // pFlash contains the information about the corresponding index of the previous sibling element to be read.
        // The readDocument method searches depending on the parent or the previous element.
        fileXML.combineValid( fileXML.readContent( FileXML::ContentType::ReadField, m_readField, pFlash, lastXmlElement ) );
    }

    // Name fuer Baum eintragen
    setName( m_readField.name );

    // @@: Vorsichtsmassregel
    // Wenn zugeordnet: nicht schreiben, dann auch bei Feld zugeordnet.
    if ( ((EntryInputChain *)parent( EntryType::InputChain ))->m_readRecord.number == 255 )
        m_readField.recordFieldNumber = 255;

    // Finding the node for this structure and update parent.
    auto oldParent = fileXML.setParentNode( lastXmlElement );

    // Daten fuer union je nach "Einlesemethode"
    switch ( fieldFunction() )
    {
        case TBFE_NORM:
        case TBFE_GET_TIME:
        case TBFE_KONSTANTE:
        case TBFE_COPY_GV_TO_FIELD:
        case TBFE_TIMEBOY_COPY_GV_TO_FIELD:
        case TBFE_SERIENNUMMER:
        case TBFE_ANALOG_WERT:
        case TBFE_ANALOG_STATUS:
        case TBFE_DIGITAL_WERT:
        case TBFE_ZAEHLER_WERT:
        case TBFE_FUNC:
        case TBFE_SPS_WERT:
        case TBFE_ENTRANCE_ZM_NUM:
        case TBFE_ENTRANCE_READER_NUM:
        case TBFE_ENTRANCE_PNR:
        case TBFE_ENTRANCE_STATUS:
        case TBFE_DEBUG:
        case TBFE_GPRS_ALIVE_COUNTER:
        case TBFE_FIRMWARE_VERSION:
        case TBFE_ONLINE_STATUS:
        case TBFE_SUMMER_WINTER_STATUS:
        case TBFE_GPS_RECORD_FIX:
        case TBFE_GPS_RECORD_VARIABLE:
        case TBFE_ENTRANCE2_IDENTIFICATION:
        case TBFE_COPY_FIELD_TO_GV:
        case TBFE_TIMEBOY_COPY_FIELD_TO_GV:
        case TBFE_FINGERPRINT_SCANN:
        case TBFE_FINGERPRINT_CARD:
        case TBFE_FINGERPRINT_TEACH:
        case TBFE_FINGERPRINT_IDENTIFICATION:
        case TBFE_FINGERPRINT_VERIFICATION:
        case TBFE_FINGERPRINT_DELETE:
        case TBFE_OPERATION:
        case TBFE_SWITCH_RELAY:
        case TBFE_SWITCH_LED:
        case TBFE_TIMER:
        case TBFE_TRANSPONDER_WRITE:
        case TBFE_GPS_DISTANCE:
        case TBFE_MOBILE_PARAMETER:
        case TBFE_SWITCH_BUZZER:
        case TBFE_KONSTANTE_TYPE_DEVICE_NAME:
        case TBFE_MOBILMASTERIV_MENUENTRY:
        case TBFE_SWITCH_COMMUNICATION:
        case TBFE_TIMER_STATE:
        case TBFE_VIBRATION_VALUE:
        case TBFE_FINGERPRINT_ADMNISTARTION:
        {
            if ( useAESFile )
            {
                m_numFieldEx = m_readField.ex.numField;
            }
            else
            {
                fileXML.combineValid( fileXML.readContent( FileXML::ContentType::ReadField_NumFieldEx, m_numFieldEx ) );
            }

            // Bei alten Setupdateien Format-String loeschen
            if ( fieldFunction() == TBFE_NORM )
            {
                if ( strcmp( (char *)m_numFieldEx.contentOfConstant, "Feld" ) == 0 )
                    strcpy( (char *)m_numFieldEx.contentOfConstant, "" );
            }

            // Bei alten Setupdateien Format-String loeschen
            if ( fieldFunction() == TBFE_COPY_GV_TO_FIELD )
            {
                if ( strcmp( (char *)m_numFieldEx.contentOfConstant, "0" ) == 0 )
                    strcpy( (char *)m_numFieldEx.contentOfConstant, "" );

                if ( strcmp( (char *)m_numFieldEx.contentOfConstant, "Feld" ) == 0 )
                    strcpy( (char *)m_numFieldEx.contentOfConstant, "" );
            }

            //###########################
            // 2004-06-04: Vergleichspruefung mit Sprung.
            // Der Strukturbereich wurde noch nicht verwendet und muss initialisiert werden.
            // Wenn der Formatstring unsinn enthaelt wird initialisiert.
            for ( size_t i = 0; i < sizeof( m_numFieldEx.compare.formatString ); ++i )
            {
                if ( m_numFieldEx.compare.formatString[i] != 0 && static_cast<unsigned char>(m_numFieldEx.compare.formatString[i]) < 0x20 )
                {
                    memset( &m_numFieldEx.compare, 0, sizeof( StructCompare ) ); // Formatstring enthaelt Zeichen auserhalb sinnvollem Bereich
                }
            }

            //############################
            // 2015-07-02: Bei der Fingerprintfunktion Scannen wird der Parameter für Mehrfachscann auf BestMatch umgelegt.
            if ( fieldFunction() == TBFE_FINGERPRINT_SCANN )
            {
                // Wenn im unteren Nibble ein Wert groesser 2 vorliegt, wird BestMatch dafür aktiviert.
                if ( (m_numFieldEx.fingerprintParam & 0x0F) >= 2 )
                {
                    m_numFieldEx.fingerprintParam &= 0xF0; // Unterer Wert kann 0, 2 - 10 für Mehrfachsann haben. Wir hier auf 0 gesetzt.
                    m_numFieldEx.fingerprintParam |= 0x10; // Flag für BestMatch
                }

            }
            break;
        }

        case TBFE_LIST_SELECT_GET:
        case TBFE_LIST_SELECT_SET:
        case TBFE_DATA_ON_CARD:
        {
            if ( useAESFile )
            {
                m_listSelectionEx = m_readField.ex.listSelection;
            }
            else
            {
                fileXML.combineValid( fileXML.readContent( FileXML::ContentType::ReadField_ListSelectionEx, m_listSelectionEx ) );
            }

            // Durch Initialisierung des file.m_flash mit 0xFF sind Bits unerwuenschterweise
            // gesetzt, hier werden sie geloescht. (16.06.2003)
            if ( m_listSelectionEx.noValueCheck & 0x80 )
            {
                m_listSelectionEx.flags = BDLB_SEL_TASTATUR;
                m_listSelectionEx.noValueCheck = 0;
            }
            // Durch Initialisierung des file.m_flash mit 0xFF sind Bits unerwuenschterweise
            // gesetzt, hier werden sie geloescht. (12.09.2003)
            if ( m_listSelectionEx.doubleColumn & 0xFE )
                m_listSelectionEx.doubleColumn = 0;

            //###########################
            // 2004-06-04: Vergleichspruefung mit Sprung.
            // Der Strukturbereich wurde noch nicht verwendet und muss initialisiert werden.
            // Wenn der Formatstring unsinn enthaelt wird initialisiert.
            for ( size_t i = 0; i < sizeof( m_listSelectionEx.compare.formatString ); ++i )
            {
                if ( m_listSelectionEx.compare.formatString[i] != 0 && static_cast<unsigned char>(m_listSelectionEx.compare.formatString[i]) < 0x20 )
                {
                    // Formatstring enthaelt Zeichen auserhalb sinnvollem Bereich
                    memset( &m_listSelectionEx.compare, 0, sizeof( StructCompare ) );

                    // Auch fuer die Funktionalitaet der Listenerweiterung, den
                    // Strukturbereich initialisieren.
                    m_listSelectionEx.selectionRecordField2 = 0;
                    m_listSelectionEx.selectionListField2 = 0;
                    m_listSelectionEx.selectionNoEntry = 0;
                    memset( m_listSelectionEx.unused_2, 0, sizeof( m_listSelectionEx.unused_2 ) );
                }
            }

            // Bereich pruefen
            // (m_ExListAuswahl.selectionNoEntry < 0) is always false due to the data type uint8_t
            if ( m_listSelectionEx.selectionNoEntry > 4 )
            {
                m_listSelectionEx.selectionNoEntry = 0;
            }

            break;
        }

        case TBFE_BESTAETIGUNG:
        {
            if ( useAESFile )
            {
                m_confirmationEx = m_readField.ex.confirmation;
            }
            else
            {
                fileXML.combineValid( fileXML.readContent( FileXML::ContentType::ReadField_ConfirmationEx, m_confirmationEx ) );
            }
            break;
        }

        case TBFE_SEND_SMS:
        case TBFE_SCRIPT:
        case TBFE_SHOW_MESSAGE:
        case TBFE_READ_VARIABLE:
        case TBFE_WRITE_VARIABLE:
        case TBFE_SIGNATURE:
        case TBFE_LONG_BARCODE:
        {
            if ( useAESFile )
            {
                m_textEx = m_readField.ex.text;
            }
            else
            {
                fileXML.combineValid( fileXML.readContent( FileXML::ContentType::ReadField_TextEx, m_textEx ) );
            }
            break;
        }

        case TBFE_SWITCH_BACKLIGHT:
        {
            if ( useAESFile )
            {
                m_dataEx = m_readField.ex.data;
            }
            else
            {
                fileXML.combineValid( fileXML.readContent( FileXML::ContentType::ReadField_DataEx, m_dataEx ) );
            }
            break;
        }

        case TBFE_GV_TO_INTERFACE:
        {
            if ( useAESFile )
            {
                m_globVarToHIDEx = m_readField.ex.globVarToHID;
            }
            else
            {
                fileXML.combineValid( fileXML.readContent( FileXML::ContentType::ReadField_GlobVarToHIDEx, m_globVarToHIDEx ) );
            }
            break;
        }

        case TBFE_SWITCH_MODULE_ENERGIE:
        {
            if ( useAESFile )
            {
                m_switchModuleEnergieEx = m_readField.ex.switchModuleEnergie;
            }
            else
            {
                fileXML.combineValid( fileXML.readContent( FileXML::ContentType::ReadField_SwitchModuleEnergieEx, m_switchModuleEnergieEx ) );
            }
            break;
        }

        case TBFE_SYSTEM_MESSAGE:
        {
            if ( useAESFile )
            {
                m_systemMessageEx = m_readField.ex.systemMessage;
            }
            else
            {
                fileXML.combineValid( fileXML.readContent( FileXML::ContentType::ReadField_SystemMessageEx, m_systemMessageEx ) );
            }
            break;
        }

        case TBFE_TAKE_PHOTO:
        {
            if ( useAESFile )
            {
                m_takePhotoEx = m_readField.ex.takePhoto;
            }
            else
            {
                fileXML.combineValid( fileXML.readContent( FileXML::ContentType::ReadField_TakePhotoEx, m_takePhotoEx ) );
            }
            break;
        }

        case TBFE_FEVER_CONTROL:
        {
            if ( useAESFile )
            {
                m_feverControlEx = m_readField.ex.feverControl;
            }
            else
            {
                fileXML.combineValid( fileXML.readContent( FileXML::ContentType::ReadField_FeverControl, m_feverControlEx ) );
            }
            break;
        }

        case TBFE_MULTI_INPUT_SOURCE:
        {
            setIconIndexes( ICON_MULTI_INPUT_CHAIN_FIELD, ICON_MULTI_INPUT_CHAIN_FIELD_SEL );

            if ( useAESFile )
            {
                m_multiInputSourceEx = m_readField.ex.multiInputSource;
                m_multiInputSourceEx.ptr = endianReverse( m_multiInputSourceEx.ptr ) - FLASH_KONFIG_START;
                // Offset abziehen, da Erweitertes - Flash hinter Setupdaten.
                m_multiInputSourceEx.ptr -= m_flashOffset;
            }
            else
            {
                fileXML.combineValid( fileXML.readContent( FileXML::ContentType::ReadField_MultiInputSource, m_multiInputSourceEx ) );
            }
            // Eintrag hat Untereinträge
            uint8_t childCount = m_multiInputSourceEx.childCount;

            if ( childCount )
            {
                std::deque<MultiInputChildType> childTypes;
                if ( m_multiInputSourceEx.keypadInput )
                {
                    childTypes.push_back( MultiInputChildType::Keypad );
                }
                if ( m_multiInputSourceEx.barcodeInput )
                {
                    childTypes.push_back( MultiInputChildType::Barcode );
                }
                if ( m_multiInputSourceEx.transponderInput )
                {
                    childTypes.push_back( MultiInputChildType::Tranponder );
                }
                if ( m_multiInputSourceEx.fingerprintIdentificationInput )
                {
                    childTypes.push_back( MultiInputChildType::FingerprintIdentification );
                }
                if ( m_multiInputSourceEx.fingerprintVerificationInput )
                {
                    childTypes.push_back( MultiInputChildType::FingerprintVerification );
                }
                //BOOST_ASSERT( childPos.size() == childCount );

                uint32_t arrayOffset = m_multiInputSourceEx.ptr;

                tinyxml2::XMLElement *lastSubElement = nullptr;
                if ( !useAESFile )
                {
                    lastSubElement = fileXML.setParentNode( lastXmlElement->Parent()->ToElement() );
                }

                // Daten aller Unterfunktionen lesen
                for ( uint8_t childNumber = 0; childNumber < childCount && childTypes.size(); ++childNumber )
                {
                    // Eintrag erstellen und Daten lesen.
                    EntryInputChainField *entry = insertMultiInputChild( childTypes.front() );
                    childTypes.pop_front();

                    if ( useAESFile )
                    {
                        uint32_t childOffset = *(uint32_t *)&file.m_flash[arrayOffset + childNumber * sizeof( uint32_t )];
                        // Byteorder anpassen und Basisoffset abziehen.
                        childOffset = endianReverse( childOffset ) - FLASH_KONFIG_START;
                        // Bei ExtendedFlash den zusätzlichen Offset abziehen.
                        childOffset -= m_flashOffset;

                        entry->readDocument( childOffset, doc );
                    }
                    else
                    {
                        entry->readDocument( pFlash + childNumber + 1, doc );
                    }
                }

                if ( !useAESFile )
                {
                    fileXML.setParentNode( lastSubElement );
                }
            }
            break;
        }

        default:
            BOOST_ASSERT( false );
    }

    // Reset parent node because the end of the structure is reached.
    fileXML.setParentNode( oldParent );

    return pFlash;
}

void EntryInputChainField::checkMembers()
{
    if ( grandParentType() == EntryType::AccessControl )
        setFieldFunction( TBFE_KONSTANTE );
    else
        setFieldFunction( TBFE_NORM );
}

void EntryInputChainField::setFieldFunction( uint8_t function )
{
    BOOST_ASSERT( function < TBFE_MAX && function != TBFE_DEBUG );

    // Wird von einer Listenmethode auf eine andere umgestellt, muessen die Listenbezuege
    // entfernt werden.
    if ( (fieldFunction() == TBFE_LIST_SELECT_SET || fieldFunction() == TBFE_LIST_SELECT_GET || fieldFunction() == TBFE_DATA_ON_CARD)
         && (function != TBFE_LIST_SELECT_SET && function != TBFE_LIST_SELECT_GET && function != TBFE_DATA_ON_CARD) )
    {
        // Bezuege der Liste entfernen
        if ( m_refList )
        {
            m_refList->decRefCount();
            m_refList = nullptr;

            if ( m_refAuswahlFeld )
            {
                m_refAuswahlFeld->decRefCount();
                m_refAuswahlFeld = nullptr;
            }
            if ( m_refDatFeld )
            {
                m_refDatFeld->decRefCount();
                m_refDatFeld = nullptr;
            }

            if ( isValidSelRef( m_refSelDFeld ) )
            {
                m_refSelDFeld->decRefCount();
            }

            m_refSelDFeld = nullptr;
            if ( m_refSelLFeld )
            {
                m_refSelLFeld->decRefCount();
                m_refSelLFeld = nullptr;
            }

            if ( isValidSelRef( m_refSelDFeld2 ) )
            {
                m_refSelDFeld2->decRefCount();
            }

            m_refSelDFeld2 = nullptr;
            if ( m_refSelLFeld2 )
            {
                m_refSelLFeld2->decRefCount();
                m_refSelLFeld2 = nullptr;
            }
        }

        m_listSelectionEx.listNumber = 0;
        m_listSelectionEx.selectionField = 0;
        m_listSelectionEx.dataField = 0;
        m_listSelectionEx.selection = 0;
        m_listSelectionEx.selectionRecordField = 0;
        m_listSelectionEx.selectionListField = 0;
        m_listSelectionEx.selectionRecordField2 = 0;
        m_listSelectionEx.selectionListField2 = 0;
    }

    // Beim Wechsel der Funktion
    if ( fieldFunction() != function )
    {
        // Standardwerte für die gewählte Feldfunktion einstellen.
        if ( function == TBFE_TAKE_PHOTO )
        {
            m_readField.behaviourAtEsc = 253; // Nächstes EK-Feld.
            m_refJumpESC = reinterpret_cast<EntryInputChainField *>(-3);
        }
        else
        {
            m_readField.behaviourAtEsc = 0; // Übergeordnetes U-Menü
            m_refJumpESC = reinterpret_cast<EntryInputChainField *>(0);
        }

        // Bei Wechsel von einer Funktion die die Textstruktur verwendet, diese löschen.
        if ( fieldFunction() == TBFE_SCRIPT || fieldFunction() == TBFE_SEND_SMS || fieldFunction() == TBFE_SHOW_MESSAGE
             || fieldFunction() == TBFE_READ_VARIABLE || TBFE_WRITE_VARIABLE )
        {
            if ( (fieldFunction() == TBFE_READ_VARIABLE && function == TBFE_WRITE_VARIABLE) ||
                 (fieldFunction() == TBFE_WRITE_VARIABLE && function == TBFE_READ_VARIABLE) )
            {
                // In diesen Fällen den eventuellen Variablennamen beibehalten.
            }
            else
            {
                std::fill_n( m_textEx.text, sizeof( m_textEx.text ), 0 );
            }
        }

        // Bei einer MultiInputSource werden Untereinträge für die Einzelnen Feldfunktionen benötigt.
        if ( function == TBFE_MULTI_INPUT_SOURCE )
        {
            // Neue Methode bereits übernehmen, da insertMultiInputChild darauf prüft.
            m_readField.type = function;

            // Eintrag für Tastatur.
            if ( m_multiInputSourceEx.keypadInput )
            {
                insertMultiInputChild( MultiInputChildType::Keypad );
            }

            // Eintrag für Barcode.
            if ( m_multiInputSourceEx.barcodeInput )
            {
                insertMultiInputChild( MultiInputChildType::Barcode );
            }

            // Eintrag für Transponder.
            if ( m_multiInputSourceEx.transponderInput )
            {
                insertMultiInputChild( MultiInputChildType::Tranponder );
            }

            // Eintrag für Fingerprint.
            if ( m_multiInputSourceEx.fingerprintIdentificationInput )
            {
                insertMultiInputChild( MultiInputChildType::FingerprintIdentification );
            }

            // Eintrag für Fingerprint.
            if ( m_multiInputSourceEx.fingerprintVerificationInput )
            {
                insertMultiInputChild( MultiInputChildType::FingerprintVerification );
            }
        }
        else if ( fieldFunction() == TBFE_MULTI_INPUT_SOURCE )
        {
            // Eintrag für Tastatur.
            if ( m_multiInputSourceEx.keypadInput )
            {
                removeMultiInputChild( MultiInputChildType::Keypad );
            }

            // Eintrag für Barcode.
            if ( m_multiInputSourceEx.barcodeInput )
            {
                removeMultiInputChild( MultiInputChildType::Barcode );
            }

            // Eintrag für Transponder.
            if ( m_multiInputSourceEx.transponderInput )
            {
                removeMultiInputChild( MultiInputChildType::Tranponder );
            }

            // Eintrag für Fingerprint.
            if ( m_multiInputSourceEx.fingerprintIdentificationInput )
            {
                removeMultiInputChild( MultiInputChildType::FingerprintIdentification );
            }

            // Eintrag für Fingerprint.
            if ( m_multiInputSourceEx.fingerprintVerificationInput )
            {
                removeMultiInputChild( MultiInputChildType::FingerprintVerification );
            }
        }
    }

    // Neue Methode übernehmen.
    m_readField.type = function;
}

uint8_t EntryInputChainField::fieldFunction()
{
    return m_readField.type;
}

///////////////////////////////////////
// Referenzierung zwischen Einträgen

void EntryInputChainField::convertReferenceToIndex()
{
    int index;
    EntryDevice *device = nullptr;
    EntryBase *handling = nullptr;
    EntryBase *main_menu = nullptr;
    EntryBase *entry = nullptr;

    // Auf Rooteintrag wechseln.
    entry = this->parent();
    while ( !entry->isType( EntryType::Device ) )
    {
        // Eintrag unter Handling?
        if ( entry->isType( EntryType::Handling ) )
        {
            handling = entry;
            main_menu = handling->first();
        }

        entry = entry->parent();
    }
    device = (EntryDevice *)entry;

    // Zeiger auf Hauptmenu sicherstellen.
    if ( !main_menu && device->rootHandling().size() )
    {
        main_menu = device->rootHandling().first();
    }


    // Zugeordnetes Tabellenfeld eintragen
    if ( m_refTableField )
    {
        m_readField.recordFieldNumber = m_refTableField->m_tableField.number;
    }
    else
    {
        m_readField.recordFieldNumber = 255;
    }

    // Sprungziel nach ESC Ermitteln und Eintragen
    if ( m_refJumpESC == reinterpret_cast<EntryBase *>(0) )
    {
        // Keine Auswahl oder Uebergeordnetes Untermenue gewaehlt.
        m_readField.behaviourAtEsc = EV_UPPER_MENUE;
    }
    else if ( m_refJumpESC == reinterpret_cast<EntryBase *>(-1) )
    {
        // Naechste Eingabekette bzw. Menue gewaehlt.
        m_readField.behaviourAtEsc = EV_NEXT_SIBLING;
    }
    else if ( m_refJumpESC == reinterpret_cast<EntryBase *>(-2) )
    {
        // Hauptmenue gewaehlt.
        m_readField.behaviourAtEsc = EV_MAIN_MENUE;
    }
    else if ( m_refJumpESC == reinterpret_cast<EntryBase *>(-3) )
    {
        // Naechstes Feld gewaehlt.
        m_readField.behaviourAtEsc = EV_NEXT_FIELD;
    }
    else
    {
        // Entweder ist das Sprungziel E-Feld, E-Kette bzw. U-Menue, F-Taste   

        // Versuch die Referenz zu E-Feld aufzuloesen.
        index = parent()->indexOf( m_refJumpESC );
        if ( index == -1 )
        {
            // Versuch die Referenz zu E-Kette bzw. U-Menue aufzuloesen.
            index = grandParent()->indexOf( m_refJumpESC );
            if ( index == -1 )
            {
                // Versuch die Referenz zu F-Taste aufzuloesen.
                if ( main_menu )
                    index = main_menu->indexOf( m_refJumpESC );

                if ( index == -1 )
                {
                    // Von Uebergeordnetes U-Menue ausgehen
                    m_readField.behaviourAtEsc = EV_UPPER_MENUE;
                }
                else
                {
                    // F-Taste
                    m_readField.behaviourAtEsc = index + 112; // 112 => F1
                }
            }
            else
            {
                // Bis zu 30 erlaubt.
                if ( index >= 30 )
                    index = 0;

                // E-Kette bzw. U-Menue
                m_readField.behaviourAtEsc = index + 200;
            }
        }
        else
        {
            // Bis zu 110 E-Felder erlaubt.
            if ( index >= 110 )
                index = -1;

            // E-Feld
            m_readField.behaviourAtEsc = index + 1;
        }
    }

    StructCompare compare;
    memset( &compare, 0, sizeof( compare ) );
    // Sprungziel von Eintrag uebernehmen
    switch ( fieldFunction() )
    {
        case TBFE_NORM:
        case TBFE_COPY_GV_TO_FIELD:
        case TBFE_COPY_FIELD_TO_GV:
        case TBFE_OPERATION:
            compare = m_numFieldEx.compare;
            break;

        case TBFE_LIST_SELECT_GET:
        case TBFE_LIST_SELECT_SET:
        case TBFE_DATA_ON_CARD:
            compare = m_listSelectionEx.compare;
            break;

        case TBFE_MULTI_INPUT_SOURCE:
            compare = m_multiInputSourceEx.compare;
            break;
    }

    // Sprungziel nach ESC Ermitteln und Eintragen
    if ( m_refJump == reinterpret_cast<EntryBase *>(0) )
    {
        // Keine Auswahl oder Uebergeordnetes Untermenue gewaehlt.
        index = EV_UPPER_MENUE;
    }
    else if ( m_refJump == reinterpret_cast<EntryBase *>(-1) )
    {
        // Naechste Eingabekette bzw. Menue gewaehlt.
        index = EV_NEXT_SIBLING;
    }
    else if ( m_refJump == reinterpret_cast<EntryBase *>(-2) )
    {
        // Hauptmenue gewaehlt.
        index = EV_MAIN_MENUE;
    }
    else if ( m_refJump == reinterpret_cast<EntryBase *>(-3) )
    {
        // Naechstes Feld gewaehlt.
        index = EV_NEXT_FIELD;
    }
    else
    {

        // Entweder ist das Sprungziel E-Feld, E-Kette bzw. U-Menue, F-Taste   

        // Versuch die Referenz zu E-Feld aufzuloesen.
        index = parent()->indexOf( m_refJump );
        if ( index == -1 )
        {
            // Versuch die Referenz zu E-Kette bzw. U-Menue aufzuloesen.
            index = grandParent()->indexOf( m_refJump );
            if ( index == -1 )
            {
                // Versuch die Referenz zu F-Taste aufzuloesen.
                if ( main_menu )
                    index = main_menu->indexOf( m_refJump );

                if ( index == -1 )
                {
                    // Von Uebergeordnetes U-Menue ausgehen
                    index = EV_UPPER_MENUE;
                }
                else
                {
                    // F-Taste
                    index += 112; // 112 => F1
                }
            }
            else
            {
                // Bis zu 30 erlaubt.
                if ( index >= 30 )
                {
                    index = 0;
                }

                // E-Kette bzw. U-Menue
                index += 200;
            }
        }
        else
        {
            // Bis zu 110 E-Felder erlaubt.
            if ( index >= 110 )
            {
                index = 0;
            }

            // E-Feld
            index++;
            // Bei den ersten Methoden wurde der pure 0 indizierte Feldindex angegeben.
            if ( (compare.type == FECT_FORMAT_STRING) || (compare.type == FECT_FORMAT_GV) )
            {
                index--;
            }
        }
    }

    // Bugzilla 1107: Das zwischeneinfügen von GV's verschiebt den Index bereits zugewiesener GV's, hier eine nachgelagerte evntl. Korrektur.
    // Bugzilla 1258: Die Prüfung von Bug 1107 darf nicht bei Verwendung der Feldfunktionen für die Timeboy-GV's durchgeführt werden.
    if ( (m_readField.globalVariable & 0xF0) != 0
         && fieldFunction() != TBFE_TIMEBOY_COPY_GV_TO_FIELD && fieldFunction() != TBFE_TIMEBOY_COPY_FIELD_TO_GV )
    {
        std::array<uint8_t, 8> defs;
        std::array<uint8_t, 8> numbers;
        int index = m_readField.globalVariable & 0x0F;
        int number = 0;
        for ( unsigned int i = 0; i < defs.size(); ++i )
        {
            if ( device->m_globalVar.names[i][0] != 0 )
            {
                defs[i] = true;
                numbers[i] = ++number;
            }
        }

        // Ist für den zugewiesenen Index eine GV definiert
        if ( defs[index] )
        {
            // Stimmt die zugewiesene Nummer?
            if ( (m_readField.globalVariable & 0xF0) != (numbers[index] << 4) )
            {
                // Korrekte Zuweisung sicherstellen.
                m_readField.globalVariable = (numbers[index] << 4) + index;
            }
        }
        else
        {
            // Zugewiesener Variablenindex gehört zu keiner GV. stiller Reset.
            m_readField.globalVariable = 0;
        }
    }

    // Sprungziel nach Vergleich Eintragen
    switch ( fieldFunction() )
    {
        case TBFE_NORM:
        case TBFE_COPY_GV_TO_FIELD:
        case TBFE_COPY_FIELD_TO_GV:
        case TBFE_OPERATION:
            m_numFieldEx.compare.jumpDestination = index;
            break;

        case TBFE_LIST_SELECT_GET:
        case TBFE_LIST_SELECT_SET:
        case TBFE_DATA_ON_CARD:
            m_listSelectionEx.compare.jumpDestination = index;
            break;

        case TBFE_MULTI_INPUT_SOURCE:
            m_multiInputSourceEx.compare.jumpDestination = index;
            break;
    }

    // Alle Angaben sind mit 0 zu initalisieren.
    m_listSelectionEx.listNumber = 0;
    m_listSelectionEx.selectionField = 0;
    m_listSelectionEx.dataField = 0;
    //m_ExListAuswahl.selection = 0;
    m_listSelectionEx.selectionRecordField = 0;
    m_listSelectionEx.selectionListField = 0;
    m_listSelectionEx.selectionRecordField2 = 0;
    m_listSelectionEx.selectionListField2 = 0;

    // Feldmethoden "Aus Liste waehlen" und "In Liste schreiben", sowie "Data on Card"
    if ( fieldFunction() == TBFE_LIST_SELECT_SET || fieldFunction() == TBFE_LIST_SELECT_GET || fieldFunction() == TBFE_DATA_ON_CARD )
    {
        int isDataOnCard = (fieldFunction() == TBFE_DATA_ON_CARD);
        EntryTableList &entryList = device->rootTableList();
        EntryAccessControl2 &entryEntrance2List = device->rootAccessControl();
        if ( m_refList )
        {
            // Es handelt sich um eine Zutrittsliste.
            if ( m_refList->parentType() == EntryType::AccessControl )
            {
                m_listSelectionEx.listNumber = entryEntrance2List.indexOf( m_refList ) | 0x80;
            }
            else
            {
                m_listSelectionEx.listNumber = entryList.indexOf( m_refList );
            }

            if ( m_refAuswahlFeld )
            {
                m_listSelectionEx.selectionField = m_refList->indexOf( m_refAuswahlFeld );
            }

            if ( m_refDatFeld )
            {
                m_listSelectionEx.dataField = m_refList->indexOf( m_refDatFeld );
            }

            if ( (m_listSelectionEx.selection & 0x01) != 0 || isDataOnCard )
            {
                if ( isValidSelRef( m_refSelDFeld ) )
                {
                    m_listSelectionEx.selectionRecordField = reinterpret_cast<EntryInputChain *>(parent())->m_refTable->indexOf( m_refSelDFeld );
                }
                else
                {
                    m_listSelectionEx.selectionRecordField = static_cast<uint8_t>( reinterpret_cast<uint64_t>( m_refSelDFeld ) ); // 0 - 255 ist fuer Variablen vorgesehen.
                }

                if ( m_refSelLFeld )
                {
                    m_listSelectionEx.selectionListField = m_refList->indexOf( m_refSelLFeld );
                }

                if ( m_listSelectionEx.selection & 0x02 )
                {
                    if ( isValidSelRef( m_refSelDFeld2 ) )
                    {
                        m_listSelectionEx.selectionRecordField2 = reinterpret_cast<EntryInputChain *>(parent())->m_refTable->indexOf( m_refSelDFeld2 );
                    }
                    else
                    {
                        m_listSelectionEx.selectionRecordField2 = static_cast<uint8_t>( reinterpret_cast<uint64_t>( m_refSelDFeld2 ) ); // 0 - 255 ist fuer Variablen vorgesehen.
                    }

                    if ( m_refSelLFeld2 )
                    {
                        m_listSelectionEx.selectionListField2 = m_refList->indexOf( m_refSelLFeld2 );
                    }
                }

                // Bei keinem Eintrag in der Liste soll gesprungen werden.
                if ( m_listSelectionEx.selectionNoEntry == 3 )
                {
                    // Sprungziel nach ESC Ermitteln und Eintragen
                    if ( m_refJumpSelNoEntry == reinterpret_cast<EntryBase *>(0) )
                    {
                        // Keine Auswahl oder Uebergeordnetes Untermenue gewaehlt.
                        index = EV_UPPER_MENUE;
                    }
                    else if ( m_refJumpSelNoEntry == reinterpret_cast<EntryBase *>(-1) )
                    {
                        // Naechste Eingabekette bzw. Menue gewaehlt.
                        index = EV_NEXT_SIBLING;
                    }
                    else if ( m_refJumpSelNoEntry == reinterpret_cast<EntryBase *>(-2) )
                    {
                        // Hauptmenue gewaehlt.
                        index = EV_MAIN_MENUE;
                    }
                    else if ( m_refJumpSelNoEntry == reinterpret_cast<EntryBase *>(-3) )
                    {
                        // Naechstes Feld gewaehlt.
                        index = EV_NEXT_FIELD;
                    }
                    else
                    {
                        // Entweder ist das Sprungziel E-Feld, E-Kette bzw. U-Menue, F-Taste   

                        // Versuch die Referenz zu E-Feld aufzuloesen.
                        index = parent()->indexOf( m_refJumpSelNoEntry );
                        if ( index == -1 )
                        {
                            // Versuch die Referenz zu E-Kette bzw. U-Menue aufzuloesen.
                            index = grandParent()->indexOf( m_refJumpSelNoEntry );
                            if ( index == -1 )
                            {
                                // Versuch die Referenz zu F-Taste aufzuloesen.
                                if ( main_menu )
                                {
                                    index = main_menu->indexOf( m_refJumpSelNoEntry );
                                }

                                if ( index == -1 )
                                {
                                    // Von Uebergeordnetes U-Menue ausgehen
                                    index = EV_UPPER_MENUE;
                                }
                                else
                                {
                                    // F-Taste
                                    index += 112; // 112 => F1
                                }
                            }
                            else
                            {
                                // Bis zu 30 erlaubt.
                                if ( index >= 30 )
                                {
                                    index = 0;
                                }

                                // E-Kette bzw. U-Menue
                                index += 200;
                            }
                        }
                        else
                        {
                            // Bis zu 110 E-Felder erlaubt.
                            if ( index >= 110 )
                            {
                                index = 0;
                            }

                            // E-Feld
                            index++;
                        }
                    }
                }

                m_listSelectionEx.jumpSelectionNoEntry = index;
            }
        }
    }

    if ( fieldFunction() == TBFE_ANALOG_WERT )
    {
        // Integralwert uebernehmen?
        if ( (m_readField.variable2 & 0x80) != 0 )
        {
            // Wert in Feld schreiben.
            if ( (m_readField.variable2 & 0x40) == 0 )
            {
                if ( m_refTableFieldIntegral )
                {
                    // Zugeordnetes Tabellenfeld eintragen
                    m_readField.variable2 &= 0xC0;
                    m_readField.variable2 |= m_refTableFieldIntegral->m_tableField.number;
                }
                else
                {
                    // Kein Feld gewaehlt.
                    m_readField.variable2 |= 0x3F;
                }
            }
        }
        else
        {
            // Komplett loeschen.
            m_readField.variable2 = 0;
        }
    }

    if ( fieldFunction() == TBFE_FEVER_CONTROL )
    {
        if ( m_refFeverControlStatusField )
        {
            m_feverControlEx.indexRecordField = m_refFeverControlStatusField->m_tableField.number;
        }
        else
        {
            m_feverControlEx.indexRecordField = 255;
        }
    }

    if ( fieldFunction() == TBFE_MULTI_INPUT_SOURCE )
    {
        if ( m_refMultiInputSourceField )
        {
            m_multiInputSourceEx.indexRecordField = m_refMultiInputSourceField->m_tableField.number;
        }
        else
        {
            m_multiInputSourceEx.indexRecordField = 255;
        }
    }
}


void EntryInputChainField::convertIndexToReference()
{
    int index;
    EntryDevice *device = nullptr;
    EntryBase *handling = nullptr;
    EntryBase *main_menu = nullptr;
    EntryBase *entry = nullptr;

    // Auf Rooteintrag wechseln.
    entry = this->parent();
    while ( !entry->isType( EntryType::Device ) )
    {
        // Eintrag unter Handling?
        if ( entry->isType( EntryType::Handling ) )
        {
            handling = entry;
            main_menu = handling->first();
        }

        entry = entry->parent();
    }

    device = reinterpret_cast<EntryDevice *>(entry);

    // Zeiger auf Hauptmenu sicherstellen.
    if ( !main_menu && device->rootHandling().size() )
        main_menu = device->rootHandling().first();

    // Die referenzierte Tabellenfeldklasse ermitteln
    if ( m_readField.recordFieldNumber != 255 )
    {
        // Ueber Parent - Tabellenrefernz auf Feld wechseln
        EntryBase *r = parent();
        if ( reinterpret_cast<EntryInputChain *>(r)->m_refTable )
        {
            m_refTableField = (EntryTableField *)((EntryInputChain *)r)->m_refTable->at( m_readField.recordFieldNumber );
            if ( m_refTableField == (EntryTableField *)-1 )
                m_refTableField = 0;

            if ( m_refTableField )
                m_refTableField->incRefCount();
        }
        else
        {
            // Wenn Parent keine Referenz, dann hat Feld keine Bezugstabelle
            m_readField.recordFieldNumber = 255;
            m_refTableField = nullptr;
        }
    }
    else
    {

     // Keine zugewiesen
        m_refTableField = nullptr;
    }

    index = m_readField.behaviourAtEsc;
    // Sprungziel nach ESC
    switch ( index )
    {
        case EV_UPPER_MENUE:
            m_refJumpESC = (EntryBase *)0;
            break;
        case EV_NEXT_SIBLING:
            m_refJumpESC = (EntryBase *)-1;
            break;
        case EV_MAIN_MENUE:
            m_refJumpESC = (EntryBase *)-2;
            break;
        case EV_NEXT_FIELD:
            m_refJumpESC = (EntryBase *)-3;
            break;
        default:
            m_refJumpESC = (EntryBase *)0;
            if ( index >= 1 && index <= 111 )
            {
                // E-Feld
                m_refJumpESC = parent()->value( index - 1 );

            }
            else if ( index >= 200 && index <= 229 )
            {
                // E-Kette bzw. Menue.
                m_refJumpESC = grandParent()->value( index - 200 );

            }
            else if ( index >= 112 && index <= 127 )
            {
                // F-Taste
                if ( main_menu )
                {
                    m_refJumpESC = main_menu->value( index - 112 ); // 112 => F1
                }
            }
            break;
    }
    if ( m_refJumpESC != (EntryBase *)0 && m_refJumpESC != (EntryBase *)-1 && m_refJumpESC != (EntryBase *)-2 && m_refJumpESC != (EntryBase *)-3 )
    {

//    m_refJumpESC->incRefCount();
    }


    m_refJump = (EntryBase *)0;
    StructCompare compare;
    memset( &compare, 0, sizeof( compare ) );
    // Sprungziel von Eintrag uebernehmen
    switch ( fieldFunction() )
    {
        case TBFE_NORM:
        case TBFE_COPY_GV_TO_FIELD:
        case TBFE_COPY_FIELD_TO_GV:
        case TBFE_OPERATION:
            compare = m_numFieldEx.compare;
            break;

        case TBFE_LIST_SELECT_GET:
        case TBFE_LIST_SELECT_SET:
        case TBFE_DATA_ON_CARD:
            compare = m_listSelectionEx.compare;
            break;

        case TBFE_MULTI_INPUT_SOURCE:
            compare = m_multiInputSourceEx.compare;
            break;
    }

    if ( (compare.type == FECT_FORMAT_STRING_2) || (compare.type == FECT_FORMAT_GV_2) )
    {

        index = compare.jumpDestination;
        // Sprungziel nach ESC
        switch ( index )
        {
            case EV_UPPER_MENUE:
                m_refJump = (EntryBase *)0;
                break;
            case EV_NEXT_SIBLING:
                m_refJump = (EntryBase *)-1;
                break;
            case EV_MAIN_MENUE:
                m_refJump = (EntryBase *)-2;
                break;
            case EV_NEXT_FIELD:
                m_refJump = (EntryBase *)-3;
                break;
            default:
                m_refJump = (EntryBase *)0;
                if ( index >= 1 && index <= 111 )
                {
                    // E-Feld
                    m_refJump = parent()->value( index - 1 );

                }
                else if ( index >= 200 && index <= 229 )
                {
                    // E-Kette bzw. Menue.
                    m_refJump = grandParent()->value( index - 200 );

                }
                else if ( index >= 112 && index <= 127 )
                {
                    // F-Taste
                    if ( main_menu )
                    {
                        m_refJump = main_menu->value( index - 112 ); // 112 => F1
                    }
                }
                break;
        }

        if ( m_refJump != (EntryBase *)0 && m_refJump != (EntryBase *)-1 && m_refJump != (EntryBase *)-2 && m_refJump != (EntryBase *)-3 )
        {
//      m_refJump->incRefCount();
        }
    }
    else if ( compare.type != FECT_NONE )
    {

        m_refJump = parent()->value( compare.jumpDestination );
        if ( m_refJump )
        {
//      m_refJump->incRefCount();
        }
    }

    // Feldmethoden "Aus Liste waehlen" und "In Liste schreiben", sowie "Data on Card".
    if ( fieldFunction() == TBFE_LIST_SELECT_SET || fieldFunction() == TBFE_LIST_SELECT_GET || fieldFunction() == TBFE_DATA_ON_CARD )
    {

        int isDataOnCard = (fieldFunction() == TBFE_DATA_ON_CARD);
        EntryTableList &entryList = device->rootTableList();
        EntryAccessControl2 &entryEntrance2List = device->rootAccessControl();
        // Handelt es sich um eine Liste der Zutrittskontrolle?
        if ( m_listSelectionEx.listNumber & 0x80 )
        {
            m_refList = (EntryTable *)entryEntrance2List.at( m_listSelectionEx.listNumber & 0x7F ); // Liste aus der ausgewaehlt wird
        }
        else
        {
            m_refList = (EntryTable *)entryList.at( m_listSelectionEx.listNumber ); // Liste aus der ausgewaehlt wird
        }

        if ( m_refList )
        {
            m_refList->incRefCount();

            m_refAuswahlFeld = (EntryTableField *)m_refList->at( m_listSelectionEx.selectionField ); // Feld aus dem ausgewaehlt wird
            if ( m_refAuswahlFeld ) m_refAuswahlFeld->incRefCount();

            m_refDatFeld = (EntryTableField *)m_refList->at( m_listSelectionEx.dataField );     // Feld das abgespeichert wird
            if ( m_refDatFeld ) m_refDatFeld->incRefCount();

            if ( (m_listSelectionEx.selection & 0x01) != 0 || isDataOnCard )
            {
                if ( m_listSelectionEx.selectionRecordField & 0x80 )
                { // 0 - 255 ist fuer Variablen vorgesehen.
                    m_refSelDFeld = reinterpret_cast<EntryTableField *>(m_listSelectionEx.selectionRecordField);  // Erstes Selektionsfeld der Datensatzbeschreibung
                }
                else if ( ((EntryInputChain *)parent())->m_refTable != nullptr )
                {
                    m_refSelDFeld = (EntryTableField *)((EntryInputChain *)parent())->m_refTable->at( m_listSelectionEx.selectionRecordField );    // Erstes Selektionsfeld der Datensatzbeschreibung
                    if ( m_refSelDFeld ) m_refSelDFeld->incRefCount();
                }
                m_refSelLFeld = (EntryTableField *)m_refList->at( m_listSelectionEx.selectionListField );    // Erstes Selektionsfeld der Liste
                if ( m_refSelLFeld ) m_refSelLFeld->incRefCount();

                if ( m_listSelectionEx.selection & 0x02 )
                {
                    if ( m_listSelectionEx.selectionRecordField2 & 0x80 )
                    {  // 0 - 255 ist fuer Variablen vorgesehen.
                        m_refSelDFeld2 = reinterpret_cast<EntryTableField *>(m_listSelectionEx.selectionRecordField2);    // Zweites Selektionsfeld der Datensatzbeschreibung
                    }
                    else if ( ((EntryInputChain *)parent())->m_refTable != nullptr )
                    {
                        m_refSelDFeld2 = (EntryTableField *)((EntryInputChain *)parent())->m_refTable->at( m_listSelectionEx.selectionRecordField2 );   // Zweites Selektionsfeld der Datensatzbeschreibung
                        if ( m_refSelDFeld2 ) m_refSelDFeld2->incRefCount();
                    }
                    m_refSelLFeld2 = (EntryTableField *)m_refList->at( m_listSelectionEx.selectionListField2 );   // Zweites Selektionsfeld der Liste
                    if ( m_refSelLFeld2 ) m_refSelLFeld2->incRefCount();
                }

                // Sprung wenn kein Eintrag in Liste gefunden.
                if ( m_listSelectionEx.selectionNoEntry == 3 )
                {

                    m_refJumpSelNoEntry = (EntryBase *)0;
                    index = m_listSelectionEx.jumpSelectionNoEntry;
                    // Sprungziel nach ESC
                    switch ( index )
                    {
                        case EV_UPPER_MENUE:
                            m_refJumpSelNoEntry = (EntryBase *)0;
                            break;
                        case EV_NEXT_SIBLING:
                            m_refJumpSelNoEntry = (EntryBase *)-1;
                            break;
                        case EV_MAIN_MENUE:
                            m_refJumpSelNoEntry = (EntryBase *)-2;
                            break;
                        case EV_NEXT_FIELD:
                            m_refJumpSelNoEntry = (EntryBase *)-3;
                            break;
                        default:
                            m_refJumpSelNoEntry = (EntryBase *)0;
                            if ( index >= 1 && index <= 111 )
                            {
// E-Feld
                                m_refJumpSelNoEntry = parent()->value( index - 1 );

                            }
                            else if ( index >= 200 && index <= 229 )
                            {
               // E-Kette bzw. Menue.
                                m_refJumpSelNoEntry = grandParent()->value( index - 200 );

                            }
                            else if ( index >= 112 && index <= 127 )
                            {
               // F-Taste
                                if ( main_menu )
                                {
                                    m_refJumpSelNoEntry = main_menu->value( index - 112 ); // 112 => F1
                                }
                            }
                    }

                    if ( m_refJumpSelNoEntry != (EntryBase *)0 && m_refJumpSelNoEntry != (EntryBase *)-1 &&
                         m_refJumpSelNoEntry != (EntryBase *)-2 && m_refJumpSelNoEntry != (EntryBase *)-3 )
                    {
//      m_refJumpSelNoEntry->incRefCount();
                    }
                }
            }
        }
    }


    if ( fieldFunction() == TBFE_ANALOG_WERT )
    {

// Die referenzierte Tabellenfeldklasse ermitteln
        if ( (m_readField.variable2 & 0x80) != 0 && (m_readField.variable2 & 0x40) == 0 )
        {

// Ueber Parent - Tabellenrefernz auf Feld wechseln
            EntryBase *r = parent();
            if ( ((EntryInputChain *)r)->m_refTable )
            {

                m_refTableFieldIntegral = (EntryTableField *)((EntryInputChain *)r)->m_refTable->at( m_readField.variable2 & 0x3F );
                if ( m_refTableFieldIntegral == (EntryTableField *)-1 ) m_refTableFieldIntegral = 0;
                if ( m_refTableFieldIntegral )
                {
                    m_refTableFieldIntegral->incRefCount();
                }
            }
            else
            {
                     // Wenn Parent keine Referenz, dann hat Feld keine Bezugstabelle
                m_readField.variable2 |= 0x3F;
                m_refTableFieldIntegral = nullptr;
            }
        }
        else
        {

             // Keine zugewiesen
            m_refTableFieldIntegral = nullptr;
        }

    }

    if ( fieldFunction() == TBFE_FEVER_CONTROL )
    {
        // Ueber Parent - Tabellenrefernz auf Feld wechseln
        EntryBase *r = parent( EntryType::InputChain );
        if ( m_feverControlEx.indexRecordField < MAX_FIELDS && ((EntryInputChain *)r)->m_refTable )
        {
            m_refFeverControlStatusField = (EntryTableField *)((EntryInputChain *)r)->m_refTable->at( m_feverControlEx.indexRecordField );
            if ( m_refFeverControlStatusField == (EntryTableField *)-1 )
            {
                m_refFeverControlStatusField = nullptr;
            }

            if ( m_refFeverControlStatusField )
            {
                m_refFeverControlStatusField->incRefCount();
            }
        }
        else
        {
            // Wenn Parent keine Referenz, dann hat Feld keine Bezugstabelle
            m_refFeverControlStatusField = nullptr;
        }
    }

    if ( fieldFunction() == TBFE_MULTI_INPUT_SOURCE )
    {
        // Ueber Parent - Tabellenrefernz auf Feld wechseln
        EntryBase *r = parent();
        if ( m_multiInputSourceEx.indexRecordField < MAX_FIELDS && ((EntryInputChain *)r)->m_refTable )
        {
            m_refMultiInputSourceField = (EntryTableField *)((EntryInputChain *)r)->m_refTable->at( m_multiInputSourceEx.indexRecordField );
            if ( m_refMultiInputSourceField == (EntryTableField *)-1 )
            {
                m_refMultiInputSourceField = nullptr;
            }

            if ( m_refMultiInputSourceField )
            {
                m_refMultiInputSourceField->incRefCount();
            }
        }
        else
        {
            // Wenn Parent keine Referenz, dann hat Feld keine Bezugstabelle
            m_refMultiInputSourceField = nullptr;
        }
    }
}

///////////////////////////////////////
// TabCtrl

int EntryInputChainField::m_pageIndex = PAGE_NO;

void EntryInputChainField::setPageIndex( int index )
{
    m_pageIndex = index;
}

int EntryInputChainField::pageIndex()
{
    return m_pageIndex;
}

///////////////////////////////////
// Daten des Eintrags

void EntryInputChainField::setDefault()
{
    setDefault( QT_TRANSLATE_NOOP_UTF8( "EntryInputChainField", "Eingabe" ) );
}

void EntryInputChainField::setDefault( const std::string &name )
{
    setName( name );

    m_iconIndex = ICON_INPUT_CHAIN_FIELD;
    m_iconIndexSelected = ICON_INPUT_CHAIN_FIELD_SEL;

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_INPUT_CHAIN_FIELD_0 );
    m_pageIndexes.push_back( PAGE_INPUT_CHAIN_FIELD_1 );
    m_pageIndexes.push_back( PAGE_INPUT_CHAIN_FIELD_2 );
    m_pageIndexes.push_back( PAGE_INPUT_CHAIN_FIELD_3 );
    m_pageIndexes.push_back( PAGE_INPUT_CHAIN_FIELD_4 );


    // Extendet
    memset( (unsigned char *)&m_confirmationEx, 0, sizeof( StructConfirmationEx ) );
    strncpy( (char *)m_confirmationEx.text1, this->name().c_str(), sizeof( m_confirmationEx.text1 ) - 1 );
    strncpy( (char *)m_confirmationEx.text2, "abschliessen", sizeof( m_confirmationEx.text2 ) - 1 );
    strncpy( (char *)m_confirmationEx.text3, "", sizeof( m_confirmationEx.text3 ) - 1 );

    memset( (unsigned char *)&m_listSelectionEx, 0, sizeof( StructListSelectionEx ) );
    m_listSelectionEx.selectionField = 1;
    m_listSelectionEx.dataField = 0;
    m_listSelectionEx.listNumber = 0;
    m_listSelectionEx.selection = FALSE;
    m_listSelectionEx.selectionRecordField = 0;
    m_listSelectionEx.selectionListField = 0;
    m_listSelectionEx.selectionBarcode = FALSE;
    m_listSelectionEx.flags = BDLB_SEL_TASTATUR;
    m_listSelectionEx.noValueCheck = 0;
    m_listSelectionEx.doubleColumn = 0;
    m_listSelectionEx.selectionRecordField2 = 0;
    m_listSelectionEx.selectionListField2 = 0;
    m_listSelectionEx.selectionNoEntry = 0;
    m_listSelectionEx.usedFontSize = 0;
    memset( m_listSelectionEx.unused_2, 0, sizeof( m_listSelectionEx.unused_2 ) );
    memset( &m_listSelectionEx.compare, 0, sizeof( m_listSelectionEx.compare ) );

    memset( (unsigned char *)&m_numFieldEx, 0, sizeof( StructNumFieldEx ) );

    EntryDevice &device = root();

    bool  boxDevice = false;
    bool  ipPCDevice = false;
    bool tbDockDevice = false;
    if ( device.hasSetupType( { SetupType::KYO_Inloc, SetupType::ZK_MasterIV, SetupType::IO_Box, SetupType::MDE_BoxIV, SetupType::Mobil_MasterIV, SetupType::Mobil_Box, SetupType::KYO_Oneloc, SetupType::KYO_Cenloc } ) )
    {
        boxDevice = true;
    }
    else if ( device.hasSetupType( { SetupType::IPC_Extension } ) )
    {
        ipPCDevice = true;
    }
    else if ( device.hasSetupType( { SetupType::Timeboy_TBDock, SetupType::Mobil_TBDock } ) )
    {
        tbDockDevice = true;
    }

    m_numFieldEx.flags = NFF_TRANSPONDER | NFF_TASTATUR;

    if ( boxDevice || ipPCDevice | tbDockDevice )
    {
        m_numFieldEx.flags = NFF_TRANSPONDER | NFF_NO_RET;
        m_listSelectionEx.flags = 0;
    }
    // else 
    // Zugelassene Eingabequellen


    // Transponder
    memset( &m_numFieldEx.tag, 0, sizeof( StructTransponder ) );
    m_numFieldEx.tag.tpSegment1 = 255; // Kein
    m_numFieldEx.tag.tpSegment2 = 255;
    m_numFieldEx.tag.tpSegment3 = 255;

    // Relais
    m_numFieldEx.relayAction = 0;
    m_numFieldEx.relayNumber = 0;
    m_numFieldEx.relayTime = 0;

    strcpy( (char *)m_numFieldEx.contentOfConstant, "" );
    memset( &m_numFieldEx.compare, 0, sizeof( m_numFieldEx.compare ) );

    // Textfeld loeschen
    memset( (unsigned char *)&m_textEx, 0, sizeof( StructTextEx ) );

    // Datenfeld loeschen
    memset( (unsigned char *)&m_dataEx, 0, sizeof( StructDataEx ) );
    if ( device.hasSetupType( { SetupType::PZE_MasterIV, SetupType::AE_MasterIV } ) )
    {
        m_dataEx.type = 4; // Fingerprint
    }
    else
    {
        m_dataEx.type = 0; // Transponder
    }
    m_dataEx.time = 1000;
    m_dataEx.whiteOK = 255;

    // Datenfeld loeschen
    memset( (unsigned char *)&m_globVarToHIDEx, 0, sizeof( StructGlobVarToHIDEx ) );

    // Datenfeld loeschen
    memset( reinterpret_cast<unsigned char *>(&m_switchModuleEnergieEx), 0, sizeof( StructSwitchModuleEnergieEx ) );

    // Datenfeld loeschen
    memset( reinterpret_cast<unsigned char *>(&m_systemMessageEx), 0, sizeof( StructSystemMessageEx ) );

    // Datenfeld loeschen
    memset( reinterpret_cast<unsigned char *>(&m_takePhotoEx), 0, sizeof( StructTakePhotoEx ) );

    // Datenfeld loeschen
    memset( reinterpret_cast<uint8_t *>(&m_feverControlEx), 0, sizeof( StructFeverControlEx ) );
    m_feverControlEx.flagUnitFahrenheit = 0;
    m_feverControlEx.flagShowTemperaturValue = 1;
    m_feverControlEx.flagValidMeasureRepeatable = 0;
    m_feverControlEx.flagInvalidMeasureRepeatable = 1;
    m_feverControlEx.measureMin = 30.0f;
    m_feverControlEx.measureMax = 42.0f;
    m_feverControlEx.measureFever = 38.0f;
    m_feverControlEx.indexRecordField = 255;

    // Datenfeld loeschen
    memset( reinterpret_cast<uint8_t *>(&m_multiInputSourceEx), 0, sizeof( StructMultiInputSourceEx ) );
    m_multiInputSourceEx.indexRecordField = 255;

    memset( (unsigned char *)&m_readField, 0, sizeof( StructReadField ) );
    m_readField.type = TBFE_NORM;
    // Namen übernehmen
    strncpy( (char *)m_readField.name, this->name().c_str(), sizeof( m_readField.name ) - 1 );
    m_readField.recordFieldNumber = 255;
    m_multiInputSourceEx.indexRecordField = 255;

    m_readField.ex.numField = m_numFieldEx;
    m_readField.behaviourAtEsc = EV_UPPER_MENUE;
    m_readField.header2Text = 0;
    m_readField.globalVariable = 0;
    m_readField.systemVariable = 0;
    m_readField.barcodeRangeFrom = 0;
    m_readField.barcodeRangeUpTo = 0;
    m_readField.inputTypeCount = 0;
    m_readField.variable1 = 0;
    m_readField.variable2 = 0;


    setInsertEnabled( false );
    setDeleteEnabled( true );

    // Pruefen der Abhaengigkeit zum Parent
    checkMembers();
}


// ########################################################
// Member: EntryInputChain

EntryInputChain::EntryInputChain( const EntryInputChain &entry )
    : EntryBase( entry )
{
    // Member kopieren
    m_operating = entry.m_operating;
    m_readRecord = entry.m_readRecord;
    m_refTable = entry.m_refTable;
    m_refJumpCR = entry.m_refJumpCR;
    m_refJumpESC = entry.m_refJumpESC;

    EntryInputChainField *child;
    EntryInputChainField *childCopy;
    // Alle Childs kopieren
    auto iterChild = entry.constBegin();
    for ( ; iterChild != entry.constEnd(); ++iterChild )
    {
        // neues anfuegen
        childCopy = (EntryInputChainField *)((EntryInputChainField *)*iterChild)->cloneEntry();
        childCopy->setParent( this );
        append( childCopy );
    }

    //##################################################
    //Beim Kopieren der Eingabeketten müssen die Referenzzeiger der Felder
    //auf die neuen Adressen angepasst werden!!!

    // Anpassung bei Eingabeketten
    // - Erhoehen des Bezugszaehlers auf Datensatzbeschreibung
    // - Referenz bei JumpCR anpassen.
    // - Referenz bei JumpESC anpassen.

    // Bezugszaehler erhoehen
    if ( this->m_refTable )
    {
        this->m_refTable->incRefCount();
    }

    int index;
    EntryDevice *device = nullptr;
    EntryBase *handling = nullptr;
    EntryBase *main_menu = nullptr;
    EntryBase *entryBase = nullptr;

    // Auf Rooteintrag wechseln.
    entryBase = entry.parent();
    while ( !entryBase->isType( EntryType::Device ) )
    {
        // Eintrag unter Handling?
        if ( entryBase->isType( EntryType::Handling ) )
        {
            handling = entryBase;
            main_menu = handling->first();
        }

        entryBase = entryBase->parent();
    }
    device = (EntryDevice *)entryBase;

    // Zeiger auf Hauptmenu sicherstellen.
    if ( !main_menu && device->rootHandling().size() )
    {
        main_menu = device->rootHandling().first();
    }

    // Sprungziel nach Bestaetigung mit CR. Uebergeordnetes U-Menue, Naechste E-Kette bzw. Menue und Hauptmenue
    if ( ((EntryInputChain *)&entry)->m_refJumpCR == (EntryBase *)0
         || ((EntryInputChain *)&entry)->m_refJumpCR == (EntryBase *)-1
         || ((EntryInputChain *)&entry)->m_refJumpCR == (EntryBase *)-2 )
    {
        // Nichts zu tun.
    }
    else
    {
        // Versuch die Referenz zu E-Feld aufzuloesen.
        auto index = ((EntryInputChain *)&entry)->indexOf( ((EntryInputChain *)&entry)->m_refJumpCR );
        if ( index == -1 )
        {
            // Versuch die Referenz zu E-Kette oder U-Menue des Uebergeordneten Menueeintrages aufzuloesen.
            index = ((EntryInputChain *)&entry)->parent()->indexOf( ((EntryInputChain *)&entry)->m_refJumpCR );
            if ( index == -1 )
            {
                // Versuch die Referenz zu F-Taste aufzuloesen.
                if ( main_menu )
                {
                    index = main_menu->indexOf( ((EntryInputChain *)&entry)->m_refJumpCR );
                }

                if ( index == -1 )
                {
                    // Umstellen auf Naechste E-Kette bzw. Menue
                    this->m_refJumpCR = (EntryBase *)-1;
                }
                else
                {
                    // Nichts zu tun. Referenz bleibt die gleiche.
                }
            }
            else
            {
                // Verweist auf die eigene Eingabekette?
                if ( ((EntryInputChain *)&entry) == ((EntryInputChain *)&entry)->m_refJumpESC )
                {
                    this->m_refJumpESC = this;
                }
                else
                {
                    // Umstellen auf Naechste E-Kette bzw. Menue
                    this->m_refJumpCR = (EntryBase *)-1;
                }
            }
        }
        else
        {
            EntryBase *copy = this->value( index );
            // Wurde gefunden?
            if ( copy )
            {
                this->m_refJumpCR = copy;
            }
        }
    }

    // Sprungziel nach Bestaetigung mit ESC. Uebergeordnetes U-Menue, Naechste E-Kette bzw. Menue und Hauptmenue
    if ( ((EntryInputChain *)&entry)->m_refJumpESC == (EntryBase *)0
         || ((EntryInputChain *)&entry)->m_refJumpESC == (EntryBase *)-1
         || ((EntryInputChain *)&entry)->m_refJumpESC == (EntryBase *)-2 )
    {
        // Nichts zu tun.
    }
    else
    {
        // Versuch die Referenz zu E-Feld aufzuloesen.
        int index;
        index = ((EntryInputChain *)&entry)->indexOf( ((EntryInputChain *)&entry)->m_refJumpESC );
        if ( index == -1 )
        {
            // Versuch die Referenz zu E-Kette oder U-Menue des Uebergeordneten Menueeintrages aufzuloesen.
            index = ((EntryInputChain *)&entry)->parent()->indexOf( ((EntryInputChain *)&entry)->m_refJumpESC );
            if ( index == -1 )
            {
                // Versuch die Referenz zu F-Taste aufzuloesen.
                if ( main_menu )
                {
                    index = main_menu->indexOf( ((EntryInputChain *)&entry)->m_refJumpESC );
                }

                if ( index == -1 )
                {
                    // Umstellen auf Naechste E-Kette bzw. Menue
                    this->m_refJumpESC = (EntryBase *)-1;
                }
                else
                {
                    // Nichts zu tun. Referenz bleibt die gleiche.
                }
            }
            else
            {
                // Verweist auf die eigene Eingabekette?
                if ( ((EntryInputChain *)&entry) == ((EntryInputChain *)&entry)->m_refJumpESC )
                {
                    this->m_refJumpESC = this;
                }
                else
                {
                    // Umstellen auf Naechste E-Kette bzw. Menue
                    this->m_refJumpESC = (EntryBase *)-1;
                }
            }
        }
        else
        {
            EntryBase *copy = this->value( index );
            // Wurde gefunden?
            if ( copy )
            {
                this->m_refJumpESC = copy;
            }
        }
    }

    // Anpassung bei Eingabekettenfelder
    // - Erhoehen des Bezugszaehlers auf Datenssatzbeschreibungsfeld
    // - Referenz auf JumpESC anpassen.
    // - Referenz auf Jump anpassen.
    // - Referenz auf JumpSelNoEntry anpassen.

    // Referenzen aller Childs anpassen
    iterChild = entry.constBegin();
    auto iterChildCopy = this->constBegin();
    for ( ; iterChild != entry.constEnd() && iterChildCopy != this->constEnd(); ++iterChild, ++iterChildCopy )
    {
        // aktuelles Child
        child = (EntryInputChainField *)*iterChild;
        // aktuelles Child der Kopie
        childCopy = (EntryInputChainField *)*iterChildCopy;

        // Sprungziel nach Bestaetigung mit ESC. Uebergeordnetes U-Menue, Naechste E-Kette bzw. U-Menue, Hauptmenue, Naechstes E-Feld.
        if ( child->m_refJumpESC == (EntryBase *)0
             || child->m_refJumpESC == (EntryBase *)-1
             || child->m_refJumpESC == (EntryBase *)-2
             || child->m_refJumpESC == (EntryBase *)-3 )
        {
            // Nichts zu tun. Referenz bleibt die gleiche.
        }
        else
        {
            // Referenz zeigt wohl auf ein Child der Eingabekette.
            index = ((EntryInputChain *)&entry)->indexOf( child->m_refJumpESC );
            if ( index == -1 )
            {
                // Versuch die Referenz zu E-Kette oder U-Menue des Uebergeordneten Menueeintrages aufzuloesen.
                index = ((EntryInputChain *)&entry)->parent()->indexOf( child->m_refJumpESC );
                if ( index == -1 )
                {
                    // Versuch die Referenz zu F-Taste aufzuloesen.
                    if ( main_menu )
                    {
                        index = main_menu->indexOf( child->m_refJumpESC );
                    }

                    if ( index == -1 )
                    {
                        // Umstellen auf Uebergeordnetes U-Menue
                        childCopy->m_refJumpESC = (EntryBase *)0;
                    }
                    else
                    {
                        // Nichts zu tun. Referenz bleibt die gleiche.
                    }
                }
                else
                {
                    // Verweist auf die eigene Eingabekette?
                    if ( ((EntryInputChain *)&entry) == child->m_refJumpESC )
                    {
                        childCopy->m_refJumpESC = this;
                    }
                    else
                    {
                        // Umstellen auf Uebergeordnetes U-Menue
                        childCopy->m_refJumpESC = (EntryBase *)0;
                    }
                }
            }
            else
            {
                EntryBase *copy = this->value( index );
                // Wurde Eintrag gefunden?
                if ( copy )
                {
                    // Neue Adresse als Referenz uebernehmen.
                    childCopy->m_refJumpESC = copy;
                }
            }
        }

        // Sprungziel bei Vergleich. Uebergeordnetes U-Menue, Naechste E-Kette bzw. U-Menue, Hauptmenue, Naechstes E-Feld.
        if ( child->m_refJump == (EntryBase *)0
             || child->m_refJump == (EntryBase *)-1
             || child->m_refJump == (EntryBase *)-2
             || child->m_refJump == (EntryBase *)-3 )
        {
            // Nichts zu tun.
        }
        else
        {
            // Referenz zeigt wohl auf ein Child der Eingabekette.
            index = ((EntryInputChain *)&entry)->indexOf( child->m_refJump );
            if ( index == -1 )
            {
                // Versuch die Referenz zu E-Kette oder U-Menue des Uebergeordneten Menueeintrages aufzuloesen.
                index = ((EntryInputChain *)&entry)->parent()->indexOf( child->m_refJump );
                if ( index == -1 )
                {
                    // Versuch die Referenz zu F-Taste aufzuloesen.
                    if ( main_menu )
                    {
                        index = main_menu->indexOf( child->m_refJump );
                    }

                    if ( index == -1 )
                    {
                        // Umstellen auf Uebergeordnetes U-Menue
                        childCopy->m_refJump = (EntryBase *)0;
                    }
                    else
                    {
                        // Nichts zu tun. Referenz bleibt die gleiche.
                    }
                }
                else
                {
                    // Verweist auf die eigene Eingabekette?
                    if ( ((EntryInputChain *)&entry) == child->m_refJump )
                    {
                        childCopy->m_refJump = this;
                    }
                    else
                    {
                        // Umstellen auf Uebergeordnetes U-Menue
                        childCopy->m_refJump = (EntryBase *)0;
                    }
                }
            }
            else
            {
                EntryBase *copy = this->value( index );
                // Wurde Eintrag gefunden?
                if ( copy )
                {
                    // Neue Adresse als Referenz uebernehmen.
                    childCopy->m_refJump = copy;
                }
            }
        }

        // Sprungziel bei keinem Eintrag in Liste. Uebergeordnetes U-Menue, Naechste E-Kette bzw. U-Menue, Hauptmenue, Naechstes E-Feld.
        if ( child->m_refJumpSelNoEntry == (EntryBase *)0
             || child->m_refJumpSelNoEntry == (EntryBase *)-1
             || child->m_refJumpSelNoEntry == (EntryBase *)-2
             || child->m_refJumpSelNoEntry == (EntryBase *)-3 )
        {
            // Nichts zu tun.
        }
        else
        {
            // Referenz zeigt wohl auf ein Child der Eingabekette.
            index = ((EntryInputChain *)&entry)->indexOf( child->m_refJumpSelNoEntry );
            if ( index == -1 )
            {
                // Versuch die Referenz zu E-Kette oder U-Menue des Uebergeordneten Menueeintrages aufzuloesen.
                index = ((EntryInputChain *)&entry)->parent()->indexOf( child->m_refJumpSelNoEntry );
                if ( index == -1 )
                {
                    // Versuch die Referenz zu F-Taste aufzuloesen.
                    if ( main_menu )
                    {
                        index = main_menu->indexOf( child->m_refJumpSelNoEntry );
                    }

                    if ( index == -1 )
                    {
                        // Umstellen auf Uebergeordnetes U-Menue
                        childCopy->m_refJumpSelNoEntry = (EntryBase *)0;
                    }
                    else
                    {
                        // Nichts zu tun. Referenz bleibt die gleiche.
                    }
                }
                else
                {
                    // Verweist auf die eigene Eingabekette?
                    if ( ((EntryInputChain *)&entry) == child->m_refJumpSelNoEntry )
                    {
                        childCopy->m_refJumpSelNoEntry = this;
                    }
                    else
                    {
                        // Umstellen auf Uebergeordnetes U-Menue
                        childCopy->m_refJumpSelNoEntry = (EntryBase *)0;
                    }
                }
            }
            else
            {
                EntryBase *copy = this->value( index );
                // Wurde Eintrag gefunden?
                if ( copy )
                {
                    // Neue Adresse als Referenz uebernehmen.
                    childCopy->m_refJumpSelNoEntry = copy;
                }
            }
        }
    }
}

///////////////////////////////////
// Childs

EntryBase *EntryInputChain::insertEntry( EntryType typeID, int pos )
{
    if ( size() >= (MAX_SUBPOINTS - 1) )
    {
        // Die maximale Anzahl Unterpunkte ( MAX_SUBPOINTS - 1 ) ist erreicht, ein weiteres Einfügen ist nicht möglich.
        return nullptr;
    }

    EntryInputChainField *entryNew = new EntryInputChainField( this );
    entryNew->setDefault( boost::str( boost::format( "%1%%2%" ) % QT_TRANSLATE_NOOP_UTF8( "EntryInputChain", "Eingabe" ) % (size() + 1) ) );

    insertEntryFinish( entryNew, pos );

    return entryNew;
}

EntryBase *EntryInputChain::cloneEntry()
{
    return reinterpret_cast<EntryBase *>(new EntryInputChain( *this ));
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryInputChain::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    unsigned int aP[MAX_SUBPOINTS];
    int i;

    // Daten von Bedien - Struktur uebernehmen
    m_operating.ptr = pFlash + sizeof( StructOperating );
    m_operating.ptr += m_flashOffset;
    m_operating.ptr = endianReverse( m_operating.ptr + FLASH_KONFIG_START );

    *(StructOperating *)&file.m_flash[pFlash] = m_operating;
    pFlash += sizeof( StructOperating );

    fileXML.writeContent( FileXML::ContentType::Operating, as_const( m_operating ) );

    // Daten von DatensatzEinlesen - Struktur uebernehmen

    // Anzahl Eintraege berechnen
    m_readRecord.fields = (unsigned char)size();
    if ( m_readRecord.fields > MAX_SUBPOINTS )
        m_readRecord.fields = MAX_SUBPOINTS;

    // Wenn keine Datensatzbeschreibung zugeordnet,
    // dann kann nicht abgespeichert werden.
    if ( m_readRecord.number == 255 )
        m_readRecord.flags |= BDEF_NICHT_SPEICHERN;
    else
        m_readRecord.flags &= ~BDEF_NICHT_SPEICHERN;

    // DatensatzEinlesen.ptr berechnen und eintragen
    m_readRecord.ptr = pFlash + sizeof( StructReadRecord );
    m_readRecord.ptr += m_flashOffset;
    m_readRecord.ptr = endianReverse( m_readRecord.ptr + FLASH_KONFIG_START );

    *(StructReadRecord *)&file.m_flash[pFlash] = m_readRecord;
    pFlash += sizeof( StructReadRecord );

    // The parent is configured outside the method and is not changed.
    fileXML.writeContent( FileXML::ContentType::ReadRecord, as_const( m_readRecord ) );

    // Pruefung auf benoetigte Mindestversion
    if ( ((m_readRecord.behaviourAtCR >= 200 && m_readRecord.behaviourAtCR <= 229) || m_readRecord.behaviourAtCR == 254)
         || ((m_readRecord.behaviourAtESC >= 200 && m_readRecord.behaviourAtESC <= 229) || m_readRecord.behaviourAtESC == 254) )
    {
        // ESC-Verhalten bei E-Menue wird erst ab Versionen x unterstuetzt.
        file.NeedStudioVersion( "04.02.02.08" );
        file.NeedFirmware2Version( "04.02.02.17" );
        file.UnsupportFirmware1Version();
    }

    // Pruefung auf benoetigte Mindestversion
    if ( m_readRecord.vehicleRelease & 0x80 )
    {
        // Fahrzeugfreigabe wird erst ab Versionen x unterstuetzt.
        file.NeedStudioVersion( "04.02.04.03" );
        file.NeedFirmware2Version( "04.02.04.10" );
        file.UnsupportFirmware1Version();
    }

    // Alle Childeintraege ablegen
    // erster freier Speicherplatz nach DatensatzEinlesen -> begin Pointerarray
    aP[0] = pFlash + sizeof( unsigned int ) * m_readRecord.fields; // Platz fuer Pointerarray reservieren

    // Finding the node for this structure and update parent.
    auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::ReadRecord ) );

    auto iterChild = constBegin();
    for ( i = 1; iterChild != constEnd(); ++iterChild, ++i )
    {
        aP[i] = (*iterChild)->writeDocument( aP[i - 1], doc );
    }

    // Reset parent node because the end of the structure is reached.
    fileXML.setParentNode( oldParent );

    // Pointer auf Unterpunkte eintragen
    for ( i = 0; i < m_readRecord.fields; ++i )
    {
        aP[i] += m_flashOffset;
        aP[i] = endianReverse( aP[i] + FLASH_KONFIG_START );
        *(unsigned int *)&file.m_flash[pFlash + i * sizeof( unsigned int )] = aP[i];
    }

    return aP[m_readRecord.fields];
}

uint32_t EntryInputChain::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    unsigned int pDat{};
    tinyxml2::XMLElement *lastXmlElement{};

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    if ( useAESFile )
    {
        // Daten von Bedien - Struktur uebernehmen
        m_operating = *(StructOperating *)&file.m_flash[pFlash];
        m_operating.ptr = endianReverse( m_operating.ptr ) - FLASH_KONFIG_START;
        // Offset abziehen, da Erweitertes - Flash hinter Setupdaten.
        m_operating.ptr -= m_flashOffset;
        pFlash += sizeof( StructOperating );

        // Daten von DatensatzEinlesen - Struktur uebernehmen
        m_readRecord = *(StructReadRecord *)&file.m_flash[pFlash];
        m_readRecord.ptr = endianReverse( m_readRecord.ptr ) - FLASH_KONFIG_START;
        // Offset abziehen, da Erweitertes - Flash hinter Setupdaten.
        m_readRecord.ptr -= m_flashOffset;
        pFlash += sizeof( StructReadRecord );
    }
    else
    {
        // pFlash contains the information about the corresponding index of the previous sibling element to be read.
        // The readDocument method searches depending on the parent or the previous element.

        // Daten von Bedien - Struktur uebernehmen
        fileXML.combineValid( fileXML.readContent( FileXML::ContentType::Operating, m_operating, pFlash, lastXmlElement ) );

        // Daten von DatensatzEinlesen - Struktur uebernehmen
        // Different entry types can occur. Therefore, indexing based on the type is not correct. Since each type is
        // preceded by an 'operating' element, this entry is used as the basis for the search.
        fileXML.combineValid( fileXML.readContent( FileXML::ContentType::ReadRecord, m_readRecord, lastXmlElement ) );
    }

    // Name fuer Baumansicht eintragen
    setName( m_readRecord.name );

    // Initalisierung eines Wertes aufgrund aelterer Setupdaten.
    SoftwarebuildVersion studioVersion;
    if ( useAESFile )
    {
        studioVersion = file.m_file_header.studioVersion;
    }
    else
    {
        studioVersion = fileXML.studioVersion();
    }

    if ( studioVersion < "04.02.04.00" )
    {
        m_readRecord.vehicleRelease = 0x40; // Fahrzeugfreigabe.
    }

    // Finding the node for this structure and update parent.
    auto oldParent = fileXML.setParentNode( lastXmlElement );

    // Daten aller Eintraege einlesen
    for ( int i = 0, skipMulti = 0; i < (int)m_readRecord.fields; ++i )
    {
        EntryInputChainField *pNeu = new EntryInputChainField( this );
        pNeu->setDefault();

        if ( useAESFile )
        {
            pDat = m_readRecord.ptr + i * sizeof( unsigned int );	// Pointer des Eintrags berechnen

            pDat = *(unsigned int *)&file.m_flash[pDat];
            pDat = endianReverse( pDat ) - FLASH_KONFIG_START;
            // Bei ExtendedFlash mit Offset
            pDat -= m_flashOffset;

            pNeu->readDocument( pDat, doc );
        }
        else
        {
            pNeu->readDocument( i + skipMulti, doc );
            if ( pNeu->fieldFunction() == TBFE_MULTI_INPUT_SOURCE )
            {
                skipMulti += pNeu->m_multiInputSourceEx.childCount;
            }
        }

        append( pNeu );
    }

    // Reset parent node because the end of the structure is reached.
    fileXML.setParentNode( oldParent );

    return pFlash;
}

// Pruefen der Member in Abhaengigkeit zum derzeitigen Parent
void EntryInputChain::checkMembers()
{
    EntryDevice &device = root();

    bool  boxDevice = false;
    bool  ipPCDevice = false;
    bool tbDockDevice = false;

    if ( device.hasSetupType( { SetupType::KYO_Inloc, SetupType::ZK_MasterIV, SetupType::IO_Box, SetupType::MDE_BoxIV, SetupType::Mobil_MasterIV, SetupType::Mobil_Box, SetupType::KYO_Oneloc, SetupType::KYO_Cenloc } ) )
    {
        boxDevice = true;
    }
    else if ( device.hasSetupType( { SetupType::IPC_Extension } ) )
    {
        ipPCDevice = true;
    }
    else if ( device.hasSetupType( { SetupType::Timeboy_TBDock, SetupType::Mobil_TBDock } ) )
    {
        tbDockDevice = true;
    }

    if ( (parentType() == EntryType::EventChains) || (parentType() == EntryType::AccessControl) )
        m_readRecord.flags = 0;
    else
    {
        if ( boxDevice || ipPCDevice || tbDockDevice )
            m_readRecord.flags = 0;
        else
            m_readRecord.flags = BDEF_ZUSATZ_RETURN;
    }
    // Ist Eingabekette unter Zutrittskontrolle
    if ( parentType() == EntryType::AccessControl )
        setDeleteEnabled( FALSE );
}

///////////////////////////////////
// Referenzierung zwischen Einträgen

void EntryInputChain::convertReferenceToIndex()
{
    int index;
    EntryDevice *device = nullptr;
    EntryBase *handling = nullptr;
    EntryBase *main_menu = nullptr;
    EntryBase *entry = nullptr;

    // Auf Rooteintrag wechseln.
    entry = this->parent();
    while ( !entry->isType( EntryType::Device ) )
    {
        // Eintrag unter Handling?
        if ( entry->isType( EntryType::Handling ) )
        {
            handling = entry;
            main_menu = handling->first();
        }

        entry = entry->parent();
    }
    device = (EntryDevice *)entry;

    // Zeiger auf Hauptmenu sicherstellen.
    if ( !main_menu && device->rootHandling().size() )
        main_menu = device->rootHandling().first();

    // Zugeordnete Tabellenummer eintragen
    if ( m_refTable )
    {

        m_readRecord.number = m_refTable->parent()->indexOf( m_refTable );
    }
    else
    {

        m_readRecord.number = 255;
    }

    // Sprungziel nach Enter Ermitteln und Eintragen
    if ( m_refJumpCR == (EntryBase *)0 )
    {
// Keine Auswahl oder Uebergeordnetes Untermenue gewaehlt.
        m_readRecord.behaviourAtCR = EV_UPPER_MENUE;

    }
    else if ( m_refJumpCR == (EntryBase *)-1 )
    {
// Naechste Eingabekette bzw. Menue gewaehlt.
        m_readRecord.behaviourAtCR = EV_NEXT_SIBLING;

    }
    else if ( m_refJumpCR == (EntryBase *)-2 )
    {
// Hauptmenue gewaehlt.
        m_readRecord.behaviourAtCR = EV_MAIN_MENUE;

    }
    else
    {

     // Entweder ist das Sprungziel F-Taste, E-Kette bzw. U-Menue oder E-Feld

     // Versuch die Referenz zu E-Feld aufzuloesen.
        index = indexOf( m_refJumpCR );
        if ( index == -1 )
        {

// Versuch die Referenz zu E-Kette bzw. U-Menue aufzuloesen.
            index = parent()->indexOf( m_refJumpCR );
            if ( index == -1 )
            {
// Versuch die Referenz zu F-Taste aufzuloesen.
                if ( main_menu )
                {
                    index = main_menu->indexOf( m_refJumpCR );
                }

                if ( index == -1 )
                {
  // Von Naechster E-Kette bzw. U-Menue ausgehen
                    m_readRecord.behaviourAtCR = EV_NEXT_SIBLING;

                }
                else
                {
                             // F-Taste
                    m_readRecord.behaviourAtCR = index + 112; // 112 => F1
                }
            }
            else
            {
                     // Bis zu 30 erlaubt.
                if ( index >= 30 ) index = 0;
                // E-Kette bzw. U-Menue
                m_readRecord.behaviourAtCR = index + 200;
            }
        }
        else
        {
             // Bis zu 110 E-Felder erlaubt.
            if ( index >= 110 ) index = -2;
            // E-Feld
            m_readRecord.behaviourAtCR = index + 2;
        }
    }
    // Sprungziel nach ESC Ermitteln und Eintragen
    if ( m_refJumpESC == (EntryBase *)0 )
    {
// Keine Auswahl oder Uebergeordnetes Untermenue gewaehlt.
        m_readRecord.behaviourAtESC = EV_UPPER_MENUE;

    }
    else if ( m_refJumpESC == (EntryBase *)-1 )
    {
// Naechste Eingabekette bzw. Menue gewaehlt.
        m_readRecord.behaviourAtESC = EV_NEXT_SIBLING;

    }
    else if ( m_refJumpESC == (EntryBase *)-2 )
    {
// Hauptmenue gewaehlt.
        m_readRecord.behaviourAtESC = EV_MAIN_MENUE;

    }
    else
    {

     // Entweder ist das Sprungziel E-Feld, E-Kette bzw. U-Menue, F-Taste   

     // Versuch die Referenz zu E-Feld aufzuloesen.
        index = indexOf( m_refJumpESC );
        if ( index == -1 )
        {

// Versuch die Referenz zu E-Kette bzw. U-Menue aufzuloesen.
            index = parent()->indexOf( m_refJumpESC );
            if ( index == -1 )
            {
// Versuch die Referenz zu F-Taste aufzuloesen.
                if ( main_menu )
                {
                    index = main_menu->indexOf( m_refJumpESC );
                }

                if ( index == -1 )
                {
  // Von Naechster E-Kette bzw. U-Menue ausgehen
                    m_readRecord.behaviourAtESC = EV_NEXT_SIBLING;

                }
                else
                {
                             // F-Taste
                    m_readRecord.behaviourAtESC = index + 112; // 112 => F1
                }
            }
            else
            {
                     // Bis zu 30 erlaubt.
                if ( index >= 30 ) index = 0;
                // E-Kette bzw. U-Menue
                m_readRecord.behaviourAtESC = index + 200;
            }
        }
        else
        {
             // Bis zu 110 E-Felder erlaubt.
            if ( index >= 110 ) index = -2;
            // E-Feld
            m_readRecord.behaviourAtESC = index + 2;
        }
    }

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertReferenceToIndex();
}

void EntryInputChain::convertIndexToReference()
{
    int index;
    EntryDevice *device = nullptr;
    EntryBase *handling = nullptr;
    EntryBase *main_menu = nullptr;
    EntryBase *entry = nullptr;

    // Auf Rooteintrag wechseln.
    entry = this->parent();
    while ( !entry->isType( EntryType::Device ) )
    {
        // Eintrag unter Handling?
        if ( entry->isType( EntryType::Handling ) )
        {
            handling = entry;
            main_menu = handling->first();
        }

        entry = entry->parent();
    }
    device = (EntryDevice *)entry;

    // Zeiger auf Hauptmenu sicherstellen.
    if ( !main_menu && device->rootHandling().size() )
        main_menu = device->rootHandling().first();

    // Die referenzierte Tabellenklasse ermitteln
    if ( m_readRecord.number != 0xFF )
    {

        // Referenz zur zugeordneten Tabelle ermitteln.
        EntryTableRecord &tableRecord = device->rootTableRecord();
        if ( tableRecord.size() > m_readRecord.number )
        {
            m_refTable = (EntryTable *)device->rootTableRecord().at( m_readRecord.number );
        }
        else
        {
            m_refTable = nullptr;
        }

        if ( m_refTable )
        {
            m_refTable->incRefCount();
        }

    }
    else
    {

     // Keine zugewiesen
        m_refTable = nullptr;
    }

    index = m_readRecord.behaviourAtCR;
    // Sprungziel nach Enter
    switch ( index )
    {
        case EV_UPPER_MENUE:
        case 1: // Ausschalten faellt weg und wird zu Uebergeordnetem Untermenue.
            m_refJumpCR = (EntryBase *)0;
            break;
        case EV_NEXT_SIBLING:
            m_refJumpCR = (EntryBase *)-1;
            break;
        case EV_MAIN_MENUE:
            m_refJumpCR = (EntryBase *)-2;
            break;
        default:
            m_refJumpCR = (EntryBase *)0;
            if ( index >= 2 && index <= 111 )
            {
// E-Feld
                m_refJumpCR = value( index - 2 );

            }
            else if ( index >= 200 && index <= 229 )
            {
// E-Kette bzw. Menue.
                m_refJumpCR = parent()->value( index - 200 );

            }
            else if ( index >= 112 && index <= 127 )
            {
// F-Taste
                m_refJumpCR = main_menu->value( index - 112 ); // 112 => F1

            }

            if ( !m_refJumpCR )
            {
// Von Naechster E-Kette bzw. U-Menue ausgehen.
                m_refJumpCR = (EntryBase *)-1;
            }
            break;
    }
    if ( m_refJumpCR != (EntryBase *)0 && m_refJumpCR != (EntryBase *)-1 && m_refJumpCR != (EntryBase *)-2 )
    {

        m_refJumpCR->incRefCount();
    }

    index = m_readRecord.behaviourAtESC;
    // Sprungziel nach ESC
    switch ( index )
    {
        case EV_UPPER_MENUE:
        case 1: // Ausschalten faellt weg und wird zu Uebergeorndetem Untermenue.
            m_refJumpESC = (EntryBase *)0;
            break;
        case EV_NEXT_SIBLING:
            m_refJumpESC = (EntryBase *)-1;
            break;
        case EV_MAIN_MENUE:
            m_refJumpESC = (EntryBase *)-2;
            break;
        default:
            m_refJumpESC = (EntryBase *)0;
            if ( index >= 2 && index <= 111 )
            {
// E-Feld
                m_refJumpESC = value( index - 2 );

            }
            else if ( index >= 200 && index <= 229 )
            {
// E-Kette bzw. Menue.
                m_refJumpESC = parent()->value( index - 200 );

            }
            else if ( index >= 112 && index <= 127 )
            {
// F-Taste
                if ( main_menu )
                    m_refJumpESC = main_menu->value( index - 112 ); // 112 => F1

            }
            if ( !m_refJumpESC )
            {
// Von Naechster E-Kette bzw. U-Menue ausgehen.
                m_refJumpESC = (EntryBase *)-1;
            }
            break;
    }
    if ( m_refJumpESC != (EntryBase *)0 && m_refJumpESC != (EntryBase *)-1 && m_refJumpESC != (EntryBase *)-2 )
    {

        m_refJumpESC->incRefCount();
    }

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertIndexToReference();
}

///////////////////////////////////
// Daten des Eintrags

void EntryInputChain::setDefault()
{
    setDefault( QT_TRANSLATE_NOOP_UTF8( "EntryInputChain", "Eingabekette" ), TRUE );
}

void EntryInputChain::setDefault( const std::string &name, bool visible )
{
    setName( name );

    m_iconIndex = ICON_INPUT_CHAIN;
    m_iconIndexSelected = ICON_INPUT_CHAIN_SEL;

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_INPUT_CHAIN_0 );

    // Bedienstruktur aufsetzen
    memset( (unsigned char *)&m_operating, 0, sizeof( StructOperating ) );
    m_operating.type = TB_DATENSATZ;
    m_operating.d1 = m_operating.d2 = m_operating.type;

    // Eingabekettenfeld aufsetzen
    memset( (unsigned char *)&m_readRecord, 0, sizeof( StructReadRecord ) );
    m_readRecord.number = 255;
    m_readRecord.fields = 1;
    m_readRecord.behaviourAtCR = EV_NEXT_SIBLING;
    m_readRecord.behaviourAtESC = EV_NEXT_SIBLING;
    m_readRecord.type = TBDA_NORM;

    strncpy( (char *)m_readRecord.name, this->name().c_str(), sizeof( m_readRecord.name ) - 1 );
    m_readRecord.header1Text = 0;
    m_readRecord.vehicleRelease = 0x40;


    // Relais
    m_readRecord.relayNum = 0;
    m_readRecord.relayTime = 3;
    m_readRecord.relayTime = endianReverse( m_readRecord.relayTime );

    m_readRecord.unused = 0;

    // Zusaetzliche Members
    setVisible( visible );
    setInsertEnabled( TRUE );
    setDeleteEnabled( TRUE );

    // Pruefen der Abhaengigkeit zum Parent
    checkMembers();
}

// ########################################################
// Member: EntryIOModule

///////////////////////////////////
// Childs

EntryBase *EntryIOModule::insertEntry( EntryType typeID, int pos )
{
    (void)typeID;
    (void)pos;

    return nullptr;
}

EntryBase *EntryIOModule::cloneEntry()
{
    // Dieser Eintrag unterstützt derzeit cloneEntry nicht.
    return nullptr;
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryIOModule::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    // Wenn IO-Module aktiviert ist, Daten übernehmen.
    if ( root().m_config.ioModuleActive )
    {
        // Die Extended - Struktur liegt als erste im erweiterten Flash.
        StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;
        unsigned int j;

        extBegin.offsetIOModule = extBegin.offsetFreeScope;
        // Nach ihr folgt die IOModul - Struktur.
        *(StructIOModule *)&file.m_flashExtended[extBegin.offsetIOModule] = m_ioModule;
        extBegin.offsetFreeScope += sizeof( StructIOModule );

        ////////////////////////////////////////
        // add information to xml file
        FileXML &fileXML = doc.fileXML();

        // The parent is configured outside the method and is not changed.
        fileXML.writeContent( FileXML::ContentType::IOModule, as_const( m_ioModule ) );

        // Wechseln der Byteorder auf Motorolla - Format
        for ( j = 0; j < 8; ++j )
        {
            m_digitalInput[j].countDivisor = endianReverse( m_digitalInput[j].countDivisor );
            m_digitalInput[j].countDivisor32Bit = endianReverse( m_digitalInput[j].countDivisor32Bit );
            m_digitalInput[j].cycle = endianReverse( m_digitalInput[j].cycle );
            m_digitalInput[j].timeout = endianReverse( m_digitalInput[j].timeout );
        }

        extBegin.offsetIOModule8DigitalInputs = extBegin.offsetFreeScope;
        // Nach ihr folgen die 8 DigitalMDE - Struktur fuer E10 - E17.
        for ( j = 0; j < 8; ++j )
        {
            *(StructDigitalInput *)&file.m_flashExtended[extBegin.offsetIOModule8DigitalInputs + j * sizeof( StructDigitalInput )] = m_digitalInput[j];
        }

        extBegin.offsetFreeScope += 8 * sizeof( StructDigitalInput );

        // Wechseln der Byteorder auf Intel - Format
        for ( j = 0; j < 8; ++j )
        {
            m_digitalInput[j].countDivisor = endianReverse( m_digitalInput[j].countDivisor );
            m_digitalInput[j].countDivisor32Bit = endianReverse( m_digitalInput[j].countDivisor32Bit );
            m_digitalInput[j].cycle = endianReverse( m_digitalInput[j].cycle );
            m_digitalInput[j].timeout = endianReverse( m_digitalInput[j].timeout );
        }

        // Extended Struktur uebernehmen
        *(StructExtendedFlash *)file.m_flashExtended = extBegin;

        // Finding the node for this structure and update parent.
        auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::IOModule ) );
        fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::DigitalInputs ) );

        // Nach ihr folgen die 8 DigitalMDE - Struktur fuer E10 - E17.
        for ( j = 0; j < 8; ++j )
        {
            // The parent is configured outside the method and is not changed.
            fileXML.writeContent( FileXML::ContentType::DigitalInput, as_const( m_digitalInput[j] ) );
        }

        // Reset parent node because the end of the structure is reached.
        fileXML.setParentNode( oldParent );
    }

    return pFlash;

}

uint32_t EntryIOModule::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    // Die Extended - Struktur liegt als erste im erweiterten Flash.

    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    int j;
    // Pruefen der Extended - Struktur
    StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    if ( useAESFile )
    {
        if ( extBegin.offsetIOModule != 0 )
        {
            m_ioModule = *(StructIOModule *)&file.m_flashExtended[extBegin.offsetIOModule];
        }

        if ( extBegin.offsetIOModule8DigitalInputs != 0 )
        {
            // Strukturen fuer Digital - Input lesen
            for ( j = 0; j < 8; ++j )
            {
                m_digitalInput[j] = *(StructDigitalInput *)&file.m_flashExtended[extBegin.offsetIOModule8DigitalInputs + j * sizeof( StructDigitalInput )];
                if ( m_digitalInput[j].countDivisor32Bit == 0xCDCDCDCD )
                {
                    // 2017-11-03: BOt
                    // Bei der Übernahme "alter" Setupdateien eventuell nicht initalisierten Strukturbereich initalisieren.
                    auto offs = offsetof( StructDigitalInput, countDivisor32Bit );
                    auto size = sizeof( m_digitalInput[j] ) - offs;
                    memset( reinterpret_cast<unsigned char *>(&m_digitalInput[j]) + offs, 0, size );
                }
            }

            // Wechseln der Byteorder auf Intel - Format
            for ( j = 0; j < 8; ++j )
            {
                m_digitalInput[j].countDivisor = endianReverse( m_digitalInput[j].countDivisor );
                m_digitalInput[j].countDivisor32Bit = endianReverse( m_digitalInput[j].countDivisor32Bit );
                m_digitalInput[j].cycle = endianReverse( m_digitalInput[j].cycle );
                m_digitalInput[j].timeout = endianReverse( m_digitalInput[j].timeout );

                // 2006-10-09: BOt
                // Zaehlerteiler 32bit eingefuehrt. Sicherstellen, dass der 32bit - Wert dem 16bit - Wert angepasst ist.
                if ( (m_digitalInput[j].countDivisor > 0) && (m_digitalInput[j].countDivisor32Bit == 0) )
                    m_digitalInput[j].countDivisor32Bit = m_digitalInput[j].countDivisor;
            }
        }
    }
    else
    {
        if ( fileXML.numberOfChildItems( FileXML::ContentType::IOModule ) > 0 )
        {
            tinyxml2::XMLElement *lastXmlElement{};
            fileXML.combineValid( fileXML.readContent( FileXML::ContentType::IOModule, m_ioModule, pFlash, lastXmlElement ) );

            if ( fileXML.isValid() )
            {
                // Finding the node for this structure and update parent.
                auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::IOModule ) );
                fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::DigitalInputs ) );

                constexpr uint32_t structItemCount = sizeof( m_digitalInput ) / sizeof( m_digitalInput[0] );
                uint32_t size = fileXML.numberOfChildItems( FileXML::ContentType::DigitalInput );

                // Check the number of digital inputs defined in the XML. If the number does not match the definition of the
                // data structure 'm_digitalE', there is an error in the XML description.
                fileXML.combineValid( size == structItemCount );
                if ( size == structItemCount )
                {
                    tinyxml2::XMLElement *lastXmlElement{};
                    for ( uint32_t i = 0; i < size; ++i )
                    {
                        fileXML.combineValid( fileXML.readContent( FileXML::ContentType::DigitalInput, m_digitalInput[i], i, lastXmlElement ) );
                        if ( !fileXML.isValid() )
                        {
                            break;
                        }
                    }
                }

                // Reset parent node because the end of the structure is reached.
                fileXML.setParentNode( oldParent );
            }
        }
    }

    return pFlash;
}

///////////////////////////////////////
// Referenzierung zwischen Einträgen

void EntryIOModule::convertReferenceToIndex()
{
    int i;
    EntryEventChains &eventChains = rootEventChains();

    // Fuer jeden digitalen Eingang.
    for ( i = 0; i < 8; ++i )
    {
        m_digitalInput[i].startLowHigh = (m_refInputChainDigital[0][i] != nullptr) ? eventChains.indexOf( m_refInputChainDigital[0][i] ) + 1 : 0;
        m_digitalInput[i].startHighLow = (m_refInputChainDigital[1][i] != nullptr) ? eventChains.indexOf( m_refInputChainDigital[1][i] ) + 1 : 0;
        m_digitalInput[i].invokeInputChainCountDivisor = (m_refInputChainDigital[2][i] != nullptr) ? eventChains.indexOf( m_refInputChainDigital[2][i] ) + 1 : 0;
        m_digitalInput[i].invokeInputChainCycle = (m_refInputChainDigital[3][i] != nullptr) ? eventChains.indexOf( m_refInputChainDigital[3][i] ) + 1 : 0;
    }

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertReferenceToIndex();
}

void EntryIOModule::convertIndexToReference()
{
    int i;
    EntryEventChains &eventChains = rootEventChains();

    // Fuer jeden digitalen Eingang.
    for ( i = 0; i < 8; ++i )
    {

        m_refInputChainDigital[0][i] = (m_digitalInput[i].startLowHigh > 0) ? (EntryInputChain *)eventChains.at( m_digitalInput[i].startLowHigh - 1 ) : nullptr;
        if ( m_refInputChainDigital[0][i] )
            m_refInputChainDigital[0][i]->incRefCount();

        m_refInputChainDigital[1][i] = (m_digitalInput[i].startHighLow > 0) ? (EntryInputChain *)eventChains.at( m_digitalInput[i].startHighLow - 1 ) : nullptr;
        if ( m_refInputChainDigital[1][i] )
            m_refInputChainDigital[1][i]->incRefCount();

        m_refInputChainDigital[2][i] = (m_digitalInput[i].invokeInputChainCountDivisor > 0) ? (EntryInputChain *)eventChains.at( m_digitalInput[i].invokeInputChainCountDivisor - 1 ) : nullptr;
        if ( m_refInputChainDigital[2][i] )
            m_refInputChainDigital[2][i]->incRefCount();

        m_refInputChainDigital[3][i] = (m_digitalInput[i].invokeInputChainCycle > 0) ? (EntryInputChain *)eventChains.at( m_digitalInput[i].invokeInputChainCycle - 1 ) : nullptr;
        if ( m_refInputChainDigital[3][i] )
            m_refInputChainDigital[3][i]->incRefCount();
    }

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertIndexToReference();
}

///////////////////////////////////
// Daten des Eintrags

void EntryIOModule::setDefault()
{
    m_iconIndex = ICON_IOMODULE;
    m_iconIndexSelected = ICON_IOMODULE_SEL;

    setName( QT_TRANSLATE_NOOP_UTF8( "EntryIOModule", "IO - Modul" ) );

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_IOMODULE_0 );
    m_pageIndexes.push_back( PAGE_IOMODULE_1 );
    m_pageIndexes.push_back( PAGE_IOMODULE_2 );
    m_pageIndexes.push_back( PAGE_IOMODULE_3 );
    m_pageIndexes.push_back( PAGE_IOMODULE_4 );
    m_pageIndexes.push_back( PAGE_IOMODULE_5 );

    // IOModul - Struct mit 0 Init
    memset( &m_ioModule, 0, sizeof( StructIOModule ) );
    defaultIOModule( &m_ioModule );

    // Strukturen fuer Digitale - Eingaenge loeschen.
    memset( m_digitalInput, 0, 8 * sizeof( StructDigitalInput ) );

    // Referenzbezug initialisieren
    memset( m_refInputChainDigital, 0, sizeof( m_refInputChainDigital ) );

    setInsertEnabled( FALSE );
    setDeleteEnabled( FALSE );
}

void EntryIOModule::defaultIOModule( StructIOModule *pIOModule )
{
    int i;
    pIOModule->voltageInputsActive = 0x00;
    pIOModule->voltageInputsShowOnScreen = 0x00;
    pIOModule->currentInputsActive = 0x00;
    pIOModule->currentInputsShowOnScreen = 0x00;

    // Setzen der default Werte.
    for ( i = 0; i < 8; ++i )
    {
        // Spannungskanaele
        sprintf( (char *)pIOModule->voltageInputsName[i], "U%i", i + 1 );
        sprintf( (char *)pIOModule->voltageInputsUnit[i], "V" );
        pIOModule->voltageInputsMinValue[i] = 0;
        pIOModule->voltageInputsMaxValue[i] = 10;

        // Stromkanaele
        sprintf( (char *)pIOModule->currentInputsName[i], "I%i", i + 1 );
        sprintf( (char *)pIOModule->currentInputsUnit[i], "mA" );
        pIOModule->currentInputsMinValue[i] = 0;
        pIOModule->currentInputsMaxValue[i] = 20;

        // Strukturen fuer Digitale - Eingaenge loeschen.
        memset( m_digitalInput, 0, sizeof( m_digitalInput ) );
    }
}


// ########################################################
// Member: EntrySignalProcessing

///////////////////////////////////////
// Childs

EntryBase *EntrySignalProcessing::insertEntry( EntryType typeID, int pos )
{
    (void)typeID;
    (void)pos;

    return nullptr;
}

EntryBase *EntrySignalProcessing::cloneEntry()
{
    // Dieser Eintrag unterstützt derzeit cloneEntry nicht.
    return nullptr;
}

///////////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntrySignalProcessing::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    // Erweiterter Flashbereich
    FileAES &file = doc.fileAES();
    int j;
    StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

    ////////////////////////////////////////
    // add information to xml file
    FileXML &fileXML = doc.fileXML();

    // Finding the node for this structure and update parent.
    auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::SignalProcessing ) );

    // Nur im Terminalmode MDE und mindestens einer aktivierten Maschine
    if ( root().m_config.operatingMode == 3 && (m_machineDefinition[0].flags & 0x01) != 0 )
    {
        // Finding the node for this structure and update parent.
        auto oldParent_2 = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::MachineDefinitions ) );

        // Strukturen der Maschinendefinitionen im naechsten freien Speicherblock ablegen
        extBegin.offsetMachineDefinitions = extBegin.offsetFreeScope;
        // Schreiben der maximal 6 Strukturen fuer die Maschinendefinitionen
        for ( j = 0; j < 6; ++j )
        {
            if ( !(m_machineDefinition[j].flags & 0x01) )
                break;

            *(StructMachineDefinition *)&file.m_flashExtended[extBegin.offsetMachineDefinitions + sizeof( unsigned char ) + j * sizeof( StructMachineDefinition )] = m_machineDefinition[j];

            // The parent is configured outside the method and is not changed.
            fileXML.writeContent( FileXML::ContentType::Machine, as_const( m_machineDefinition[j] ) );
        }

        // Reset parent node because the end of the structure is reached.
        fileXML.setParentNode( oldParent_2 );

        if ( j )
        {
            file.m_flashExtended[extBegin.offsetMachineDefinitions] = (unsigned char)j;
            extBegin.offsetFreeScope += (j + 1) * sizeof( StructMachineDefinition );
        }
    }

    // Bei Verwendung des Vibrationssensors, diesen eintragen.
    if ( m_vibrationSensor.useAs != 0 )
    {
        // Struktur des Vibrationssensors im naechsten freien Speicherblock ablegen.
        extBegin.offsetVibrationSensor = extBegin.offsetFreeScope;

        *(StructVibrationSensor *)&file.m_flashExtended[extBegin.offsetVibrationSensor] = m_vibrationSensor;

        // The parent is configured outside the method and is not changed.
        fileXML.writeContent( FileXML::ContentType::VibrationSensor, as_const( m_vibrationSensor ) );

        extBegin.offsetFreeScope += sizeof( StructVibrationSensor );
    }

    // Schreiben der Erweiterten Struktur
    *(StructExtendedFlash *)file.m_flashExtended = extBegin;

    // Alle Unterpunkte int das Speicherimage eintragen
    auto iterChild = constBegin();
    for ( ; iterChild != constEnd(); ++iterChild )
        pFlash = (*iterChild)->writeDocument( pFlash, doc );

    // Eintrag für Systemmeldungen schreiben.
    pFlash = m_systemMessages->writeDocument( pFlash, doc );

    // Reset parent node because the end of the structure is reached.
    fileXML.setParentNode( oldParent );

    return pFlash;
}

uint32_t EntrySignalProcessing::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    // Erweiterten Struktur lesen
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    // Finding the node for this structure and update parent.
    auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::SignalProcessing ) );

    // Maschinendefinitionen vorhanden?
    if ( useAESFile )
    {
        if ( extBegin.offsetMachineDefinitions != 0 )
        {
            // Die maximal 6 Strukturen fuer Maschinendefintionen lesen.
            for ( int j = 0; j < file.m_flashExtended[extBegin.offsetMachineDefinitions] && j < 6; ++j )
            {
                m_machineDefinition[j] = *(StructMachineDefinition *)&file.m_flashExtended[extBegin.offsetMachineDefinitions + sizeof( unsigned char ) + j * sizeof( StructMachineDefinition )];
            }
        }
    }
    else
    {
        // Finding the node for this structure and update parent.
        if ( fileXML.numberOfChildItems( FileXML::ContentType::MachineDefinitions ) )
        {
            auto oldParent_2 = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::MachineDefinitions ) );

            tinyxml2::XMLElement *lastXmlElement{};
            constexpr size_t numberOfItems = sizeof( m_machineDefinition ) / sizeof( m_machineDefinition[0] );

            // Die maximal 6 Strukturen fuer Maschinendefintionen lesen.
            bool interimResult{ false };
            for ( int i = 0; i < static_cast<int>(numberOfItems); ++i )
            {
                interimResult = fileXML.readContent( FileXML::ContentType::Machine, m_machineDefinition[i], i, lastXmlElement );
                // There is no fixed number of elements, so the presence of an element must be taken into account.
                if ( lastXmlElement != nullptr )
                {
                    fileXML.combineValid( interimResult );
                }
                else
                {
                    break;
                }
            }

            // Reset parent node because the end of the structure is reached.
            fileXML.setParentNode( oldParent_2 );
        }
    }

    if ( extBegin.size > sizeof_StructExtendedFlash_1 )
    {
        // Vibrationssensor vorhanden?
        if ( useAESFile )
        {
            if ( extBegin.offsetVibrationSensor != 0 ) // ToDo HK reset?
            {
                m_vibrationSensor = *(StructVibrationSensor *)&file.m_flashExtended[extBegin.offsetVibrationSensor]; // ToDo HK reset? )
            }
        }
        else
        {
            if ( fileXML.numberOfChildItems( FileXML::ContentType::VibrationSensor ) > 0 )
            {
                fileXML.combineValid( fileXML.readContent( FileXML::ContentType::VibrationSensor, m_vibrationSensor ) );
            }
        }
    }


    // Alle Unterpunkte aus Flash lesen
    for ( auto iterChild = constBegin(); iterChild != constEnd(); ++iterChild )
    {
        pFlash = (*iterChild)->readDocument( pFlash, doc );
    }

    // Eintrag für Systemmeldungen lesen
    pFlash = m_systemMessages->readDocument( pFlash, doc );

    // Reset parent node because the end of the structure is reached.
    fileXML.setParentNode( oldParent );

    return pFlash;
}

///////////////////////////////////////
// Referenzierung zwischen Einträgen

void EntrySignalProcessing::convertReferenceToIndex()
{
    int i;
    EntryEventChains &eventChains = rootEventChains();
    EntryDevice &device = root();

    // Fuer den Vibrationssensor
    for ( i = 0; i < 2; ++i )
        m_vibrationSensor.invokeInputChain[i] = (m_refInputChainVibration[i] != nullptr) ? eventChains.indexOf( m_refInputChainVibration[i] ) + 1 : 0;

    // Fuer GPRS-Alive Ereigniskette
    device.m_extended.gprs.invokeInputChain = (m_refInputChainGPRS != nullptr) ? eventChains.indexOf( m_refInputChainGPRS ) + 1 : 0;

    // Für GPS Ereignisketten.
    device.m_extended.gps.invokeInputChainStart = (m_refInputChainGPS[0] != nullptr) ? eventChains.indexOf( m_refInputChainGPS[0] ) + 1 : 0;
    device.m_extended.gps.invokeInputChainStop = (m_refInputChainGPS[1] != nullptr) ? eventChains.indexOf( m_refInputChainGPS[1] ) + 1 : 0;
    device.m_extended.gps.invokeInputChainCycle = (m_refInputChainGPS[2] != nullptr) ? eventChains.indexOf( m_refInputChainGPS[2] ) + 1 : 0;
    device.m_extended.gps.invokeInputChainCourse = (m_refInputChainGPS[3] != nullptr) ? eventChains.indexOf( m_refInputChainGPS[3] ) + 1 : 0;
    device.m_extended.gps.invokeInputChainMove = (m_refInputChainGPS[4] != nullptr) ? eventChains.indexOf( m_refInputChainGPS[4] ) + 1 : 0;

    // Fuer die Timeboy-Anbindung
    device.m_extended.adaption.invokeInputChainPush = (m_refInputChainAdaption[0] != nullptr) ? eventChains.indexOf( m_refInputChainAdaption[0] ) + 1 : 0;
    device.m_extended.adaption.invokeInputChainPull = (m_refInputChainAdaption[1] != nullptr) ? eventChains.indexOf( m_refInputChainAdaption[1] ) + 1 : 0;

    // Systemmeldung ist noch kein Child.
    m_systemMessages->convertReferenceToIndex();

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertReferenceToIndex();
}

void EntrySignalProcessing::convertIndexToReference()
{
    int i;
    EntryEventChains &eventChains = rootEventChains();
    EntryDevice &device = root();

    // Fuer die analogen Eingaenge
    for ( i = 0; i < 2; ++i )
    {
        m_refInputChainVibration[i] = (m_vibrationSensor.invokeInputChain[i] > 0) ? (EntryInputChain *)eventChains.at( m_vibrationSensor.invokeInputChain[i] - 1 ) : nullptr;
        if ( m_refInputChainVibration[i] )
            m_refInputChainVibration[i]->incRefCount();
    }


    // Fuer GPRS-Alive Eingabekette
    m_refInputChainGPRS = (device.m_extended.gprs.invokeInputChain > 0) ? (EntryInputChain *)eventChains.at( device.m_extended.gprs.invokeInputChain - 1 ) : nullptr;
    if ( m_refInputChainGPRS )
        m_refInputChainGPRS->incRefCount();


    // gps
    m_refInputChainGPS[0] = (device.m_extended.gps.invokeInputChainStart > 0) ? (EntryInputChain *)eventChains.at( device.m_extended.gps.invokeInputChainStart - 1 ) : nullptr;
    if ( m_refInputChainGPS[0] )
        m_refInputChainGPS[0]->incRefCount();

    m_refInputChainGPS[1] = (device.m_extended.gps.invokeInputChainStop > 0) ? (EntryInputChain *)eventChains.at( device.m_extended.gps.invokeInputChainStop - 1 ) : nullptr;
    if ( m_refInputChainGPS[1] )
        m_refInputChainGPS[1]->incRefCount();

    m_refInputChainGPS[2] = (device.m_extended.gps.invokeInputChainCycle > 0) ? (EntryInputChain *)eventChains.at( device.m_extended.gps.invokeInputChainCycle - 1 ) : nullptr;
    if ( m_refInputChainGPS[2] )
        m_refInputChainGPS[2]->incRefCount();

    m_refInputChainGPS[3] = (device.m_extended.gps.invokeInputChainCourse > 0) ? (EntryInputChain *)eventChains.at( device.m_extended.gps.invokeInputChainCourse - 1 ) : nullptr;
    if ( m_refInputChainGPS[3] )
        m_refInputChainGPS[3]->incRefCount();

    m_refInputChainGPS[4] = (device.m_extended.gps.invokeInputChainMove > 0) ? (EntryInputChain *)eventChains.at( device.m_extended.gps.invokeInputChainMove - 1 ) : nullptr;
    if ( m_refInputChainGPS[4] )
        m_refInputChainGPS[4]->incRefCount();


    // Fuer die Timeboyanbindung
    m_refInputChainAdaption[0] = (device.m_extended.adaption.invokeInputChainPush > 0) ? (EntryInputChain *)eventChains.at( device.m_extended.adaption.invokeInputChainPush - 1 ) : nullptr;
    if ( m_refInputChainAdaption[0] )
        m_refInputChainAdaption[0]->incRefCount();

    m_refInputChainAdaption[1] = (device.m_extended.adaption.invokeInputChainPull > 0) ? (EntryInputChain *)eventChains.at( device.m_extended.adaption.invokeInputChainPull - 1 ) : nullptr;
    if ( m_refInputChainAdaption[1] )
        m_refInputChainAdaption[1]->incRefCount();

    // Systemmeldung ist noch kein Child.
    m_systemMessages->convertIndexToReference();

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertIndexToReference();
}

///////////////////////////////////////
// Daten des Eintrags

void EntrySignalProcessing::setDefault()
{
    int i;

    m_iconIndex = ICON_MDE;
    m_iconIndexSelected = ICON_MDE_SEL;

    setName( QT_TRANSLATE_NOOP_UTF8( "EntrySignalProcessing", "Signalverarbeitung" ) );

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_VIBRATION_SENSOR_0 );
    m_pageIndexes.push_back( PAGE_GPRS_0 );
    m_pageIndexes.push_back( PAGE_GPS_0 );
    m_pageIndexes.push_back( PAGE_TIMEBOY_ADAPTION_0 );
    m_pageIndexes.push_back( PAGE_MACHINES_0 );

    // Vibrationssensor
    memset( &m_vibrationSensor, 0, sizeof( StructVibrationSensor ) );

    // Setzen der default Werte des Vibrationssensors.
    //     sprintf( (char*)m_vibrationSensor.name, "a-E%i", 1 );
    //     sprintf( (char*)m_vibrationSensor.unit, "V" );
    //     m_vibrationSensor.minLimit = 0;
    //     m_vibrationSensor.maxLimit = 10;
    //     m_vibrationSensor.minInput = 0;
    //     m_vibrationSensor.maxInput = 10;

    m_vibrationSensor.level = 0;
    m_vibrationSensor.timeout = 0;

    // Referenzbezüge initialisieren
    m_refInputChainVibration.fill( nullptr );

    // Referenzbezüge initialisieren
    m_refInputChainGPRS = nullptr;
    m_refInputChainGPS.fill( nullptr );
    m_refInputChainAdaption.fill( nullptr );

    // Strukturen fuer Maschinendefinition loeschen.
    memset( m_machineDefinition, 0, 6 * sizeof( StructMachineDefinition ) );
    // Setzen der default Namen.
    for ( i = 0; i < 6; ++i )
        strncpy( (char *)m_machineDefinition[i].name, boost::str( boost::format( "%1%%2%" ) % QT_TRANSLATE_NOOP_UTF8( "EntrySignalProcessing", "Maschine" ) % (i + 1) ).c_str(), sizeof( m_machineDefinition[i].name ) - 1 );

    // Digitale Eingänge
    EntryDigialInputs *digitalInputs = new EntryDigialInputs( this );
    digitalInputs->setDefault();
    append( digitalInputs );

    // Analoge Eingänge
    EntryAnalogInputs *analogInputs = new EntryAnalogInputs( this );
    analogInputs->setDefault();
    append( analogInputs );

    // Timer
    EntryTimers *timers = new EntryTimers( this );
    timers->setDefault();
    append( timers );

    // SystemMessages auch auf Default setzen.
    m_systemMessages->setDefault();

    setInsertEnabled( false );
    setDeleteEnabled( false );
}

///////////////////////////////////////
// Schnittstelle des Eintrags
EntryDigialInputs &EntrySignalProcessing::rootDigitalInputs()
{
    return (EntryDigialInputs &)*at( 0 );
}

EntryAnalogInputs &EntrySignalProcessing::rootAnalogInputs()
{
    return (EntryAnalogInputs &)*at( 1 );
}

EntryTimers &EntrySignalProcessing::rootTimers()
{
    return (EntryTimers &)*at( 2 );
}

EntrySystemMessages &EntrySignalProcessing::rootSystemMessages()
{
    return (EntrySystemMessages &)*m_systemMessages;
}

// ########################################################
// Member: EntryDigitalInputs

///////////////////////////////////
// Childs

EntryBase *EntryDigialInputs::insertEntry( EntryType typeID, int pos )
{
    int indexBase = pos;

    // Maximaleintragsgröße abfragen.
    if ( size() >= 16 )
    {
        // Es sind maximal (size() * 2) Digitale Eingänge anlegbar.
        return nullptr;
    }

    if ( indexBase == -1 )
        indexBase = size();

    EntryDigitalInput *entryNew = new EntryDigitalInput( this );
    entryNew->setDefault( indexBase * 2 );

    insertEntryFinish( entryNew, pos );

    if ( pos != -1 )
    {
        indexBase = 0;
        // Allen Unterpunkten den basisindex und dadurch auch Namen zuweisen.
        auto iterChild = constBegin();
        for ( ; iterChild != constEnd(); ++iterChild )
        {
            ((EntryDigitalInput *)*iterChild)->setIndexBase( indexBase * 2 );
            indexBase++;
        }
    }

    return entryNew;
}

EntryBase *EntryDigialInputs::cloneEntry()
{
    // Dieser Eintrag unterstützt derzeit cloneEntry nicht.
    return nullptr;
}

void EntryDigialInputs::deleteEntry( int pos )
{
    // Loeschanforderung an Basisklasse weiterleiten
    EntryBase::deleteEntry( pos );

    if ( pos != -1 )
    {
        int indexBase = 0;
        // Allen Unterpunkten den basisindex und dadurch auch Namen zuweisen.
        auto iterChild = constBegin();
        for ( ; iterChild != constEnd(); ++iterChild )
        {
            ((EntryDigitalInput *)*iterChild)->setIndexBase( indexBase * 2 );
            indexBase++;
        }
    }

}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryDigialInputs::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    // Nur vorhandene Einträge speichern.
    if ( size() )
    {
        int fakeCount = 0;
        StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

        // Anlegen der Rahmenstruktur, für die Datenablage.

        // ################################################
        // Strukturen der Digitalen Eingaenge im naechsten freien Speicherblock ablegen
        extBegin.offsetInternDigitalInputs = extBegin.offsetFreeScope;

        // initialisieren des OffsetArrays
        memset( &file.m_flashExtended[extBegin.offsetInternDigitalInputs], 0, sizeof( StructHeaderOffsetArray ) );
        // Daten des OffsetArrays eintragen.
        StructHeaderOffsetArray *headerOffsetArray = (StructHeaderOffsetArray *)&file.m_flashExtended[extBegin.offsetInternDigitalInputs];
        headerOffsetArray->size = sizeof( StructHeaderOffsetArray );
        headerOffsetArray->count = size() * 2;
        headerOffsetArray->offsetArray = headerOffsetArray->size;
        // initialisieren
        memset( &file.m_flashExtended[extBegin.offsetInternDigitalInputs + headerOffsetArray->offsetArray], 0, headerOffsetArray->count * sizeof( unsigned int ) );
        // Vordefinieren der Offsets für das OffsetArray
        for ( uint32_t i = 0; i < headerOffsetArray->count; ++i )
        {
            *(unsigned int *)&file.m_flashExtended[extBegin.offsetInternDigitalInputs + headerOffsetArray->offsetArray + i * sizeof( unsigned int )] = (headerOffsetArray->count * sizeof( unsigned int )) + i * sizeof( StructDigitalInput );
            // Abwärtskompatiblilität durch setzen der "alten" Offsets auf die entsprechenden Stellen herstellen.
            if ( i == 0 ) // Alter Offset auf die ersten 4 Digitalen Eingänge
                extBegin.offsetIntern4DigitalInputs = extBegin.offsetInternDigitalInputs + headerOffsetArray->offsetArray + (headerOffsetArray->count * sizeof( unsigned int ));
            else if ( i == 4 ) // Alter Offset auf die zweiten 4 Digitalen Eingänge
                extBegin.offsetIntern4DigitalInputsPlus = extBegin.offsetInternDigitalInputs + headerOffsetArray->offsetArray + (headerOffsetArray->count * sizeof( unsigned int )) + i * sizeof( StructDigitalInput );
        }

        // Abwärtskompatiblilität durch freistellen von 2 zusätlichen Eingängen herstellen, damit die älteren Softwarestände nicht auf ungültige Strukturdaten stoßen.
        if ( headerOffsetArray->count == 2 || headerOffsetArray->count == 6 )
            fakeCount = 2;

        // initialisieren des Ablagespeichers für die Datenstrukturen
        memset( &file.m_flashExtended[extBegin.offsetInternDigitalInputs + headerOffsetArray->offsetArray + headerOffsetArray->count * sizeof( unsigned int )], 0, (headerOffsetArray->count + fakeCount) * sizeof( StructDigitalInput ) );

        // Den nächsten freie Bereich angeben.
        extBegin.offsetFreeScope += headerOffsetArray->size + headerOffsetArray->count * sizeof( unsigned int ) + (headerOffsetArray->count + fakeCount) * sizeof( StructDigitalInput );

        // ################################################
        // Strukturen der erweiterten Digitalen Eingaenge im naechsten freien Speicherblock ablegen
        extBegin.offsetInternDigitalInputsExt = extBegin.offsetFreeScope;

        // initialisieren des OffsetArrays
        memset( &file.m_flashExtended[extBegin.offsetInternDigitalInputsExt], 0, sizeof( StructHeaderOffsetArray ) );
        // Daten des OffsetArrays eintragen.
        headerOffsetArray = (StructHeaderOffsetArray *)&file.m_flashExtended[extBegin.offsetInternDigitalInputsExt];
        headerOffsetArray->size = sizeof( StructHeaderOffsetArray );
        headerOffsetArray->count = size() * 2;
        headerOffsetArray->offsetArray = headerOffsetArray->size;
        // initialisieren
        memset( &file.m_flashExtended[extBegin.offsetInternDigitalInputsExt + headerOffsetArray->offsetArray], 0, headerOffsetArray->count * sizeof( unsigned int ) );
        // Vordefinieren der Offsets für das OffsetArray
        for ( uint32_t i = 0; i < headerOffsetArray->count; ++i )
        {
            *(unsigned int *)&file.m_flashExtended[extBegin.offsetInternDigitalInputsExt + headerOffsetArray->offsetArray + i * sizeof( unsigned int )] = (headerOffsetArray->count * sizeof( unsigned int )) + i * sizeof( StructDigitalInputExt );
            // Abwärtskompatiblilität wird in EntryDigitalInput::writeDocument, durch paralleles Eintragen der ersten 8 Strukturen in den Erweiterten Bereich hergestellt.
        }

        // initialisieren des Ablagespeichers für die Datenstrukturen
        memset( &file.m_flashExtended[extBegin.offsetInternDigitalInputsExt + headerOffsetArray->offsetArray + headerOffsetArray->count * sizeof( unsigned int )], 0, headerOffsetArray->count * sizeof( StructDigitalInputExt ) );

        // Den nächsten freie Bereich angeben.
        extBegin.offsetFreeScope += headerOffsetArray->size + headerOffsetArray->count * sizeof( unsigned int ) + headerOffsetArray->count * sizeof( StructDigitalInputExt );

        // Schreiben der Erweiterten Struktur
        *(StructExtendedFlash *)file.m_flashExtended = extBegin;

        ////////////////////////////////////////
        // add information to xml file
        FileXML &fileXML = doc.fileXML();

        // Finding the node for this structure and update parent.
        auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::DigitalInputs ) );

        // Alle Unterpunkte in Flash eintragen und die jeweiligen Pointer sichern
        auto iterChild = constBegin();
        for ( ; iterChild != constEnd(); ++iterChild )
        {
            (*iterChild)->writeDocument( pFlash, doc );
        }

        // Reset parent node because the end of the structure is reached.
        fileXML.setParentNode( oldParent );
    }

    return pFlash;
}

uint32_t EntryDigialInputs::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    uint32_t size{};
    StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    if ( useAESFile )
    {
        // Bei aktuellen Setupdateien wird mit der Signalverarbeitung in einen neuen Bereich gewechselt.
        if ( extBegin.size > sizeof_StructExtendedFlash_1 &&extBegin.offsetInternDigitalInputs )
        {
            // Daten des OffsetArrays lesen.
            StructHeaderOffsetArray *headerOffsetArray = (StructHeaderOffsetArray *)&file.m_flashExtended[extBegin.offsetInternDigitalInputs];
            // Anzahl ermitteln / pro Seite 2 Digitale Eingänge
            size = headerOffsetArray->count / 2;
        }
        else if ( extBegin.offsetIntern4DigitalInputs )
        {
            if ( extBegin.offsetIntern4DigitalInputsPlus )
                size = 4;
            else
                size = 2;
        }

        // Daten aller Eintraege einlesen
        for ( uint32_t i = 0; i < size; ++i )
        {
            EntryDigitalInput *pNeu = new EntryDigitalInput( this );
            pNeu->setDefault( i * 2 );

            pNeu->readDocument( pFlash, doc );
            append( pNeu );
        }
    }
    else
    {
        // Finding the node for this structure and update parent.
        auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::DigitalInputs ) );

        size = fileXML.numberOfChildItems( FileXML::ContentType::DigitalInput ) / 2;

        // Reading all analog inputs.
        for ( uint32_t i = 0; i < size; ++i )
        {
            EntryDigitalInput *pInput = new EntryDigitalInput( this );
            pInput->setDefault( i );

            pInput->readDocument( i, doc );
            append( pInput );
        }

        // Reset parent node because the end of the structure is reached.
        fileXML.setParentNode( oldParent );
    }

    return pFlash;
}

///////////////////////////////////
// Daten des Eintrags

void EntryDigialInputs::setDefault()
{
    // Icon's
    m_iconIndex = ICON_DIGITAL_INPUTS;
    m_iconIndexSelected = ICON_DIGITAL_INPUTS_SEL;

    // Name
    setName( QT_TRANSLATE_NOOP_UTF8( "EntryDigialInputs", "Digitale Eingänge" ) );

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_BLANK );

    setInsertEnabled( true );
    setDeleteEnabled( false );
}

///////////////////////////////////////
// Schnittstelle des Eintrags
StructDigitalInput EntryDigialInputs::digitalInput( int index ) const
{
    StructDigitalInput res;
    memset( &res, 0, sizeof( StructDigitalInput ) );

    // Alle Unterpunkte in Flash eintragen und die jeweiligen Pointer sichern
    auto iterChild = constBegin();
    for ( ; iterChild != constEnd(); ++iterChild )
    {
        if ( ((EntryDigitalInput *)(*iterChild))->indexBase() == index || (((EntryDigitalInput *)(*iterChild))->indexBase() + 1) == index )
        {
            res = ((EntryDigitalInput *)(*iterChild))->m_digitalInput[index % 2];
            break;
        }
    }

    return res;
}

StructDigitalInputExt EntryDigialInputs::digitalInputExt( int index ) const
{
    StructDigitalInputExt res;
    memset( &res, 0, sizeof( StructDigitalInputExt ) );

    // Alle Unterpunkte in Flash eintragen und die jeweiligen Pointer sichern
    auto iterChild = constBegin();
    for ( ; iterChild != constEnd(); ++iterChild )
    {
        if ( ((EntryDigitalInput *)(*iterChild))->indexBase() == index || (((EntryDigitalInput *)(*iterChild))->indexBase() + 1) == index )
        {
            res = ((EntryDigitalInput *)(*iterChild))->m_digitalInputExt[index % 2];
            break;
        }
    }

    return res;
}

int EntryDigialInputs::digitalInputCount() const
{
    return size() * 2; // Pro Eintrag sind 2 Eingänge vorhanden.
}

// ########################################################
// Member: EntryDigitalInput

///////////////////////////////////
// Childs

EntryBase *EntryDigitalInput::insertEntry( EntryType typeID, int pos )
{
    (void)typeID;
    (void)pos;

    return nullptr;
}

EntryBase *EntryDigitalInput::cloneEntry()
{
    // Dieser Eintrag unterstützt derzeit cloneEntry nicht.
    return nullptr;
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryDigitalInput::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    int i;
    StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

    // Wechseln der Byteorder auf Motorolla - Format
    for ( i = 0; i < 2; ++i )
    {
        m_digitalInput[i].countDivisor = endianReverse( m_digitalInput[i].countDivisor );
        m_digitalInput[i].countDivisor32Bit = endianReverse( m_digitalInput[i].countDivisor32Bit );
        m_digitalInput[i].cycle = endianReverse( m_digitalInput[i].cycle );
        m_digitalInput[i].timeout = endianReverse( m_digitalInput[i].timeout );
        m_digitalInput[i].debounceHigh = endianReverse( m_digitalInput[i].debounceHigh );
        m_digitalInput[i].debounceLow = endianReverse( m_digitalInput[i].debounceLow );
    }

    // _indexBase ist Basisindex der Digitalen Strukturen
    // 0 -> E1 + E2
    // 2 -> E3 + E4
    // 4 -> E5 + E6
    // ...

    // OffsetArray der Digitalen Eingänge
    StructHeaderOffsetArray *headerOffsetArray = (StructHeaderOffsetArray *)&file.m_flashExtended[extBegin.offsetInternDigitalInputs];
    unsigned int offsetFlash = *(unsigned int *)&file.m_flashExtended[extBegin.offsetInternDigitalInputs + headerOffsetArray->offsetArray + m_indexBase * sizeof( unsigned int )];
    // Zeiger auf die 2 Strukturen fuer die Digitalen Eingaenge
    StructDigitalInput *digitalInput = (StructDigitalInput *)&file.m_flashExtended[extBegin.offsetInternDigitalInputs + headerOffsetArray->offsetArray + offsetFlash];

    // Schreiben der 2 Strukturen fuer die Digitalen Eingaenge
    for ( i = 0; i < 2; ++i )
    {
        digitalInput[i] = m_digitalInput[i];
    }

    // Wechseln der Byteorder auf Intel - Format
    for ( i = 0; i < 2; ++i )
    {
        m_digitalInput[i].countDivisor = endianReverse( m_digitalInput[i].countDivisor );
        m_digitalInput[i].countDivisor32Bit = endianReverse( m_digitalInput[i].countDivisor32Bit );
        m_digitalInput[i].cycle = endianReverse( m_digitalInput[i].cycle );
        m_digitalInput[i].timeout = endianReverse( m_digitalInput[i].timeout );
        m_digitalInput[i].debounceHigh = endianReverse( m_digitalInput[i].debounceHigh );
        m_digitalInput[i].debounceLow = endianReverse( m_digitalInput[i].debounceLow );
    }

    // OffsetArray der erweiterten Digitalen Eingänge
    headerOffsetArray = (StructHeaderOffsetArray *)&file.m_flashExtended[extBegin.offsetInternDigitalInputsExt];
    offsetFlash = *(unsigned int *)&file.m_flashExtended[extBegin.offsetInternDigitalInputsExt + headerOffsetArray->offsetArray + m_indexBase * sizeof( unsigned int )];
    // Zeiger auf die 2 Strukturen fuer die erweiterten Digitalen Eingaenge
    StructDigitalInputExt *digitalInputExt = (StructDigitalInputExt *)&file.m_flashExtended[extBegin.offsetInternDigitalInputsExt + headerOffsetArray->offsetArray + offsetFlash];

    // Schreiben der 2 Strukturen fuer die erweiterten Digitalen Eingaenge
    for ( i = 0; i < 2; ++i )
    {
        digitalInputExt[i] = m_digitalInputExt[i];
    }

    ////////////////////////////////////////
    // add information to xml file
    FileXML &fileXML = doc.fileXML();

    for ( i = 0; i < 2; ++i )
    {
        // The parent is configured outside the method and is not changed.
        fileXML.writeContent( FileXML::ContentType::DigitalInput, as_const( m_digitalInput[i] ) );

        // Finding the node for this structure and update parent.
        auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::DigitalInput ) );

        fileXML.writeContent( FileXML::ContentType::DigitalInputEx, as_const( m_digitalInputExt[i] ) );

        // Reset parent node because the end of the structure is reached.
        fileXML.setParentNode( oldParent );
    }

    // Die ersten 8 Digitalen Eingänge werden zwecks Abwärtskompatibilität in die bisher vorhandenen Strukturen ablegegt.
    if ( m_indexBase <= 6 )
    {
        // Lesen der erweiterten Parameter - Struktur
        // Diese Struktur enthaelt informationen auf die beim Baumaufbau von anderen Eintraegen zurueckgegriffen wird,
        // somit muss sie vor dem Aufbau geladen sein.
        EntryDevice &device = root();
        for ( i = 0; i < 2; ++i )
        {
            device.m_extended.digitalExtInput[m_indexBase + i] = m_digitalInputExt[i];
        }
    }

    return pFlash;
}

uint32_t EntryDigitalInput::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    // Erweiterter Flashbereich
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();

    // Erweiterten Struktur lesen
    int i;
    StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    // _indexBase ist Basisindex der Digitalen Strukturen
    // 0 -> E1 + E2
    // 2 -> E3 + E4
    // 4 -> E5 + E6
    // ...

    // Strukturen der Digitalen Eingänge lesen
    StructDigitalInput *digitalInput = nullptr;
    StructDigitalInputExt *digitalInputExt = nullptr;

    if ( useAESFile )
    {
    // Die die Offsets der aktuellen TExtededFlash-Struktur vor?
        if ( extBegin.size > sizeof_StructExtendedFlash_1 &&extBegin.offsetInternDigitalInputs )
        {
            // OffsetArray der Digitalen Eingänge
            StructHeaderOffsetArray *headerOffsetArray = (StructHeaderOffsetArray *)&file.m_flashExtended[extBegin.offsetInternDigitalInputs];
            unsigned int offsetFlash = *(unsigned int *)&file.m_flashExtended[extBegin.offsetInternDigitalInputs + headerOffsetArray->offsetArray + m_indexBase * sizeof( unsigned int )];
            // Zeiger auf die 2 Strukturen fuer die Digitalen Eingaenge
            digitalInput = (StructDigitalInput *)&file.m_flashExtended[extBegin.offsetInternDigitalInputs + headerOffsetArray->offsetArray + offsetFlash];

            // OffsetArray der erweiterten Digitalen Eingänge
            headerOffsetArray = (StructHeaderOffsetArray *)&file.m_flashExtended[extBegin.offsetInternDigitalInputsExt];
            offsetFlash = *(unsigned int *)&file.m_flashExtended[extBegin.offsetInternDigitalInputsExt + headerOffsetArray->offsetArray + m_indexBase * sizeof( unsigned int )];
            // Zeiger auf die 2 Strukturen fuer die erweiterten Digitalen Eingaenge
            digitalInputExt = (StructDigitalInputExt *)&file.m_flashExtended[extBegin.offsetInternDigitalInputsExt + headerOffsetArray->offsetArray + offsetFlash];
        }
        else
        {
            // Bei den Vorgängerversionen der Setupdatei muss über die folgenden Offsets gelesen werden.
            if ( m_indexBase >= 0 && m_indexBase <= 2 )
            {
                // Bereits vorhanden?
                if ( extBegin.offsetIntern4DigitalInputs != 0 )
                    digitalInput = (StructDigitalInput *)&file.m_flashExtended[extBegin.offsetIntern4DigitalInputs + m_indexBase * sizeof( StructDigitalInput )];
            }
            else if ( m_indexBase >= 4 && m_indexBase <= 6 )
            {
                // Bereits vorhanden?
                if ( extBegin.offsetIntern4DigitalInputsPlus != 0 )
                    digitalInput = (StructDigitalInput *)&file.m_flashExtended[extBegin.offsetIntern4DigitalInputsPlus + (m_indexBase - 4) * sizeof( StructDigitalInput )];
            }
            else
            {
                BOOST_ASSERT( false );
            }
        }

        // Strukturzeiger zugewiesen?
        if ( digitalInput != nullptr )
        {
            // Die 2 Strukturen fuer Digitale Eingaenge lesen.
            for ( i = 0; i < 2; ++i )
                m_digitalInput[i] = digitalInput[i];

            // Wechseln der Byteorder auf Intel - Format
            for ( i = 0; i < 2; ++i )
            {
                m_digitalInput[i].countDivisor = endianReverse( m_digitalInput[i].countDivisor );
                m_digitalInput[i].countDivisor32Bit = endianReverse( m_digitalInput[i].countDivisor32Bit );
                m_digitalInput[i].cycle = endianReverse( m_digitalInput[i].cycle );
                m_digitalInput[i].timeout = endianReverse( m_digitalInput[i].timeout );
                m_digitalInput[i].debounceHigh = endianReverse( m_digitalInput[i].debounceHigh );
                m_digitalInput[i].debounceLow = endianReverse( m_digitalInput[i].debounceLow );

                //@@: 2006-10-09
                // Zaehlerteiler 32bit eingefuehrt. Sicherstellen, dass der 32bit - Wert dem 16bit - Wert angepasst ist.
                if ( (m_digitalInput[i].countDivisor > 0) && (m_digitalInput[i].countDivisor32Bit == 0) )
                    m_digitalInput[i].countDivisor32Bit = m_digitalInput[i].countDivisor;
            }
        }

        if ( digitalInputExt != nullptr )
        {
            // Die 2 Strukturen für erweiterte Digitale Eingaenge lesen.
            for ( i = 0; i < 2; ++i )
            {
                m_digitalInputExt[i] = digitalInputExt[i];
                if ( m_digitalInputExt[i].cavity == 0 || m_digitalInputExt[i].strokes == 0 )
                {
                    m_digitalInputExt[i].cavity = 1;
                    m_digitalInputExt[i].strokes = 1;
                }
            }
        }
        else if ( m_indexBase <= 6 )    // Lesen der erweiterten Strukturen der Digitalen Eingänge.
        {
            // Lesen der erweiterten Parameter - Struktur
            // Diese Struktur enthaelt informationen auf die beim Baumaufbau von anderen Eintraegen zurueckgegriffen wird,
            // somit muss sie vor dem Aufbau geladen sein.
            EntryDevice &device = root();
            for ( i = 0; i < 2; ++i )
            {
                m_digitalInputExt[i] = device.m_extended.digitalExtInput[m_indexBase + i];
                if ( m_digitalInputExt[i].cavity == 0 || m_digitalInputExt[i].strokes == 0 )
                {
                    m_digitalInputExt[i].cavity = 1;
                    m_digitalInputExt[i].strokes = 1;
                }
            }
        }
    }
    else
    {
        tinyxml2::XMLElement *lastXmlElement{};

        // pFlash contains the index of the current digital input pair
        for ( i = 0; i < 2; ++i )
        {
            fileXML.combineValid( fileXML.readContent( FileXML::ContentType::DigitalInput, m_digitalInput[i], (pFlash * 2 + i), lastXmlElement ) );

            // Finding the node for this structure and update parent.
            auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::DigitalInput ) );

            fileXML.combineValid( fileXML.readContent( FileXML::ContentType::DigitalInputEx, m_digitalInputExt[i] ) );

            // Reset parent node because the end of the structure is reached.
            fileXML.setParentNode( oldParent );
        }
    }

    return pFlash;
}

void EntryDigitalInput::checkMembers()
{
}

///////////////////////////////////////
// Referenzierung zwischen Einträgen

void EntryDigitalInput::convertReferenceToIndex()
{
    int i;
    EntryEventChains &eventChains = rootEventChains();

    // Fuer jeden digitalen Eingang.
    for ( i = 0; i < 2; ++i )
    {
        m_digitalInput[i].startLowHigh = (m_refInputChainDigital[i][0] != nullptr) ? eventChains.indexOf( m_refInputChainDigital[i][0] ) + 1 : 0;
        m_digitalInput[i].startHighLow = (m_refInputChainDigital[i][1] != nullptr) ? eventChains.indexOf( m_refInputChainDigital[i][1] ) + 1 : 0;
        m_digitalInput[i].invokeInputChainCountDivisor = (m_refInputChainDigital[i][2] != nullptr) ? eventChains.indexOf( m_refInputChainDigital[i][2] ) + 1 : 0;
        m_digitalInput[i].invokeInputChainCycle = (m_refInputChainDigital[i][3] != nullptr) ? eventChains.indexOf( m_refInputChainDigital[i][3] ) + 1 : 0;
    }

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertReferenceToIndex();
}

void EntryDigitalInput::convertIndexToReference()
{
    int i;
    EntryEventChains &eventChains = rootEventChains();

    // Fuer jeden digitalen Eingang.
    for ( i = 0; i < 2; ++i )
    {
        m_refInputChainDigital[i][0] = (m_digitalInput[i].startLowHigh > 0) ? (EntryInputChain *)eventChains.at( m_digitalInput[i].startLowHigh - 1 ) : nullptr;
        if ( m_refInputChainDigital[i][0] )
        {
            m_refInputChainDigital[i][0]->incRefCount();
        }
        m_refInputChainDigital[i][1] = (m_digitalInput[i].startHighLow > 0) ? (EntryInputChain *)eventChains.at( m_digitalInput[i].startHighLow - 1 ) : nullptr;
        if ( m_refInputChainDigital[i][1] )
        {
            m_refInputChainDigital[i][1]->incRefCount();
        }
        m_refInputChainDigital[i][2] = (m_digitalInput[i].invokeInputChainCountDivisor > 0) ? (EntryInputChain *)eventChains.at( m_digitalInput[i].invokeInputChainCountDivisor - 1 ) : nullptr;
        if ( m_refInputChainDigital[i][2] )
        {
            m_refInputChainDigital[i][2]->incRefCount();
        }
        m_refInputChainDigital[i][3] = (m_digitalInput[i].invokeInputChainCycle > 0) ? (EntryInputChain *)eventChains.at( m_digitalInput[i].invokeInputChainCycle - 1 ) : nullptr;
        if ( m_refInputChainDigital[i][3] )
        {
            m_refInputChainDigital[i][3]->incRefCount();
        }
    }

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertIndexToReference();
}

///////////////////////////////////////
// Daten des Eintrags

void EntryDigitalInput::setDefault()
{
    setDefault( 0 );
}

void EntryDigitalInput::setDefault( int indexBase )
{
    m_iconIndex = ICON_DIGITAL_INPUT;
    m_iconIndexSelected = ICON_DIGITAL_INPUT_SEL;

    // Offset auf die Digitalen Eingänge
    // 0 -> E1 / E2
    // 2 -> E3 / E4
    // 4 -> E5 / E6
    // ...
    setIndexBase( indexBase );

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_DIGITAL_INPUT_0 );

    // Strukturen fuer Digitale - Eingaenge loeschen.
    memset( m_digitalInput, 0, 2 * sizeof( StructDigitalInput ) );

    // Strukturen fuer Digitale - Eingaenge loeschen (Erweitert)
    memset( m_digitalInputExt, 0, 2 * sizeof( StructDigitalInputExt ) );
    for ( int i = 0; i < 2; ++i )
    {
        // Kavität und Hubzahl initalisieren.
        m_digitalInputExt[i].cavity = 1;
        m_digitalInputExt[i].strokes = 1;
    }

    // Referenzbezug initialisieren
    memset( m_refInputChainDigital, 0, sizeof( m_refInputChainDigital ) );

    setInsertEnabled( false );
    setDeleteEnabled( true );
}

// ########################################################
// Member: EntryAnalogInputs

///////////////////////////////////
// Childs

EntryBase *EntryAnalogInputs::insertEntry( EntryType typeID, int pos )
{
    int indexBase = pos;

    // Maximaleintragsgröße abfangen.
    if ( size() >= 32 )
    {
        // Es sind maximal size() Analoge Eingänge anlegbar.
        return nullptr;
    }

    if ( indexBase == -1 )
        indexBase = size();

    EntryAnalogInput *entryNew = new EntryAnalogInput( this );
    entryNew->setDefault( indexBase );

    insertEntryFinish( entryNew, pos );

    if ( pos != -1 )
    {
        indexBase = 0;
        // Allen Unterpunkten den basisindex und dadurch auch Namen zuweisen.
        auto iterChild = constBegin();
        for ( ; iterChild != constEnd(); ++iterChild )
        {
            ((EntryAnalogInput *)*iterChild)->setIndexBase( indexBase );
            indexBase++;
        }
    }

    return entryNew;
}

EntryBase *EntryAnalogInputs::cloneEntry()
{
    // Dieser Eintrag unterstützt derzeit cloneEntry nicht.
    return nullptr;
}

void EntryAnalogInputs::deleteEntry( int pos )
{
    // Loeschanforderung an Basisklasse weiterleiten
    EntryBase::deleteEntry( pos );

    if ( pos != -1 )
    {
        int indexBase = 0;
        // Allen Unterpunkten den basisindex und dadurch auch Namen zuweisen.
        auto iterChild = constBegin();
        for ( ; iterChild != constEnd(); ++iterChild )
        {
            ((EntryAnalogInput *)*iterChild)->setIndexBase( indexBase );
            indexBase++;
        }
    }
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryAnalogInputs::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    // Nur vorhandene Einträge speichern.
    if ( size() )
    {
        int fakeCount = 0;
        StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

        // Anlegen der Rahmenstruktur, für die Datenablage.

        // ################################################
        // Strukturen der Analogen Eingaenge im naechsten freien Speicherblock ablegen
        extBegin.offsetInternAnalogInputs = extBegin.offsetFreeScope;

        // initialisieren des OffsetArrays
        memset( &file.m_flashExtended[extBegin.offsetInternAnalogInputs], 0, sizeof( StructHeaderOffsetArray ) );
        // Daten des OffsetArrays eintragen.
        StructHeaderOffsetArray *headerOffsetArray = (StructHeaderOffsetArray *)&file.m_flashExtended[extBegin.offsetInternAnalogInputs];
        headerOffsetArray->size = sizeof( StructHeaderOffsetArray );
        headerOffsetArray->count = size();
        headerOffsetArray->offsetArray = headerOffsetArray->size;
        // initialisieren
        memset( &file.m_flashExtended[extBegin.offsetInternAnalogInputs + headerOffsetArray->offsetArray], 0, headerOffsetArray->count * sizeof( unsigned int ) );
        // Vordefinieren der Offsets für das OffsetArray
        for ( uint32_t i = 0; i < headerOffsetArray->count; ++i )
        {
            *(unsigned int *)&file.m_flashExtended[extBegin.offsetInternAnalogInputs + headerOffsetArray->offsetArray + i * sizeof( unsigned int )] = (headerOffsetArray->count * sizeof( unsigned int )) + i * sizeof( StructAnalogInput );
            // Abwärtskompatiblilität durch setzen des "alten" Offsets auf die entsprechende Stelle herstellen.
            if ( i == 0 ) // Alter Offset auf die ersten 2 Digitalen Eingänge
                extBegin.offsetIntern2AnalogInputs = extBegin.offsetInternAnalogInputs + headerOffsetArray->offsetArray + (headerOffsetArray->count * sizeof( unsigned int ));
        }

        // Abwärtskompatiblilität durch freistellen von 1 zusätlichen Eingang herstellen, damit die älteren Softwarestände nicht auf ungültige Strukturdaten stoßen.
        if ( headerOffsetArray->count == 1 )
            fakeCount = 1;

        // initialisieren des Ablagespeichers für die Datenstrukturen
        memset( &file.m_flashExtended[extBegin.offsetInternAnalogInputs + headerOffsetArray->offsetArray + headerOffsetArray->count * sizeof( unsigned int )], 0, (headerOffsetArray->count + fakeCount) * sizeof( StructAnalogInput ) );

        // Den nächsten freie Bereich angeben.
        extBegin.offsetFreeScope += headerOffsetArray->size + headerOffsetArray->count * sizeof( unsigned int ) + (headerOffsetArray->count + fakeCount) * sizeof( StructAnalogInput );

        // Schreiben der Erweiterten Struktur
        *(StructExtendedFlash *)file.m_flashExtended = extBegin;

        ////////////////////////////////////////
        // add information to xml file
        FileXML &fileXML = doc.fileXML();

        // Finding the node for this structure and update parent.
        auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::AnalogInputs ) );

        // Alle Unterpunkte in Flash eintragen und die jeweiligen Pointer sichern
        auto iterChild = constBegin();
        for ( ; iterChild != constEnd(); ++iterChild )
        {
            (*iterChild)->writeDocument( pFlash, doc );
        }

        // Reset parent node because the end of the structure is reached.
        fileXML.setParentNode( oldParent );
    }

    return pFlash;
}

uint32_t EntryAnalogInputs::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    uint32_t size{};
    StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    if ( useAESFile )
    {
        // Bei aktuellen Setupdateien wird mit der Signalverarbeitung in einen neuen Bereich gewechselt.
        if ( extBegin.size > sizeof_StructExtendedFlash_1 &&extBegin.offsetInternAnalogInputs )
        {
            // Daten des OffsetArrays lesen.
            StructHeaderOffsetArray *headerOffsetArray = (StructHeaderOffsetArray *)&file.m_flashExtended[extBegin.offsetInternAnalogInputs];
            // Anzahl ermitteln
            size = headerOffsetArray->count;
        }
        else if ( extBegin.offsetIntern2AnalogInputs )
        {
            size = 2;
        }

        // Daten aller Eintraege einlesen
        for ( uint32_t i = 0; i < size; ++i )
        {
            EntryAnalogInput *pNeu = new EntryAnalogInput( this );
            pNeu->setDefault( i );

            pNeu->readDocument( pFlash, doc );
            append( pNeu );
        }
    }
    else
    {
        // Finding the node for this structure and update parent.
        auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::AnalogInputs ) );

        size = fileXML.numberOfChildItems( FileXML::ContentType::AnalogInput );

        // Reading all analog inputs.
        for ( uint32_t i = 0; i < size; ++i )
        {
            EntryAnalogInput *pInput = new EntryAnalogInput( this );
            pInput->setDefault( i );

            pInput->readDocument( i, doc );
            append( pInput );
        }

        // Reset parent node because the end of the structure is reached.
        fileXML.setParentNode( oldParent );
    }

    return pFlash;
}

///////////////////////////////////
// Daten des Eintrags

void EntryAnalogInputs::setDefault()
{
    // Icon's
    m_iconIndex = ICON_ANALOG_INPUTS;
    m_iconIndexSelected = ICON_ANALOG_INPUTS_SEL;

    // Name
    setName( QT_TRANSLATE_NOOP_UTF8( "EntryAnalogInputs", "Analoge Eingänge" ) );

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_BLANK );

    setInsertEnabled( true );
    setDeleteEnabled( false );
}

////////////////////////////////////
// Schnittstelle des Eintrags
StructAnalogInput EntryAnalogInputs::analogInput( int index ) const
{
    StructAnalogInput res;
    memset( &res, 0, sizeof( StructAnalogInput ) );

    // Alle Unterpunkte in Flash eintragen und die jeweiligen Pointer sichern
    auto iterChild = constBegin();
    for ( ; iterChild != constEnd(); ++iterChild )
    {
        if ( ((EntryAnalogInput *)(*iterChild))->indexBase() == index )
        {
            res = ((EntryAnalogInput *)(*iterChild))->m_analogInput;
            break;
        }
    }

    return res;
}

int EntryAnalogInputs::analogInputCount() const
{
    return size();
}

// ########################################################
// Member: EntryAnalogInput

///////////////////////////////////
// Childs

EntryBase *EntryAnalogInput::insertEntry( EntryType typeID, int pos )
{
    (void)typeID;
    (void)pos;

    return nullptr;
}

EntryBase *EntryAnalogInput::cloneEntry()
{
    // Dieser Eintrag unterstützt derzeit cloneEntry nicht.
    return nullptr;
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryAnalogInput::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    // Erweiterter Flashbereich
    StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

    // OffsetArray der Digitalen Eingänge
    StructHeaderOffsetArray *headerOffsetArray = (StructHeaderOffsetArray *)&file.m_flashExtended[extBegin.offsetInternAnalogInputs];
    unsigned int offsetFlash = *(unsigned int *)&file.m_flashExtended[extBegin.offsetInternAnalogInputs + headerOffsetArray->offsetArray + m_indexBase * sizeof( unsigned int )];
    // Zeiger auf die Struktur fuer den Analogen Eingang
    StructAnalogInput *analogInput = (StructAnalogInput *)&file.m_flashExtended[extBegin.offsetInternAnalogInputs + headerOffsetArray->offsetArray + offsetFlash];

    // Schreiben der Struktur fuer den Analogen Eingang
    *analogInput = m_analogInput;

    ////////////////////////////////////////
    // add information to xml file
    FileXML &fileXML = doc.fileXML();
    fileXML.writeContent( FileXML::ContentType::AnalogInput, as_const( m_analogInput ) );

    return pFlash;
}

uint32_t EntryAnalogInput::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    // Erweiterten Struktur lesen
    StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    if ( useAESFile )
    {
    // Struktur des Analogen Eingangs lesen
        StructAnalogInput *analogInput = nullptr;

        // Die die Offsets der aktuellen TExtededFlash-Struktur vor?
        if ( extBegin.size > sizeof_StructExtendedFlash_1 &&extBegin.offsetInternAnalogInputs )
        {
            // OffsetArray der Digitalen Eingänge
            StructHeaderOffsetArray *headerOffsetArray = (StructHeaderOffsetArray *)&file.m_flashExtended[extBegin.offsetInternAnalogInputs];
            unsigned int offsetFlash = *(unsigned int *)&file.m_flashExtended[extBegin.offsetInternAnalogInputs + headerOffsetArray->offsetArray + m_indexBase * sizeof( unsigned int )];
            // Zeiger auf die Struktur fuer den Analogen Eingang
            analogInput = (StructAnalogInput *)&file.m_flashExtended[extBegin.offsetInternAnalogInputs + headerOffsetArray->offsetArray + offsetFlash];
        }
        else
        {
            // Strukturen des Analogen Eingangs
            if ( m_indexBase >= 0 && m_indexBase <= 1 )
            {
                // Bereits vorhanden?
                if ( extBegin.offsetIntern2AnalogInputs != 0 )
                    analogInput = (StructAnalogInput *)&file.m_flashExtended[extBegin.offsetIntern2AnalogInputs + m_indexBase * sizeof( StructAnalogInput )];
            }
            else
            {
                BOOST_ASSERT( false );
            }

        }

        // Strukturzeiger zugewiesen?
        if ( analogInput != nullptr )
        {
            // Die Struktur für den Analogen Eingang lesen.
            m_analogInput = *analogInput;
        }
    }
    else
    {
        tinyxml2::XMLElement *lastXmlElement{};
        // pFlash contains the index of the current analog input
        fileXML.combineValid( fileXML.readContent( FileXML::ContentType::AnalogInput, m_analogInput, pFlash, lastXmlElement ) );
    }

    return pFlash;
}

void EntryAnalogInput::checkMembers()
{
}

///////////////////////////////////////
// Referenzierung zwischen Einträgen

void EntryAnalogInput::convertReferenceToIndex()
{
    int i;
    EntryEventChains &eventChains = rootEventChains();

    // Fuer die analogen Eingaenge
    for ( i = 0; i < 5; ++i )
    {
        m_analogInput.invokeInputChain[i] = (m_refInputChainAnalog[i] != nullptr) ? eventChains.indexOf( m_refInputChainAnalog[i] ) + 1 : 0;
    }

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertReferenceToIndex();
}

void EntryAnalogInput::convertIndexToReference()
{
    int i;
    EntryEventChains &eventChains = rootEventChains();

    // Fuer die analogen Eingaenge
    for ( i = 0; i < 5; ++i )
    {
        m_refInputChainAnalog[i] = (m_analogInput.invokeInputChain[i] > 0) ? (EntryInputChain *)eventChains.at( m_analogInput.invokeInputChain[i] - 1 ) : nullptr;
        if ( m_refInputChainAnalog[i] )
            m_refInputChainAnalog[i]->incRefCount();
    }

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertIndexToReference();
}

///////////////////////////////////////
// Daten des Eintrags

void EntryAnalogInput::setDefault()
{
    setDefault( 0 );
}

void EntryAnalogInput::setDefault( int indexBase )
{
    m_iconIndex = ICON_ANALOG_INPUT;
    m_iconIndexSelected = ICON_ANALOG_INPUT_SEL;

    // Offset auf die Analogen Eingänge
    // 0 -> E1
    // 1 -> E2
    // 2 -> E3
    // ...
    setIndexBase( indexBase );

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_ANALOG_INPUT_0 );

    // Strukturen fuer Analoge - Eingaenge loeschen.
    memset( &m_analogInput, 0, sizeof( StructAnalogInput ) );

    // Setzen der default Werte des Analogkanals
    sprintf( (char *)m_analogInput.name, "a-E%i", (indexBase + 1) );
    sprintf( (char *)m_analogInput.unit, "V" );
    m_analogInput.minLimit = 0;
    m_analogInput.maxLimit = 10;
    m_analogInput.minInput = 0;
    m_analogInput.maxInput = 10;

    // Referenzbezug initialisieren
    memset( m_refInputChainAnalog, 0, sizeof( m_refInputChainAnalog ) );

    setInsertEnabled( false );
    setDeleteEnabled( true );
}

// ########################################################
// Member: EntryTimers

///////////////////////////////////
// Childs

EntryBase *EntryTimers::insertEntry( EntryType typeID /*= Type_id::Base*/, int pos /*= -1*/ )
{
    int indexBase = pos;

    // Maximaleintragsgröße abfangen.
    if ( size() >= 32 )
    {
        // Es sind maximal 32 Timer anlegbar.
        return nullptr;
    }

    if ( indexBase == -1 )
        indexBase = size();

    EntryTimer *entryNew = new EntryTimer( this );
    entryNew->setDefault( indexBase );

    insertEntryFinish( entryNew, pos );

    if ( pos != -1 )
    {
        indexBase = 0;
        // Allen Unterpunkten den basisindex und dadurch auch Namen zuweisen.
        auto iterChild = constBegin();
        for ( ; iterChild != constEnd(); ++iterChild )
        {
            ((EntryTimer *)*iterChild)->setIndexBase( indexBase );
            indexBase++;
        }
    }

    return entryNew;
}

EntryBase *EntryTimers::cloneEntry()
{
    // Dieser Eintrag unterstützt derzeit cloneEntry nicht.
    return nullptr;
}

void EntryTimers::deleteEntry( int pos )
{
    // Loeschanforderung an Basisklasse weiterleiten
    EntryBase::deleteEntry( pos );

    if ( pos != -1 )
    {
        int indexBase = 0;
        // Allen Unterpunkten den basisindex und dadurch auch Namen zuweisen.
        auto iterChild = constBegin();
        for ( ; iterChild != constEnd(); ++iterChild )
        {
            ((EntryTimer *)*iterChild)->setIndexBase( indexBase );
            indexBase++;
        }
    }
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryTimers::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    // Nur vorhandene Einträge speichern.
    if ( size() )
    {
        StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

        // Anlegen der Rahmenstruktur, für die Datenablage.

        // ################################################
        // Strukturen der Timer im naechsten freien Speicherblock ablegen
        extBegin.offsetTimers = extBegin.offsetFreeScope;

        // initialisieren des OffsetArrays
        memset( &file.m_flashExtended[extBegin.offsetTimers], 0, sizeof( StructHeaderOffsetArray ) );
        // Daten des OffsetArrays eintragen.
        StructHeaderOffsetArray *headerOffsetArray = (StructHeaderOffsetArray *)&file.m_flashExtended[extBegin.offsetTimers];
        headerOffsetArray->size = sizeof( StructHeaderOffsetArray );
        headerOffsetArray->count = size();
        headerOffsetArray->offsetArray = headerOffsetArray->size;
        // initialisieren
        memset( &file.m_flashExtended[extBegin.offsetTimers + headerOffsetArray->offsetArray], 0, headerOffsetArray->count * sizeof( unsigned int ) );
        // Vordefinieren der Offsets für das OffsetArray
        for ( uint32_t i = 0; i < headerOffsetArray->count; ++i )
            *(unsigned int *)&file.m_flashExtended[extBegin.offsetTimers + headerOffsetArray->offsetArray + i * sizeof( unsigned int )] = (headerOffsetArray->count * sizeof( unsigned int )) + i * sizeof( StructTimer );

        // initialisieren des Ablagespeichers für die Datenstrukturen
        memset( &file.m_flashExtended[extBegin.offsetTimers + headerOffsetArray->offsetArray + headerOffsetArray->count * sizeof( unsigned int )], 0, headerOffsetArray->count * sizeof( StructTimer ) );

        // Den nächsten freie Bereich angeben.
        extBegin.offsetFreeScope += headerOffsetArray->size + headerOffsetArray->count * sizeof( unsigned int ) + headerOffsetArray->count * sizeof( StructTimer );

        // Schreiben der Erweiterten Struktur
        *(StructExtendedFlash *)file.m_flashExtended = extBegin;

        ////////////////////////////////////////
        // add information to xml file
        FileXML &fileXML = doc.fileXML();

        // Finding the node for this structure and update parent.
        auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::Timers ) );

        // Alle Unterpunkte in Flash eintragen und die jeweiligen Pointer sichern
        auto iterChild = constBegin();
        for ( ; iterChild != constEnd(); ++iterChild )
        {
            (*iterChild)->writeDocument( pFlash, doc );
        }

        // Reset parent node because the end of the structure is reached.
        fileXML.setParentNode( oldParent );
    }

    // Wegen Abwärtskompatibilität die Ersten vier Timer auch in der erweiteren Struktur ablegen.
    if ( size() < 4 )
    {
        // Hier wird sichergestellt, dass unbenutzte Timer auch in der erweiterten Struktur gelöscht werden. Ist wegen Abwärtskompatibilität nötig.

        // Schreiben der erweiterten Parameter - Struktur
        // Diese Struktur enthaelt informationen auf die beim Baumaufbau von anderen Eintraegen zurueckgegriffen wird,
        // somit muss sie vor dem Aufbau geladen sein.
        EntryDevice &device = root();
        // Nulltimer
        StructTimer timer;
        memset( &timer, 0, sizeof( StructTimer ) );

        for ( int i = size(); i < 4; ++i )
            device.m_extended.timer[i] = timer;
    }

    return pFlash;
}

uint32_t EntryTimers::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    uint32_t size{};
    StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    if ( useAESFile )
    {
        // Bei aktuellen Setupdateien wird mit der Signalverarbeitung in einen neuen Bereich gewechselt.
        if ( extBegin.size > sizeof_StructExtendedFlash_1 &&extBegin.offsetTimers )
        {
            // Daten des OffsetArrays lesen.
            StructHeaderOffsetArray *headerOffsetArray = (StructHeaderOffsetArray *)&file.m_flashExtended[extBegin.offsetTimers];
            // Anzahl ermitteln
            size = headerOffsetArray->count;
        }
        else
        {
            // Lesen der erweiterten Parameter - Struktur
            // Diese Struktur enthaelt informationen auf die beim Baumaufbau von anderen Eintraegen zurueckgegriffen wird,
            // somit muss sie vor dem Aufbau geladen sein.
            EntryDevice &device = root();

            for ( uint32_t i = 0; i < 4; ++i )
            {
                if ( !std::all_of( (uint8_t *)&device.m_extended.timer[i], (uint8_t *)&device.m_extended.timer[i] + sizeof( StructTimer ), []( uint8_t v )
                {
                    return v == 0;
                } ) )
                {
                    size = i + 1;
                }
            }
        }

        // Daten aller Eintraege einlesen
        for ( uint32_t i = 0; i < size; ++i )
        {
            EntryTimer *pNeu = new EntryTimer( this );
            pNeu->setDefault( i );

            pNeu->readDocument( pFlash, doc );
            append( pNeu );
        }
    }
    else
    {
        // Finding the node for this structure and update parent.
        auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::Timers ) );

        size = fileXML.numberOfChildItems( FileXML::ContentType::Timer );

        // Reading all analog inputs.
        for ( uint32_t i = 0; i < size; ++i )
        {
            EntryTimer *pInput = new EntryTimer( this );
            pInput->setDefault( i );

            pInput->readDocument( i, doc );
            append( pInput );
        }

        // Reset parent node because the end of the structure is reached.
        fileXML.setParentNode( oldParent );
    }

    return pFlash;
}

///////////////////////////////////
// Daten des Eintrags

void EntryTimers::setDefault()
{
    // Icon's
    m_iconIndex = ICON_TIMERS;
    m_iconIndexSelected = ICON_TIMERS_SEL;

    // Name
    setName( QT_TRANSLATE_NOOP_UTF8( "EntryTimers", "Timer" ) );

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_TIMER_0 );

    setInsertEnabled( true );
    setDeleteEnabled( false );
}

////////////////////////////////////
// Schnittstelle des Eintrags
StructTimer EntryTimers::timer( int index ) const
{
    StructTimer res;
    memset( &res, 0, sizeof( StructTimer ) );

    // Alle Unterpunkte in Flash eintragen und die jeweiligen Pointer sichern
    auto iterChild = constBegin();
    for ( ; iterChild != constEnd(); ++iterChild )
    {
        if ( ((EntryTimer *)(*iterChild))->indexBase() == index )
        {
            res = ((EntryTimer *)(*iterChild))->m_timer;
            break;
        }
    }

    return res;
}

int EntryTimers::timerCount() const
{
    return size();
}

// ########################################################
// Member: EntryTimer

///////////////////////////////////
// Childs

EntryBase *EntryTimer::insertEntry( EntryType typeID, int pos )
{
    (void)typeID;
    (void)pos;

    return nullptr;
}

EntryBase *EntryTimer::cloneEntry()
{
    // Dieser Eintrag unterstützt derzeit cloneEntry nicht.
    return nullptr;
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryTimer::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    // Erweiterter Flashbereich
    StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

    // OffsetArray der Digitalen Eingänge
    StructHeaderOffsetArray *headerOffsetArray = (StructHeaderOffsetArray *)&file.m_flashExtended[extBegin.offsetTimers];
    unsigned int offsetFlash = *(unsigned int *)&file.m_flashExtended[extBegin.offsetTimers + headerOffsetArray->offsetArray + m_indexBase * sizeof( unsigned int )];
    // Zeiger auf die Struktur fuer den Analogen Eingang
    StructTimer *timer = (StructTimer *)&file.m_flashExtended[extBegin.offsetTimers + headerOffsetArray->offsetArray + offsetFlash];

    // Schreiben der Struktur fuer den Timer
    *timer = m_timer;

    ////////////////////////////////////////
    // add information to xml file
    FileXML &fileXML = doc.fileXML();
    fileXML.writeContent( FileXML::ContentType::Timer, as_const( m_timer ) );

    // Wegen Abwärtskompatibilität die Ersten vier Timer auch in der erweiteren Struktur ablegen.
    if ( m_indexBase >= 0 && m_indexBase <= 3 )
    {
        // Schreiben der erweiterten Parameter - Struktur
        // Diese Struktur enthaelt informationen auf die beim Baumaufbau von anderen Eintraegen zurueckgegriffen wird,
        // somit muss sie vor dem Aufbau geladen sein.
        EntryDevice &device = root();

        device.m_extended.timer[m_indexBase] = m_timer;
    }

    return pFlash;
}

uint32_t EntryTimer::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    // Erweiterten Struktur lesen
    StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    if ( useAESFile )
    {
        // Struktur des Timers lesen
        StructTimer *timer = nullptr;

        // Die die Offsets der aktuellen TExtededFlash-Struktur vor?
        if ( extBegin.size > sizeof_StructExtendedFlash_1 &&extBegin.offsetTimers )
        {
            // OffsetArray der Digitalen Eingänge
            StructHeaderOffsetArray *headerOffsetArray = (StructHeaderOffsetArray *)&file.m_flashExtended[extBegin.offsetTimers];
            unsigned int offsetFlash = *(unsigned int *)&file.m_flashExtended[extBegin.offsetTimers + headerOffsetArray->offsetArray + m_indexBase * sizeof( unsigned int )];
            // Zeiger auf die Struktur fuer Timers
            timer = (StructTimer *)&file.m_flashExtended[extBegin.offsetTimers + headerOffsetArray->offsetArray + offsetFlash];
        }
        else
        {
            // Lesen der erweiterten Parameter - Struktur
            // Diese Struktur enthaelt informationen auf die beim Baumaufbau von anderen Eintraegen zurueckgegriffen wird,
            // somit muss sie vor dem Aufbau geladen sein.
            EntryDevice &device = root();

            if ( m_indexBase >= 0 && m_indexBase <= 3 )
            {
                timer = &device.m_extended.timer[m_indexBase];
            }
            else
            {
                BOOST_ASSERT( false );
            }
        }

        // Strukturzeiger zugewiesen?
        if ( timer != nullptr )
        {
            // Die Struktur für den Timer lesen.
            m_timer = *timer;
        }
    }
    else
    {
        tinyxml2::XMLElement *lastXmlElement{};
        // pFlash contains the index of the current timer
        fileXML.combineValid( fileXML.readContent( FileXML::ContentType::Timer, m_timer, pFlash, lastXmlElement ) );
    }

    return pFlash;
}

void EntryTimer::checkMembers()
{
}

///////////////////////////////////////
// Referenzierung zwischen Einträgen

void EntryTimer::convertReferenceToIndex()
{
    EntryEventChains &eventChains = rootEventChains();

    m_timer.invokeInputChain = (m_refInputChainTimer != nullptr) ? eventChains.indexOf( m_refInputChainTimer ) + 1 : 0;

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertReferenceToIndex();
}

void EntryTimer::convertIndexToReference()
{
    EntryEventChains &eventChains = rootEventChains();

    m_refInputChainTimer = (m_timer.invokeInputChain > 0) ? (EntryInputChain *)eventChains.at( m_timer.invokeInputChain - 1 ) : nullptr;
    if ( m_refInputChainTimer )
        m_refInputChainTimer->incRefCount();

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertIndexToReference();
}

///////////////////////////////////////
// Daten des Eintrags

void EntryTimer::setDefault()
{
    setDefault( 0 );
}

void EntryTimer::setDefault( int indexBase )
{
    m_iconIndex = ICON_TIMER;
    m_iconIndexSelected = ICON_TIMER_SEL;

    // Offset auf die Analogen Eingänge
    // 0 -> T1
    // 1 -> T2
    // 2 -> T3
    // ...
    setIndexBase( indexBase );

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_TIMER_1 );

    // Strukturen fuer Analoge - Eingaenge loeschen.
    memset( &m_timer, 0, sizeof( StructTimer ) );

    // Referenzbezug 
    m_refInputChainTimer = 0;

    setInsertEnabled( false );
    setDeleteEnabled( true );
}

// ########################################################
// Member: EntrySystemMessages

///////////////////////////////////
// Childs

EntryBase *EntrySystemMessages::insertEntry( EntryType typeID, int pos )
{
    int indexBase = pos;

    // Maximaleintragsgröße abfangen.
    if ( size() >= 32 )
    {
        // Es sind maximal 32 Systemmeldungen anlegbar.
        return nullptr;
    }

    if ( indexBase == -1 )
        indexBase = size();

    EntrySystemMessage *entryNew = new EntrySystemMessage( this );
    entryNew->setDefault( indexBase );

    insertEntryFinish( entryNew, pos );

    if ( pos != -1 )
    {
        indexBase = 0;
        // Allen Unterpunkten den basisindex und dadurch auch Namen zuweisen.
        auto iterChild = constBegin();
        for ( ; iterChild != constEnd(); ++iterChild )
        {
            ((EntrySystemMessage *)*iterChild)->setIndexBase( indexBase );
            indexBase++;
        }
    }

    return entryNew;
}

EntryBase *EntrySystemMessages::cloneEntry()
{
    // Dieser Eintrag unterstützt derzeit cloneEntry nicht.
    return nullptr;
}

void EntrySystemMessages::deleteEntry( int pos )
{
    // Loeschanforderung an Basisklasse weiterleiten
    EntryBase::deleteEntry( pos );

    if ( pos != -1 )
    {
        int indexBase = 0;
        // Allen Unterpunkten den basisindex und dadurch auch Namen zuweisen.
        auto iterChild = constBegin();
        for ( ; iterChild != constEnd(); ++iterChild )
        {
            ((EntrySystemMessage *)*iterChild)->setIndexBase( indexBase );
            indexBase++;
        }
    }
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntrySystemMessages::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    // Nur vorhandene Einträge speichern.
    if ( size() )
    {
        StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

        // Anlegen der Rahmenstruktur, für die Datenablage.

        // ################################################
        // Strukturen der Systemmeldungen im naechsten freien Speicherblock ablegen
        extBegin.offsetSystemMessages = extBegin.offsetFreeScope;

        // initialisieren des OffsetArrays
        memset( &file.m_flashExtended[extBegin.offsetSystemMessages], 0, sizeof( StructHeaderOffsetArray ) );
        // Daten des OffsetArrays eintragen.
        StructHeaderOffsetArray *headerOffsetArray = (StructHeaderOffsetArray *)&file.m_flashExtended[extBegin.offsetSystemMessages];
        headerOffsetArray->size = sizeof( StructHeaderOffsetArray );
        headerOffsetArray->count = size();
        headerOffsetArray->offsetArray = headerOffsetArray->size;
        // initialisieren
        memset( &file.m_flashExtended[extBegin.offsetSystemMessages + headerOffsetArray->offsetArray], 0, headerOffsetArray->count * sizeof( unsigned int ) );
        // Vordefinieren der Offsets für das OffsetArray
        for ( uint32_t i = 0; i < headerOffsetArray->count; ++i )
            *(unsigned int *)&file.m_flashExtended[extBegin.offsetSystemMessages + headerOffsetArray->offsetArray + i * sizeof( unsigned int )] = (headerOffsetArray->count * sizeof( unsigned int )) + i * sizeof( StructSystemMessage );

        // initialisieren des Ablagespeichers für die Datenstrukturen
        memset( &file.m_flashExtended[extBegin.offsetSystemMessages + headerOffsetArray->offsetArray + headerOffsetArray->count * sizeof( unsigned int )], 0, headerOffsetArray->count * sizeof( StructSystemMessage ) );

        // Den nächsten freie Bereich angeben.
        extBegin.offsetFreeScope += headerOffsetArray->size + headerOffsetArray->count * sizeof( unsigned int ) + headerOffsetArray->count * sizeof( StructSystemMessage );

        // Schreiben der Erweiterten Struktur
        *(StructExtendedFlash *)file.m_flashExtended = extBegin;

        ////////////////////////////////////////
        // add information to xml file
        FileXML &fileXML = doc.fileXML();

        // Finding the node for this structure and update parent.
        auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::SystemMessages ) );

        // Alle Unterpunkte in Flash eintragen und die jeweiligen Pointer sichern
        auto iterChild = constBegin();
        for ( ; iterChild != constEnd(); ++iterChild )
            (*iterChild)->writeDocument( pFlash, doc );

        // Reset parent node because the end of the structure is reached.
        fileXML.setParentNode( oldParent );
    }

    return pFlash;
}

uint32_t EntrySystemMessages::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    uint32_t size{};
    StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    if ( useAESFile )
    {
    // Bei aktuellen Setupdateien wird mit der Signalverarbeitung in einen neuen Bereich gewechselt.
        if ( extBegin.size > sizeof_StructExtendedFlash_1 &&extBegin.offsetSystemMessages )
        {
            // Daten des OffsetArrays lesen.
            StructHeaderOffsetArray *headerOffsetArray = (StructHeaderOffsetArray *)&file.m_flashExtended[extBegin.offsetSystemMessages];
            // Anzahl ermitteln
            size = headerOffsetArray->count;
        }

        // Daten aller Eintraege einlesen
        for ( uint32_t i = 0; i < size; ++i )
        {
            EntrySystemMessage *pNeu = new EntrySystemMessage( this );
            pNeu->setDefault( i );

            pNeu->readDocument( pFlash, doc );
            append( pNeu );
        }
    }
    else
    {
        // Finding the node for this structure and update parent.
        auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::SystemMessages ) );

        size = fileXML.numberOfChildItems( FileXML::ContentType::SystemMessage );

        // Reading all analog inputs.
        for ( uint32_t i = 0; i < size; ++i )
        {
            EntrySystemMessage *pInput = new EntrySystemMessage( this );
            pInput->setDefault( i );

            pInput->readDocument( i, doc );
            append( pInput );
        }

        // Reset parent node because the end of the structure is reached.
        fileXML.setParentNode( oldParent );
    }

    return pFlash;
}

///////////////////////////////////
// Daten des Eintrags

void EntrySystemMessages::setDefault()
{
    // Icon's
    m_iconIndex = ICON_TIMERS;
    m_iconIndexSelected = ICON_TIMERS_SEL;

    // Name
    setName( QT_TRANSLATE_NOOP_UTF8( "EntrySystemMessages", "Systemmeldung" ) );

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_BLANK );

    setInsertEnabled( true );
    setDeleteEnabled( false );
}

////////////////////////////////////
// Schnittstelle des Eintrags
StructSystemMessage EntrySystemMessages::systemMessage( int index ) const
{
    StructSystemMessage res;
    memset( &res, 0, sizeof( StructSystemMessage ) );

    // Alle Unterpunkte in Flash eintragen und die jeweiligen Pointer sichern
    auto iterChild = constBegin();
    for ( ; iterChild != constEnd(); ++iterChild )
    {
        if ( ((EntrySystemMessage *)(*iterChild))->indexBase() == index )
        {
            res = ((EntrySystemMessage *)(*iterChild))->m_systemMessage;
            break;
        }
    }

    return res;
}

int EntrySystemMessages::systemMessageCount() const
{
    return size();
}

bool EntrySystemMessages::exists( EnumSystemMessageType type, UnionSystemMessageId id )
{
    bool found = false;
    // Ggf. vorhandenen Eintrag der Systemmeldung finden.
    for ( uint32_t pos = 0; found == false && pos < size(); ++pos )
    {
        const EntrySystemMessage *item = reinterpret_cast<const EntrySystemMessage *>(at( pos ));
        if ( item->m_systemMessage.type == type )
        {
            switch ( type )
            {
                case EnumSystemMessageType::Information:
                    found = item->m_systemMessage.id.information == id.information;
                    break;

                case EnumSystemMessageType::Error:
                    found = item->m_systemMessage.id.error == id.error;
                    break;

                default:
                    BOOST_ASSERT( false ); // Es müssen alle typen berücksichtigt werden!!
            }
        }
    }

    return found;
}

EntrySystemMessage *EntrySystemMessages::add( EnumSystemMessageType type, UnionSystemMessageId id )
{
    EntrySystemMessage *systemMessage = nullptr;
    BOOST_ASSERT( exists( type, id ) == false );

    insertEntry();
    systemMessage = const_cast<EntrySystemMessage *>(reinterpret_cast<const EntrySystemMessage *>(last()));
    systemMessage->m_systemMessage.type = type;
    systemMessage->m_systemMessage.id = id;

    return systemMessage;
}

void EntrySystemMessages::remove( EnumSystemMessageType type, UnionSystemMessageId id )
{
    bool found = false;
    // Ggf. vorhandenen Eintrag entfernen.
    for ( uint32_t pos = 0; found == false && pos < size(); ++pos )
    {
        const EntrySystemMessage *item = reinterpret_cast<const EntrySystemMessage *>(at( pos ));
        if ( item->m_systemMessage.type == type )
        {
            switch ( type )
            {
                case EnumSystemMessageType::Information:
                    found = item->m_systemMessage.id.information == id.information;
                    break;

                case EnumSystemMessageType::Error:
                    found = item->m_systemMessage.id.error == id.error;
                    break;

                default:
                    BOOST_ASSERT( false ); // Es müssen alle typen berücksichtigt werden!!
            }

            if ( found )
            {
                deleteEntry( pos );
            }
        }
    }
}

EntrySystemMessage *EntrySystemMessages::systemMessage( EnumSystemMessageType type, UnionSystemMessageId id )
{
    EntrySystemMessage *entry = nullptr;
    // Ggf. vorhandenen Eintrag der Systemmeldung finden.
    for ( uint32_t pos = 0; entry == nullptr && pos < size(); ++pos )
    {
        const EntrySystemMessage *item = reinterpret_cast<const EntrySystemMessage *>(at( pos ));
        if ( item->m_systemMessage.type == type )
        {
            switch ( type )
            {
                case EnumSystemMessageType::Information:
                    if ( item->m_systemMessage.id.information == id.information )
                        entry = const_cast<EntrySystemMessage *>(item);
                    break;

                case EnumSystemMessageType::Error:
                    if ( item->m_systemMessage.id.error == id.error )
                        entry = const_cast<EntrySystemMessage *>(item);
                    break;

                default:
                    BOOST_ASSERT( false ); // Es müssen alle typen berücksichtigt werden!!
            }
        }
    }

    return entry;
}

// ########################################################
// Member: EntrySystemMessage

///////////////////////////////////
// Childs

EntryBase *EntrySystemMessage::insertEntry( EntryType typeID, int pos )
{
    (void)typeID;
    (void)pos;

    return nullptr;
}

EntryBase *EntrySystemMessage::cloneEntry()
{
    // Dieser Eintrag unterstützt derzeit cloneEntry nicht.
    return nullptr;
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntrySystemMessage::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    // Erweiterter Flashbereich
    StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

    // OffsetArray der Systemmeldungen
    StructHeaderOffsetArray *headerOffsetArray = (StructHeaderOffsetArray *)&file.m_flashExtended[extBegin.offsetSystemMessages];
    unsigned int offsetFlash = *(unsigned int *)&file.m_flashExtended[extBegin.offsetSystemMessages + headerOffsetArray->offsetArray + m_indexBase * sizeof( unsigned int )];
    // Zeiger auf die Struktur fuer die Systemmeldung
    StructSystemMessage *message = (StructSystemMessage *)&file.m_flashExtended[extBegin.offsetSystemMessages + headerOffsetArray->offsetArray + offsetFlash];

    // Schreiben der Struktur fuer die Systemmeldung
    *message = m_systemMessage;

    ////////////////////////////////////////
    // add information to xml file
    // The parent is configured outside the method and is not changed.
    doc.fileXML().writeContent( FileXML::ContentType::SystemMessage, as_const( m_systemMessage ) );

    return pFlash;
}

uint32_t EntrySystemMessage::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    // Erweiterten Struktur lesen
    StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    if ( useAESFile )
    {
        // Die die Offsets der aktuellen TExtededFlash-Struktur vor?
        if ( extBegin.size > sizeof_StructExtendedFlash_1 &&extBegin.offsetSystemMessages )
        {
            // OffsetArray der Systemmeldungen
            StructHeaderOffsetArray *headerOffsetArray = (StructHeaderOffsetArray *)&file.m_flashExtended[extBegin.offsetSystemMessages];
            unsigned int offsetFlash = *(unsigned int *)&file.m_flashExtended[extBegin.offsetSystemMessages + headerOffsetArray->offsetArray + m_indexBase * sizeof( unsigned int )];
            // Zeiger auf die Struktur fuer Systemmeldung
            m_systemMessage = *(StructSystemMessage *)&file.m_flashExtended[extBegin.offsetSystemMessages + headerOffsetArray->offsetArray + offsetFlash];
        }
    }
    else
    {
        tinyxml2::XMLElement *lastXmlElement{};
        // pFlash contains the index of the current message
        fileXML.combineValid( fileXML.readContent( FileXML::ContentType::SystemMessage, m_systemMessage, pFlash, lastXmlElement ) );
    }

    return pFlash;
}

void EntrySystemMessage::checkMembers()
{
}

///////////////////////////////////////
// Referenzierung zwischen Einträgen

void EntrySystemMessage::convertReferenceToIndex()
{
    EntryEventChains &eventChains = rootEventChains();

    m_systemMessage.invokeInputChain = (m_refInputChain != nullptr) ? eventChains.indexOf( m_refInputChain ) + 1 : 0;

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertReferenceToIndex();
}

void EntrySystemMessage::convertIndexToReference()
{
    EntryEventChains &eventChains = rootEventChains();

    m_refInputChain = (m_systemMessage.invokeInputChain > 0) ? (EntryInputChain *)eventChains.at( m_systemMessage.invokeInputChain - 1 ) : nullptr;
    if ( m_refInputChain )
        m_refInputChain->incRefCount();

    // Basisklasse ruft alle Childs auf.
    EntryBase::convertIndexToReference();
}

///////////////////////////////////////
// Daten des Eintrags

void EntrySystemMessage::setDefault()
{
    setDefault( 0 );
}

void EntrySystemMessage::setDefault( int indexBase )
{
    m_iconIndex = ICON_TIMER;
    m_iconIndexSelected = ICON_TIMER_SEL;

    // Offset auf die Analogen Eingänge
    // 0 -> T1
    // 1 -> T2
    // 2 -> T3
    // ...
    setIndexBase( indexBase );

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_TIMER_1 );

    // Strukturen fuer Systemmeldung loeschen.
    memset( &m_systemMessage, 0, sizeof( StructSystemMessage ) );

    // Referenzbezug 
    m_refInputChain = 0;

    setInsertEnabled( false );
    setDeleteEnabled( true );
}

// ########################################################
// Member: EntryEventChains

///////////////////////////////////
// Childs

EntryBase *EntryEventChains::insertEntry( EntryType typeID, int pos )
{
    EntryBase *entryNew;

    switch ( typeID )
    {
        case EntryType::InputChain:
            entryNew = new EntryInputChain( this );
            ((EntryInputChain *)entryNew)->setDefault( boost::str( boost::format( "%1%%2%" ) % QT_TRANSLATE_NOOP_UTF8( "EntryEventChains", "Eingabekette" ) % (size() + 1) ) );
            break;

        default:
            entryNew = new EntryInputChain( this );
            entryNew->setDefault();
            break;
    }

    insertEntryFinish( entryNew, pos );

    return entryNew;
}

EntryBase *EntryEventChains::cloneEntry()
{
    // Dieser Eintrag unterstützt derzeit cloneEntry nicht.
    return nullptr;
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryEventChains::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    // Erweiterter Flashbereich
    FileAES &file = doc.fileAES();
    int i;
    StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

    unsigned int aP[MAX_SUBPOINTS], PArrayStart;
    int Unterpunkte;

    // Anzahl Unterpunkte berechnen und eintragen
    Unterpunkte = (int)size();
    if ( Unterpunkte > MAX_SUBPOINTS )
        Unterpunkte = MAX_SUBPOINTS;

    // Eingabeketten in naechsten Extendedbereich eintragen
    extBegin.offsetEventChains = extBegin.offsetFreeScope;

    // Anzahl in erste Byte eintragen
    *(unsigned int *)(&file.m_flashExtended[extBegin.offsetEventChains]) = Unterpunkte;
    changeOrderInt32( &file.m_flashExtended[extBegin.offsetEventChains] );

    // erster freier Speicherplatz -> begin Pointerarray
    PArrayStart = sizeof( unsigned int );
    // Platz für Pointerarray reservieren
    aP[0] = PArrayStart + sizeof( unsigned int ) * Unterpunkte;

    m_flashOffset = FLASH_SIZE_SU;

    // Extended für Flash ausgeben.
    unsigned char *flashMem = file.m_flash;
    file.m_flash = &file.m_flashExtended[extBegin.offsetEventChains];

        ////////////////////////////////////////
        // add information to xml file
    FileXML &fileXML = doc.fileXML();

    // Finding the node for this structure and update parent.
    auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::EventChains ) );

    auto iterChild = constBegin();
    for ( i = 1; iterChild != constEnd() && (i <= MAX_SUBPOINTS); ++iterChild, ++i )
    {
        aP[i] = (*iterChild)->writeDocument( aP[i - 1], doc );
    }

    // Reset parent node because the end of the structure is reached.
    fileXML.setParentNode( oldParent );

    file.m_flash = flashMem;

    m_flashOffset = 0;

    // Pointer auf naechsten freien Speicher im file.m_flashExtended
    extBegin.offsetFreeScope = extBegin.offsetEventChains + aP[i - 1];

    // Pointer auf Unterpunkte eintragen
    for ( i = 0; i < Unterpunkte; ++i )
    {
        // 16kByte fuer FlashOffset aufaddieren
        aP[i] += FLASH_SIZE_SU;
        aP[i] = endianReverse( aP[i] + FLASH_KONFIG_START );
        *(unsigned int *)(&file.m_flashExtended[extBegin.offsetEventChains + PArrayStart + i * sizeof( unsigned int )]) = aP[i];
    }

    // Schreiben der Erweiterten Struktur
    *(StructExtendedFlash *)file.m_flashExtended = extBegin;

    return pFlash;
}

uint32_t EntryEventChains::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    // Erweiterten Struktur lesen
    const StructExtendedFlash &extBegin = file.m_extended_header;

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    if ( useAESFile )
    {
        if ( extBegin.offsetEventChains != 0 )
        {
            unsigned int Unterpunkte;
            unsigned int i;
            unsigned int pEK;

            Unterpunkte = *(unsigned int *)&file.m_flashExtended[extBegin.offsetEventChains];
            Unterpunkte = endianReverse( Unterpunkte );

            if ( Unterpunkte != 0 )
            {
                m_flashOffset = FLASH_SIZE_SU;
                for ( i = 0; i < Unterpunkte; ++i )
                {
                    // Pointer des Eintrags berechnen
                    pEK = extBegin.offsetEventChains + sizeof( unsigned int ) + i * sizeof( unsigned int );
                    pEK = *reinterpret_cast<unsigned int *>(&file.m_flashExtended[pEK]);
                    pEK = endianReverse( pEK ) - FLASH_KONFIG_START;
                    // 16kByte fuer FlashOffset subtrahieren
                    pEK -= FLASH_SIZE_SU;

                    EntryInputChain *pNeu = new EntryInputChain( this );
                    pNeu->setDefault();
                    // Extended fuer Flash ausgeben!!
                    unsigned char *flashMem = file.m_flash;
                    file.m_flash = &file.m_flashExtended[extBegin.offsetEventChains];

                    pNeu->readDocument( pEK, doc );

                    file.m_flash = flashMem;
                    append( pNeu );
                }
                m_flashOffset = 0;
            }
        }
    }
    else
    {
        if ( fileXML.numberOfChildItems( FileXML::ContentType::EventChains ) > 0 )
        {
            // Finding the node for this structure and update parent.
            auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::EventChains ) );

            uint32_t size = fileXML.numberOfChildItems( FileXML::ContentType::Operating );
            for ( uint32_t i = 0; i < size; ++i )
            {
                EntryInputChain *pInput = new EntryInputChain( this );
                pInput->setDefault();
                pInput->readDocument( i, doc );
                append( pInput );
            }

            // Reset parent node because the end of the structure is reached.
            fileXML.setParentNode( oldParent );
        }
    }

    return pFlash;
}

///////////////////////////////////
// Daten des Eintrags

void EntryEventChains::setDefault()
{
    setName( QT_TRANSLATE_NOOP_UTF8( "EntryEventChains", "Ereignisketten" ) );

    m_iconIndex = ICON_INPUT_CHAINS;
    m_iconIndexSelected = ICON_INPUT_CHAINS_SEL;

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_BLANK );

    setInsertEnabled( TRUE );
    setDeleteEnabled( FALSE );
}

// ########################################################
// Memeber: EntryAccessControl2 (Zutrittskontrolle 2)

///////////////////////////////////
// Childs

EntryBase *EntryAccessControl2::insertEntry( EntryType typeID, int pos )
{
    (void)typeID;
    (void)pos;

    return nullptr;
}

EntryBase *EntryAccessControl2::cloneEntry()
{
    // Dieser Eintrag unterstützt derzeit cloneEntry nicht.
    return nullptr;
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryAccessControl2::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    int i;
    unsigned int Unterpunkte;
    // Anzahl Unterpunkte berechnen und eintragen
    Unterpunkte = (unsigned int)size();
    if ( Unterpunkte > MAX_SUBPOINTS )
        Unterpunkte = MAX_SUBPOINTS;

    // Eintraegen der 7 Tabellen
    if ( Unterpunkte < 7 )
    {
        // Tabellen sind anzulegen
        setDefaultTable();
        // Eingabeketten sind Anzulegen
        setDefaultInputChain();
        // Erneut ermitteln, da neue angelegt.
        Unterpunkte = (unsigned int)size();
    }

    EntryDevice &entry_device = root();
    // Parameter werden von ZK1 & ZK2 verwendet und duerfen,
    // nur von der eingestellten gesetzt werden. (sonst Ueberschreibung)
    if ( entry_device.m_config.enableEntrance2 )
    {
        entry_device.m_config.useExternalReader = m_useExternalReader;
        entry_device.m_config.entranceMainMenu = m_entranceMainMenu;
        entry_device.m_config.entranceFlag = m_accessControlFlag;
        entry_device.m_config.enableCut = m_enableCut;
        entry_device.m_config.cutFrom = m_cutAt;
        entry_device.m_config.cutUpTo = m_cutTo;

        // Transponderstruktur schreiben
        entry_device.m_config.transponderAccessControl = m_transponder;

        // Lesen der Extended - Struktur
        StructExtendedFlash &extended_header = file.m_extended_header;

        // Listen der Zutrittskontrolle (Version 2) eintragen.
        extended_header.offsetAccessControlTablesV2 = extended_header.offsetFreeScope;


        unsigned int aP[MAX_SUBPOINTS], PArrayStart;

        // Abhaengig davon ob die Erweiterte ZK gefordert ist,
        // die Tabelle Action2 mit ablegen.
        int tablesCount = 7;
        int ekCount = 1;
        EntryDevice *device = reinterpret_cast<EntryDevice *>(parent());
        if ( device->m_extended.acExtension )
        {
            // Auch Action2 in die Datei ablegen.
            tablesCount = 8;

            // Erweiterte ZK wird erst ab Versionen x unterstuetzt.
            file.NeedStudioVersion( "04.02.03.01" );
            file.NeedLibraryVersion( "04.02.03.01" );
            file.NeedFirmware2Version( "04.02.03.00" );
            file.UnsupportFirmware1Version();
        }

        // Anzahl in erste Byte eintragen
        *(unsigned int *)&file.m_flashExtended[extended_header.offsetAccessControlTablesV2] = tablesCount;

        // erster freier Speicherplatz -> begin Pointerarray
        PArrayStart = sizeof( unsigned int );
        // Platz fuer Pointerarray reservieren
        aP[0] = PArrayStart + sizeof( unsigned int ) * tablesCount;

        unsigned char *flashMem = file.m_flash;
        file.m_flash = &file.m_flashExtended[extended_header.offsetAccessControlTablesV2];

        ////////////////////////////////////////
        // add information to xml file
        FileXML &fileXML = doc.fileXML();

        // Finding the node for this structure and update parent.
        auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::AccessControl ) );

        auto iterChild = constBegin();
        for ( i = 1; iterChild != constEnd() && (i <= tablesCount); ++iterChild, ++i )
        {
            // Index der Tabelle eintragen
            ((EntryTable *)*iterChild)->m_table.number = i - 1;

            // Extended fuer Flash ausgeben
            aP[i] = (*iterChild)->writeDocument( aP[i - 1], doc );
        }

        file.m_flash = flashMem;

        // Pointer auf naechsten freien Speicher im file.m_flashExtended
        extended_header.offsetFreeScope = extended_header.offsetAccessControlTablesV2 + aP[i - 1];

        // Pointer auf Unterpunkte eintragen
        for ( i = 0; i < tablesCount; ++i )
        {
            *(unsigned int *)&file.m_flashExtended[extended_header.offsetAccessControlTablesV2 + PArrayStart + i * sizeof( unsigned int )] = aP[i];
            *(unsigned int *)&file.m_flashExtended[extended_header.offsetAccessControlTablesV2 + PArrayStart + i * sizeof( unsigned int )] += FLASH_KONFIG_START; // Absoluter Speicheroffset.
        }

        // EK's der Zutrittskontrolle (Version 2) eintragen.
        // Eintraegen der Eingabeketten
        extended_header.offsetAccessControlEventChainsV2 = extended_header.offsetFreeScope;

        // Auf die Erste EK wechseln.
        iterChild = constBegin();
        for ( ; iterChild != constEnd(); ++iterChild )
            if ( (*iterChild)->isType( EntryType::InputChain ) )
                break;

        // Anzahl EK's als erstes eintragen
        *(unsigned int *)&file.m_flashExtended[extended_header.offsetAccessControlEventChainsV2] = ekCount;

        // erster freier Speicherplatz -> begin Pointerarray
        PArrayStart = sizeof( unsigned int );
        // Platz für Pointerarray reservieren
        aP[0] = PArrayStart + sizeof( unsigned int ) * ekCount;

        // Extended fuer Flash ausgeben
        flashMem = file.m_flash;
        file.m_flash = &file.m_flashExtended[extended_header.offsetAccessControlEventChainsV2];

        // Finding the node for this structure and update parent.
        fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::Access ) );

        // Alle Unterpunkte in Flash eintragen und die jeweiligen Pointer sichern
        for ( i = 1; (i <= (int)ekCount) && iterChild != constEnd(); ++iterChild, ++i )
        {
            aP[i] = (*iterChild)->writeDocument( aP[i - 1], doc );
        }

        file.m_flash = flashMem;

        // Reset parent node because the end of the structure is reached.
        fileXML.setParentNode( oldParent );

        // Pointer auf naechsten freien Speicher im file.m_flashExtended
        extended_header.offsetFreeScope = extended_header.offsetAccessControlEventChainsV2 + aP[i - 1];

        // Pointer auf Unterpunkte eintragen
        for ( i = 0; i < ekCount; ++i )
        {
            *(unsigned int *)&file.m_flashExtended[extended_header.offsetAccessControlEventChainsV2 + PArrayStart + i * sizeof( unsigned int )] = aP[i];
            *(unsigned int *)&file.m_flashExtended[extended_header.offsetAccessControlEventChainsV2 + PArrayStart + i * sizeof( unsigned int )] += FLASH_KONFIG_START; // Absoluter Speicheroffset.
        }

        // Schreiben der Erweiterten Struktur
        file.m_extended_header = extended_header;
    }

    return pFlash;
}

uint32_t EntryAccessControl2::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    int i;
    EntryDevice &entry_device = root();

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    if ( entry_device.m_config.enableEntrance2 )
    {
        m_useExternalReader = entry_device.m_config.useExternalReader;
        m_entranceMainMenu = entry_device.m_config.entranceMainMenu;
        m_accessControlFlag = entry_device.m_config.entranceFlag;
        m_enableCut = entry_device.m_config.enableCut;
        m_cutAt = entry_device.m_config.cutFrom;
        m_cutTo = entry_device.m_config.cutUpTo;

        // Transponderstruktur lesen
        m_transponder = entry_device.m_config.transponderAccessControl;

        // 2012_06_28(BO): Bugbereinigung in 04.02.03
        // Sitzt das bit 0x20, dann wird neu initialisiert.
        SoftwareVersion studioVersion;
        if ( useAESFile )
        {
            studioVersion = file.m_file_header.studioVersion;
        }
        else
        {
            studioVersion = fileXML.studioVersion();
        }

        if ( studioVersion == "04.02.03" && (m_transponder.tpMode1 & 0x20) != 0 )
        {
            // Transponder
            memset( &m_transponder, 0, sizeof( StructTransponder ) );
            m_transponder.tpSegment1 = 255; // Kein
            m_transponder.tpSegment2 = 255;
            m_transponder.tpSegment3 = 255;
        }

        // Erweiterten Struktur lesen
        StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;

        // 2016_11_17(BO): Flag wird ab der Version 04.03.09 als gesetzt vorgegeben.
        // Es war in ab der Version 04.03.07 im Standard nicht gesetzt.
        if ( studioVersion < "04.03.09" && ((entry_device.m_extended.readerPollingFlags & 0x02) != 0x02) )
        {
            entry_device.m_extended.readerPollingFlags |= 0x02;
        }

        unsigned int Unterpunkte;
        unsigned int pEK{};
        tinyxml2::XMLElement *oldParent = fileXML.parentNode();

        // Strukturen und Eingabeketten fuer Zutrittskontrolle lesen.

        if ( (extBegin.offsetAccessControlTablesV2 != 0) || !useAESFile )
        {
            if ( studioVersion < "04.03.12" )
            {
                // Änderungen der Strukturdaten, wie z. B. Feldtypen, überschreiben die aktuell hinterlegten Listendefinitionen.
                setDefaultTable();
            }
            else
            {
                // Eintraegen der ZK-Tabellen lesen.
                if ( useAESFile )
                {
                    Unterpunkte = *(unsigned int *)&file.m_flashExtended[extBegin.offsetAccessControlTablesV2];
                }
                else
                {
                    // Finding the node for this structure and update parent.
                    fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::AccessControl ) );
                    Unterpunkte = fileXML.numberOfChildItems( FileXML::ContentType::RecordDescription );
                    fileXML.combineValid( (Unterpunkte > 0) && (Unterpunkte <= 8) );
                }

                if ( Unterpunkte == 0 )
                {
                    fileXML.setParentNode( oldParent );
                    return pFlash;
                }

                // Es koennen nur 8 ZK-Tabellen verarbeitet werden.
                if ( Unterpunkte > 8 )
                {
                    Unterpunkte = 8;
                }

                for ( i = 0; i < (int)Unterpunkte; ++i )
                {
                    // Pointer des Eintrags berechnen
                    if ( useAESFile )
                    {
                        pEK = extBegin.offsetAccessControlTablesV2 + sizeof( unsigned int ) + i * sizeof( unsigned int );
                        pEK = *(unsigned int *)&file.m_flashExtended[pEK];
                        pEK -= FLASH_KONFIG_START; // Absoluter Speicheroffset.
                    }

                    EntryTable *pNeu = new EntryTable( this );
                    pNeu->setDefault( QT_TRANSLATE_NOOP_UTF8( "EntryAccessControl2", "Liste" ) );

                    // Index der Tabelle eintragen (Wird durch gekuerzte Struktur nicht mit abgelegt)
                    pNeu->m_table.number = i;

                    // Extended fuer Flash ausgeben!!
                    if ( useAESFile )
                    {
                        unsigned char *flashMem = file.m_flash;
                        file.m_flash = &file.m_flashExtended[extBegin.offsetAccessControlTablesV2];

                        pNeu->readDocument( pEK, doc );

                        file.m_flash = flashMem;
                    }
                    else
                    {
                        pNeu->readDocument( i, doc );
                    }
                    append( pNeu );

                    // Bei ZK-Erweiterung wird Action im Standard ausgeblended.
                    if ( Unterpunkte == 8 && i == 6 )
                    {
                        pNeu->setVisible( FALSE );
                    }
                }

                // Fehlt die Action2-Tabelle in der ZK-Erweiterung?
                if ( Unterpunkte == 7 )
                {
                    setDefaultAction2Table(); // anfuegen
                }

            }
        }

        // Eintraegen der Eingabeketten
        if ( (extBegin.offsetAccessControlEventChainsV2 != 0) || !useAESFile )
        {
            // Eintraegen der ZK-Tabellen lesen.
            if ( useAESFile )
            {
                Unterpunkte = *(unsigned int *)&file.m_flashExtended[extBegin.offsetAccessControlEventChainsV2];
            }
            else
            {
                // Finding the node for this structure and update parent.
                fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::Access ) );
                Unterpunkte = fileXML.numberOfChildItems( FileXML::ContentType::Operating );
                fileXML.combineValid( Unterpunkte <= 1 );
            }

            if ( Unterpunkte == 0 )
            {
                fileXML.setParentNode( oldParent );
                return pFlash;
            }

            // Es kann nur maximal eine EK Verarbeitet werden.
            if ( Unterpunkte > 1 )
            {
                Unterpunkte = 1;
            }

            for ( i = 0; i < (int)Unterpunkte; ++i )
            {
                // Pointer des Eintrags berechnen
                if ( useAESFile )
                {
                    pEK = extBegin.offsetAccessControlEventChainsV2 + sizeof( unsigned int ) + i * sizeof( unsigned int );
                    pEK = *(unsigned int *)&file.m_flashExtended[pEK];
                    pEK -= FLASH_KONFIG_START; // Absoluter Speicheroffset.
                }

                EntryInputChain *pNeu = new EntryInputChain( this );
                pNeu->setDefault();

                if ( useAESFile )
                {
                    // Extended fuer Flash ausgeben!!
                    unsigned char *flashMem = file.m_flash;
                    file.m_flash = &file.m_flashExtended[extBegin.offsetAccessControlEventChainsV2];

                    pNeu->readDocument( pEK, doc );

                    file.m_flash = flashMem;
                }
                else
                {
                    pNeu->readDocument( i, doc );
                }
                append( pNeu );
            }
        }

        if ( !useAESFile )
        {
            // Reset parent node because the end of the structure is reached.
            fileXML.setParentNode( oldParent );
        }
    }

    return pFlash;
}

///////////////////////////////////
// Daten des Eintrags

void EntryAccessControl2::setDefault()
{
    m_iconIndex = ICON_ACCESS_CONTROL;
    m_iconIndexSelected = ICON_ACCESS_CONTROL_SEL;

    setName( QT_TRANSLATE_NOOP_UTF8( "EntryAccessControl2", "Zutrittskontrolle 2" ) );

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_ACCESS_CONTROL2_0 );
    m_pageIndexes.push_back( PAGE_ACCESS_CONTROL2_1 );

    // Eintrag soll nicht im Baum angezeigt werden.
    setVisible( FALSE );

    m_accessControlFlag = 0x01;
    m_entranceMainMenu = 0;
    m_cutAt = 0;
    m_cutTo = 0;
    m_enableCut = FALSE;

    //	m_ucDoorAlarmTime = 10;
    //	m_ucDoorMonitoring = FALSE;
    //	m_ucDoorSignalTime = 3;

    // 2016.01.14.BOt: Bis Version 04.03.07 war TS (Wert 0) Default.
    m_useExternalReader = 2; // PHG-Crypt

    // Transponder
    memset( &m_transponder, 0, sizeof( StructTransponder ) );
    m_transponder.tpSegment1 = 255; // Kein
    m_transponder.tpSegment2 = 255;
    m_transponder.tpSegment3 = 255;

    setInsertEnabled( FALSE );
    setDeleteEnabled( FALSE );
}

void EntryAccessControl2::setDefaultTable()
{
    // Tabelle: Reader
    EntryTable *pTable = new EntryTable( this );
    pTable->setDefault( "Reader" );
    append( pTable );
    // Id	Nummer	4	Int	Primärschlüssel.
    EntryTableField *pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "Id", TF_NUM, 5 );
    pTable->append( pTableField );
    // ZM	Nummer	4	Int	Gibt die zugehörige Master ID an.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "ZM", TF_NUM, 5 );
    pTable->append( pTableField );
    // TM	Nummer	3	Int	Gibt die Busnummer des Leser an.(intern 010..160, extern 011..161)
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "TM", TF_NUM, 4 );
    pTable->append( pTableField );
    // RefLocation	Nummer	4	Int	Definiert die Raum ID, zu dem der ZL Zutritt verschafft.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "RefLocation", TF_NUM, 5 );
    pTable->append( pTableField );
    // RefAction	Nummer	4	Int	Aktion, die bei gültigem Zutritt ausgelöst wird.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "RefAction", TF_NUM, 5 );
    pTable->append( pTableField );
    // PinGeneral	Nummer	8	Int	Mit der hier hinterlegten PIN erhält man ohne AW Zutritt.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "PinGeneral", TF_NUM, 9 );
    pTable->append( pTableField );


    // Tabelle: Identification
    pTable = new EntryTable( this );
    pTable->setDefault( "Identification" );
    append( pTable );
    // Id	Text	20	ASCII	Gibt die zu prüfende Ausweisnummer an.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "Id", TF_A_NUM, 21 );
    pTable->append( pTableField );
    // Group	Nummer	4	Int	Gibt die BG an, in welcher der Ausweis mit anderen enthalten ist.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "Group", TF_NUM, 5 );
    pTable->append( pTableField );
    // Pin	Nummer	8	Int	Definiert die PIN. Falls 0 dann keine Abfrage
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "Pin", TF_NUM, 9 );
    pTable->append( pTableField );
    // Menace	Nummer	4	Int	Definiert die Bedrohungsziffern (Erweiterungsziffern zum PIN). Falls 0 dann keine Abfrage
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "Menace", TF_NUM, 5 );
    pTable->append( pTableField );
    // ActiveStart	Text	10	Date	Sperrbereich des AW (Ab wann ist der AW gültig).
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "ActiveStart", TF_A_NUM, 11 );
    pTable->append( pTableField );
    // DateEnd	Text	10	Date	Sperrbereich des AW (Bis wann ist der AW gültig).
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "ActiveEnd", TF_A_NUM, 11 );
    pTable->append( pTableField );
    // ActiveGeneral	Nummer	1	Int	Globale Sperrung des AW.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "ActiveGeneral", TF_NUM, 2 );
    pTable->append( pTableField );


    // Tabelle: Location
    pTable = new EntryTable( this );
    pTable->setDefault( "Location" );
    append( pTable );
    // Id	Nummer	4	Int	Gibt den Raum an.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "Id", TF_NUM, 5 );
    pTable->append( pTableField );
    // RefGroup	Nummer	4	Int	Die BG für den Raum.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "RefGroup", TF_NUM, 5 );
    pTable->append( pTableField );
    // RefTime	Nummer	4	Int	Das entsprechende Zeitmodell.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "RefTime", TF_NUM, 5 );
    pTable->append( pTableField );
    // RefTimePin		Nummer	4	Int	Zeitmodell, wann eine PIN eingegeben werden muß.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "RefTimeNoPin", TF_NUM, 5 );
    pTable->append( pTableField );


    // Tabelle: Time
    pTable = new EntryTable( this );
    pTable->setDefault( "Time" );
    append( pTable );
    // Id	Nummer	4	Int	ID des Zeitmodell.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "Id", TF_NUM, 5 );
    pTable->append( pTableField );
    // Weekdays	Nummer	7	Weekday	Definiert die Wochentage, an den das Zeitmodell gilt. (Format: Max. 7 Stellen 1 - 7 z.B. 134567= Montag, Mittwoch bis Sonntag)
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "Weekdays", TF_NUM, 8 );
    pTable->append( pTableField );
    // TimeStart	Text	5	Time	Von wann gültig. (Format 24h: HH:MM)
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "TimeStart", TF_A_NUM, 6 );
    pTable->append( pTableField );
    // TimeEnd	Text	5	Time	Bis wann gültig.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "TimeEnd", TF_A_NUM, 6 );
    pTable->append( pTableField );


    // Tabelle: Holiday
    pTable = new EntryTable( this );
    pTable->setDefault( "Holiday" );
    append( pTable );
    // Day	Text	10	Date	Datum des Sperrtages. (Format: YYYY-MM-DD)
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "Day", TF_A_NUM, 11 );
    pTable->append( pTableField );
    // RefGroup	Nummer	4	Int	Gibt die BG für den Sperrtag an.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "RefGroup", TF_NUM, 5 );
    pTable->append( pTableField );
    // RefTime	Nummer	4	Int	Gibt das abweichende Zeitmodell an.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "RefTime", TF_NUM, 5 );
    pTable->append( pTableField );


    // Tabelle: Event
    pTable = new EntryTable( this );
    pTable->setDefault( "Event" );
    append( pTable );
    // RefReader	Nummer	4	Int	Master, an dem das TM ist.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "RefReader", TF_NUM, 5 );
    pTable->append( pTableField );
    // PortIn	Nummer	1	Int	Gibt den DI am TM an.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "PortIn", TF_A_NUM, 2 );
    pTable->append( pTableField );
    // RefAction	Nummer	4	Int	ID der Aktion, die ausgeführt werden soll.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "RefAction", TF_NUM, 5 );
    pTable->append( pTableField );
    // RefTime	Nummer	4	Int	Zeitmodell, wann der DI gültig ist.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "RefTime", TF_NUM, 5 );
    pTable->append( pTableField );


    // Tabelle: Action
    pTable = new EntryTable( this );
    pTable->setDefault( "Action" );
    append( pTable );
    // Id	Nummer	4	Int	ID der Aktion.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "Id", TF_NUM, 5 );
    pTable->append( pTableField );
    // RefReader	Nummer	4	Int	Master, an dem das Ereignis ausgelöst werden soll.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "RefReader", TF_NUM, 5 );
    pTable->append( pTableField );
    // PortOut	Nummer	1	Int	Welcher Ausgang wird geschaltet.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "PortOut", TF_A_NUM, 2 );
    pTable->append( pTableField );
    // Elapse	Nummer	3	Int	Dauer des Schaltens in Sekunden, 0 -> dauernd
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "Elapse", TF_NUM, 4 );
    pTable->append( pTableField );
    // RefTime	Nummer	4	Int	Zeitmodell, wann das Relais auslösen soll.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "RefTime", TF_NUM, 5 );
    pTable->append( pTableField );

    // Tabelle Action2
    setDefaultAction2Table();
}

void EntryAccessControl2::setDefaultAction2Table()
{
    // Tabelle: Action2
    EntryTable *pTable = new EntryTable( this );
    pTable->setDefault( "Action2" );
    append( pTable );
    // Id	Nummer	4	Int	ID der Aktion.
    EntryTableField *pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "Id", TF_NUM, 5 );
    pTable->append( pTableField );
    // RefGroup	Nummer	4	Int	Gibt die BG fuer die Aktion an.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "RefGroup", TF_NUM, 5 );
    pTable->append( pTableField );
    // RefTime	Nummer	4	Int	Zeitmodell, wann das Relais ausloesen soll.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "RefTime", TF_NUM, 5 );
    pTable->append( pTableField );
    // RefReaderRelais	Nummer	4	Int	Master, an dem das Ereignis ausgeloest werden soll.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "RefReaderRelais", TF_NUM, 5 );
    pTable->append( pTableField );
    // PortOut	Nummer	1	Int	Welcher Ausgang wird geschaltet.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "PortOut", TF_A_NUM, 2 );
    pTable->append( pTableField );
    // Elapse	Nummer	6	Int	Dauer des Schaltens in Sekunden, 0 -> dauernd
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "Elapse", TF_A_NUM, 7 );
    pTable->append( pTableField );
    // RefReaderLED	Nummer	4	Int	Master, an dem das Ereignis ausgelöst werden soll.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "RefReaderLED", TF_NUM, 5 );
    pTable->append( pTableField );
    // RefTimeRelais	Nummer	4	Int	Zeitmodell, wann das Relais auslösen soll.
    pTableField = new EntryTableField( pTable );
    pTableField->setDefault( "RefTimeRelais", TF_NUM, 5 );
    pTable->append( pTableField );
    // Zusatzinfo in Tabelle eintragen
    pTable->m_table.count = pTable->size();
}

void EntryAccessControl2::setDefaultInputChain()
{
    // Eingabekette Zutritt
    EntryInputChain *pInputChain = new EntryInputChain( this );
    pInputChain->setDefault( QT_TRANSLATE_NOOP_UTF8( "EntryAccessControl2", "Zutritt" ), TRUE );
    append( pInputChain );
}



// ########################################################
// Memeber: EntryEnergieSchemes (Energieverwaltung)

///////////////////////////////////
// Childs

EntryBase *EntryEnergieSchemes::insertEntry( EntryType typeID, int pos )
{
    (void)typeID;
    (void)pos;

    return nullptr;
}

EntryBase *EntryEnergieSchemes::cloneEntry()
{
    // Dieser Eintrag unterstützt derzeit cloneEntry nicht.
    return nullptr;
}

///////////////////////////////////
// Serialisierung der Eintragsdaten

uint32_t EntryEnergieSchemes::writeDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    EntryDevice &device = root();
    // Wenn die Energieverwaltung nicht aktiviert ist, dann vor Speicherung die Standardprofile sicherstellen.
    // Abhängig von den Setupeinstellungen können sich ggf. notwendige Anpassungen ergeben. Zum Beispiel wirkt sich der 
    // eingestellte Transponderlesertyp auf die Energieverwaltung aus.
    if ( (device.m_extended.powerActiv & PAF_ENABLED) != PAF_ENABLED )
        setDefaultEnergieScheme();

    // Die Extended - Struktur liegt als erste im erweiterten Flash.
    StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;
    int idx;

    // Strukturen für die Energieverwaltung eintragen.
    extBegin.offsetEnergieSchemas = extBegin.offsetFreeScope;
    PowerSetupNode *powerSetupNode = (PowerSetupNode *)&file.m_flashExtended[extBegin.offsetEnergieSchemas];

    // Hauptstruktur für Energieverwaltung eintragen.
    *powerSetupNode = m_powerSetupNode;
    extBegin.offsetFreeScope += sizeof( PowerSetupNode );

    ////////////////////////////////////////
    // add information to xml file
    FileXML &fileXML = doc.fileXML();

    // The parent is configured outside the method and is not changed.
    fileXML.writeContent( FileXML::ContentType::PowerSetupNode, as_const( m_powerSetupNode ) );

    // Finding the node for this structure and update parent.
    auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::PowerSetupNode ) );

    // Jede einzelne Struktur pro aktivem Energieschema eintragen.
    powerSetupNode->ptrProfileArray = sizeof( PowerSetupNode );
    PowerSetup *ptrProfileArray = (PowerSetup *)&file.m_flashExtended[extBegin.offsetEnergieSchemas + powerSetupNode->ptrProfileArray];
    for ( idx = 0; idx < powerSetupNode->countProfile && idx < POWER_PROFIL_COUNT; ++idx )
    {
        // Struktur eintragen
        ptrProfileArray[idx] = m_powerSetup[idx];

        // The parent is configured outside the method and is not changed.
        fileXML.writeContent( FileXML::ContentType::PowerSetup, as_const( m_powerSetup[idx] ) );
    }
    extBegin.offsetFreeScope += powerSetupNode->countProfile * sizeof( PowerSetup );


    if ( powerSetupNode->countProfile > 1 )
    {
        int idxEvent = 0;
        // Jede einzelne Struktur pro aktivem Ereignis eintragen.
        powerSetupNode->ptrEventArray = powerSetupNode->ptrProfileArray + powerSetupNode->countProfile * sizeof( PowerSetup );
        PowerEvent *ptrEventArray = (PowerEvent *)&file.m_flashExtended[extBegin.offsetEnergieSchemas + powerSetupNode->ptrEventArray];
        // Zuerst alle aktiven Systemtimer eintragen
        // Die ersten Timer sind die TypA Timer, die restlichen TypB.
        for ( idx = 0; idx < POWER_EVENT_TIMER_A_MAX; ++idx )
        {
            if ( m_powerEventTimerA[idx].profile == 0xFF ) continue; // Nicht aktive übergehen.
            // Struktur eintragen
            ptrEventArray[idxEvent] = m_powerEventTimerA[idx];

            // The parent is configured outside the method and is not changed.
            fileXML.writeContent( FileXML::ContentType::PowerEvent, as_const( m_powerEventTimerA[idx] ) );

            idxEvent++;
        }
        // Alle weiteren aktiven Ereignisse eintragen
        for ( idx = 0; idx < POWER_EVENT_MAX; ++idx )
        {
            if ( m_powerEvent[idx].profile == 0xFF ) continue; // Nicht aktive übergehen.
            // Struktur eintragen
            ptrEventArray[idxEvent] = m_powerEvent[idx];

            // The parent is configured outside the method and is not changed.
            fileXML.writeContent( FileXML::ContentType::PowerEvent, as_const( m_powerEvent[idx] ) );

            idxEvent++;
        }

        // Anzahl der Ereignisse eintragen
        powerSetupNode->countEvents = idxEvent;
    }
    else
    {
        // Kein Umschalten zwischen Schemen moegleich somit keine Ereignisse notwendig.
        powerSetupNode->ptrEventArray = 0;
        powerSetupNode->countEvents = 0;
    }

    extBegin.offsetFreeScope += powerSetupNode->countEvents * sizeof( PowerEvent );

    // Memotext eintragen
    extBegin.offsetEnergieMemo = extBegin.offsetFreeScope;
    strcpy( (char *)&file.m_flashExtended[extBegin.offsetEnergieMemo], m_powerMemo.memo );
    extBegin.offsetFreeScope += (unsigned int)strlen( (char *)&file.m_flashExtended[extBegin.offsetEnergieMemo] ) + 1;

    // The parent is configured outside the method and is not changed.
    fileXML.writeContent( FileXML::ContentType::PowerMemo, as_const( m_powerMemo ) );

    // Extended Struktur uebernehmen
    *(StructExtendedFlash *)file.m_flashExtended = extBegin;

    // Reset parent node because the end of the structure is reached.
    fileXML.setParentNode( oldParent );

    // An update of the main node of the energy schema configuration is necessary, since the number of active
    // entries is only determined during the write process.
    tinyxml2::XMLElement *lastXmlElement{};
    fileXML.updateContent( FileXML::ContentType::PowerSetupNode, as_const( *powerSetupNode ), lastXmlElement );

    return pFlash;
}

uint32_t EntryEnergieSchemes::readDocument( uint32_t pFlash, DocumentStudio &doc )
{
    FileAES &file = doc.fileAES();
    FileXML &fileXML = doc.fileXML();
    // Die Extended - Struktur liegt als erste im erweiterten Flash.
    StructExtendedFlash extBegin = *(StructExtendedFlash *)file.m_flashExtended;
    int idx;
    EntryDevice &device = root();

    // Reset the event structures to delete any existing event definition.
    constexpr int timerSize = sizeof( m_powerEventTimerA ) / sizeof( m_powerEventTimerA[0] );
    for ( int i = 0; i < timerSize; ++i )
    {
        m_powerEventTimerA[i] = PowerEvent{};
    }
    constexpr int eventSize = sizeof( m_powerEvent ) / sizeof( m_powerEvent[0] );
    for ( int i = 0; i < eventSize; ++i )
    {
        m_powerEvent[i] = PowerEvent{};
    }

    // Flag to indicate the processing of an AES file.
    // This represents the original processing.
    bool useAESFile = (doc.currentFileType() != FileType::XML);

    // Handelt es sich um eine alte Version der Energieverwalungsstrukturen?
    if ( useAESFile )
    {
        if ( extBegin.offsetEnergieSchemas != 0 && ((PowerSetupNode *)&file.m_flashExtended[extBegin.offsetEnergieSchemas])->version == 1 )
        {
            if ( (device.m_extended.powerActiv & PAF_ENABLED) == PAF_ENABLED && pFlash == 1 )
            {
                // Es wird versucht eine veraltete Profildatei zu laden.
                return FALSE;
            }

            // Wenn der Expertenmodus aktiviert wurde, dann eine Meldung ueber den Verwurf der Energiestrukturen informieren!
            if ( (device.m_extended.powerActiv & (PAF_ENABLED | PAF_EXPERT)) == (PAF_ENABLED | PAF_EXPERT) )
            {
                //QString msg;
                //msg = QT_TRANSLATE_NOOP_UTF8( "EntryEnergieSchemes","Die im Setup verwendete Version der Energieverwaltung\nwird durch das vorliegende DatafoxStudioIV nicht mehr unterstützt!\n\n");
                //msg += QT_TRANSLATE_NOOP_UTF8( "EntryEnergieSchemes","Erst bei Abspeicherung der Datei, werden die enthaltenen Profile\nmit den vorliegenden Standardprofilen ersetzt.");
                //QMessageBox::warning( ownerWidget(), "DatafoxStudioIV", msg );

                // Aktuelle Einstellung resetten
                device.m_extended.powerActiv &= ~PAF_EXPERT;
            }

            return pFlash;
        }

        // Falls keine Daten verfügbar sind oder Version nicht gleich  2, abbruch.
        if ( extBegin.offsetEnergieSchemas == 0 || ((PowerSetupNode *)&file.m_flashExtended[extBegin.offsetEnergieSchemas])->version != 2 )
        {
            // Wenn der Expertenmodus aktiviert wurde, dann eine Meldung ueber den Verwurf der Energiestrukturen informieren!
            if ( (device.m_extended.powerActiv & (PAF_ENABLED | PAF_EXPERT)) == (PAF_ENABLED | PAF_EXPERT) )
            {
                //QString msg;
                //msg = QT_TRANSLATE_NOOP_UTF8( "EntryEnergieSchemes","Die vorliegende Version der Energieverwaltung wird\ndurch diese Version des DatafoxStudioIV nicht unterstützt!\n\n");
                //msg += QT_TRANSLATE_NOOP_UTF8( "EntryEnergieSchemes","Erst bei Abspeicherung der Datei, werden die enthaltenen Profile\nmit den vorliegenden Standardprofilen ersetzt.\n\n");
                //msg += QT_TRANSLATE_NOOP_UTF8( "EntryEnergieSchemes","Bitte verwenden Sie eine aktuellere Version des DatafoxStudioIV zur Bearbeitung des Setups.");
                //QMessageBox::warning( ownerWidget(), "DatafoxStudioIV", msg );

                // Expertenmodus entfernen
                device.m_extended.powerActiv &= ~PAF_EXPERT;
            }

            return pFlash;
        }

        // Strukturen für die Energieverwaltung lesen.
        m_powerSetupNode = *(PowerSetupNode *)&file.m_flashExtended[extBegin.offsetEnergieSchemas];

        // Jede einzelne Struktur pro aktivem Energieschema lesen.
        PowerSetup *ptrProfileArray = (PowerSetup *)&file.m_flashExtended[extBegin.offsetEnergieSchemas + m_powerSetupNode.ptrProfileArray];
        for ( idx = 0; idx < m_powerSetupNode.countProfile && idx < POWER_PROFIL_COUNT; ++idx )
        {
            // Struktur eintragen
            m_powerSetup[idx] = ptrProfileArray[idx];
        }

        int idxEventTimerA = 0, idxEvent = 0;
        // Jede einzelne Struktur pro aktivem Ereignis lesen.
        PowerEvent *ptrEventArray = (PowerEvent *)&file.m_flashExtended[extBegin.offsetEnergieSchemas + m_powerSetupNode.ptrEventArray];
        for ( idx = 0; idx < m_powerSetupNode.countEvents; ++idx )
        {
            // Struktur je nach Typ übernehmen
            if ( ptrEventArray[idx].type == POWER_EVENT_TIMER_A )
            {
                // Timerevent übernehmen
                // Für jedes mögliche Profil existiert maximal 1 Timer TypA
                // (ptrEventArray[idx].profile >= 0) is always true due to the data type uint8_t
                if ( idxEventTimerA < POWER_EVENT_TIMER_A_MAX && ptrEventArray[idx].profile < POWER_EVENT_TIMER_A_MAX )
                {
                    //m_powerEventTimerA[idxEventTimerA] = ptrEventArray[idx];
                    m_powerEventTimerA[ptrEventArray[idx].profile] = ptrEventArray[idx];
                    idxEventTimerA++;
                }
            }
            else
            {
                // Event übernehmen
                if ( idxEvent < POWER_EVENT_MAX )
                {
                    m_powerEvent[idxEvent] = ptrEventArray[idx];
                    idxEvent++;
                }
            }
        }

        // Memotext zuweisen
        if ( extBegin.offsetEnergieMemo != 0 )
        {
            strncpy( (char *)m_powerMemo.memo, (char *)&file.m_flashExtended[extBegin.offsetEnergieMemo], sizeof( m_powerMemo.memo ) - 1 );
        }
    }
    else
    {
        fileXML.combineValid( fileXML.readContent( FileXML::ContentType::PowerSetupNode, m_powerSetupNode ) );

        // Finding the node for this structure and update parent.
        auto oldParent = fileXML.setParentNode( fileXML.nodeByType( FileXML::ContentType::PowerSetupNode ) );

        uint32_t size = fileXML.numberOfChildItems( FileXML::ContentType::PowerSetup );
        // Check the number of power setup configurations defined in the XML. If the number does not match the definition of the
        // 'm_powerSetupNode.countProfile', there is an error in the XML description.
        fileXML.combineValid( size == m_powerSetupNode.countProfile );

        if ( fileXML.isValid() )
        {
            tinyxml2::XMLElement *lastItem{};
            for ( uint8_t i = 0; i < m_powerSetupNode.countProfile && i < POWER_PROFIL_COUNT; ++i )
            {
                fileXML.combineValid( fileXML.readContent( FileXML::ContentType::PowerSetup, m_powerSetup[i], i, lastItem ) );
            }
        }

        size = fileXML.numberOfChildItems( FileXML::ContentType::PowerEvent );
        // Check the number of power events defined in the XML. If the number does not match the definition of the
        // 'm_powerSetupNode.countEvents', there is an error in the XML description.
        fileXML.combineValid( size == m_powerSetupNode.countEvents );

        if ( fileXML.isValid() )
        {
            PowerEvent tmp;
            int idxEventTimerA{};
            int idxEvent{};
            tinyxml2::XMLElement *lastItem{};
            for ( uint8_t i = 0; i < m_powerSetupNode.countEvents; ++i )
            {
                fileXML.combineValid( fileXML.readContent( FileXML::ContentType::PowerEvent, tmp, i, lastItem ) );
                if ( tmp.type == POWER_EVENT_TIMER_A )
                {
                    // Timerevent übernehmen
                    // Für jedes mögliche Profil existiert maximal 1 Timer TypA
                    // (tmp.profile >= 0) is always true due to the data type uint8_t
                    if ( idxEventTimerA < POWER_EVENT_TIMER_A_MAX && tmp.profile < POWER_EVENT_TIMER_A_MAX )
                    {
                        m_powerEventTimerA[tmp.profile] = tmp;
                        idxEventTimerA++;
                    }
                }
                else
                {
                    // Event übernehmen
                    if ( idxEvent < POWER_EVENT_MAX )
                    {
                        m_powerEvent[idxEvent] = tmp;
                        idxEvent++;
                    }
                }
            }
        }

        fileXML.combineValid( fileXML.readContent( FileXML::ContentType::PowerSetup, m_powerMemo ) );

        // Reset parent node because the end of the structure is reached.
        fileXML.setParentNode( oldParent );
    }

    return pFlash;
}

///////////////////////////////////
// Daten des Eintrags

void EntryEnergieSchemes::setDefault()
{
    m_iconIndex = ICON_ENERGIE_SCHEMES;
    m_iconIndexSelected = ICON_ENERGIE_SCHEMES_SEL;

    setName( QT_TRANSLATE_NOOP_UTF8( "EntryEnergieSchemes", "Energieverwaltung" ) );

    // Zugeordnete Pages zum Eintrag.
    m_pageIndexes.clear();
    m_pageIndexes.push_back( PAGE_ENERGIE_SCHEMES_0 );
    m_pageIndexes.push_back( PAGE_ENERGIE_SCHEMES_1 );
    m_pageIndexes.push_back( PAGE_ENERGIE_SCHEMES_2 );
    m_pageIndexes.push_back( PAGE_ENERGIE_SCHEMES_3 );
    m_pageIndexes.push_back( PAGE_ENERGIE_SCHEMES_4 );
    m_pageIndexes.push_back( PAGE_ENERGIE_SCHEMES_5 );

    // Standardschema setzen
    setDefaultEnergieScheme();

    // Eintrag soll nicht im Baum angezeigt werden.
    setVisible( FALSE );

    setInsertEnabled( FALSE );
    setDeleteEnabled( FALSE );
}

void EntryEnergieSchemes::setDefaultEnergieScheme()
{
    EntryDevice &device = root();
    SetupType setupType = device.setupType();
    // Eingestellter Tranponderleser.
    unsigned char transponderType = device.m_config.transponderType & 0x3F;
    unsigned int powerActive = 0x0007FFFF;
    // Alle Stukturdaten initialisieren.
    memset( &m_powerSetupNode, 0, sizeof( PowerSetupNode ) );
    memset( m_powerSetup, 0, POWER_PROFIL_COUNT * sizeof( PowerSetup ) );
    memset( m_powerEventTimerA, 0, POWER_EVENT_TIMER_A_MAX * sizeof( PowerEvent ) );
    memset( m_powerEvent, 0, POWER_EVENT_MAX * sizeof( PowerEvent ) );

    // Ausblenden nicht benötigter Komponenten.
    switch ( device.setupType() )
    {
        case SetupType::PZE_MasterIV:                // PZE-MasterIV
            break;

        case SetupType::EVO_2_8_Pure:                // EVO 2.8 Pure / EVO 3.5 Pure
        case SetupType::EVO_4_3:                     // EVO 4.3 / EVO 4.6 FlexKey
        case SetupType::EVO_3_5_Universal:           // EVO 3.5 Universal
            powerActive &= ~PAF_USB_HOST;
            break;

        case SetupType::ZK_MasterIV:                 // ZK-MasterIV
        case SetupType::MDE_BoxIV:                   // MDE-BoxIV
        case SetupType::Mobil_MasterIV:              // Mobil-MasterIV
            powerActive &= ~PAF_DISPLAY;
            powerActive &= ~PAF_INPUT_MANUAL_FINGERPRINT;
            powerActive &= ~PAF_INPUT_MANUAL_TRANSPONDER;
            break;

        case SetupType::IO_Box:            // IO-Box
        case SetupType::KYO_Inloc:         // KYO Inloc
        case SetupType::Mobil_Box:         // Mobil-Box
        case SetupType::KYO_Oneloc:        // KYO Oneloc
        case SetupType::KYO_Cenloc:        // KYO Cenloc
            powerActive &= ~PAF_INPUT_MANUAL_FINGERPRINT;
            powerActive &= ~PAF_USB_HOST;
            break;

        case SetupType::Timeboy_TBDock:    // Dockingstation-V2
        case SetupType::Mobil_TBDock:      // Fahrzeugdatenlogger-V2
            powerActive &= ~PAF_EXTERNAL_ENERGIE_12V;
            powerActive &= ~PAF_EXTERNAL_ENERGIE_5V;
            powerActive &= ~PAF_INPUT_MANUAL_FINGERPRINT;
            powerActive &= ~PAF_INPUT_MANUAL_TRANSPONDER;
            powerActive &= ~PAF_USB_HOST;
            break;

        case SetupType::IPC_Extension: // IPC-Extensions
            powerActive &= ~PAF_EXTERNAL_ENERGIE_12V;
            powerActive &= ~PAF_EXTERNAL_ENERGIE_5V;
            powerActive &= ~PAF_USB_HOST;
            powerActive &= ~PAF_DISPLAY;
            powerActive &= ~PAF_MMC;
            break;

        case SetupType::BDE_MasterIV:               // BDE-MasterIV
            break;

        case SetupType::AE_MasterIV:                // AE-MasterIV
            break;

        case SetupType::Timeboy_MasterIV:           // TimeboyIV
        case SetupType::Timeboy_Mobil_PZE:          // Timeboy-Mobil-PZE
            powerActive &= ~PAF_EXTERNAL_ENERGIE_12V;
            powerActive &= ~PAF_EXTERNAL_ENERGIE_5V;
            powerActive &= ~PAF_INPUT_MANUAL_FINGERPRINT;
            powerActive &= ~PAF_USB_HOST;
            powerActive &= ~PAF_MMC;
            break;

        case SetupType::Exklusiv_LineIV:          // Exklusiv-Line IV
        case SetupType::Flex_MasterIV:            // Flex-MasterIV
            powerActive &= ~PAF_USB_HOST;
            break;

        default:
            BOOST_ASSERT( false );
    }

    // Hauptstruktur initialisieren
    m_powerSetupNode.version = 2; // Version ist 2, Version 1 bis zur 04.02.01.07 vom 14.02.2011 
    m_powerSetupNode.active = 0; // Aktive Schema ist 1
    m_powerSetupNode.countProfile = 2; // Anzahl der aktiven Profile
    m_powerSetupNode.countEvents = 0; // Anzahl der aktiven Ereignisse

    int idx = 0;
    // Namen der Schemen zuweisen
    for ( ; idx < 2; ++idx )
    {
        // Alle sind im Standard auf "Ein"
        m_powerSetup[idx].cpu.item = POWER_ITEM_NR_ON;
        m_powerSetup[idx].timer.item = POWER_ITEM_NR_ON;
        m_powerSetup[idx].inputs.item = POWER_ITEM_NR_ON;
        m_powerSetup[idx].deviceGlobal.item = POWER_ITEM_NR_ON;
        m_powerSetup[idx].U5V.item = POWER_ITEM_NR_ON;
        m_powerSetup[idx].U12V.item = POWER_ITEM_NR_ON;
        m_powerSetup[idx].communication.item = POWER_ITEM_NR_ON;
        m_powerSetup[idx].displayLight.item = POWER_ITEM_NR_ON;
        m_powerSetup[idx].displayText.item = POWER_ITEM_NR_ON;
        m_powerSetup[idx].fingerprint.item = POWER_ITEM_NR_ON;
        m_powerSetup[idx].transponder.item = POWER_ITEM_NR_ON;
        m_powerSetup[idx].gps.item = POWER_ITEM_NR_ON;
        m_powerSetup[idx].usbHost.item = POWER_ITEM_NR_ON;
        m_powerSetup[idx].mmc.item = POWER_ITEM_NR_ON;

        switch ( idx )
        {
            case 0: // Betrieb
                strcpy( reinterpret_cast<char *>(m_powerSetup[idx].name), QT_TRANSLATE_NOOP_UTF8( "EntryEnergieSchemes", "Betrieb" ) );

                // Aenderung des Profiles abhaengig vom gewaehlten Geraetetyp
                if ( device.hasSetupType( SetupType::Timeboy_MasterIV ) )
                {
                    m_powerSetup[idx].displayLight.item = POWER_ITEM_NR_SERVICE_TIMEOUT;
                    m_powerSetup[idx].displayLight.time = 10;
                    m_powerSetup[idx].displayLight.value = 1;
                    m_powerSetup[idx].transponder.item = POWER_ITEM_NR_CHAIN_FIELD;
                }
                else if ( device.hasSetupType( SetupType::Flex_MasterIV ) )
                {
                    m_powerSetup[idx].U5V.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].U12V.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].displayLight.item = POWER_ITEM_NR_SERVICE_TIMEOUT;
                    m_powerSetup[idx].displayLight.time = 15;
                    m_powerSetup[idx].displayLight.value = 1;
                    m_powerSetup[idx].fingerprint.item = POWER_ITEM_NR_CHAIN_FIELD;
                    m_powerSetup[idx].transponder.item = POWER_ITEM_NR_CHAIN_FIELD;
                }
                else if ( device.hasSetupType( SetupType::Timeboy_Mobil_PZE ) )
                {
                    m_powerSetup[idx].transponder.item = POWER_ITEM_NR_CHAIN_FIELD;
                }
                else if ( device.hasSetupType( { SetupType::EVO_4_3, SetupType::EVO_2_8_Pure, SetupType::EVO_3_5_Universal } ) )
                {
                    m_powerSetup[idx].displayLight.item = POWER_ITEM_NR_SERVICE_TIMEOUT;
                    m_powerSetup[idx].displayLight.time = 3600;
                    m_powerSetup[idx].displayLight.value = 80;
                }
                else if ( device.hasSetupType( SetupType::KYO_Cenloc ) )
                {
                    m_powerSetup[idx].displayLight.item = POWER_ITEM_NR_SERVICE_TIMEOUT;
                    m_powerSetup[idx].displayLight.time = 300;
                    m_powerSetup[idx].displayLight.value = 20;
                    m_powerSetup[idx].transponder.item = POWER_ITEM_NR_CHAIN_FIELD;
                }

                break;

            case 1: // Sparmodus

                strcpy( reinterpret_cast<char *>(m_powerSetup[idx].name), QT_TRANSLATE_NOOP_UTF8( "EntryEnergieSchemes", "Sparmodus" ) );

                m_powerSetup[idx].cpu.item = POWER_ITEM_NR_SLEEP;
                m_powerSetup[idx].cpu.interval = 10;

                if ( device.hasSetupType( { SetupType::EVO_4_3, SetupType::EVO_2_8_Pure, SetupType::EVO_3_5_Universal } ) )
                {
                    m_powerSetup[idx].cpu.item = POWER_ITEM_NR_ON;
                    m_powerSetup[idx].cpu.interval = 0;
                    m_powerSetup[idx].displayLight.item = POWER_ITEM_NR_SERVICE_TIMEOUT;
                    m_powerSetup[idx].displayLight.time = 10;
                    m_powerSetup[idx].displayLight.value = 50;
                    m_powerSetup[idx].fingerprint.item = POWER_ITEM_NR_CHAIN_FIELD;
                    m_powerSetup[idx].transponder.item = POWER_ITEM_NR_CHAIN_FIELD;
                    m_powerSetup[idx].transponder.time = 4;
                    m_powerSetup[idx].transponder.interval = 0;
                }

                if ( device.hasSetupType( { SetupType::PZE_MasterIV, SetupType::AE_MasterIV } ) )
                {
                    m_powerSetup[idx].cpu.item = POWER_ITEM_NR_ON;
                    m_powerSetup[idx].cpu.interval = 0;
                    m_powerSetup[idx].displayLight.item = POWER_ITEM_NR_SERVICE_TIMEOUT;
                    m_powerSetup[idx].displayLight.time = 3;
                    m_powerSetup[idx].displayLight.value = 1;
                    m_powerSetup[idx].fingerprint.item = POWER_ITEM_NR_CHAIN_FIELD;
                    m_powerSetup[idx].transponder.item = POWER_ITEM_NR_CHAIN_FIELD;
                    m_powerSetup[idx].transponder.time = 4;
                    m_powerSetup[idx].transponder.interval = 0;
                }

                if ( device.hasSetupType( SetupType::IPC_Extension ) )
                {
                    m_powerSetup[idx].timer.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].U5V.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].U12V.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].gps.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].fingerprint.item = POWER_ITEM_NR_CHAIN_FIELD;
                    m_powerSetup[idx].transponder.item = POWER_ITEM_NR_CHAIN_FIELD;
                }

                if ( device.hasSetupType( { SetupType::Timeboy_MasterIV, SetupType::Timeboy_Mobil_PZE, SetupType::Flex_MasterIV } ) )
                {
                    m_powerSetup[idx].gps.item = POWER_ITEM_NR_GPS_CYCLE;
                    m_powerSetup[idx].gps.interval = 10800;
                    m_powerSetup[idx].gps.time = 180;
                }

                if ( device.hasSetupType( { SetupType::Mobil_MasterIV, SetupType::Mobil_Box, SetupType::Mobil_TBDock } ) )
                {
                    m_powerSetup[idx].U5V.item = POWER_ITEM_NR_TOFF;
                    m_powerSetup[idx].U5V.value = 1;
                    m_powerSetup[idx].U5V.time = 36000;
                    m_powerSetup[idx].U12V.item = POWER_ITEM_NR_TOFF;
                    m_powerSetup[idx].U12V.value = 1;
                    m_powerSetup[idx].U12V.time = 36000;
                    m_powerSetup[idx].gps.item = POWER_ITEM_NR_GPS_CYCLE;
                    m_powerSetup[idx].gps.interval = 10800;
                    m_powerSetup[idx].gps.time = 300;
                }

                if ( device.hasSetupType( SetupType::Mobil_TBDock ) )
                {
                    m_powerSetup[idx].timer.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].displayLight.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].displayText.item = POWER_ITEM_NR_OFF;
                }

                if ( device.hasSetupType( SetupType::Flex_MasterIV ) )
                {
                    m_powerSetup[idx].timer.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].U5V.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].U12V.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].displayLight.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].displayText.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].fingerprint.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].transponder.item = POWER_ITEM_NR_OFF;
                }

                if ( device.hasSetupType( SetupType::KYO_Inloc ) )
                {
                    m_powerSetup[idx].timer.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].U5V.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].U12V.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].displayLight.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].displayText.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].gps.item = POWER_ITEM_NR_OFF;
                }

                if ( device.hasSetupType( SetupType::KYO_Cenloc ) )
                {
                    m_powerSetup[idx].timer.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].U5V.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].U12V.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].displayLight.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].displayText.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].gps.item = POWER_ITEM_NR_OFF;
                }

                if ( device.hasSetupType( SetupType::ZK_MasterIV ) )
                {
                    m_powerSetup[idx].timer.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].U5V.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].U12V.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].gps.item = POWER_ITEM_NR_OFF;
                }

                if ( device.hasSetupType( SetupType::Mobil_Box ) )
                {
                    m_powerSetup[idx].displayLight.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].displayText.item = POWER_ITEM_NR_OFF;
                }

                if ( device.hasSetupType( { SetupType::Timeboy_MasterIV, SetupType::Timeboy_Mobil_PZE } ) )
                {
                    m_powerSetup[idx].timer.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].inputs.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].displayLight.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].displayText.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].transponder.item = POWER_ITEM_NR_OFF;
                }

                if ( device.hasSetupType( SetupType::IO_Box ) )
                {
                    m_powerSetup[idx].U5V.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].U12V.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].displayLight.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].displayText.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].gps.item = POWER_ITEM_NR_OFF;
                }

                if ( device.hasSetupType( SetupType::Timeboy_TBDock ) )
                {
                    m_powerSetup[idx].timer.item = POWER_ITEM_NR_OFF;
                    m_powerSetup[idx].U5V.item = POWER_ITEM_NR_TOFF;
                    m_powerSetup[idx].U5V.value = 1;
                    m_powerSetup[idx].U5V.time = 36000;
                    m_powerSetup[idx].U12V.item = POWER_ITEM_NR_TOFF;
                    m_powerSetup[idx].U12V.value = 1;
                    m_powerSetup[idx].U12V.time = 36000;
                    m_powerSetup[idx].gps.item = POWER_ITEM_NR_GPS_CYCLE;
                    m_powerSetup[idx].gps.interval = 10800;
                    m_powerSetup[idx].gps.time = 300;
                }

                break;
        }

        // 2014_06_18 (BO) Version 04.03.03.04
        // Im Profil für den Betrieb, muss bei Verwendung eines Legic-Lesers die Einstellung für Transponder von Standard "Feldfunktion" auf "Ein" umgestellt werden.
        // Der Verbaute Lesertyp bei den Stationären Geräten benötigt 5 Sekunden um Betriebsbereit zu sein, somit ist diese Standardeinstellung
        // bei diesem Transponderleser nicht sinnvoll.
        if ( (powerActive & PAF_INPUT_MANUAL_TRANSPONDER)
             && (m_powerSetup[idx].transponder.item == POWER_ITEM_NR_CHAIN_FIELD)
             && (transponderType == TRANSPONDER_LEGIC || transponderType == TRANSPONDER_LEGIC_ADVANT)
             && !device.hasSetupType( { SetupType::Timeboy_MasterIV, SetupType::Timeboy_Mobil_PZE } ) )
        {
            m_powerSetup[idx].transponder.item = POWER_ITEM_NR_ON;
            m_powerSetup[idx].transponder.time = 0;
            m_powerSetup[idx].transponder.interval = 0;
        }
    }

    idx = 0;
    // Alle Ereignisse Deaktivieren
    for ( ; idx < POWER_EVENT_TIMER_A_MAX; ++idx )
    {
        m_powerEventTimerA[idx].profile = 0xFF;
    }

    idx = 0;
    // Alle Ereignisse Deaktivieren
    for ( ; idx < POWER_EVENT_MAX; ++idx )
    {
        m_powerEvent[idx].profile = 0xFF;
    }

    // Bei PZE und AE wird bei USV-Betrieb auf Sparmodus und Netzbetrieb auf Betrieb geschaltet.
    if ( device.hasSetupType( { SetupType::PZE_MasterIV, SetupType::AE_MasterIV, SetupType::EVO_4_3, SetupType::EVO_2_8_Pure, SetupType::EVO_3_5_Universal } ) )
    {
        m_powerEvent[0].type = POWER_EVENT_NR_BATTERY;
        m_powerEvent[0].profile = 0x01;

        m_powerEvent[1].type = POWER_EVENT_NR_LINE_IN;
        m_powerEvent[1].profile = 0x00;
    }

    // Bei Timeboy und Flex-Master werden die Ereignisse fuer Ein-/Ausschalter belegt.
    if ( device.hasSetupType( { SetupType::Timeboy_MasterIV, SetupType::Timeboy_Mobil_PZE, SetupType::Flex_MasterIV } ) )
    {
        m_powerEvent[0].type = POWER_EVENT_NR_KEY_TIMEOUT_1;
        m_powerEvent[0].profile = 0x01;
        m_powerEvent[0].time = 45;

        m_powerEvent[1].type = POWER_EVENT_NR_SWITCH_ON;
        m_powerEvent[1].profile = 0x00;

        m_powerEvent[2].type = POWER_EVENT_NR_SWITCH_OFF;
        m_powerEvent[2].profile = 0x01;
    }

    if ( device.hasSetupType( { SetupType::Mobil_MasterIV, SetupType::Mobil_Box, SetupType::Mobil_TBDock } ) )
    {
        m_powerEvent[0].type = POWER_EVENT_NR_MDE_START;
        m_powerEvent[0].profile = 0;
        m_powerEvent[0].channel = 1;
        m_powerEvent[0].time = 0;

        m_powerEvent[1].type = POWER_EVENT_NR_MDE_STOP;
        m_powerEvent[1].profile = 1;
        m_powerEvent[1].channel = 1;
        m_powerEvent[1].time = 300;
    }

    if ( device.hasSetupType( SetupType::Timeboy_TBDock ) )
    {
        m_powerEvent[0].type = POWER_EVENT_NR_TB_PLUGIN;
        m_powerEvent[0].profile = 0;
        //m_powerEvent[0].channel = 1;
        //m_powerEvent[0].time    = 0;

        m_powerEvent[1].type = POWER_EVENT_NR_TB_PLUGOUT;
        m_powerEvent[1].profile = 1;
        //m_powerEvent[1].channel = 1;
        //m_powerEvent[1].time    = 300;
    }


    // Default Memotext aufbauen
    time_t tm_time;
    struct tm *ptm;
    tm_time = time( nullptr );
    ptm = localtime( &tm_time );

    memset( m_powerMemo.memo, 0, sizeof( m_powerMemo.memo ) );
    sprintf( m_powerMemo.memo, "Angelegt am %02i%02i_%02i_%02i %02i:%02i:%02i",
        (ptm->tm_year + 1900) / 100, (ptm->tm_year + 1900) % 100, ptm->tm_mon + 1, ptm->tm_mday,
             ptm->tm_hour, ptm->tm_min, ptm->tm_sec );

         // Werte in den Geraeteeintrag uebernehmen
    device.m_extended.powerActiv = powerActive;
}

void EntryEnergieSchemes::verifyEnergieSchemes()
{
    EntryDevice &device = root();
    unsigned int powerActive = (device.m_extended.powerActiv & (PAF_ENABLED | PAF_EXPERT)); // Nur Aktiv und Expert-bit uebernehmen
    powerActive |= 0x000FFFFF; // Alle bekannten Optionen aktiv.

    switch ( device.setupType() )
    {
        case SetupType::PZE_MasterIV:               // PZE-MasterIV
            break;

        case SetupType::EVO_2_8_Pure:      // EVO 2.8 Pure / EVO 3.5 Pure
        case SetupType::EVO_4_3:           // EVO 4.3 / EVO 4.6 FlexKey
        case SetupType::EVO_3_5_Universal: // EVO 3.5 Universal
            powerActive &= ~PAF_USB_HOST;
            break;

        case SetupType::ZK_MasterIV:                // ZK-MasterIV
        case SetupType::MDE_BoxIV:               // MDE-BoxIV
        case SetupType::Mobil_MasterIV:             // Mobil-MasterIV
            powerActive &= ~PAF_DISPLAY;
            powerActive &= ~PAF_INPUT_MANUAL_FINGERPRINT;
            powerActive &= ~PAF_INPUT_MANUAL_TRANSPONDER;
            break;

        case SetupType::IO_Box:            // IO-Box
        case SetupType::KYO_Inloc:         // ZK-Box
        case SetupType::Mobil_Box:         // Mobil-Box
        case SetupType::KYO_Oneloc:        // KYO-Oneloc
        case SetupType::KYO_Cenloc:        // KYO-Cenloc
            powerActive &= ~PAF_INPUT_MANUAL_FINGERPRINT;
            powerActive &= ~PAF_USB_HOST;
            break;

        case SetupType::Timeboy_TBDock:    // Dockingstation-V2
        case SetupType::Mobil_TBDock:      // Fahrzeugdatenlogger-V2
            powerActive &= ~PAF_EXTERNAL_ENERGIE_12V;
            powerActive &= ~PAF_EXTERNAL_ENERGIE_5V;
            powerActive &= ~PAF_INPUT_MANUAL_FINGERPRINT;
            powerActive &= ~PAF_INPUT_MANUAL_TRANSPONDER;
            powerActive &= ~PAF_USB_HOST;
            break;

        case SetupType::IPC_Extension:    // IPC-Extensions
            powerActive &= ~PAF_EXTERNAL_ENERGIE_12V;
            powerActive &= ~PAF_EXTERNAL_ENERGIE_5V;
            powerActive &= ~PAF_USB_HOST;
            powerActive &= ~PAF_DISPLAY;
            powerActive &= ~PAF_MMC;
            break;

        case SetupType::BDE_MasterIV:               // BDE-MasterIV
            break;

        case SetupType::AE_MasterIV:                // AE-MasterIV
            break;

        case SetupType::Timeboy_MasterIV:           // TimeboyIV
        case SetupType::Timeboy_Mobil_PZE: // Timeboy-Mobil-PZE
            powerActive &= ~PAF_EXTERNAL_ENERGIE_12V;
            powerActive &= ~PAF_EXTERNAL_ENERGIE_5V;
            powerActive &= ~PAF_INPUT_MANUAL_FINGERPRINT;
            powerActive &= ~PAF_USB_HOST;
            powerActive &= ~PAF_MMC;
            break;

        case SetupType::Exklusiv_LineIV:          // Exklusiv-Line IV
        case SetupType::Flex_MasterIV:              // Flex-MasterIV
            powerActive &= ~PAF_USB_HOST;
            break;

        default:
            BOOST_ASSERT( false );
    }

    // Sicherstellen, dass zwei Profile aktiv sind.

    int idx = 0;
    // Sicherstellen, dass die nicht nutzbaren Module auf Ein stehen.
    for ( ; idx < POWER_PROFIL_COUNT; ++idx )
    {
        if ( (powerActive & PAF_INPUT_SIGNAL_TIMER) == 0 && m_powerSetup[idx].timer.item != POWER_ITEM_NR_OFF )
        {
            memset( &m_powerSetup[idx].timer, 0, sizeof( PowerModule ) );
            m_powerSetup[idx].timer.item = POWER_ITEM_NR_ON;
        }

        if ( (powerActive & PAF_INPUT_SIGNAL_DIGITAL) == 0 && m_powerSetup[idx].inputs.item != POWER_ITEM_NR_OFF )
        {
            memset( &m_powerSetup[idx].inputs, 0, sizeof( PowerModule ) );
            m_powerSetup[idx].inputs.item = POWER_ITEM_NR_ON;
        }

        if ( (powerActive & PAF_EXTERNAL_ENERGIE_5V) == 0 && m_powerSetup[idx].U5V.item != POWER_ITEM_NR_OFF )
        {
            memset( &m_powerSetup[idx].U5V, 0, sizeof( PowerModule ) );
            m_powerSetup[idx].U5V.item = POWER_ITEM_NR_ON;
        }

        if ( (powerActive & PAF_EXTERNAL_ENERGIE_12V) == 0 && m_powerSetup[idx].U12V.item != POWER_ITEM_NR_OFF )
        {
            memset( &m_powerSetup[idx].U12V, 0, sizeof( PowerModule ) );
            m_powerSetup[idx].U12V.item = POWER_ITEM_NR_ON;
        }

        if ( (powerActive & PAF_DISPLAY) == 0 && m_powerSetup[idx].displayLight.item != POWER_ITEM_NR_OFF )
        {
            memset( &m_powerSetup[idx].displayLight, 0, sizeof( PowerModule ) );
            m_powerSetup[idx].displayLight.item = POWER_ITEM_NR_ON;
        }

        if ( (powerActive & PAF_DISPLAY) == 0 && m_powerSetup[idx].displayText.item != POWER_ITEM_NR_OFF )
        {
            memset( &m_powerSetup[idx].displayText, 0, sizeof( PowerModule ) );
            m_powerSetup[idx].displayText.item = POWER_ITEM_NR_ON;
        }

        if ( (powerActive & PAF_INPUT_MANUAL_FINGERPRINT) == 0 && m_powerSetup[idx].fingerprint.item != POWER_ITEM_NR_OFF )
        {
            memset( &m_powerSetup[idx].fingerprint, 0, sizeof( PowerModule ) );
            m_powerSetup[idx].fingerprint.item = POWER_ITEM_NR_ON;
        }

        if ( (powerActive & PAF_INPUT_MANUAL_TRANSPONDER) == 0 && m_powerSetup[idx].transponder.item != POWER_ITEM_NR_OFF )
        {
            memset( &m_powerSetup[idx].transponder, 0, sizeof( PowerModule ) );
            m_powerSetup[idx].transponder.item = POWER_ITEM_NR_ON;
        }

        if ( (powerActive & PAF_MOVEMENT_GPS) == 0 && m_powerSetup[idx].gps.item != POWER_ITEM_NR_OFF )
        {
            memset( &m_powerSetup[idx].gps, 0, sizeof( PowerModule ) );
            m_powerSetup[idx].gps.item = POWER_ITEM_NR_ON;
        }

        if ( (powerActive & PAF_USB_HOST) == 0 && m_powerSetup[idx].usbHost.item != POWER_ITEM_NR_OFF )
        {
            memset( &m_powerSetup[idx].usbHost, 0, sizeof( PowerModule ) );
            m_powerSetup[idx].usbHost.item = POWER_ITEM_NR_ON;
        }

        if ( (powerActive & PAF_MMC) == 0 && m_powerSetup[idx].mmc.item != POWER_ITEM_NR_OFF )
        {
            memset( &m_powerSetup[idx].mmc, 0, sizeof( PowerModule ) );
            m_powerSetup[idx].mmc.item = POWER_ITEM_NR_ON;
        }
    }

    // Sicherstellen, dass die nicht nutzbaren Ereignisse auf Aus stehen.
    idx = 0;
    bool remove;
    for ( ; idx < POWER_EVENT_MAX; ++idx )
    {
        remove = false;

        switch ( m_powerEvent[idx].type )
        {
            case POWER_EVENT_NR_KEY_PRESSED:           ///< Eingabeereignis(Tastatur, Biokey, Transponder, Barcode)
            case POWER_EVENT_NR_KEY_TIMEOUT_1:         ///< Eingabe TimeOut 1
            case POWER_EVENT_NR_KEY_TIMEOUT_2:         ///< Eingabe TimeOut 2
                break;

            case POWER_EVENT_NR_USB:                   ///< Stromversorgung über USB aktiviert
                remove = true;
                break;

            case POWER_EVENT_NR_BATTERY:               ///< Stromversorgung über Batterie aktiviert
            case POWER_EVENT_NR_LINE_IN:               ///< Stromversorgung über Netzteil aktiviert
            case POWER_EVENT_NR_BAT_LOW:               ///< Batteriezustand niedrig
            case POWER_EVENT_NR_BAT_EMPTY:             ///< Batteriezustand leer
                break;

            case POWER_EVENT_NR_SWITCH_ON:             ///< Ein-/Ausschalter schaltet EIN
            case POWER_EVENT_NR_SWITCH_OFF:            ///< Ein-/Ausschalter schaltet AUS
                if ( !device.hasSetupType( { SetupType::BDE_MasterIV, SetupType::AE_MasterIV, SetupType::Timeboy_MasterIV, SetupType::Timeboy_Mobil_PZE, SetupType::Exklusiv_LineIV } ) )
                {
                    remove = true;
                }
                break;

            case POWER_EVENT_NR_TB_PLUGIN:             ///< Timeboy-Steck-Event
            case POWER_EVENT_NR_TB_PLUGOUT:            ///< Timeboy-Zieh-Event
                if ( device.hasSetupType( { SetupType::Timeboy_MasterIV, SetupType::Timeboy_Mobil_PZE, SetupType::IPC_Extension } ) )
                {
                    remove = true;
                }
                break;

            case POWER_EVENT_NR_MDE_START:             ///< Start-Erkennung MDE
            case POWER_EVENT_NR_MDE_STOP:              ///< Stopp-Erkennung MDE
                break;

            case POWER_EVENT_NR_ANALOG_OVER_VALUE:     ///< Analogwertüberwachung Messwert Überschreitung
            case POWER_EVENT_NR_ANALOG_UNDER_VALUE:    ///< Analogwertüberwachung Messwert Unterschreitung
                if ( device.hasSetupType( { SetupType::Timeboy_MasterIV, SetupType::Timeboy_Mobil_PZE } ) )
                {
                    remove = true;
                }
                break;

            case POWER_EVENT_NR_MOVE_START:            ///< Bewegungserkennung Start
            case POWER_EVENT_NR_MOVE_STOP:             ///< Bewegungserkennung Stopp
            case POWER_EVENT_NR_COMMUNICATION:         ///< Kommunikation
            case POWER_EVENT_NR_COMMUNICATION_TIMEOUT: ///< KommunikationsTimeOut
                break;
        }

        // Soll das Ereigniss geloescht werden?
        if ( remove )
        {
            memset( &m_powerEvent[idx], 0, sizeof( PowerEvent ) );
            m_powerEvent[idx].profile = 0xFF;
        }
    }

    // Werte in den Geraeteeintrag uebernehmen
    device.m_extended.powerActiv = powerActive;
}

